Package: munsellinterpol
Type: Package
Title: Interpolate Munsell Renotation Data from Hue Value/Chroma to
        CIE/RGB
Version: 3.4-0
Encoding: UTF-8
Date: 2026-07-03
Authors@R: c(person(given = "Jose", family = "Gama", role = c("aut","trl"), email = "rxprtgama@gmail.com"),
        person(given = "Paul", family = "Centore", role = c("aut","cph"), email = "centore@99main.com"),
        person(given = "Glenn", family = "Davis", role = c("aut","cre"), email = "gdavis@gluonics.com")  )
Description: Methods for interpolating data in the Munsell color system following the ASTM D-1535 standard. Hues and chromas with decimal values can be interpolated and converted to/from the Munsell color system and CIE xyY, CIE XYZ, CIE Lab, CIE Luv, or RGB.  Includes ISCC-NBS color block lookup.  Based on the work by Paul Centore, "The Munsell and Kubelka-Munk Toolbox".
License: GPL (>= 3)
LazyLoad: yes
LazyData: yes
Depends: R (>= 3.5.0)
Imports: rootSolve, spacesRGB, spacesXYZ (>= 1.6-0), logger
Suggests: microbenchmark, mgcv, knitr, rmarkdown, colorSpec (>= 1.8-0),
        flextable, OxSR
Author: Jose Gama [aut, trl],
  Paul Centore [aut, cph],
  Glenn Davis [aut, cre]
Maintainer: Glenn Davis <gdavis@gluonics.com>
Repository: CRAN
NeedsCompilation: no
VignetteBuilder: knitr
BuildVignettes: yes
Packaged: 2026-07-03 03:25:11 UTC; Glenn
Date/Publication: 2026-07-03 14:30:02 UTC

#  exportPattern("^[[:alpha:]]+")


# functions written by Jose Gama and Glenn Davis
export( LabToMunsell,   LabtoMunsell,
        LuvToMunsell,   LuvtoMunsell,
        MunsellHVC,
        MunsellToLab,
        MunsellToLuv,
        MunsellTosRGB,
        MunsellToxyY,
        MunsellToXYZ,
        sRGBtoMunsell,
        xyYtoMunsell,
        XYZtoMunsell,
        IsWithinMacAdamLimits,
        MunsellSpecToColorlabFormat,
        ColorlabFormatToMunsellSpec
        )

# functions by Jose Gama, and used in package 'colorscience'
#export( # srgb2xyz,
#        xyz2srgb )

# functions added by Glenn Davis
export( HVCfromMunsellName,
        HueStringFromNumber,
        MunsellNameFromHVC,
        HueNumberFromString,
        VfromY,
        YfromV,
        plotLociHC,
        plotPatchesH,
        ColorBlockFromMunsell,
#       sectionOptimals,
        RGBtoMunsell,
        MunsellToRGB,
        NickersonColorDifference,
        roundHVC
        )

#   export for testing only
# export( process_white )

import( logger )

importFrom(rootSolve,multiroot)
#   importFrom(arrangements,combinations)
#   importFrom(geometry, delaunayn, tsearchn)
importFrom(stats, approx, spline, splinefun, predict, residuals, formula, lm, polym, coef )
importFrom(graphics, abline, grid, lines, polygon, par, plot,
                plot.new, plot.window, points, rect, segments, text, title)
importFrom(utils, data, read.table, str, flush.console, object.size, combn )

The Munsell xy data was copied from:

http://www.rit-mcsl.org/MunsellRenotation/all.dat
http://www.rit-mcsl.org/MunsellRenotation/real.dat

Also see:
http://www.rit.edu/science/pocs/renotation


The xy data for Value >= 1 is taken from:

    Newhall, Sidney M., Dorothy Nickerson, Deane B. Judd.
    Final Report of the O.S.A. Subcommitte on the Spacing of the Munsell Colors.
    Journal of the Optical Society of America.
    Vol. 33.  No. 7.  pp. 385-418.
    July 1943.

The xy data for Value < 1 is taken from:

    Judd, Deane B. and Gunter Wyszecki.
    Extension of the Munsell Renotation System to Very Dark Colors.
    Journal of the Optical Society of America.
    Vol. 46.  No. 4.  pp. 281-284.
    April 1956.
    
In addition, the file all.dat has extrapolated xy data
that goes to higher Chromas than the data in these 2 references.
Unfortunately the method of extrapolation is not documented.

## ----setup, echo=FALSE, results="hide"--------------------------------------------------
require("knitr",quietly=TRUE)
opts_chunk$set(fig.path="figs/ag2-", fig.align="center",
  fig.width=7, fig.height=7, comment="")
knit_hooks$set(output = function(x, options) {
  paste('\\begin{Soutput}\n', x, '\\end{Soutput}\n', sep = '')
})
options(width=90)
par( omi=c(0,0,0,0), mai=c(0.2,0.2,0.2,0.2) )
if(!file.exists("figs")) dir.create("figs")

## ----packs, echo=TRUE, message=FALSE----------------------------------------------------
library( munsellinterpol )

## ----lee11, echo=TRUE, fig.pos="H", fig.height=5, out.width='1.0\\linewidth', fig.cap='Loci of Constant Hue and Chroma for Value 8/'----
par( omi=c(0,0,0,0), mai=c(0.41,0.7,0.6,0.2) )
plotLociHC( value=8 )

## ----lee12, echo=TRUE, fig.pos="H", fig.height=5, out.width='1.0\\linewidth', fig.cap='Loci of Constant Hue and Chroma for Value 8/'----
par( omi=c(0,0,0,0), mai=c(0.41,0.7,0.6,0.2) )
plotLociHC( value=8, hue=seq(5,55,by=2.5), chroma=seq(2,34,by=2) )

## ----lee13, echo=TRUE, fig.pos="H", fig.height=5, out.width='1.0\\linewidth', fig.cap='Loci of Constant Hue and Chroma for Value 0.4/'----
par( omi=c(0,0,0,0), mai=c(0.41,0.7,0.6,0.2) )
plotLociHC( value=0.4, hue=seq(80,90,by=2.5), chroma=seq(2,12,by=2) )

## ----lee14, echo=TRUE, fig.pos="H", fig.height=5, out.width='1.0\\linewidth', fig.cap='Loci of Constant Hue and Chroma for Value 8/, with bilinear interpolation'----
par( omi=c(0,0,0,0), mai=c(0.41,0.7,0.6,0.2) )
plotLociHC( value=8, hcinterp='bilinear' )

## ----finish, echo=FALSE, results="asis"-------------------------------------------------
knit_hooks$set(output = function(x, options) { x })
toLatex(sessionInfo(), locale=FALSE)

## ----setup, include=FALSE-------------------------------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)
options( width=100 )

## ----echo=TRUE, message=FALSE---------------------------------------------------------------------
library(munsellinterpol)
library(spacesRGB)  # for converting to RGB and plotting the patches
library(spacesXYZ)  # for xyY<->XYZ and Chromatic Adaptation Transform

## ----echo=TRUE------------------------------------------------------------------------------------
path = system.file( 'extdata/dental.txt', package='munsellinterpol' )
dental = read.table( path, header=TRUE, sep='\t', stringsAsFactors=FALSE )
dental

## ----echo=TRUE,  fig.width=7.2, fig.height=5.2,  fig.show='hold'----------------------------------
xyY = as.matrix( dental[ c('x','y','Y') ] )
XYZ = XYZfromxyY( xyY ) / 100
# adapt from Illuminant C to the whitepoint of sRGB, which is D65
# make the Chromatic Adaptation Transform
theCAT = spacesXYZ::CAT( 'C', getWhiteXYZ('sRGB',which='display') )
XYZ = adaptXYZ( theCAT, XYZ )
# create data.frame obj for plotting
obj = expand.grid( LEFT=1:6, TOP=1:4 )
obj$WIDTH = 0.9
obj$HEIGHT = 0.9
obj$RGB = RGBfromXYZ( XYZ, space='sRGB' )$RGB     # convert to sRGB
rownames(obj) = rownames(dental)
# plot as square patches
par( omi=c(0,0,0,0), mai=c(0.1,0.1,0.1,0.1) )
plotPatchesRGB( obj, which='signal', labels="bottomleft", adj=c(-0.2,-0.5), cex=0.7 )

## ----echo=TRUE------------------------------------------------------------------------------------
Lab  = as.matrix( dental[ c('L','a','b') ] )
XYZ = XYZfromxyY( xyY )
Lab2 = LabfromXYZ( XYZ/100, 'C' )     # recompute Lab
HVC  = HVCfromMunsellName( dental$Munsell )
HVC2 = XYZtoMunsell( XYZ )            # recompute Munsell HVC
comp = data.frame( row.names=rownames(dental) )
comp$Y  = dental$Y
comp$L  = Lab[ ,1]
comp$L2 = round(Lab2[ ,1],4)
comp$Ldiff = round( comp$L - comp$L2, 4 )
comp$DeltaE = round( DeltaE( Lab, Lab2 ), 4 )   # DeltaE is the pairwise color difference
comp$Munsell = dental$Munsell
comp$Munsell2 = MunsellNameFromHVC( HVC2, format='f', digits=2 )
comp$NickersonCD = round( NickersonColorDifference( HVC, HVC2 ), 4 )
comp

## ----echo=TRUE------------------------------------------------------------------------------------
HVC3 = HVC2
HVC3[ ,2] = VfromY( dental$Y, which='MgO' )
comp$Munsell2 = NULL
comp$NickersonCD = NULL
comp$Munsell3 = MunsellNameFromHVC( HVC3, format='f', digits=2 )
comp$NickersonCD = round( NickersonColorDifference( HVC, HVC3 ), 4 )
comp

## ----echo=TRUE------------------------------------------------------------------------------------
obj = data.frame( row.names=rownames(dental) )
obj$Munsell2 = MunsellNameFromHVC( HVC2, format='f', digits=2 )
block = ColorBlockFromMunsell( HVC2 )
obj[[ "ISCC-NBS Name" ]] = block$Name
obj

## ----echo=FALSE, results='asis'-------------------------------------------------------------------
sessionInfo()

## ----setup, include=FALSE-------------------------------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)
options( width=100 )

## ----echo=TRUE, message=TRUE----------------------------------------------------------------------
library(munsellinterpol)

## ----echo=TRUE, warning=TRUE----------------------------------------------------------------------
load_soil <- function( soil_raw=OxSR::soil_refle, subset=c( "a4", "a6", "a8", "a11", "a14", "a15" ),
                                         wave=380:780 )
    {
    #   locate only the wavelengths in OxSR::soil_refle that are relevant for color
    #   the given data.frame OxSR::soil_refle goes way into the infra-red and these are not needed
    #   We already know that all relevant multiples of 1nm are in soil_raw
    idx     = match( wave, soil_raw$wavelength_nm )

    #   extract just a subset of samples with high variability
    #   the reflectances in OxSR::soil_refle are percentages, so divide by 100    
    mat     = as.matrix(soil_raw)[ idx, subset ] / 100
    
    #   convert the matrix mat to a colorSpec object
    soil_spec = colorSpec::colorSpec( mat, wavelength=wave, quantity='reflectance' )
    
    #   add neutral gray as a test sample
    gray      = colorSpec::neutralMaterial( 0.18, wavelength=wave )
    soil_spec = colorSpec::bind( soil_spec, gray )
   
    return( soil_spec ) 
    }

## ----echo=TRUE, warning=TRUE,  fig.width=7.3, fig.height=5.2,  fig.show='hold'--------------------
soil_spec = load_soil()
par( omi=c(0,0,0,0), mai=c(0.6,0.6,0.3,0.3) )
plot( soil_spec, legend="topleft" )

## ----echo=TRUE, warning=TRUE, results="hide"------------------------------------------------------
soil_data <- function( soil_spec, CCT )
    {
    wave    = colorSpec::wavelength(soil_spec) 
    
    #   make illuminant from the CCT. This is the reference illuminant in IES Standard TM-30.
    illum   = colorSpec::referenceSpectraTM30( CCT, wavelength=wave )
    
    #   make scanner from this illuminant and the standard x,y,z responsivities of the human eye
    scanner = colorSpec::product( illum, "soil", colorSpec::xyz1931.1nm, wavelength=wave )
    
    #   scale so the Perfect Reflecting Diffuser has Y=100; which is conventional for Munsell work
    scanner = colorSpec::calibrate( scanner, response=c(NA,100,NA), method='scaling' )
    
    #   compute white point of illum
    white   = colorSpec::product( colorSpec::neutralMaterial(1,wavelength=wave), scanner )
    
    #   compute XYZ for all soil samples; this is XYZ under illum
    XYZ     = colorSpec::product( soil_spec, scanner )
    
    #   compute Lab for all soil samples, under illum
    Lab     = spacesXYZ::LabfromXYZ( XYZ, white )

    #   convert Lab to Munsell HVC, which automatically does a chromatic adaptation to Illuminant C
    HVC     = LabtoMunsell( Lab, white=white )   #  the default is to use the original NBS variant of C
    rownames(HVC)   = colorSpec::specnames(soil_spec)
    
    # create output data.frame and add 4 columns: HVC, Munsell notation, ISCC-NBS color name, and Lab
    out = data.frame( row.names=rownames(HVC) )
    out$HVC     = HVC
    out$Munsell = MunsellNameFromHVC( HVC )
    out[[ "ISCC-NBS Name" ]] = ColorBlockFromMunsell( HVC )$Name
    out$Lab     = Lab
    
    return( out )
    }

## ----echo=TRUE, message=TRUE----------------------------------------------------------------------
dat_soil = soil_data( soil_spec, CCT=6500 )
dat_soil

## ----echo=TRUE, message=TRUE----------------------------------------------------------------------
dat_rnd     = roundHVC( dat_soil$HVC, books="soil" )
dat_rnd$HVC = NULL     # delete matrix HVC, because it is redundant in this context
dat_rnd

## ----echo=TRUE, message=TRUE, fig.width=7.3, fig.height=5.2,  fig.show='hold'---------------------
par( omi=c(0,0,0,0), mai=c(0.4,0.4,0.25,0) )
plotPatchesH( "2.5YR",   value=c(2.5,3:8), chroma=c(1,2,3,4,6,8) )
text( c(4,4,3), c(5,4,4), c("a6","a8","a15"), adj=c(0.5,0.5), col="white" )

## ----echo=TRUE, message=TRUE----------------------------------------------------------------------
tbl = data.frame( row.names=1:(2*nrow(dat_soil)) )
tbl$Sample  = as.character( rbind(rownames(dat_soil),'') )
tbl$Precise = as.character( rbind('',MunsellNameFromHVC( dat_soil$HVC, digits=3 )) )
tbl$Rounded = as.character( rbind('',dat_rnd$MunsellRounded ) )

library( flextable )
myrt <- regulartable( tbl )
myrt <- width( myrt, j=c(2,3), width=2 )
myrt <- align( myrt, j=1:3, align='center', part='all' )
myrt <- hline( myrt, i=seq( 2, nrow(tbl)-2, by=2 ), border=fp_border_default( color="black", width=2 ) )
myrt <- hrule( myrt, rule = "exact" )
idxcolor <- seq(1,nrow(tbl)-1,by=2)
myrt <- height( myrt, i=idxcolor, height=1 )
myrt <- bg( myrt, i=idxcolor, j=2, bg=rgb( round(MunsellTosRGB(dat_soil$HVC)$RGB), max=255 ) )
myrt <- bg( myrt, i=idxcolor, j=3, bg=rgb( round(MunsellTosRGB(dat_rnd$MunsellRounded)$RGB), max=255 ) )

## ----echo=FALSE, message=TRUE---------------------------------------------------------------------
myrt

## ----echo=FALSE, results='asis'-------------------------------------------------------------------
sessionInfo()

## ----setup, include=FALSE-------------------------------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)
options( width=100 )
library(munsellinterpol)

## ----echo=TRUE------------------------------------------------------------------------------------
MunsellToxyY( '4.2RP 5.5/8' )

## ----echo=TRUE------------------------------------------------------------------------------------
xyY = MunsellToxyY( '4.2RP 5.5/8' )$xyY
xyYtoMunsell( xyY )

## ----echo=TRUE,  message=TRUE, results='hold', fig.width=7, fig.height=6,  fig.show='hold'--------
par( omi=c(0,0,0,0), mai=c(0.6,0.7,0.4,0.2) )
plotPatchesH( "10GY", back='#f7f7f7' )

## ----echo=TRUE,  message=TRUE, results='hold', fig.width=7, fig.height=6,  fig.show='hold'--------
par( omi=c(0,0,0,0), mai=c(0.6,0.7,0.4,0.2) )
plotPatchesH( "10GY", space='AdobeRGB', back='#f7f7f7' )

## ----echo=TRUE,  message=TRUE, results='hold', fig.width=7, fig.height=6,  fig.show='hold'--------
par( omi=c(0,0,0,0), mai=c(0.6,0.7,0.6,0.2) )
plotLociHC( value=8 )

## ----echo=TRUE,  message=TRUE, results='hold', fig.width=7, fig.height=6,  fig.show='hold'--------
par( omi=c(0,0,0,0), mai=c(0.6,0.7,0.6,0.2) )
plotLociHC( value=8, coords='ab' )

## ----echo=TRUE------------------------------------------------------------------------------------
path = system.file( 'extdata/PantoneCoY.txt', package='munsellinterpol' )
pantone = read.table( path, header=TRUE, sep='\t', strings=FALSE )
pantone

## ----echo=TRUE------------------------------------------------------------------------------------
pantone$Year = NULL ;   pantone$Code = NULL
RGB = as.matrix( pantone[ , c('R','G','B') ] )
HVC = RGBtoMunsell( RGB, space='sRGB' )
pantone$Munsell = MunsellNameFromHVC( HVC )
block = ColorBlockFromMunsell( HVC )
pantone[[ "ISCC-NBS Name" ]] = block$Name
pantone[[ "ISCC-NBS Centroid" ]] = block$Centroid
pantone

## ----echo=TRUE------------------------------------------------------------------------------------
color.pant = rgb( RGB, max=255 )
color.cent = rgb( MunsellToRGB( block$Centroid, space='sRGB' )$RGB, max=255 )
tbl = data.frame( row.names=1:nrow(pantone) )
tbl[[ "Pantone Name" ]]  = pantone$Name
tbl[[ "Pantone Color" ]] = '' ;    tbl[[ "Centroid Color" ]] = ''
tbl[[ "ISCC-NBS Name" ]] = block$Name
library( flextable )
myrt <- regulartable( tbl )
myrt <- align( myrt, j=4, align='left', part='all' )
myrt <- align( myrt, j=2:3, align='center', part='all' )
myrt <- hrule( myrt, rule='exact' )
myrt <- height( myrt, height=0.75 )
myrt <- width( myrt, j=c(1,4), width=2 ) ; myrt <- width( myrt, j=2:3, width=2.5 )
myrt <- fontsize( myrt, size=14, part='all' ) ; myrt <- fontsize( myrt, size=16, part='header' )
for( i in 1:nrow(tbl) )
{ myrt <- bg(myrt, i=i, j=2, bg=color.pant[i]) ; myrt <- bg(myrt, i=i, j=3, bg=color.cent[i]) }
myrt

## ----echo=FALSE, results='asis'-------------------------------------------------------------------
sessionInfo()

## ----setup, echo=FALSE, results="hide"--------------------------------------------------
require("knitr",quietly=TRUE)
opts_chunk$set(fig.path="figs/ag2-", fig.align="center",
  fig.width=7, fig.height=7, comment="")
knit_hooks$set(output = function(x, options) {
  paste('\\begin{Soutput}\n', x, '\\end{Soutput}\n', sep = '')
})
options(width=90)
par( omi=c(0,0,0,0), mai=c(0.2,0.2,0.2,0.2) )
if(!file.exists("figs")) dir.create("figs")

## ----packs, echo=TRUE, message=FALSE----------------------------------------------------
library( munsellinterpol )

## ----lee11, echo=TRUE, fig.pos="H", fig.height=5, out.width='1.0\\linewidth', fig.cap='Value V  vs.  Luminance Factor Y'----
par( omi=c(0,0,0,0), mai=c(0.5,0.5,0.1,0.1) )
plot( c(0,100), c(0,10), type='n', xlab='', ylab='', las=1, tcl=0,
        lab=c(10,8,7), mgp=c(3,0.25,0) )
title( xlab='Y', line=1.5 ) ;  title( ylab='Value', line=1.5 )
grid( lty=1 ) ;  abline( h=0, v=0 )
V = seq( 0, 10, by=0.125 )
color = unlist( list(ASTM='black',OSA='black',MgO='black',Munsell='red',Priest='blue') )
for( w in names(color) )
    lines( YfromV(V,w), V, col=color[w], lty=ifelse(w=='MgO',2,1), lwd=0.75 )
lty = ifelse( names(color)=='MgO', 2, 1 )
legend( "bottomright", names(color), bty='n', lty=lty, lwd=1.5, col=color, inset=0.1 )

## ----lee12, echo=TRUE, fig.pos="H", fig.height=3.5, out.width='1.0\\linewidth', fig.cap='The Difference between ASTM and OSA Definitions of Value'----
par( omi=c(0,0,0,0), mai=c(0.5,1,0.1,0.1) )
Y = seq( 0, 100, by=0.5 )
delta = VfromY(Y,'OSA') - VfromY(Y,'ASTM')
plot( range(Y), range(delta), type='n', xlab='', ylab='', las=1, tcl=0,
        lab=c(10,8,7), mgp=c(3,0.25,0) )
title( xlab='Y', line=1.5 ) ;  title( ylab='{OSA Value} - {ASTM Value}', line=3 )
grid( lty=1 ) ;  abline( h=0, v=0 )
lines( Y, delta, lwd=0.75 )

## ----Lightness, echo=TRUE, message=FALSE------------------------------------------------

Lightness_from_linear  <-  function( Y )  {
ifelse( Y < (24/116)^3, (116/12)^3 * Y, 116*Y^(1/3) - 16 )
}

## ----lee13, echo=TRUE, fig.pos="H", fig.height=5, out.width='1.0\\linewidth', fig.cap='Comparing Lightness and Munsell Value'----
par( omi=c(0,0,0,0), mai=c(0.5,0.75,0.1,0.1) )
Y = (0:100)/100
L = Lightness_from_linear( Y )
plot( range(Y), range(L), type='n', xlab='', ylab='', las=1, tcl=0,
        lab=c(10,8,7), mgp=c(3,0.25,0) )
title( xlab='Y   (luminance factor)', line=1.5 ); title( ylab='Lightness', line=2 )
grid( lty=1 ) ;  abline( h=0, v=0 )
lines( Y, L, lwd=0.75 )
V = VfromY( 100 * Y, 'ASTM' )
lines( Y, 10*V, lty=2 )
legend(  "bottomright", c("Lightness (CIE)","10*Value (ASTM)"), lty=c(1,2),
            bty='n', inset=0.1 )

## ----lee14, echo=TRUE, fig.pos="H", fig.height=3, out.width='1.0\\linewidth', fig.cap='Comparing Lightness and Munsell Value'----
par( omi=c(0,0,0,0), mai=c(0.5,0.75,0.1,0.1) )
quotient = L / V
plot( range(Y), range(quotient,na.rm=T), type='n', xlab='', ylab='', las=1, tcl=0,
        lab=c(10,8,7), mgp=c(3,0.25,0) )
title( xlab='Y   (luminance factor)', line=1.5 )
title( ylab='Lightness / Value', line=3 )
grid( lty=1 ) ;  abline( h=0, v=0 )
lines( Y, quotient )

## ----finish, echo=FALSE, results="asis"-------------------------------------------------
knit_hooks$set(output = function(x, options) { x })
toLatex(sessionInfo(), locale=FALSE)

library( munsellinterpol )

options( width=144 )

#
#   checkColorLookup()
#   performs lookup with ColorBlockFromMunsell() and verifies that the color block number is correct
#
#   return value:  TRUE or FALSE
#
testColorBlock <- function()
    {
    Number  = ColorBlockFromMunsell( CentroidsISCCNBS$MunsellSpec )$Number
    
    #   Number[1] = 57L  test the test !
    
    #   compute the number of classification errors
    idx = which( Number != CentroidsISCCNBS$Number )
    
    out = (length(idx) == 0)
    
    if( ! out )
        {
        mess    = sprintf( "testColorBlock().  There are %d errors\n", length(idx) )
        cat(mess)
        df  = cbind( CentroidsISCCNBS[ idx, ], NumberLookup=Number[idx] )
        print( df )
        }

    return( out )
    }
    
    
if( ! testColorBlock() )    stop( "testColorBlock() failed !\n" )


cat(  "Passed all ISCC-NBS tests !\n" )

#
#   xyY             a numeric Nx3 matrix, or a vector that can be converted to one,
#                   with xyY in each row
#   xyC             xy  2-vector for illuminant C
#                   can also be a string: 'NBS', 'JOSA', 'NTSC', or 'ASTM'
#   hcinterp        'bicubic' or 'bilinear'
#   vinterp         'cubic' or 'linear'
#   VfromY          'ASTM' or 'NBS' or 'MUNSELL'.  'MGO' is not allowed here
#   warn            log a warning if there were any inversion failures
#   perf            add performance data to returned output
#
#   return value    a data.frame with columns
#                   xyY             the input
#                   HVC             a matrix with the same size as xyY, with (Hue,Value,Chroma) in the rows
#                   SAMPLE_NAME     Munsell notation for HVC, a string

xyYtoMunsell  <-  function( xyY,
                            xyC='NBS',
                            hcinterp='bicubic',
                            vinterp='cubic',
                            VfromY='ASTM',
                            rtol=1.e-8,
                            atol=1.e-6,
                            warn=TRUE,
                            perf=FALSE )
    {
    p   = 'rootSolve'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }

    time_start  = gettime()

    xyY = prepareNx3(xyY)
    if( is.null(xyY) )  return(NULL)

    if( is.null(colnames(xyY)) )
        colnames(xyY)   = c( 'x', 'y', 'Y' )

    #   process xyC
    xyC  = process_xyC( xyC )
    if( any(is.na(xyC)) )
        {
        log_level( ERROR, "xyC is invalid." )
        return(NULL)
        }

    # process hcinterp
    hcinterp    = process_hcinterp(hcinterp)
    if( is.na(hcinterp) )
        {
        log_level( ERROR, "hcinterp='%s' is invalid.", hcinterp )
        return(NULL)
        }

    #   assign function pointer
    if( hcinterp == 'bicubic' )
        hcinterpfun = bicubicCardinal  # mybicubic
    else if( hcinterp == 'bilinear' )
        hcinterpfun = bilinearApprox    # mybilinear



    # process vinterp
    vinterp    = process_vinterp(vinterp)
    if( is.na(vinterp) )
        {
        log_level( ERROR, "vinterp='%s' is invalid.", as.character(vinterp) )
        return(NULL)
        }


    #   process VfromY
    full    = c( 'ASTM', 'NBS', 'MUNSELL', 'PRIEST' )
    idx     = pmatch( toupper(VfromY), full )
    if( is.na(idx) )
        {
        log_level( ERROR, "VfromY='%s' is invalid. ('MgO' is not allowed in this function).", VfromY )
        return(NULL)
        }

    # allocate arrays for mapping H,C -> x and H,C -> y
    # Value 7 has the full extent, so just use that as a template
    xlookup = array( NA_real_, dim=dim(p.LookupList[[7]]$x) )
    ylookup = array( NA_real_, dim=dim(p.LookupList[[7]]$y) )

    H.vector    = attr( p.LookupList, "H.vector" )   # ; cat( 'H.vector=',H.vector,'\n')
    V.vector    = attr( p.LookupList, "V.vector" )
    C.vector    = attr( p.LookupList[[7]], "C.vector" )

    #   fill neutrals at Chroma=0, these are the same for all hues and values
    xlookup[ ,1] = xyC[1]
    ylookup[ ,1] = xyC[2]



    #   define the "forward" function, which we will try to find the root of
    #   AB          input 2-vector.  It changes on each iteration.
    #   value       the Munsell Value.  It remains constant throughout the iteration
    #   vlevel      vector of value indices (not the values) where HC is interpolated.
    #               This is 1-based  It depends only on value. It remains constant throughout the iteration
    #   xy_target   we are trying to find HC that maps to xy_target.  It remains constant throughout the iteration
    forwardfun <- function( AB,    value, vlevel, xy_target )
        {
        evaluation <<- evaluation + 1
        #cat( "######  evaluation = ", evaluation, " ###########\n", file=stderr() )
        # cat( 'HC=', HC, "  value=", value, '\n' )

        HC  = unlist( HCfromAB( AB[1], AB[2] ) )

        #if( HC[2] < 0 )
        #    log_level( WARN, "Chroma = %g < 0.  hue=%g value=%g.  evaluation=%d.",
        #                HC[2], HC[1], value, evaluation )


        iH  = findInterval( HC[1], H.vector, all.inside=TRUE )  #; cat( 'iH=',iH,'\n')
        iC  = findInterval( HC[2], C.vector, all.inside=TRUE )

        if( hcinterp == 'bicubic' )
            {
            hseq    = max(iH-1,1) : min(iH+2,length(H.vector))      # 4 points for bicubic
            cseq    = max(iC-1,1) : min(iC+2,length(C.vector))      # 4 points for bicubic
            }
        else
            {
            hseq    = iH : (iH+1)      # 2 points for bilinear
            cseq    = iC : (iC+1)      # 2 points for bilinear
            }

        #   iterate over 4x4 or 2x2 grid, and compute missing values
        #print(hseq)
        #print(cseq)

        gridpoints = 0

        #   extract just the values that are relevant to the iteration, usually either 2 or 4 of them
        #   V.vector will not be used again
        Vsub    = V.vector[ vlevel ]

        for( ih in hseq )
            {
            for( ic in cseq )
                {
                if( ! is.na( xlookup[ih,ic] ) ) next

                #   xlookup[ih,ic] has been computed yet, so compute it now
                gridpoints = gridpoints + 1

                xvec    = rep( NA_real_, length(Vsub) )  # length(V.vector)
                yvec    = xvec

                for( k in 1:length(vlevel) )      # v in vlevel )
                    {
                    iV   = vlevel[k]
                    icc  = min( ic, ncol(p.LookupList[[iV]]$x) )
                    xvec[k] = p.LookupList[[iV]]$x[ih,icc]
                    yvec[k] = p.LookupList[[iV]]$y[ih,icc]
                    }

                fmask   = is.finite(xvec)
                count   = sum( fmask )    #; print(count)

                if( vinterp == 'linear'   ||  (p.vinterpOverride && value<2)  )
                    {
                    # print( xvec ) ; print( V.vector ) ; print( value )
                    if( count < 1 )
                        {
                        #   need 1 finite value
                        if( evaluation == 1 )
                            mess = sprintf( "forwardfun(). The initial guess for root is outside the range of the HVC -> xyY lookup table.  AB=%g,%g   HC=%g,%g",
                                                 AB[1], AB[2], HC[1], HC[2] )
                        else
                            mess = sprintf( "forwardfun(). The root-finder has wandered outside the range of the HVC -> xyY lookup table. evaluation=%d. AB=%g,%g   HC=%g,%g",
                                                evaluation, AB[1], AB[2], HC[1], HC[2] )
                        stop( mess, call.=FALSE )
                        }

                    if( count == 2 )  #  all(fmask)
                        {
                        #   interpolate between those 2
                        xlookup[ih,ic]  <<- approx( Vsub, xvec, xout=value )$y
                        ylookup[ih,ic]  <<- approx( Vsub, yvec, xout=value )$y
                        }
                    else
                        {
                        #   just use the one
                        #   this happens for 1 out of 994 optimals
                        idx = which( fmask )
                        xlookup[ih,ic]  <<- xvec[ idx ]
                        ylookup[ih,ic]  <<- yvec[ idx ]
                        }
                    }
                else if( vinterp == 'cubic' )
                    {
                    if( count < 2 )
                        {
                        #   need 2 finite values
                        if( evaluation == 1 )
                            mess = sprintf( "forwardfun(). The initial guess for root is outside the range of the HVC -> xyY lookup table.   AB=%g,%g   HC=%g,%g",
                                                AB[1], AB[2], HC[1], HC[2]  )
                        else
                            mess = sprintf( "forwardfun(). The root-finder has wandered outside the range of the HVC -> xyY lookup table. evaluation=%d.  AB=%g,%g  HC=%g,%g",
                                                evaluation, AB[1], AB[2], HC[1], HC[2] )
                        stop( mess, call.=FALSE )
                        }

                    xyvec =  cbind(xvec,yvec)
                    rownames(xyvec) = as.character(Vsub)
                    #print( xyvec )

                    if( all(fmask) )
                        #   the usual case
                        xy  = splineCardinal( Vsub, xyvec, xout=value )
                    else
                        {
                        #   despite the dilation, count==3 still happens for 11 out of 994 optimals
                        #   TODO:  another workaround might be shrinking the polym() model, as already done for value=0.2
                        # log_level( TRACE, "cubic  count=%d   evaluation=%d   value=%g", count, evaluation, value )
                        xy  = splineCardinal( Vsub[fmask], xyvec[fmask, ], xout=value )
                        }

                    #xlookup[ih,ic]  <<- spline( Vsub, xvec, xout=value, method='natural' )$y
                    #ylookup[ih,ic]  <<- spline( Vsub, yvec, xout=value, method='natural' )$y

                    xlookup[ih,ic]  <<- xy[1]
                    ylookup[ih,ic]  <<- xy[2]
                    }
                }
            }


        #   the grids xlookup and ylookup are ready, now interpolate in this V-plane
        xy      = numeric(2)
        xy[1]   = hcinterpfun( H.vector, C.vector, xlookup, HC[1], HC[2] )$z
        xy[2]   = hcinterpfun( H.vector, C.vector, ylookup, HC[1], HC[2] )$z


        #cat( 'vlevel=', vlevel, '\n' )
        #cat( 'xy_target=', xy_target, '\n' )
        #cat( 'gridpoints=', gridpoints, '\n' )
        #cat( 'xy=', xy, '\n' )

        delta   =  xy - xy_target
        #cat( 'delta=', sum(abs(delta)), '\n' )

        return(delta)
        }



    HVC = array( NA_real_, dim=dim(xyY) )
    colnames(HVC)   = c( 'H', 'V', 'C' )

    HVC[ ,2]    = VfromY( xyY[ ,3], which=VfromY )

    n   = nrow(xyY) #; print(n)

    evaluations.vec     = rep( NA_integer_, n )
    iterations.vec      = rep( NA_integer_, n )
    estim.precis.vec    = rep( NA_real_, n )
    time.elapsed.vec    = rep( NA_real_, n )

    for( i in 1:n )
        {
        time_begin  = gettime()  # as.double( Sys.time() )

        xy_target   = xyY[i,1:2]
        value       = HVC[i,2]

        if( is.na( value ) )    next

        if( 0 < value  &&  any( is.na(xy_target) ) )    next    # invalid, so ignore this one

        if( value==0  ||  all( xy_target == xyC ) )
            {
            #   neutrals are easy.
            #   Note that value==0 must be pure black !
            #   Note that xy_target must be equal to xyC exactly, with no tolerance;
            #       if the difference is extremely small we depend on approxABfromxyY() to set ABstart=(0,0)
            #       and the rootSolve::multiroot() tolerances to find a root with C=0.
            #   Set both H and C to 0
            HVC[ i, c(1,3) ] = 0
            time.elapsed.vec[i] = (gettime()  - time_begin)
            next
            }

        delta   = abs(V.vector - value)     #; print( delta )
        iV  = which.min( delta )
        if( delta[iV] < 5.e-9 )
            {
            #   this is so close to a V-plane, that this is probably where xyY[i, ] came from
            value       = V.vector[iV]
            HVC[i,2]    = value
            }

        if( value < 0  ||  10 < value )
            #   this is out of range
            next

        iV  = match(value,V.vector)

        if( is.na(iV) )
            {
            #   value is *between* 2 V-planes. This is the usual case
            #   erase the lookup tables, but only where Chroma > 0
            xlookup[ ,-1] = NA_real_    #; print(xlookup)
            ylookup[ ,-1] = NA_real_

            iV  = findInterval( value, V.vector, all.inside=TRUE )
            if( vinterp == 'linear'  || (p.vinterpOverride && value<2) )
                vlevel = iV : (iV+1)
            else if( vinterp == 'cubic' )
                vlevel = max(iV-1,1) : min(iV+2, length(V.vector))
            }
        else
            {
            #   value is *exactly* on a V-plane
            #   not likely but since all the action will be on this plane, we can optimize it
            #   cat( "i=", i, "  Exact value=", value, '\n', file=stderr() )
            nC  = dim(p.LookupList[[iV]]$x)[2]
            xlookup[ ,1:nC] = p.LookupList[[iV]]$x
            ylookup[ ,1:nC] = p.LookupList[[iV]]$y
            vlevel  = iV    # vlevel will never be accessed, but this is what it would be if it were
            }

        #   make an initial guess for AB
        evaluation = 0

        #HCstart  = approxHCfromxyY( xyY[i, ] , c(xyC,100) )
        #ABstart = unlist( ABfromHC( HCstart[1], HCstart[2] ) )

        ABstart = approxABfromxyY( xyY[i, ], c(xyC,100), V.vector )

        res = try( rootSolve::multiroot( forwardfun, ABstart, rtol=rtol, atol=atol, ctol=0, verbose=FALSE,
                                            value=value, vlevel=vlevel, xy_target=xy_target ),  silent=F )

        if( inherits(res,"try-error" ) )   # class(res) == "try-error" )
           {
           #    failed to find the root
           # print(res)
           HCstart = unlist( HCfromAB( ABstart[1], ABstart[2] ) )

           log_level( DEBUG, "rootSolve::multiroot() failed.  ABstart=%g,%g.   HCstart=%g,%g.  value=%g",
                                         ABstart[1], ABstart[2], HCstart[1], HCstart[2], value )
           next
           }

        time.elapsed.vec[i] = (gettime() - time_begin)    # ;as.double( Sys.time() ) - time_begin

        #   convert from AB back to HC
        HVC[ i, c(1,3) ] = unlist( HCfromAB( res$root[1], res$root[2] ) )

        #   cat( "Success !  iterations=", res$iter, '\n', file=stderr() )

        iterations.vec[i]   = res$iter
        evaluations.vec[i]  = evaluation
        estim.precis.vec[i] = res$estim.precis

        log_level( TRACE, "i=%d   ABstart=%g,%g  xy_target=%g,%g  iters=%d  estim.precis=%g.",
                        i, ABstart[1], ABstart[2], xy_target[1], xy_target[2], res$iter, res$estim.precis )
        }


    #   hue wrap-around to interval (0,100]
    hue         = HVC[ ,1] %% 100
    mask        = (hue==0)  &  (HVC[ ,3]!=0)    # make exception to hue<-100 when Chroma==0
    hue[mask]   = 100
    HVC[ , 1]   = hue

    SAMPLE_NAME = MunsellNameFromHVC(HVC)

    #   copy rownames from xyY to HVC, unless NULL then use SAMPLE_NAME
    rownames(HVC)   = rownames(xyY)
    if( is.null(rownames(HVC)) )    rownames(HVC) = SAMPLE_NAME

    out = data.frame( row.names=1:n, stringsAsFactors=FALSE )

    out$xyY         = xyY
    out$HVC         = HVC
    out$SAMPLE_NAME = SAMPLE_NAME

    if( perf )
        {
        out$time.elapsed    = time.elapsed.vec
        out$iterations      = iterations.vec
        out$evaluations     = evaluations.vec
        out$estim.precis    = estim.precis.vec
        }


    #   print( str(out) )

    if( FALSE )
        {
        time_elapsed = gettime() - time_start
        log_level( INFO, "Processed %d samples in %g seconds.  %g/sample.  [mean(time.elapsed)=%g].",
                                    n, time_elapsed, time_elapsed/n,  mean(out$time.elapsed,na.rm=TRUE) )
        }

    if( warn )
        {
        mask    = is.na(HVC[ ,1])  |  is.na(HVC[ ,3])
        if( any(mask)  )
            {
            log_level( WARN, "%d samples, out of %d, could not be mapped; HVC row set to NA.",
                                                sum(mask), n )
            }
        }

    return( out )
    }



#   this one uses global variable p.InversionCoeffs
approxABfromxyY <- function( xyY, xyY.white, V.vector )
    {
    p   = 'spacesXYZ'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }

    V   = VfromY( xyY[3] )
    iV  = which.min( abs(V - V.vector) )

    coeffs  = p.InversionCoeffs[[iV]]
    vars    = colnames(coeffs)[1]

    if( grepl( "a,b", vars, fixed=TRUE ) )
        {
        #   polynomial in a and b
        Yp          = YfromV( V.vector[iV] )    # use Y in the closest V plane, which is where the model was computed
        XYZ         = spacesXYZ::XYZfromxyY( c(xyY[1],xyY[2],Yp) )
        XYZ.white   = spacesXYZ::XYZfromxyY( xyY.white )

        Lab = spacesXYZ::LabfromXYZ( XYZ, XYZ.white )
        a   = Lab[2]
        b   = Lab[3]
        a2  = a*a
        b2  = b*b
        term    = c( a, a2, a2*a, b, a*b, a2*b, b2, a*b2, b2*b )    # compare order here with polym() in makeInversionCoeffs()
        }
    else if( grepl( "xdelta,ydelta", vars, fixed=TRUE ) )
        {
        #   polynomial in xdelta and ydelta
        xd  = xyY[1] - xyY.white[1]
        yd  = xyY[2] - xyY.white[2]

        xd2 = xd*xd
        yd2 = yd*yd

        term    = c( xd, xd2, xd2*xd, yd, xd*yd, xd2*yd, yd2, xd*yd2, yd2*yd )    # compare order here with polym() in makeInversionCoeffs()
        }
    else
        log_level( FATAL, "vars='%s' is invalid.", vars )

    AB  = numeric(2)
    for( k in 1:2 )
        AB[k]   = sum( coeffs[k, ] * term )

    return( AB )
    }

IsWithinMacAdamLimits <- function( xyY, Illuminant='C' )
    {
    p   = 'spacesXYZ'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }
        
    xyY = prepareNx3( xyY )    
    if( is.null(xyY) )  return(NULL)
    
    full    = names( p.OptimalHull )
    idx     = pmatch( toupper(Illuminant), full )
    if( is.na(idx) )
        {
        log_level( ERROR, "Illuminant='%s' is invalid.", Illuminant )
        return(NULL)
        }
    Illuminant  = full[idx]
    
    hull    = p.OptimalHull[[ Illuminant ]]

    #   out = is.finite( geometry::tsearchn( hull$XYZ, hull$tessellation, spacesXYZ::XYZfromxyY(xyY) )$idx )
    
    out = inside( hull, spacesXYZ::XYZfromxyY(xyY) )
    
    return( out )
    }
    

if( FALSE )
{    
#   xyY     Nx3 matrix of points on the optimal boundary
#
#   returns a list with 2 components
#       XYZ             the given points xyY -> XYZ, an Nx3 matrix
#       tesselation     as returned from geometry::delaunayn
#
makeOptimalHull <- function( xyY )
    {
    if( ! requireNamespace( 'spacesXYZ', quietly=TRUE ) )   return( NULL )
    
    if( ! requireNamespace( 'geometry', quietly=TRUE ) )    return( NULL )    
    
    XYZ = spacesXYZ::XYZfromxyY( xyY )
    
    #   option QJ is to joggle the input and avoid degenerate simplices
    #   without this option there are 2 degenerate ones for Illuminant C, and these generate a warning message for every query
    tessellation = geometry::delaunayn( XYZ, options='QJ' ) 
    
    out = list( XYZ=XYZ, tessellation=tessellation )
    
    return( out )
    }
}


    
    
#   Y       for the Y-plane defining the section.  in the interval (0,100)
#   n       number of points in the section to return
#   illum   'C' or 'D65'
#
#   returns a data.frame with n rows and these columns
#       XYZ     an optimal color, which is defined by 
#       lambda0 
#       lambda1
sectionOptimals  <-  function( Y, n=100, illum='C' )
    {
    ok  = (0 < Y)  &  (Y < 100)
    if( ! ok )
        {
        log_level( WARN, "Y=%g is invalid; it must be in interval (0,100).", Y )
        return(NULL)
        }
        
    ok  = illum %in% colnames( p.ACDs )
    if( ! ok )   
        {
        log_level( ERROR, "illum='%s' is invalid.", illum )
        return(NULL)
        }

    
    wave    = p.ACDs[[1]]   # wavelengths are column #1
    
    if( ! identical( wave, p.xyz1931$Wavelength ) )
        {
        log_level( FATAL, "illuminant '%s' and xyz have different wavelengths.", illum )
        return(NULL)
        }
    
    funlist = makeReparamFunctionList( wave, p.ACDs[[illum]], as.matrix(p.xyz1931[ ,2:4]) )
    

    Y0  = ( 0:(n-1) ) / n
    Y1  = (Y0 + Y/100) %% 1
    
    omega0  = funlist$omega.from.integral[[2]]( Y0 )
    omega1  = funlist$omega.from.integral[[2]]( Y1 )
    
    bandstop    = (omega1 < omega0)
    
    XYZ.white   = numeric(3)
    XYZ =   matrix( NA_real_, n, 3 )
    
    for( j in 1:3 )
        {
        XYZ.white[j]    = 100 * funlist$integral.from.omega[[j]]( 1 )
        XYZ[ ,j]        = 100 * ( funlist$integral.from.omega[[j]]( omega1 ) - funlist$integral.from.omega[[j]]( omega0 ) )
        
        #   correct the bandstops
        XYZ[bandstop,j] = XYZ[bandstop,j]  +  XYZ.white[j]
        }
        
    #   force Y to be exact
    XYZ[ ,2]    = Y
        
    out         = data.frame( row.names=1:n )
    out$XYZ     = XYZ
    out$lambda0 = funlist$lambda.from.omega( omega0 )
    out$lambda1 = funlist$lambda.from.omega( omega1 )
    
    return(out)
    }
    
    

#   x	a vector containing the x coordinates of the rectangular data grid.  Must be increasing.
#   y	a vector containing the y coordinates of the rectangular data grid.  Must be increasing.
#   z	a matrix containing the z[i,j] data values for the grid points (x[i],y[j]).
#   x0	vector of x coordinates, at which to interpolate
#   y0	vector of y coordinates, at which to interpolate
#
#   Warning:  If x and y are not regular, then the spline is not C^1

bicubicCardinal  <- function( x, y, z, x0, y0 )
    {    
    #   check validity
    lenx    = length(x)
    if( lenx != nrow(z) )
        {
        log_level( ERROR, "length(x)=%d != %d=nrow(z)", length(x), nrow(z) )
        return(NULL)
        }

    leny    = length(y)        
    if( leny != ncol(z) )
        {
        log_level( ERROR, "length(y)=%d != %d=ncol(z)", length(y), ncol(z) )
        return(NULL)
        }
    
    n   = length(y0)
    if( length(x0) != n )
        {
        log_level( ERROR, "length(x0)=%d != %d=length(y0)", length(x0), n )
        return(NULL)
        }
        
    zout    = rep( NA_real_, n )
    
    i   = findInterval( x0, x, rightmost.closed=FALSE ) #; print(i)
    j   = findInterval( y0, y, rightmost.closed=FALSE ) 
    
    zz  = rep( NA_real_, leny )
    
    for( k in 1:n )
        {
        i0  = i[k]
        if( i0==0  || x[lenx] < x0[k] )    next    #   outside the rectangle
        
        j0  = j[k]
        if( j0==0  || y[leny] < y0[k] )    next    #   outside the rectangle
        
        xexact  = (x0[k] == x[i0])
        yexact  = (y0[k] == y[j0])
        
        if( xexact  &&  yexact )
            {
            #   exactly on a grid point.  If one of the neighbors is NA it does not matter
            zout[k] = z[ i0, j0 ]   #; log_level( INFO, "Exactly on grid point %g,%g", x0[k], y0[k] )
            next
            }
        
        if( xexact )
            {
            #   on a gridline, so it reduces to splineCardinal()
            zout[k] = splineCardinal( y, z[i0, ], y0[k] )
            next
            }

        if( yexact )
            {
            #   on a gridline, so it reduces to splineCardinal()
            zout[k] = splineCardinal( x, z[ ,j0], x0[k] )
            next
            }         

        jseq    =  max(j0-1,1) :  min(j0+2,leny)  # ; print(jseq)

        #for( jj in jseq )
        #    zz[jj]  = splineCardinal( x, z[ ,jj], x0[k] )
        
        res = splineCardinal( x, z[ ,jseq], x0[k] )     #; print(zz)
        if( is.null(res) )  next

        z4  = zz
        z4[ jseq ]  = res
        
        zout[k] = splineCardinal( y, z4, y0[k] )
        }
    
    return( list( x=x0, y=y0, z=zout ) )    
    }
    


#   x	a vector containing the x coordinates of the rectangular data grid.  Must be increasing.
#   y	a vector containing the y coordinates of the rectangular data grid.  Must be increasing.
#   z	a matrix containing the z[i,j] data values for the grid points (x[i],y[j]).
#   x0	vector of x coordinates, at which to interpolate
#   y0	vector of y coordinates, at which to interpolate
#
#   the name comes from the fact that this calls stats::approx()

bilinearApprox  <- function( x, y, z, x0, y0 )
    {    
    #   check validity
    lenx    = length(x)
    if( lenx != nrow(z) )
        {
        log_level( ERROR, "length(x)=%d != %d=nrow(z)", length(x), nrow(z) )
        return(NULL)
        }

    leny    = length(y)        
    if( leny != ncol(z) )
        {
        log_level( ERROR, "length(y)=%d != %d=ncol(z)", length(y), ncol(z) )
        return(NULL)
        }
    
    n   = length(y0)
    if( length(x0) != n )
        {
        log_level( ERROR, "length(x0)=%d != %d=length(y0)", length(x0), n )
        return(NULL)
        }
        
    zout    = rep( NA_real_, n )
    
    i   = findInterval( x0, x, rightmost.closed=FALSE ) #; print(i)
    j   = findInterval( y0, y, rightmost.closed=FALSE ) 

    for( k in 1:n )
        {
        i0  = i[k]
        if( i0==0  || x[lenx] < x0[k] )    next    #   outside the rectangle
        
        j0  = j[k]
        if( j0==0  || y[leny] < y0[k] )    next    #   outside the rectangle
        
        xexact  = (x0[k] == x[i0])
        yexact  = (y0[k] == y[j0])
        
        if( xexact  &&  yexact )
            {
            #   exactly on a grid point.  If one of the neighbors is NA it does not matter
            zout[k] = z[ i0, j0 ]   #; log_level( INFO, "Exactly on grid point %g,%g", x0[k], y0[k] )
            next
            }
        
        if( xexact )
            {
            #   on a gridline, so it reduces to approx()
            zout[k] = stats::approx( y, z[i0, ], y0[k] )$y
            next
            }

        if( yexact )
            {
            #   on a gridline, so it reduces to splineCardinal()
            zout[k] = stats::approx( x, z[ ,j0], x0[k] )$y
            next
            }         

        jseq    =  j0 : min(j0+1,leny)  # ; print(jseq)

        zz  = rep( NA_real_, leny )
            
        for( jj in jseq )
            zz[jj]  = stats::approx( x, z[ ,jj], x0[k] )$y
                
        zout[k] = stats::approx( y, zz, y0[k] )$y
        }
    
    return( list( x=x0, y=y0, z=zout ) )    
    }
    

    
    
    
    
    


#   x       N-vector of points, must be increasing.  If the spline is to be C^1, then x must be regular
#   y       N-vector giving values at x.   y can also be an NxP matrix.
#           y can have NA values and it will still work, as long as the subset actually referenced are not NA
#   xout    M-vector of new points at which to evaluate
#
#   return value:
#       if y is an N-vector, then a M-vector 
#       if y is an NxP matrix, then an MxP matrix
#
#   Warning:  If x is not regular, then the spline is not C^1
#
splineCardinal <- function( x, y, xout )
    {
    C   = matrix( c(-0.5,1,-0.5,0,  1.5,-2.5,0,1,  -1.5,2,0.5,0,  0.5,-0.5,0,0), 4, 4 )
    
    #a   = 0.5      this is the only a that has the linear-accuracy property
    #C   = matrix( c(-a,2*a,-a,0,  2-a,a-3,0,1,   a-2,3-2*a,a,0,  a,-a,0,0 ), 4, 4 )
    
    forcemat    = is.null( dim(y) )
    if( forcemat ) 
        #   make y into a matrix with 1 column
        dim(y)  = c(length(y),1)
        
    n   = length(x)
    
    if( nrow(y) != n )
        {
        log_level( ERROR, "length(x) = %d  !=  %d = nrow(y).", length(x), nrow(y) )
        return(NULL)
        }
        
    if( n == 1 )    return( y ) # special case,  maybe shouldn't waste the time checking

    p   = ncol(y)
    
    m   = length(xout)
    out = array( NA_real_, dim=c(m,p) )
    
    idx = findInterval( xout, x, all.inside=TRUE )
    
    for( k in 1:m )
        {
        j   = idx[k]    #; print(j)
        
        u   = (xout[k] - x[j]) / (x[j+1] - x[j])
        u2  = u*u
        U   = c( u*u2, u2, u, 1 )       #; print(U)

        jseq    = pmin( pmax( (j-1):(j+2), 1 ), n )
        Y   = y[ jseq, ,drop=FALSE ]
        
        if( any(is.na(Y)) ) next
        
        #   fixup the ends - by linear extrapolation
        if( j == 1 )
            Y[1, ]    = 2*y[1, ] - y[2, ]     # y[1] - (y[2]-y[1])
        if( j == n-1 )
            Y[4, ]    = 2*y[n, ] - y[n-1, ]   # y[n] + (y[n] - y[n-1])
        
        #print( U %*% C ) 
        #print( Y )
        
        out[k, ]  = (U %*% C) %*% Y
        }
        
    if( forcemat )  
        #   y was originally a plain vector,
        #   so force out to be a plain vector too
        dim(out) = NULL
    
    return( out )
    }

makeSurface3D <- function( interp='bicubic', steps=32  )
    {
    x   = seq( -2, 2, by=1/steps )
    y   = x
    
    Z       = matrix(0,5,5)
    Z[3,3]  = 1
    
    zout    = matrix( NA_real_, length(x), length(y) )
    
    if( interp == 'bicubic' )
        interpfun   = bicubicCardinal
    else
        interpfun   = bilinearApprox
    
    for( i in 1:length(x) )
        {
        zout[i, ]   = interpfun( -2:2, -2:2, Z, rep(x[i],length(y)), y )$z
        }
        
    return( invisible( list( x=x, y=y, z=zout ) ) )
    }
    
    

plotEllipse <- function( asp=3 )
    {
    theta   = seq( 0, 2*pi, len=21 )
    
    x   = asp * cos( theta )
    y   = sin( theta )
    
    plot( x, y, type='n' )
    abline( h=0, v=0 )
    
    thetafine   = seq( 0, 2*pi, len=121 )
    
    xyfine  = splineCardinal( theta, cbind(x,y), thetafine )
    #   yfine   = splineCardinal( theta, y, thetafine )
    
    lines( xyfine[ ,1], xyfine[ ,2], col='red' )
    
    return(TRUE)
    }
    

#   MunsellSpec     a character vector of Munsell names
#                   or a numeric matrix with 3 columns, with HVC in each row
#                   or a single plain numeric HVC 3-vector
#   xyC             xy 2-vector for illuminant C
#                   can also be a string: 'NBS', 'JOSA', 'NTSC', or 'ASTM'
#   hcinterp        'bicubic' or 'bilinear'
#   vinterp         'cubic' or 'linear'
#   space           'xy' or 'ab'
#   YfromV          'ASTM' or 'NBS' or 'MUNSELL'. 'MGO' is not allowed here

#   return value    a data.frame

MunsellToxyY  <-  function( MunsellSpec, 
                            xyC='NBS',
                            hcinterp='bicubic',
                            vinterp='cubic',
                            YfromV='ASTM',
                            warn=TRUE )
    {
    time_start  = gettime()

    if( is.character(MunsellSpec) )
        {
        HVC         = HVCfromMunsellName(MunsellSpec)
        SAMPLE_NAME = MunsellSpec
        }
    else  
        {
        HVC = prepareNx3( MunsellSpec )
        if( is.null(HVC) )  return(NULL)

        SAMPLE_NAME = MunsellNameFromHVC( HVC, digits=3 )
        #rownames(HVC)   = MunsellNameFromHVC( HVC, digits=3 )
        
        if( is.null(colnames(HVC)) )
            colnames(HVC)   = c( "H", "V", "C" )            
        }

    #   process xyC
    xyC  = process_xyC( xyC )
    if( any(is.na(xyC)) )
        {
        log_level( ERROR, "xyC is invalid." )
        return(NULL)
        }     

    # process hcinterp
    hcinterp    = process_hcinterp(hcinterp)
    if( is.na(hcinterp) )
        {
        log_level( ERROR, "hcinterp='%s' is invalid.", hcinterp )
        return(NULL)
        }  

    #   assign function pointer
    if( hcinterp == 'bicubic' )
        hcinterpfun = bicubicCardinal   # mybicubic
    else if( hcinterp == 'bilinear' )
        hcinterpfun = bilinearApprox    # mybilinear
    
    
    # process vinterp
    vinterp     = process_vinterp(vinterp)
    if( is.na(vinterp) )
        {
        log_level( ERROR, "vinterp='%s' is invalid.", vinterp )
        return(NULL)
        }  

    # process YfromV
    full    = c( 'ASTM', 'NBS', 'MUNSELL', 'PRIEST' )
    idx     = pmatch( toupper(YfromV), full )
    if( is.na(idx) )
        {
        log_level( ERROR, "YfromV='%s' is invalid. ('MgO' is not allowed in this function).", YfromV )
        return(NULL)
        }  
    
    #   ready to process data
    
    #   test any element of the lookup tables to see whether illuminant C has changed
    same    =  xyC[1]==p.LookupList[[1]]$x[1,1]  &&  xyC[2]==p.LookupList[[1]]$y[1,1]
    if( is.na(same)  ||  ! same )
        {
        base::unlockBinding( "p.LookupList", asNamespace('munsellinterpol') )
    
        # fill all the lookup tables with x and y for illuminant C  
        log_level( TRACE, "filling p.LookupList with xy=%g,%g for illuminant C.", xyC[1], xyC[2] )
        for( k in 1:length(p.LookupList) )
            {
            p.LookupList[[k]]$x[ ,1] <<- xyC[1]
            p.LookupList[[k]]$y[ ,1] <<- xyC[2]
            }
        }

    n   = nrow(HVC)
    
    xyY = matrix( NA_real_, n, 3 )
    colnames( xyY ) = c( 'x', 'y', 'Y' )
    
    #   do all Y coordinates in only 1 call
    xyY[ ,3]    = YfromV( HVC[ ,2], which=YfromV  )  #*  (whiteC[3]/100)
                
    
    V.vector    = attr( p.LookupList, "V.vector" )
    H.vector    = attr( p.LookupList, "H.vector" )

    for( i in 1:n )
        {
        chroma  = HVC[i,3]
        if( ! is.finite(chroma) || chroma<0 )
            {
            #   log_level( ERROR, "Failed to map sample %d, of %d. chroma=%g.", i, n, chroma )
            next
            }
            
        if( chroma == 0 )
            {
            #   neutrals are a special case
            xyY[i,1:2] = xyC[1:2]                
            next
            }
                
        hue     = HVC[i,1]
        value   = HVC[i,2]
        
        if( value == 0 )
            {
            #   we must have 0<chroma, and this is physically unrealizable
            next
            }
        
        iV = match( value, V.vector )
        if( ! is.na(iV) )
            {
            C.vector    = attr( p.LookupList[[iV]], "C.vector" )
    
            #   special case, the input point is exactly on one of the constant Value planes, so only 1 HC interpolation required.
            #   But this will probably happen a lot
            xyY[i,1]    = hcinterpfun( H.vector, C.vector, p.LookupList[[iV]]$x, hue, chroma )$z
            xyY[i,2]    = hcinterpfun( H.vector, C.vector, p.LookupList[[iV]]$y, hue, chroma )$z
            
            #if( any( is.na(xyY[i, ]) ) )
            #    log_level( ERROR, "Failed to map sample %d, of %d. Exactly on V plane.", i, n )
                
            next
            }
            
        vinterp.i   = vinterp            
        
        if( p.vinterpOverride  &&  value < 2 )
            vinterp.i   = 'linear'
            
        #   value is *between* V-planes, so a lot more work is needed
        iV  = findInterval( value, V.vector, all.inside=TRUE )
        if( vinterp.i == 'linear' )
            vlevel = iV : (iV+1)
        else if( vinterp.i == 'cubic' )
            vlevel = max(iV-1,1) : min(iV+2, length(V.vector))      

        xymat   = matrix( NA, length(vlevel), 2 )
        for( k in 1:length(vlevel) )
            {
            iV  = vlevel[k]
            
            C.vector    = attr( p.LookupList[[iV]], "C.vector" )

            xymat[k,1]    = hcinterpfun( H.vector, C.vector, p.LookupList[[iV]]$x, hue, chroma )$z
            xymat[k,2]    = hcinterpfun( H.vector, C.vector, p.LookupList[[iV]]$y, hue, chroma )$z
            }
            
        vvec    = V.vector[vlevel]                  
    
        if( any( is.na(xymat) ) )
            {
            #log_level( ERROR, "Failed to map sample %d, of %d. Between V planes.", i, n )     
            #rownames( xymat ) = as.character(vvec)          #; colnames(xymat) = c('x','y') ;       print( xymat )            
            #print( xymat )            
            next
            }
            
        #   now interpolate along each column of xymat
        if( vinterp.i == 'linear' )
            {
            xyY[i,1]    = stats::approx( vvec, xymat[ ,1], xout=value )$y
            xyY[i,2]    = stats::approx( vvec, xymat[ ,2], xout=value )$y
            }
        else if( vinterp.i == 'cubic' )
            {
            #xyY[i,1]    = spline( vvec, xymat[ ,1], xout=value, method='natural' )$y
            #xyY[i,2]    = spline( vvec, xymat[ ,2], xout=value, method='natural' )$y

            xyY[i,1:2]    = splineCardinal( vvec, xymat, xout=value )
            }
        }
    

    out = data.frame( row.names=1:n, stringsAsFactors=FALSE )
    
    out$SAMPLE_NAME = SAMPLE_NAME
    out$HVC         = HVC
    
    rownames(xyY)   = rownames(HVC)
    if( is.null(rownames(xyY) ) )   rownames(xyY)   = SAMPLE_NAME
    out$xyY         = xyY
    
    if( FALSE  &&   100 <= n )
        {
        time_elapsed = gettime() - time_start
        log_level( INFO, "Processed %d samples in %g seconds.  %g/sample.",
                            n, time_elapsed, time_elapsed/n )               
        }
    

    if( warn )
        {
        #   check for mapping failure
        mask    = is.na(xyY[ ,1])  |  is.na(xyY[ ,2])         
        
        if( any(mask,na.rm=TRUE) )
            {
            log_level( WARN, "%d samples, out of %d, could not be mapped; xy set to NA.", sum(mask), n )
            }
            
        #   check for xy outside the triangle
        mask    = xyY[ ,1]<0  |  xyY[ ,2]<0  |  1<xyY[ ,1]+xyY[ ,2]
        mask[ is.na(mask) ] = FALSE
        
        if( any(mask) )
            {
            res = log_level( WARN, "%d samples, out of %d, were mapped to xy outside the triangle.", sum(mask), n )
            
            if( ! is.null(res)  &&  sum(mask) <= 10 )
                {
                #  since res is not NULL, a logging event actually occurred, and the number of problems is small
                print( out[mask, ] )
                }
            }            
        }
    
    #   print( str(out) )
    


    return(out)
    }
    
#   xyC     single string or numeric 2-vector
#
#   returns valid numeric 2-vector, or NA in case that xyC is invalid

process_xyC <- function( xyC )
    {
    if( is.character(xyC) )
        {
        idx = pmatch( toupper(xyC), rownames(p.xyC) )
        
        if( is.na(idx) )
            return(NA)

        xyC   = p.xyC[ idx, ]
        }
        
    ok  = is.numeric(xyC)  &&  length(xyC) == 2
    if( ! ok )
        return(NA)
  
        
    #   if( length(whiteC) == 2 )    whiteC = c( whiteC, 100 )
    
    return( xyC )
    }
    
    
    
process_hcinterp <- function( hcinterp )
    {    
    full    = c('bicubic','bilinear')
    
    idx = pmatch( tolower(hcinterp), full )
    if( is.na(idx) )
        return(NA)

    return( full[idx] )
    }

process_vinterp <- function( vinterp )
    {        
    full   = c('cubic','linear')
    idx = pmatch( tolower(vinterp), full )
    if( is.na(idx) )
        return(NA)
        
    return( full[idx] )
    }

        

#   p.LookupList    = NULL

makeLookupList <- function( Munsell2xy, xyC=c(0.3101,0.3163), kfactor=c(0.7,0.5) )
    {   
    time_start  = gettime()
    
    Hue.vector      = seq(2.5,100,by=2.5)   #  H is always a multiple of 2.5. later this vector will change
            
    Value.vector    = sort( unique(Munsell2xy$V) )
    Value.vector    = c( 0, Value.vector )  # put 0 at beginning, so we can interpolate between 0 and 0.2
    #   Value.vector    = c(0,0.2,0.4,0.6,0.8,1,2,3,4,5,6,7,8,9,10)
    
    #   put 0 at beginning of chroma.vector,
    #   because we will put the neutral x,y there, for Illuminant C
    Chroma.vector   = seq(0,50,by=2) 
    
    #   allocate 4D array, full of NAs
    Lookup4D = array( NA_real_, dim=c(length(Hue.vector),length(Value.vector),length(Chroma.vector),2) )
    
    dimnames(Lookup4D)[[1]] = as.character( Hue.vector )    
    dimnames(Lookup4D)[[2]] = as.character( Value.vector )
    dimnames(Lookup4D)[[3]] = as.character( Chroma.vector )
    dimnames(Lookup4D)[[4]] = c('x','y')
    
    #   filter out very dark samples that are not real
    #mask    = Munsell2xy$real  |  (1 <= Munsell2xy$V )
    #dfsub   = Munsell2xy[ mask, ]
    
    #   copy Munsell2xy to Lookup4D[].  Some of these may be overwritten later   
    for( i in 1:nrow(Munsell2xy) )
        {
        row = Munsell2xy[i, ]
        
        iH  = match( row$H, Hue.vector )
        iV  = match( row$V, Value.vector )
        iC  = match( row$C, Chroma.vector )        
        
        Lookup4D[ iH, iV, iC,   ] = c( row$x, row$y )
        }    


    
    Lookup4D = extrapGAM( Lookup4D, Munsell2xy, kfactor=kfactor  )

    
    #print( str(Lookup4D) ) ;  print( object.size(Lookup4D) )
    #   return(FALSE)
    
    #   count the total number of non-NA points in Lookup4D
    gridcount   = integer( length(Value.vector) )
    names(gridcount)    = as.character(Value.vector)
    
    for( iV in 2:length(Value.vector) )
        gridcount[iV]   = sum( is.finite(Lookup4D[ ,iV, ,'x',drop=FALSE]) )
    
    gridcount   = c( gridcount, sum(gridcount) )
    names(gridcount)[ length(gridcount) ]   = 'sum'
    cat( "Total gridcount:\n" )
    print( gridcount )

    
    
    #   now copy from array Lookup4D[] to a list with an item for each Munsell Value
        
    out = vector( length(Value.vector), mode='list' )
    
    names(out) = as.character(Value.vector)

     #   there are are only 40 hues, but we make 43 for easier wraparound (periodicity)
    Hue.vector      = seq(-2.5,102.5,by=2.5) 
        
    attr( out, "V.vector" ) = Value.vector
    attr( out, "H.vector" ) = Hue.vector
    attr( out, "C.vector" ) = Chroma.vector
    
    #   initialize with arrays of NAs
   
    dummy   = array( NA_real_, dim=c( length(Hue.vector), length(Chroma.vector)  ) )
    rownames(dummy) = HueStringFromNumber( Hue.vector )
    colnames(dummy) = as.character( Chroma.vector )
    
    for( iV in 1:length(out) )
        {
        out[[iV]] = list()  # list( x=dummy, y=dummy )    
        
        dummy[ 3:42, ]  = Lookup4D[ ,iV, , 'x' ]
        out[[iV]]$x     = dummy        
        
        dummy[ 3:42, ]  = Lookup4D[ ,iV, , 'y' ]
        out[[iV]]$y     = dummy        
        }
    
        
    #   copy data from V=0.2 to V=0
    out[[1]]$x = out[[2]]$x
    out[[1]]$y = out[[2]]$y
    
    #   trim away high Chroma, to reduce memory
    for( iV in 1:length(out) )
        {
        x   = out[[iV]]$x  #; print( str(x) )
        
        cmax    = NA        
        for( iC in 2:ncol(x) )
            {
            if( all( is.na(x[ ,iC]) ) )
                {
                cmax = iC - 1
                break
                }
            }
            
        if( is.na(cmax) )
            {
            log_level( INFO, "At V=%g, there is data up to the maximum C=%g.",
                            Value.vector[iV], Chroma.vector[length(Chroma.vector)] )
            attr( out[[iV]], "C.vector" ) = Chroma.vector            
            }
        else
            {
            log_level( INFO, "At V=%g, there is data up to C=%g.",
                                Value.vector[iV], Chroma.vector[cmax] )
                
            out[[iV]]$x = out[[iV]]$x[ , 1:cmax ]
            out[[iV]]$y = out[[iV]]$y[ , 1:cmax ]
            
            attr( out[[iV]], "C.vector" ) = Chroma.vector[ 1:cmax ]            
            }
            
            
        #   duplicate data for Hue wraparound
        out[[iV]]$x[1, ] = out[[iV]]$x[41, ]
        out[[iV]]$y[1, ] = out[[iV]]$y[41, ]
            
        out[[iV]]$x[2, ] = out[[iV]]$x[42, ]
        out[[iV]]$y[2, ] = out[[iV]]$y[42, ]    

        out[[iV]]$x[43, ] = out[[iV]]$x[3, ]
        out[[iV]]$y[43, ] = out[[iV]]$y[3, ]    
        
        #   replace NAs by 0s
        #out[[iV]]$x[ is.na(out[[iV]]$x) ] = 0
        #out[[iV]]$y[ is.na(out[[iV]]$y) ] = 0      
        }
        
        
    # fill Chroma=0 with xyC
    for( k in 1:length(out) )
        {
        out[[k]]$x[ ,1] = xyC[1]
        out[[k]]$y[ ,1] = xyC[2]
        }        
    
    time_elapsed  = gettime()  -  time_start
    
    log_level( INFO, "Made list in %g seconds.", time_elapsed )
    
    return( invisible(out) )
    }

    
extrapGAM  <-  function( Lookup4D, Munsell2xy, kfactor=c(0.9,0.5)  )
    {
    if( ! requireNamespace( 'mgcv', quietly=TRUE ) )        return( NULL )
    
    if( ! requireNamespace( 'spacesXYZ', quietly=TRUE ) )   return(NULL)
        
    
    time_start  = gettime()
    
    Hue.vector      = as.numeric( dimnames(Lookup4D)[[1]] )
    Value.vector    = as.numeric( dimnames(Lookup4D)[[2]] )    
    Chroma.vector   = as.numeric( dimnames(Lookup4D)[[3]] )
    
    
    out = Lookup4D
    
    Mask3D  = makeRequiredMask3D( Munsell2xy )    
    
    samples   = 0
    
    log_level( INFO, "Processing colors with V < 1..."  )
    
    #   for value < 1 make a special data.frame, and make only 1 model from it
    #   use a subset of points for V<1, plus all points for V==1
    mask    = Munsell2xy$V < 1  
    mask    = mask &  Munsell2xy$real  #  |  (75 <= Munsell2xy$H  &  Munsell2xy$H <= 100) )
    mask    = mask | (Munsell2xy$V == 1)

    data.dark   = Munsell2xy[ mask, ]    

    #   add rectangular coordinates A,B
    tmp = ABfromHC( data.dark$H, data.dark$C )
    data.dark   = cbind( data.dark, A = tmp$A, B = tmp$B )
    
    #   add ab coordinates
    
    white.C = spacesXYZ::XYZfromxyY( c( p.xyC['NBS',],100) )
    Y       = YfromV( data.dark$V )
    XYZ     = spacesXYZ::XYZfromxyY( cbind( data.dark$x, data.dark$y, Y ) )
    Lab     = spacesXYZ::LabfromXYZ( XYZ, white.C )
    
    data.dark$a = Lab[ ,2]
    data.dark$b = Lab[ ,3]
    
    #Luv     = xyz2luv( XYZ, white.C )
    #data.dark$u = Luv[ ,2]
    #data.dark$v = Luv[ ,3]
    
    
    #   print( str(data.dark) )
    if( TRUE )
    {
    k   = round( kfactor[1] * nrow(data.dark) )
    
    form    = sprintf( "x ~ s( A, B, V,  bs='tp', k=%d )", k )
    log_level( INFO, "Computing model '%s'...", form ) ; flush.console();
    gam.x = mgcv::gam( formula(form), data=data.dark )  

    form    = sprintf( "y ~ s( A, B, V,  bs='tp', k=%d )", k )
    log_level( INFO, "Computing model '%s'...", form ) ; flush.console();    
    gam.y = mgcv::gam( formula(form), data=data.dark )    
    
    log_level( INFO, "Very Dark"  )
    log_level( INFO, "x residual range = [%g,%g]", range(residuals(gam.x))[1],  range(residuals(gam.x))[2] )
    log_level( INFO, "y residual range = [%g,%g]", range(residuals(gam.y))[1],  range(residuals(gam.y))[2] )
    }
    
    
    
    for( value in c(0.2,0.4,0.6,0.8) )
        {
        iV  = match( value, Value.vector )
        
        if( FALSE )
        {
        dfv = Munsell2xy[ Munsell2xy$V==value, ]

        #   make mask of the ones to keep for modeling
        mask.model  =   dfv$real    #   |  (75 <= dfv$H  &  dfv$H <= 100)
        
        dfmodel     = dfv[ mask.model, ]
        
        #   add rectangular coordinates A,B
        tmp = ABfromHC( dfmodel$H, dfmodel$C )
        dfmodel  = cbind( dfmodel, A=tmp$A, B=tmp$B )
        
        k   = round( kfactor[1] * nrow(dfmodel) )
        
        form    = sprintf( "x ~ s( H, C,  bs='tp', k=%d )", k )
        log_level( INFO, "value=%g.  Computing model '%s'...", value, form ) ; flush.console();
        gam.x = mgcv::gam( formula(form), data=dfmodel )  

        form    = sprintf( "y ~ s( H, C,  bs='tp', k=%d )", k )
        log_level( INFO, "value=%g.  Computing model '%s'...", value, form ) ; flush.console();    
        gam.y = mgcv::gam( formula(form), data=dfmodel )    

        log_level( INFO, "x residual range = [%g,%g]", range(residuals(gam.x))[1],  range(residuals(gam.x))[2] )
        log_level( INFO, "y residual range = [%g,%g]", range(residuals(gam.y))[1],  range(residuals(gam.y))[2] )
        }
        
        
        #log_level( INFO, "For value=%g, found %d real samples.", value, nrow(dfreal) )        
        
        #   make 2D mask of the ones used for modeling
        mask2D  = matrix( FALSE, length(Hue.vector), length(Chroma.vector) )
        
        for( i in 1:nrow(data.dark) )
            {
            if( data.dark$V[i] == value )
                {
                iH  = match( data.dark$H[i], Hue.vector )
                iC  = match( data.dark$C[i], Chroma.vector )
                mask2D[iH,iC] = TRUE
                }
            }
        
        #   make a mask of iH,iC points to be extrapolated.  Preserve the ones used for modeling.
        #   preserve the entries used
        mask_extrap = Mask3D[ , iV, ]  &  (! mask2D)    # mask_extrap is 2D
        
        for( iC in 2:length(Chroma.vector) )    # start at 2 to ensure that neutral is NOT extrapolated
            {
            mask    = mask_extrap[ , iC]            # mask is 1D horizontal slice of constant C
            
            if( ! any(mask) )
                {
                #   log_level( INFO, "For value=%g, and iC=%d, no extrapolation needed.", value, iC )        
                next    # no extrapolation needed here
                }
                
            newdata = data.frame( H=Hue.vector[mask], V=value, C=Chroma.vector[iC] )
            
            tmp = ABfromHC( newdata$H, newdata$C )            
            newdata = cbind( newdata, A = tmp$A, B = tmp$B )
            
            #L   = Lightness_from_linear( YfromV( newdata$V ) / 100 )
            x   = predict( gam.x, newdata=newdata )
            y   = predict( gam.y, newdata=newdata )
            
            #Luv = cbind( L, u, v )
            #xyY = XYZ2xyY( luv2xyz( Luv, white.C ) )
            xyY = cbind( x, y, YfromV( newdata$V ) / 100 )
            
            mask.na = is.na( x )
            if( any(mask.na) )
                {
                log_level( WARN, "For value=%g and chroma=%g,  %d of the extrapolated xy's are NA.",
                                        value,  Chroma.vector[iC], sum(mask.na) )
                df  = newdata[mask.na, , drop=FALSE]
                #   df$Luv  = Luv[mask.na, , drop=FALSE]
                df$xyY  = xyY[mask.na, , drop=FALSE]
                print(df)
                }
            
            
            out[mask,iV,iC,'x']  =  x   # predict( gam.x, newdata=newdata )
            out[mask,iV,iC,'y']  =  y   # predict( gam.b, newdata=newdata )
            
            samples = samples + sum(mask)
            }   
        }
    
    log_level( INFO, "Extrapolated %d very dark samples in %g seconds.", samples, gettime() - time_start )  

    #   for V >= 1 do one value at a time
    log_level( INFO, "Processing colors with V >= 1..." )
    
    for( value in 1:10 )
        {
        iV  = match( value, Value.vector )

        #   for value >=1 Munsell2xy has good data
        dfsub   = Munsell2xy[ Munsell2xy$V==value,  ]  #; print(value) ; print( str(dfsub) )
        
        #   add rectangular coordinates A,B
        tmp = ABfromHC( dfsub$H, dfsub$C )                 
        dfsub   = cbind( dfsub, A = tmp$A, B = tmp$B ) 
            
        k   = round( kfactor[2] * nrow(dfsub) )  # - 10
        
        mess    = sprintf( "x ~ s( A, B, bs='tp', k=%d )", k )        
        gam.x   = mgcv::gam( formula(mess), data=dfsub )
        mess    = sprintf( "y ~ s( A, B, bs='tp', k=%d )", k )          
        gam.y   = mgcv::gam( formula(mess), data=dfsub )
        
        ran.x   = range(residuals(gam.x))
        ran.y   = range(residuals(gam.y))
        log_level( INFO, "value=%g.  x residual range = [%g,%g]", value, ran.x[1], ran.x[2] ) 
        log_level( INFO, "value=%g.  y residual range = [%g,%g]", value, ran.y[1], ran.y[2] ) 
        flush.console(); 
        
        #   do not extrapolate if a value is already present
        mask_extrap = Mask3D[ , iV, ]  &  is.na( Lookup4D[ ,iV, ,'x'] )    # mask_extrap is 2D
                      
        for( iC in 2:length(Chroma.vector) )
            {
            chroma  = Chroma.vector[iC]

            mask    = mask_extrap[ ,iC]         # mask is 1D
            
            if( ! any(mask) )   next    # no interpolation needed here
            
            newdata = data.frame( H=Hue.vector[mask], C=chroma )    #V=value,
            
            #   add rectangular coordinates A,B
            tmp = ABfromHC( newdata$H, newdata$C )              
            newdata   = cbind( newdata, A = tmp$A, B = tmp$B )       
                           
            out[mask,iV,iC,'x']  =  predict( gam.x, newdata=newdata )
            out[mask,iV,iC,'y']  =  predict( gam.y, newdata=newdata )
            
            samples = samples + sum(mask)
            }        
        }

    log_level( INFO, "Extrapolated %d samples in %g seconds.", samples, gettime() - time_start )
    
    return( out )
    }
    
    

        
#   returns a logical 3D array
#   for given HVC, the entry is TRUE if an xy value is required for the HVC

makeRequiredMask3D <- function( Munsell2xy )
    {   
    time_start  = gettime()
    
    Hue.vector      = seq(2.5,100,by=2.5)  
            
    Value.vector    = sort( unique(Munsell2xy$V) )
    Value.vector    = c( 0, Value.vector )  # put 0 at beginning, so we can interpolate between 0 and 0.2
    #   Value.vector    = c(0,0.2,0.4,0.6,0.8,1,2,3,4,5,6,7,8,9,10)
    
    #   put 0 at beginning of chroma.vector,
    #   because we will put the neutral x,y there, for Illuminant C
    Chroma.vector   = seq(0,50,by=2) 
    
    nH  = length(Hue.vector)
    nV  = length(Value.vector)
    nC  = length(Chroma.vector)
    
    #   allocate 3D array, and copy Munsell2xy to it
    Mask3D  = array( FALSE, dim=c( nH, nV, nC ) )
    
    dimnames(Mask3D)[[1]] = as.character( Hue.vector )    
    dimnames(Mask3D)[[2]] = as.character( Value.vector )
    dimnames(Mask3D)[[3]] = as.character( Chroma.vector )        
    
    
    if( TRUE )
        #   use the entire all.dat
        dfsub   = Munsell2xy
    else
        #   only the real ones
        dfsub   = Munsell2xy[ Munsell2xy$real,  ]   # ; cat( "nrow(dfsub) = ", nrow(dfsub), '\n' )
        
    
    for( i in 1:nrow(dfsub) )
        {
        row = dfsub[i, ]
        
        iH  = match( row$H, Hue.vector )    # H is always a multiple of 2.5
        iV  = match( row$V, Value.vector )
        iC  = match( row$C, Chroma.vector )  
        
        Mask3D[ iH, iV, iC  ] = TRUE
        }    

        
    #   now dilate with mask
    out = Mask3D
    
    huedelta    = 1     # 2
    chromadelta = 1     # 2
    valuedelta  = 2     # 1 fails when hcinterp='bicub',  vinterp='lin'
    for( iV in 2:nV  )
        {
        mask00  = Mask3D[ , iV, ]
        
        # rotate and translate mask00
        #   hue
        for( h in -huedelta:huedelta )    
            {
            hrot    =  (((1:nH) + h-1) %% nH)  +  1
            maskrot = mask00[ hrot, , drop=FALSE ]
            
            # chroma
            for( j in 0:chromadelta ) 
                {
                cshift  = pmax( (1-j) : (nC-j), 1 )
                mask    = maskrot[  , cshift, drop=FALSE ]
                
                #   copy to valuedelta levels above and below
                vlevel  = max( iV-valuedelta, 2 )  :  min( iV+valuedelta,nV )
                for( v in vlevel )
                    {
                    out[  , v,  ] = mask  |  out[  , v,  ]
                    }
                }
            }
        }

    log_level( INFO, "Made extrapolation mask in %g seconds.",
                            gettime() - time_start )
                            
    count   = integer( nV )                     
    for( iV in 1:nV )
        count[iV]   = sum( out[ ,iV, ] )
        
    count           = c( count, sum(count) )
    names(count)    = c(as.character(Value.vector),'Total')
    print( count )
    
    return( invisible(out) )
    }
    
    

library( munsellinterpol )
library( microbenchmark )
library( spacesXYZ )

printf <- function( msg, ... )
    {    
    mess = sprintf( msg[1], ... )    # should this really be msg[1] ?
    cat( mess, '\n' )   #, file=stderr() )
    }
    
gettime <- function()
    {
    return( microbenchmark::get_nanotime() * 1.e-9 )
    }    
    
ABfromHC <- function( H, C )
    {
    theta   = H * pi/50
    list( A = C*cos(theta), B = C*sin(theta) )
    }
    
makeABab <- function( Munsell2xy, value )
    {
    dfsub  = Munsell2xy[ Munsell2xy$V == value , ]    # &  Munsell2xy$real
    
    tmp = ABfromHC( dfsub$H, dfsub$C )
    dfsub$A    = tmp$A
    dfsub$B    = tmp$B
    
    XYZ.C   = spacesXYZ::XYZfromxyY( c( 0.3101,0.3163, 100 ) )   
    
    xyY = cbind( dfsub$x, dfsub$y, YfromV(value) )
    XYZ = spacesXYZ::XYZfromxyY( xyY )
    Lab = spacesXYZ::LabfromXYZ( XYZ, XYZ.C )    
    
    dfsub$a    = Lab[ ,2]
    dfsub$b    = Lab[ ,3]
    
    return( dfsub )
    }
        
    
testPoly <- function()
    {
    df5 = makeABab( munsellinterpol::Munsell2xy, 5 ) 
    
    mod.poly    = lm( A ~ polym(a,b,degree=3,raw=TRUE) + 0, data=df5 )  #   but polym prediction does not work for singletons
    
    form    = "A ~ a + I(a^2) + I(a^3) + b + I(a*b) + I(a^2*b) + I(b^2) + I(a*b^2) + I(b^3) + 0"
    
    mod.terms   = lm( formula(form), data=df5 )    
    
    #   coefficients should be identical
    ok  = identical( as.numeric(coef(mod.terms)) , as.numeric(coef(mod.poly)) )
    if( ! ok )
        {
        printf( "direct and polynomial coefficients are not identical." )
        return(FALSE)
        }
    
    
    coef    = coef(mod.terms) #;  print( coef )

    count   = nrow(df5)
    pred1   = numeric( count )
    
    time_start  = gettime()
    
    # predict A for each row, one at a time
    for( i in 1:count )
        pred1[i] = predict( mod.terms, newdata=list( a=df5$a[i], b=df5$b[i] ) )
    
    time_elapsed    = gettime() - time_start
    
    printf( "Predicted %d points using predict() (one at a time) in %g seconds.  %g/point.", count, time_elapsed, time_elapsed/count )
    
    
    #   do it again by direct calc, which is much faster
    pred2   = numeric( count )    
    time_start  = gettime()
    for( i in 1:count )
        {
        a   = df5$a[i]
        b   = df5$b[i]
        a2  = a*a
        b2  = b*b
        term    = c( a, a2, a2*a, b, a*b, a2*b, b2, a*b2, b2*b )
        pred2[i] = sum( coef * term )
        }
    
    time_elapsed    = gettime() - time_start
    
    printf( "Predicted %d points using sum(coef*term) (one at a time) in %g seconds.  %g/point.", count, time_elapsed, time_elapsed/count )
        
    #   now compare pred1[] to pred2[]
    ran = range( pred1 - pred2 )
    printf( "Difference range = [%g,%g].", ran[1], ran[2] )
    
    tol = 5.e-14
    
    return(  all( abs(ran) < tol )  )
    }    
    
    
    
    
{
if( testPoly() )    
    printf( "testPoly() succeeded !"  )
else
    stop( "testPoly() failed !", call.=FALSE )
}

printf( "Passed all polynomial tests !"  )
    

library( munsellinterpol )

options( width=144 )


#   return value:  TRUE or FALSE
#
testRounding <- function( count=100 )
    {
    set.seed(0)
    
    #   make hues that are in the book
    countbad    = 0
    
    #   make vector of Hues that actually appear in the soil book
    hueseq  = c( seq(5,25,by=2.5), seq(30,75,by=5) )
    
    for( hue in hueseq )
        {
        mess = sprintf( "-------------------  Hue %s  ---------------------\n", HueStringFromNumber(hue) )
        cat( mess )
        
        HVC = cbind( hue, runif(count,min=2,max=8), runif(count,min=1,max=8) )
        
        dat = roundHVC( HVC, book='soil' )
        
        #   hopefully all the rounded hues are hue
        HVCrnd  = HVCfromMunsellName( dat$MunsellRnd )
        maskbad = HVCrnd[ ,1] != hue
        
        if( any(maskbad) )
            {
            print( dat[maskbad, ] )
            countbad    = countbad + sum(maskbad)
            
            # return(FALSE)
            }
        }
        
    #   calculate fraction of bad ones
    fractionbad = countbad / ( count*length(hueseq) )
        
        
    cat( "bad count: ", countbad, " of ",  count*length(hueseq),  "    Fraction bad = ", fractionbad, '\n' )
    
    return( fractionbad < 0.02 )
    }
    
    
if( ! testRounding() )    stop( "testRounding() failed !\n" )


cat(  "Passed rounding test !\n" )

makeDataForPrediction <- function( Munsell2xy, value, p.LookupList )
    {
    p   = 'spacesXYZ'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }
        
    dfsub  = Munsell2xy[ Munsell2xy$V == value , ]    # &  Munsell2xy$real
    
    if( nrow(dfsub) == 0 )  return(NULL)
    
    #   add the illuminant as an aimpoint, which will also affect the model a little bit
    xyC = p.xyC[ 'NBS', ]
    dfsub   = rbind( data.frame(H=0,V=value,C=0,x=xyC[1],y=xyC[2],real=TRUE), dfsub  )
        
    #   add a problem xy point, for debugging
    # dfsub   = rbind( dfsub, data.frame(H=18.75,V=1.488,C=11.69,  x=0.71812,y=0.28188,real=FALSE) )
        
    V.vector    = attr( p.LookupList, "V.vector" )
    H.vector    = attr( p.LookupList, "H.vector" )        
    
    if( value < 1 )
        {
        #   replace xy at non-reals by p.LookupList xy instead.
        #   in some cases they might be the same.        
        iV  = match( value, V.vector )
        
        C.vector    = attr( p.LookupList[[iV]], "C.vector" )
        
        for( j in 1:nrow(dfsub) )
            {
            if( dfsub$real[j] )   next

            iH  = match( dfsub$H[j], H.vector )
            iC  = match( dfsub$C[j], C.vector )
            dfsub$x[j]   = p.LookupList[[iV]]$x[ iH, iC ]
            dfsub$y[j]   = p.LookupList[[iV]]$y[ iH, iC ]                    
            }

        mask    = is.na(dfsub$x) 
        if( FALSE  &&  any(mask) )
            {
            mess    = sprintf( "makeDataForPrediction() WARN.  for value=%g, %d of the rows have x==NA.",
                                        value, sum(mask) )
            cat( mess,'\n', file=stderr() )
            print( dfsub[ mask, ] )
            }
        }    
        
        
        
    #   add A,B
    tmp = ABfromHC( dfsub$H, dfsub$C )
    dfsub$A    = tmp$A
    dfsub$B    = tmp$B
           
    #   add Y    
    dfsub$Y = YfromV(dfsub$V) 
        
    #   add a,b
    XYZ.C   = spacesXYZ::XYZfromxyY( c( xyC, 100 ) ) 

    xyY = cbind( dfsub$x, dfsub$y, dfsub$Y )
    XYZ = spacesXYZ::XYZfromxyY( xyY )
    Lab = spacesXYZ::LabfromXYZ( XYZ, XYZ.C )    
    
    #   dfsub$L    = Lab[ ,1]
    dfsub$a    = Lab[ ,2]
    dfsub$b    = Lab[ ,3]
    
    #   add offset from xyC
    dfsub$xdelta    = dfsub$x - xyC[1]
    dfsub$ydelta    = dfsub$y - xyC[2]
    
    return( dfsub )
    }
    
#   
#   data    as returned from makeDataForPrediction()
#    
#   return value
#       a data.frame with predicted (estimated) columns for AB, HC, xy, and ab.  And these attributes
#       "coeffs"            2x9 matrix of coefficients
addPredictions <- function( data, warn=TRUE )
    {
    p   = 'spacesXYZ'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }

    #   get value from first row
    value   = data$V[1]
    
    out = data    
    
    if( 1 <= value )
        {
        #   use a,b to predict A,B
        modA = lm( A ~ polym(a,b,degree=3,raw=TRUE) + 0, data=data )    
        modB = lm( B ~ polym(a,b,degree=3,raw=TRUE) + 0, data=data )           
        #modA = lm( A ~ a + b + I(a*b) + I(a^2) + I(b^2) + I(a^3) + I(a^2*b) + I(a*b^2) + I(b^3) + 0, data=data )    
        #modB = lm( B ~ a + b + I(a*b) + I(a^2) + I(b^2) + I(a^3) + I(a^2*b) + I(a*b^2) + I(b^3) + 0, data=data )    
        #print( coef(modA) )
        #print( coef(modB) )     
        }
    else
        {
        #   use xdelta,ydelta to predict A,B  
        modA = lm( A ~ polym(xdelta,ydelta,degree=3,raw=TRUE) + 0, data=data )    
        modB = lm( B ~ polym(xdelta,ydelta,degree=3,raw=TRUE) + 0, data=data )   
        }
        
    coeffs  = rbind( stats::coef(modA), stats::coef(modB) )        
        
    colnames(coeffs)    = gsub( ' ', '', names( stats::coef(modA) ) )      # in this case, space are an annoyance
    rownames(coeffs)    = c('A','B')
    #   print( coeffs )
    attr( out, "coeffs" )   = coeffs

    
    
    
    #   add predicted AB
    out$A.pred  = predict( modA, newdata=data )
    out$B.pred  = predict( modB, newdata=data )  
    
    if( warn  &&  any( is.na(out$A.pred) ) )
        log_level( WARN, "%d of A.pred are NA", sum(is.na(out$A.pred)) )

    #   add predicted HC
    tmp = HCfromAB( out$A.pred, out$B.pred )
    out$H.pred  = tmp$H
    out$C.pred  = tmp$C
    
    #   add predicted xy
    HVC = cbind( out$H.pred, value, out$C.pred )
    
    xyY = MunsellToxyY( HVC, warn=FALSE )$xyY
    out$x.pred  = xyY[ ,1]
    out$y.pred  = xyY[ ,2]
    
    if( warn  &&  any( is.na(out$x.pred) ) )
        log_level( WARN, "%d of x.pred are NA", sum(is.na(out$x.pred)) )
    
    #   add predicted ab
    xyC     = p.xyC['NBS', ]    
    XYZ.C   = spacesXYZ::XYZfromxyY( c( xyC, 100 ) )

    xyY = cbind( out$x.pred, out$y.pred, data$Y )
    XYZ = spacesXYZ::XYZfromxyY( xyY )
    Lab = spacesXYZ::LabfromXYZ( XYZ, XYZ.C )    

    out$a.pred = Lab[ ,2]
    out$b.pred = Lab[ ,3]
    
    mask    = is.na(out$a.pred) 
    if( warn  &&  any(mask) )
        {
        log_level( WARN, "%d of a.pred are NA", sum(mask) )
        print( out[mask, ] )
        }
        
    #   print( str(out) )        
        
    return(out)
    }
    
    

    
#   for each of the 15 values, make models for A and B, as polynomials in a and b    
#   return value:
#       list of 15 2x9 matrices.  [[iV]]['A' or 'B'][ index coefficient ]     # 3D array  15x2x9  [iV]['A' or 'B'][ index coefficient ]   
#
makeInversionCoeffs  <-  function( Munsell2xy, p.LookupList, warn=TRUE )
    {
    time_start  = gettime()
    
    V.vector    = sort( unique(Munsell2xy$V) )
    V.vector    = c( 0, V.vector )    
    
    #   triangular  = 4*(4+1)/2     # 10 terms in *full* polynomial, including intercept. like bowling pins.
    
    out = vector( length(V.vector), mode='list' )       #array( NA_real_, dim=c( length(V.vector), 2, triangular-1 ) )
    names(out)  = as.character(V.vector)
    
    #   AB  = c('A','B')
    #   dimnames(out)  = list( as.character(V.vector), AB, NULL )
    
    #   in this loop, start at 2 to skip Value = 0, for which we have no data
    for( iV in 2:length(V.vector) )
        {
        value   = V.vector[iV]
        
        dfv = makeDataForPrediction( Munsell2xy, value, p.LookupList )

        dfv = addPredictions( dfv, warn=warn )
        
        out[[iV]]   = attr( dfv, 'coeffs' )
        
        if( iV == 2 )
            # shrink a little bit because of problems in the very dark purple area
            out[[iV]]   = 0.8 * out[[iV]]  
        }
        
    #   correct the last name
    #   dimnames(out)[[3]] = names( stats::coef(mod.poly) )
    
    #   data for Value=0 is not available
    #   copy from Value=0.2 to Value=0
    out[[1]]    = out[[2]]
    
    time_elapsed    = gettime() - time_start
    mess    = sprintf( "made inversion coeffs in %g seconds.\n", time_elapsed )
    cat(mess)
    
    return(out)
    }

#
#   colorBlockFromMunsell()
#
#   MunsellSpec     a numeric Nx3 matrix, with HVC in each row
#                   or a single plain numeric HVC 3-vector
#                   or a character vector of Munsell names
#
#   if MunsellSpec is a 3-vector 
#       MunsellSpec[1]     Munsell hue, on the ASTM D1535 100 point circular scale.  values are wrapped to [1,101)
#       MunsellSpec[2]     Munsell value, must be between 0 and 10
#       MunsellSpec[3]     Munsell chroma, must be non-negative
#
#   if MunsellSpec is an Nx3 matrix
#       each row is defined as above
#
#   if MunsellSpec is an character N-vector of Munsell notations,
#       it is converted to an Nx3 matrix
#
#   return value:
#        data.frame with columns: HVC, ISCC-NBS Number, ISCC-NBS Name
#
#   the function requires these global data.frames:
#       p.System_ISCCNBS    this one is private
#       CentroidsISCCNBS    this one is public
#
#   author:  Glenn Davis


ColorBlockFromMunsell  <-  function( MunsellSpec )
    {    
    if( is.character(MunsellSpec) )
        {
        HVC     = HVCfromMunsellName(MunsellSpec)
        if( is.null(HVC) )  return(NULL)
        
        mask    = is.na( HVC[ ,1] )
        if( any(mask) )
            log_level( ERROR, "%d (out of %d) Munsell notations are invalid.", sum(mask), nrow(HVC) )
        
        rnames  = MunsellSpec
        }
    else  
        {
        HVC = prepareNx3( MunsellSpec )
        if( is.null(HVC) )  return(NULL)
        
        rnames  = rownames(HVC)
        
        if( is.null(rnames) )
            {
            rnames  = MunsellNameFromHVC( HVC, digits=2 )
            rownames(HVC)   = rnames
            }
            
        colnames(HVC)   = c('H','V','C')
        }
    
    n   = nrow(HVC)
    
    if( any(is.na(rnames)) ||  anyDuplicated(rnames) ) rnames  = 1:n   # rnames will be used later
    
    number      = rep( NA_integer_, n )
    name        = rep( NA_character_, n )
    centroid    = rep( NA_character_, n )
    vmax        = 10                
    
    for( i in 1:n )
        {
        hvc = HVC[i, ]
    
        valid   = all( is.finite(hvc) )  &&  (0 <= hvc[2])  &&  (hvc[2] <= vmax)  &&  (0 <= hvc[3])
        
        if( ! valid )   next
        
        #   wrap hue to the interval [1,101)   (101 is not included)
        hvc[1] = ((hvc[1] - 1 ) %% 100) + 1
            
        if( hvc[2] == vmax )   hvc[2] = vmax - 1.e-6     # because upper comparison below is strict
        
        #   do 6-way comparison.  
        #   Note upper comparisons are strict, and lower comparisons are not strict.
        #   So a point on a boundary is in only 1 block.
        mask.H  = p.System_ISCCNBS$Hmin <= hvc[1]  &  hvc[1] < p.System_ISCCNBS$Hmax
        mask.V  = p.System_ISCCNBS$Vmin <= hvc[2]  &  hvc[2] < p.System_ISCCNBS$Vmax
        mask.C  = p.System_ISCCNBS$Cmin <= hvc[3]  &  hvc[3] < p.System_ISCCNBS$Cmax
            
        theRow  = p.System_ISCCNBS[ mask.H & mask.V & mask.C,  , drop=FALSE]
        
        if( nrow(theRow) != 1 )
            {
            log_level( ERROR, "(internal).  Expected exactly 1 match for hvc=%g,%g,%g, but found %d.\n", 
                                    hvc[1],hvc[2],hvc[3], nrow(theRow) )
            next
            }
            
        number[i]   = theRow$Number
        name[i]     = munsellinterpol::CentroidsISCCNBS$Name[ theRow$Number ]           # we use the shortcut that rownumber is actually ISCC-NBS Number !
        centroid[i] = munsellinterpol::CentroidsISCCNBS$MunsellSpec[ theRow$Number ]    # we use the shortcut that rownumber is actually ISCC-NBS Number !
        }
        
    out = data.frame( row.names=rnames )
    out$HVC         = HVC
    out$Number      = number
    out$Name        = name
    out$Centroid    = centroid
    
    return( out )
    }
    
    

plotLociHC  <-  function( value=5, hue=seq(2.5,100,by=2.5), chroma='auto', coords='xy', main="Value %g/", est=FALSE, ... )
    {
    ok  = is.numeric(value)  &&  1<=length(value)  &&  all( 0<value  &  value<=10 )
    if( ! ok )
        {
        log_level( ERROR, "All Values must be numeric and in the interval (0,10]." )
        return(FALSE)
        }

    if( is.character(hue) )
        {
        hue = HueNumberFromString(hue)
        if( all( is.na(hue) ) )
            {
            log_level( "All character hues are invalid." )
            return(FALSE)
            }
        }

    ok  = is.numeric(hue)  &&  1<=length(hue)
    if( ! ok )
        {
        log_level( ERROR, "All Hues must be numeric." )
        return(FALSE)
        }

    hue = (hue %% 100)
    hue[ hue == 0 ] = 100

    ok  = (is.numeric(chroma)  &&  1<=length(chroma)  &&  all( 0 < chroma ))
    ok  = ok || (is.character(chroma) && chroma[1] == 'auto')
    if( ! ok )
        {
        log_level( ERROR, "All Chromas must be numeric and positive, or the string 'auto'." )
        return(FALSE)
        }

    if( ! (coords %in% c('xy','ab','AB') ) )
        {
        log_level( ERROR, "coords = '%s' is invalid.", coords )
        return(FALSE)
        }

    Value.vec   = sort( unique(munsellinterpol::Munsell2xy$V) )

    for( val in value )
        {
        main.val    = sprintf( main, val )

        if( is.character(chroma)  &&  chroma[1] == 'auto' )
            {
            idx     = which.min( abs(Value.vec - val) )
            vnear   = min( Value.vec[idx], 9 )
            mask    = (munsellinterpol::Munsell2xy$V == vnear)  &  (munsellinterpol::Munsell2xy$real)
            if( any(mask) )
                {
                chromamax   = max( munsellinterpol::Munsell2xy$C[mask] )
                chromavec  =  seq( 2, chromamax, by=2 )
                }
            else
                {
                log_level( ERROR, "Cannot determine maximum chroma for value=%g.", val )
                next
                }
            }
        else
            chromavec  = chroma

        if( coords == 'xy' )
            out = plotLociHC.xy( value, hue, chromavec, main.val, est, ...  )
        else if( coords == 'ab' )
            out = plotLociHC.ab( value, hue, chromavec, main.val, est, ...  )
        else if( coords == 'AB' )
            out = plotLociHC.AB( value, hue, chromavec, main.val, est, ...  )

        if( ! out ) break
        }

    return( invisible(out) )
    }



#   value       a *single* value this time
plotLociHC.xy  <-  function( value, hue, chroma, main, est, ... )
    {
    p   = 'spacesXYZ'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }

    df  = expand.grid( H=hue, C=chroma )

    HVC = as.matrix( cbind( df$H, value, df$C ) )

    #   add the illuminant at the end
    HVC = rbind( HVC, c(0,10,0) )

    xyY = MunsellToxyY( HVC, warn=FALSE, ... )$xyY

    xy_white    = xyY[ nrow(xyY), 1:2 ]

    xlim    = range( xyY[ ,1], na.rm=TRUE )
    ylim    = range( xyY[ ,2], na.rm=TRUE )



    plot( xlim, ylim, type='n', xlab='', ylab='', las=1, asp=1, lab=c(10, 10, 7), tcl=0, mgp=c(3,0.25,0) )
    title( xlab="x", line=1.25 )
    title( ylab="y", line=2.5 )
    grid( lty=1, col='gray70' )
    abline( h=0, v=0 )
    #   lines( c(0,1), c(1,0), lty=2 )

    #   draw the MacAdam limits
    tmp = sectionOptimals( YfromV(value) )
    if( ! is.null(tmp) )
        {
        xyY = spacesXYZ::xyYfromXYZ( tmp$XYZ )
        polygon( xyY[ ,1], xyY[ ,2], border='red', lwd=0.5 )
        }

    #   draw the inverted U
    xyz = p.xyz1931[  , 2:4 ]
    xyY = spacesXYZ::xyYfromXYZ( as.matrix(xyz) )
    polygon( xyY[ ,1], xyY[ ,2], border='blue', lwd=0.5 )


    #   put a small point at Illuminant C
    #xyY = MunsellToxyY( 'N 10/', warn=FALSE, ... )$xyY
    #xy_white = xyY[1:2]
    points( xy_white[1], xy_white[2], pch=19, cex=0.25 )

    #   draw radials
    chroma_min  = min( chroma, 1 )
    chroma_rad  = seq( chroma_min, max(chroma), by=0.25 )
    #   n   = length(chroma_rad)

    for( h in hue )
        {
        HVC = cbind( h, value, chroma_rad )

        xyY = MunsellToxyY( HVC, warn=FALSE, ...  )$xyY

        lines( xyY[ ,1], xyY[ ,2], lwd=0.5 )   #, border='black' )

        #   put a label at last point that is not NA
        #   print(  xyY[ ,1] )
        idx     = which( is.finite(xyY[ ,1]) )  #; print( is.finite(xyY[ ,1]) ) ; print(idx)
        m   = length(idx)
        xy      = xyY[idx[m],1:2]
        #offset  = xy - xy_white
        offset  = xy - xyY[idx[m-1],1:2]

        adj = 0.5 - 0.7 * sign(offset)
        text( xy[1], xy[2], HueStringFromNumber(h), adj=adj, cex=0.75, xpd=NA )   #, border='black' )
        }

    #   draw ovoids
    huevec  = seq( 0, 100, by=2.5/8 )
    #   n       = length(huevec)

    for( ch in chroma )
        {
        HVC = cbind( huevec, value, ch )

        xyY = MunsellToxyY( HVC, warn=FALSE, ... )$xyY

        lines( xyY[ ,1], xyY[ ,2], lwd=0.5 )   #, border='black' )
        }


    # label the hue lines
    #   n   = length(hue)
    #ch  = 1.05 * max(chroma)
    #HVC = cbind( hue, value, ch )
    #xyY = MunsellToxyY( HVC, warn=FALSE, ... )$xyY
    #text( xyY[ ,1], xyY[ ,2], HueStringFromNumber(hue), adj=c(0.5,0.5), cex=0.75 )   #, border='black' )


    #   mark the points in the lookup table
    mask    = (munsellinterpol::Munsell2xy$V == value)  &  (munsellinterpol::Munsell2xy$H %in% hue)  &  (munsellinterpol::Munsell2xy$C %in% chroma)

    if( any(mask) )
        {
        if( est )
            {
            data   = makeDataForPrediction( munsellinterpol::Munsell2xy, value, p.LookupList )  #; print( str(dfsub) )
            if( is.null(data) )    return(FALSE)

            #   print( dfsub )

            data   = addPredictions( data, warn=FALSE )

            #   for debugging
            #   print( data[ nrow(data), ] )

            segments( data$x, data$y, data$x.pred, data$y.pred, col='red' )
            points( data$x.pred, data$y.pred, pch=19, cex=0.4, col='red' )
            }

        dfsub   = munsellinterpol::Munsell2xy[ mask,  ]
        pch     = ifelse( dfsub$real, 19, 1 )
        points( dfsub$x, dfsub$y, pch=pch, cex=0.5, lwd=0.25 )   #, border='black' )
        }


    mess = sprintf( "%s \n[Chromas = %s]", main, paste0(chroma,collapse=', ') )
    #mess    = main
    title( main=mess )

    return(TRUE)
    }



#   value   a *single* value
#   hue     can be a vector
#   chroma  can be a vector

plotLociHC.ab  <-  function( value, hue, chroma, main, est, ...  )
    {
    p   = 'spacesXYZ'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }

    # par( omi=rep(0.5,4) )

    XYZ.C   = spacesXYZ::XYZfromxyY( c( p.xyC['NBS', ], 100 ) )

    df  = expand.grid( H=hue, C=chroma )

    HVC = as.matrix( cbind( df$H, value, df$C ) )

    #   add the illuminant at the end
    HVC = rbind( HVC, c(0,10,0) )

    xyY = MunsellToxyY( HVC, warn=FALSE, ... )$xyY

    #   xy_white    = xyY[ nrow(xyY), 1:2 ]

    #   convert to XYZ
    XYZ     = spacesXYZ::XYZfromxyY( xyY )
    XYZ.C   = XYZ[ nrow(XYZ), ]

    Lab     = spacesXYZ::LabfromXYZ( XYZ, XYZ.C )

    xlim    = range( Lab[ ,2], na.rm=TRUE )
    ylim    = range( Lab[ ,3], na.rm=TRUE )


    plot( xlim, ylim, type='n', xlab='', ylab='', las=1, asp=1, lab=c(10, 10, 7),  tcl=0, mgp=c(3,0.25,0)  )
    title( xlab="a", line=1.25 )
    title( ylab="b", line=2.5 )
    grid( lty=1, col='gray70' )
    abline( h=0, v=0 )

    #   draw the MacAdam limits
    tmp = sectionOptimals( YfromV(value) )
    if( ! is.null(tmp) )
        {
        Lab = spacesXYZ::LabfromXYZ( tmp$XYZ, XYZ.C )
        polygon( Lab[ ,2], Lab[ ,3], border='red', lwd=0.5 )
        }

    #   draw radials
    # cat( "drawing radials...\n", file=stderr() )

    chromavec  = seq( 1, max(chroma), by=0.25 )
    #   n   = length(chroma)

    for( h in hue )
        {
        HVC = cbind( h, value, chromavec  )

        xyY = MunsellToxyY( HVC, warn=FALSE, ...  )$xyY
        XYZ = spacesXYZ::XYZfromxyY( xyY )
        Lab = spacesXYZ::LabfromXYZ( XYZ, XYZ.C )

        lines( Lab[ ,2], Lab[ ,3], lwd=0.5 )   #, border='black' )

        #   put a label at the end
        idx     = which( is.finite(Lab[ ,2]) )  #; print( is.finite(xyY[ ,1]) ) ; print(idx)
        m       = length(idx)
        ab      = Lab[idx[m],2:3]
        offset  = ab - Lab[idx[m-1],2:3]

        adj = 0.5 - 0.7 * sign(offset)
        text( ab[1], ab[2], HueStringFromNumber(h), adj=adj, cex=0.75, xpd=NA )   #, border='black' )

        #  if( any(is.na(XYZ)) )   print( xyY )
        }



    #   draw ovoids
    # cat( "drawing ovoids...\n", file=stderr() )
    huevec  = seq( 0, 100, by=2.5/8 )
    #   n       = length(huevec)
    for( ch in chroma )
        {
        HVC = cbind( huevec, value, ch )

        xyY = MunsellToxyY( HVC, warn=FALSE, ... )$xyY
        XYZ = spacesXYZ::XYZfromxyY( xyY )
        Lab = spacesXYZ::LabfromXYZ( XYZ, XYZ.C )

        lines( Lab[ ,2], Lab[ ,3], lwd=0.5  )   #, border='black' )
        }


    #   mark the points in the lookup table

    data   = makeDataForPrediction( munsellinterpol::Munsell2xy, value, p.LookupList )  #; print( str(dfsub) )

    if( ! is.null(data) )
        {
        mask    =  (data$H %in% hue)  &  (data$C %in% chroma)

        if( any(mask)  &&  est )
            {
            data   = addPredictions( data, warn=FALSE )

            dfsub   = data[ mask, , drop=FALSE ]

            segments( dfsub$a, dfsub$b, dfsub$a.pred, dfsub$b.pred, col='red' )
            points( dfsub$a.pred, dfsub$b.pred, pch=19, cex=0.4, col='red' )
            }

        pch     = ifelse( data$real, 19, 1 )
        points( data$a, data$b, pch=pch, cex=0.5, lwd=0.25 )   #, border='black' )
        }


    mess = sprintf( "%s \n[Chromas=%s]", main, paste0( chroma, collapse=', ' ) )
    title( main=mess )

    return( invisible(TRUE) )
    }



plotLociHC.AB  <-  function( value, hue, chroma, main, est, ...    )
    {
    chroma.max  = max(chroma)

    xlim    = c( -chroma.max, chroma.max )
    ylim    = c( -chroma.max, chroma.max )

    plot( xlim, ylim, type='n', xlab='A', ylab='B', las=1, asp=1, lab=c(10, 10, 7) )


    #   draw radials
    for( theta in (hue * pi/50) )
        lines( chroma.max * c(0,cos(theta)), chroma.max * c(0,sin(theta)), col="#eeeeee" )   #, border='black' )

    #   draw circles
    theta   = seq( 0, 2*pi, len=101 )
    circle  = matrix( c(cos(theta),sin(theta)), length(theta), 2 )
    for( ch in chroma )
        lines( ch * circle[ ,1], ch * circle[ ,2], col="#eeeeee" )

    grid( lty=1 )
    abline( h=0, v=0 )


    data   = makeDataForPrediction( munsellinterpol::Munsell2xy, value, p.LookupList )  #; print( str(dfsub) )

    if( ! is.null(data) )
        {
        mask    =  (data$H %in% hue)  &  (data$C %in% chroma)

        if( est  &&  any(mask)  )
            {
            #    plot inversion estimates
            data   = addPredictions( data, warn=FALSE )

            dfsub   = data[ mask, , drop=FALSE ]

            segments( dfsub$A, dfsub$B, dfsub$A.pred, dfsub$B.pred, col='red' )
            points( dfsub$A.pred, dfsub$B.pred, pch=19, cex=0.4, col='red' )
            }

        pch     = ifelse( data$real, 19, 1 )
        points( data$A, data$B, pch=pch, cex=0.5, lwd=0.25 )
        }


    if( 0 )
    {
    #   plot the "crude" prediction
    dfreal   = dfsub[ dfsub$real, ]

    A.crude = dfsub$a / 5.5
    B.crude = dfsub$b / 5.5
    segments( dfreal$A, dfreal$B, A.crude, B.crude, col='blue' )
    points( A.crude, B.crude, pch=19, cex=0.4, col='blue' )

    #   plot the real points
    points( dfreal$A, dfreal$B, pch=19, cex=0.4 )

    #   plot the non-real points
    dfnonreal   = dfsub[ ! dfsub$real, ]
    points( dfnonreal$A, dfnonreal$B, pch=1, cex=0.4, col='cyan' )
    }


    mess = sprintf( "%s \n[Chromas=%s]", main, paste( chroma, collapse=' ' ) )
    title( main=mess )

    return(TRUE)
    }


plotPatchesH  <-  function( hue, space='sRGB', adapt='Bradford', background='gray50',
                                main="Hue %s  (H=%g)      [%s   adapt=%s]", value=NULL, chroma=NULL, ...  )
    {
    if( is.character(hue) )
        {
        hue = HueNumberFromString(hue)
        if( any( is.na(hue) ) )
            {
            log_level( "One or more of the character hues is invalid." )
            return( invisible(FALSE) )
            }
        }

    ok  = is.numeric(hue)  &&  1<=length(hue)
    if( ! ok )
        {
        log_level( ERROR, "All Hues must be numeric." )
        return( invisible(FALSE) )
        }

    hue = (hue %% 100)
    hue[ hue == 0 ] = 100

    if( ! is.null(value) )
        {
        #   check value
        ok  = is.numeric(value)  &&  0<length(value)  &&  0<=value[1]  &&  all( 0 < diff(value) )  &&  value[length(value)]<=10
        if( ! ok )
            {
            log_level( ERROR, "argument 'value' is invalid.  It must be strictly increasing with all values in [0,10]." )
            return( invisible(FALSE) )
            }
        }

    if( ! is.null(chroma) )
        {
        #   check chroma
        ok  = is.numeric(chroma)  &&  0<length(chroma)  &&  0<=chroma[1]  &&  all( 0 < diff(chroma) )
        if( ! ok )
            {
            log_level( ERROR, "argument 'chroma' is invalid.  It must be non-negative and strictly increasing." )
            return( invisible(FALSE) )
            }
        }

    for( h in hue )
        {
        main.single = sprintf( main, HueStringFromNumber(h), h, space, adapt )

        out = plotPatchesH.single( h, space=space, adapt=adapt, background=background, main=main.single,
                                        valuevec=value, chromavec=chroma, ... )

        if( ! out ) break
        }

    return( invisible(out) )
    }




#   hue is a single number
plotPatchesH.single <- function( hue, space='sRGB', adapt='Bradford', background='gray50', main='',
                                        valuevec=NULL, chromavec=NULL, ... )
    {
    #   we might have to compute HVC twice, once from real.dat and once as a grid
    makegrid    = ! is.null(valuevec)  ||  ! is.null(chromavec)

    if( is.null(valuevec) )
        valuevec = 0:10

    if( is.null(chromavec) )
        {
        #   find the closest hue page
        huevec  = sort( unique( munsellinterpol::Munsell2xy$H ) )   #; print( huevec )

        delta   = abs(hue - huevec)
        delta   = pmin( delta, 100 - delta )
        hue.near    = huevec[ which.min( delta ) ]  #; print( hue.near )

        #   find all the patches on the hue.near page, using samples in the file real.dat
        #   only look at real patches, and Value>=1
        mask    = munsellinterpol::Munsell2xy$H == hue.near  &  munsellinterpol::Munsell2xy$real  &  1 <= munsellinterpol::Munsell2xy$V
        dfsub   = munsellinterpol::Munsell2xy[ mask, ]

        HVC     = cbind( hue, dfsub$V, dfsub$C )

        if( ! makegrid )
            #   the HVC we are making now is the final one, but it currently has no neutrals, so add them
            HVC = rbind( cbind(0,valuevec,0), HVC )

        #   find the largest Chroma in the hue.near page, using samples in  the file real.dat
        chromamax   = max( HVC[ ,3] )

        chromavec   = seq( 0, chromamax, by=2 )
        }

    if( makegrid )
        {
        #   ignore previous HVC and create a new one
        #   the previous one was only needed to get chromavec
        df  = expand.grid( valuevec, chromavec )

        HVC = cbind( hue, df[ ,1], df[ ,2] )
        }

    colnames(HVC)  = c('H','V','C')

    #   the next call might create WARNs, so suppress them
    tmp     = MunsellToRGB( HVC, space=space, adapt=adapt, warn=FALSE, ... )

    #cat( "Bad xyY's = ", sum( is.na(tmp$xyY[ ,1]) ), '\n' )
    #print( tmp )

    xyY     = tmp$xyY          # this is adapted to illuminant C

    #   only keep ones that were successfully transformed, and then within MacAdamLimits
    tol     = 0.004     # adjusted for "N 10/", which is problematic, it's right on the boundary
    mask    = is.finite( xyY[ ,1] )   &   (IsWithinMacAdamLimits( xyY, 'C' )$delta <= tol)     #;   print( sum(mask) )

    # the sample N 10/ is problematic, it's right on the boundary
    #idx = which( HVC[ ,2]==10  &  HVC[ ,3]==0 )
    #if( 0 < length(idx) )   mask[ idx[1] ]  = TRUE

    #   throw away the ones outside the limits
    dfsub       = tmp[ mask, , drop=FALSE ]
    dfsub$HVC   = HVC[ mask, , drop=FALSE ]

    # print( dfsub )
    # cat( "samples inside MacAdamLimits = ", nrow(dfsub), '\n' )

    dfsub$color = grDevices::rgb( round(dfsub$RGB), maxColorValue=255 )    #    print( str(dfsub) )

    #   setup the plot
    xlim    = c( 0.5, length(chromavec)+0.5 )
    ylim    = c( 0.5, length(valuevec)+0.5 )

    bg.prev = par( bg=background )
    par( mgp=c(0, 0.5, 0) )
    plot.new()
    plot.window( xlim, ylim )


    #   set up margins
    margin  = 0.1

    size    = grDevices::dev.size('in')

    scale.x = size[1] / (xlim[2] - xlim[1])
    scale.y = size[2] / (ylim[2] - ylim[1])


    if( scale.y < scale.x )
        {
        margin.x    = margin * (scale.y / scale.x)
        margin.y    = margin
        }
    else
        {
        margin.x    = margin
        margin.y    = margin * (scale.x / scale.y)
        }

    #   draw axes
    lines( c(xlim[1],xlim[2]) - 0, c(ylim[1],ylim[1]) - 0, xpd=NA )   # X axis
    lines( c(xlim[1],xlim[1]) - 0, c(ylim[1],ylim[2]) - 0, xpd=NA )   # Y axis

    title( xlab='Chroma', line=0.5 )
    title( ylab='Value', line=0.5 )

    title( main=main, line=0 )

    #   x labels
    text( 1:length(chromavec), rep(ylim[1],length(chromavec)), sprintf( "/%g", chromavec ), adj=c(0.5,1.5), cex=0.8, xpd=NA )

    #   y labels  =  Value labels
    ht  = graphics::strheight('0')
    text( rep(xlim[1] - ht/5, length(valuevec)), 1:length(valuevec), sprintf( "%g/", valuevec ), adj=c(1,0.5), cex=0.8, xpd=NA )

    if( nrow(dfsub) == 0 )
        {
        #   nothing to draw
        par( bg=bg.prev )   # restore previous background
        return(FALSE)
        }

    #   make small dictionaries for Value and Chroma
    xfromchroma = 1:length(chromavec)
    names(xfromchroma)  = as.character(chromavec)

    yfromvalue  = 1:length(valuevec)
    names(yfromvalue)  = as.character(valuevec)

    if( FALSE &&  chromavec[1] == 0 )
        {
        #   draw neutrals
        HVC = cbind( 0, valuevec, 0 )
        dfneutral   = MunsellToRGB( HVC, space=space, adapt=adapt, ... )
        dfneutral$HVC   = HVC
        dfneutral$color = grDevices::rgb( round(dfneutral$RGB), maxColorValue=255 )   #    print( dfneutral )

        n       = nrow( dfneutral )
        left    = rep( 0.5 + margin.x, n )
        right   = rep( 1.5 - margin.x, n )
        y       = yfromvalue[ as.character(dfneutral$HVC[ ,2]) ]
        bottom  = y - 1/2 + margin.y
        top     = y + 1/2 - margin.y
        rect( left, bottom, right, top, col=dfneutral$color, border=NA )
        }


    #   draw the HVCs
    x   = xfromchroma[ as.character(dfsub$HVC[ ,3]) ]
    y   = yfromvalue[ as.character(dfsub$HVC[ ,2]) ]

    dfsub$left    = x - 1/2 + margin.x
    dfsub$right   = x + 1/2 - margin.x
    dfsub$bottom  = y - 1/2 + margin.y
    dfsub$top     = y + 1/2 - margin.y
    color   = ifelse( dfsub$OutOfGamut, NA, dfsub$color )
    border  = ifelse( dfsub$OutOfGamut, 'gray30', NA )
    rect( dfsub$left, dfsub$bottom, dfsub$right, dfsub$top, col=color, border=border )

    #   draw RGB labels inside the out-of-RGB-gamuts
    dfout   = dfsub[ dfsub$OutOfGamut, , drop=FALSE ]
    if( 0 < nrow(dfout) )
        {
        RGB     = round( dfout$RGB )
        label   = sprintf( "R %3d\nG %3d\nB %3d", RGB[ ,1], RGB[ ,2], RGB[ ,3] )

        w   = graphics::strwidth( label[1], family="mono" )
        h   = graphics::strheight( label[1], family="mono" )
        w0  = graphics::strwidth( "0", family="mono" )

        cex = min( (1 - 2*margin.x - w0) / w, (1 - 2*margin.y - w0) / h )

        text( (dfout$left+dfout$right)/2, (dfout$top+dfout$bottom)/2, label, cex=cex, adj=c(0.5,0.5), family="mono" )
        }

    par( bg=bg.prev )   # restore previous background

    return(TRUE)
    }




plotLookupDF <- function( Munsell2xy )
    {
    Value.vec   = sort( unique(Munsell2xy$V) )

    huelim    = c(0,100)
    chromalim    = c(0,50)

    hueseq      = seq( 0, 100, by=2.5 )
    chromaseq   = seq( 0, 50, by=2 )
    for( v in Value.vec )
        {
        plot.new()
        plot.window( huelim, chromalim ) #,  mgp=c(2,2,2)  )
        #   par( mai=c( 0.5, 0.5, 0.5, 0 ) )
        #par( mgp=c(1/2,3,1/2) )
        title( xlab="Hue", ylab="Chroma", main=sprintf( "Value %g", v ) )

        segments(  0, chromaseq, 100, chromaseq, col='gray' )
        segments(  hueseq, 0, hueseq, 50, col='gray' )
        rect( huelim[1], chromalim[1], huelim[2], chromalim[2], border='black' )

        text( -1, chromaseq, as.character(chromaseq), adj=c(1,1/2), cex=0.75 )
        text( hueseq, -1, HueStringFromNumber(hueseq), adj=c(1,1/2), cex=0.75, srt=90, xpd=NA )

        dfsub   =   Munsell2xy[ Munsell2xy$V==v,  ]

        pch = ifelse( dfsub$real, 19, 1 )
        points( dfsub$H, dfsub$C, pch=pch )
        #   break
        }

    return(TRUE)
    }

plotLookupList <- function( LookupList, Munsell2xy, value=NULL )
    {
    Value.vec   = attr( LookupList, 'V.vector' )
    Hue.vec     = attr( LookupList, 'H.vector' )
    Chroma.vec  = attr( LookupList, 'C.vector' )

    huelim    = c(0,100)
    chromalim    = c(0,50)

    hueseq      = seq( 0, 100, by=2.5 )
    chromaseq   = seq( 0, 50, by=2 )

    if( is.null(value) )
        vvec    =  1:length(LookupList)
    else
        {
        vvec    = match( value, Value.vec )
        vvec    = vvec[ is.finite(vvec) ]
        }

    for( iV in vvec )
        {
        par( mai=c( 0.5, 0.5, 0.75, 0.25 ) )
        #par( mgp=c(1/2,3,1/2) )
        plot.new()
        plot.window( huelim, chromalim ) #,  mgp=c(2,2,2)  )

        # plot.default( huelim, chromalim, las=1, xlab='', ylab='', type='n', lab=c(10,8,7),  tcl=0, mgp=c(3, 0.25, 0) )

        title( xlab="Hue", line=0.25  )
        title( ylab="Chroma", line=0  )

        value   = Value.vec[iV]
        title(  main=sprintf( "Value %g", value ) )

        segments(  0, chromaseq, 100, chromaseq, col='gray' )
        segments(  hueseq, 0, hueseq, 50, col='gray' )
        rect( huelim[1], chromalim[1], huelim[2], chromalim[2], border='black' )

        text( -1, chromaseq, as.character(chromaseq), adj=c(1,1/2), cex=0.75 )
        text( 100+1, chromaseq, as.character(chromaseq), adj=c(0,1/2), cex=0.75 )

        hueskip = seq(0,100,by=5)
        text( hueskip, -1, hueskip, adj=c(0.5,1), cex=0.75 )
        text( hueseq, chromalim[2]+0.5, HueStringFromNumber(hueseq), adj=c(0,1/2), cex=0.75, srt=90, xpd=NA )

        dfsub   = Munsell2xy[ Munsell2xy$V==value,  ]
        pch = ifelse( dfsub$real, 19, 1 )
        points( dfsub$H, dfsub$C, pch=pch )

        dfsub   = dfsub[ dfsub$H==100, ]
        pch = ifelse( dfsub$real, 19, 1 )
        points( rep(0,length(dfsub$C)), dfsub$C, pch=pch )

        idx = which( ! is.na(LookupList[[iV]]$x), arr.ind=TRUE )  #; print( idx )
        hue     = Hue.vec[ idx[ ,1] ]
        chroma  = Chroma.vec[ idx[ ,2] ]
        mask    = 0 <= hue  &  hue <= 100
        points( hue[mask], chroma[mask], pch=4 )
        #   break

        #dfsub   = subset( Munsell2xy, V==v )

        #pch = ifelse( dfsub$real, 19d, 1 )
        #points( dfsub$H, dfsub$C, pch=pch )
        #   break
        }

    return(TRUE)
    }

#   x   mask is TRUE
#       solid disk, real
#   o   open circle, not real
plotMask3D <- function( Mask3D, Munsell2xy )
    {
    Hue.vec     = as.numeric( dimnames(Mask3D)[[1]] )
    Value.vec   = as.numeric( dimnames(Mask3D)[[2]] )
    Chroma.vec  = as.numeric( dimnames(Mask3D)[[3]] )

    huelim    = c(0,100)
    chromalim    = c(0,50)

    hueseq      = seq( 0, 100, by=2.5 )
    chromaseq   = seq( 0, 50, by=2 )


    for( iV in 1:length(Value.vec) )
        {
        plot.new()
        plot.window( huelim, chromalim ) #,  mgp=c(2,2,2)  )
        #   par( mai=c( 0.5, 0.5, 0.5, 0 ) )
        #par( mgp=c(1/2,3,1/2) )

        value   = Value.vec[iV]
        title( xlab="Hue", ylab="Chroma", main=sprintf( "Value %g", value ) )

        segments(  0, chromaseq, 100, chromaseq, col='gray' )
        segments(  hueseq, 0, hueseq, 50, col='gray' )
        rect( huelim[1], chromalim[1], huelim[2], chromalim[2], border='black' )

        text( -1, chromaseq, as.character(chromaseq), adj=c(1,1/2), cex=0.75 )
        text( hueseq, -1, HueStringFromNumber(hueseq), adj=c(1,1/2), cex=0.75, srt=90, xpd=NA )


        dfsub   = Munsell2xy[ Munsell2xy$V==value,  ]

        pch = ifelse( dfsub$real, 19, 1 )
        points( dfsub$H, dfsub$C, pch=pch )

        idx = which( Mask3D[ , iV, ], arr.ind=TRUE )
        H   = Hue.vec[ idx[ ,1] ]
        C   = Chroma.vec[ idx[ ,2] ]
        points( H, C, pch=4 )
        }

    return(TRUE)
    }

## ----setup, include=FALSE-------------------------------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)
options( width=100 )

## ----echo=TRUE, message=TRUE----------------------------------------------------------------------
library(munsellinterpol)

## ----echo=TRUE, warning=TRUE----------------------------------------------------------------------
load_soil <- function( soil_raw=OxSR::soil_refle, subset=c( "a4", "a6", "a8", "a11", "a14", "a15" ),
                                         wave=380:780 )
    {
    #   locate only the wavelengths in OxSR::soil_refle that are relevant for color
    #   the given data.frame OxSR::soil_refle goes way into the infra-red and these are not needed
    #   We already know that all relevant multiples of 1nm are in soil_raw
    idx     = match( wave, soil_raw$wavelength_nm )

    #   extract just a subset of samples with high variability
    #   the reflectances in OxSR::soil_refle are percentages, so divide by 100    
    mat     = as.matrix(soil_raw)[ idx, subset ] / 100
    
    #   convert the matrix mat to a colorSpec object
    soil_spec = colorSpec::colorSpec( mat, wavelength=wave, quantity='reflectance' )
    
    #   add neutral gray as a test sample
    gray      = colorSpec::neutralMaterial( 0.18, wavelength=wave )
    soil_spec = colorSpec::bind( soil_spec, gray )
   
    return( soil_spec ) 
    }

## ----echo=TRUE, warning=TRUE,  fig.width=7.3, fig.height=5.2,  fig.show='hold'--------------------
soil_spec = load_soil()
par( omi=c(0,0,0,0), mai=c(0.6,0.6,0.3,0.3) )
plot( soil_spec, legend="topleft" )

## ----echo=TRUE, warning=TRUE, results="hide"------------------------------------------------------
soil_data <- function( soil_spec, CCT )
    {
    wave    = colorSpec::wavelength(soil_spec) 
    
    #   make illuminant from the CCT. This is the reference illuminant in IES Standard TM-30.
    illum   = colorSpec::referenceSpectraTM30( CCT, wavelength=wave )
    
    #   make scanner from this illuminant and the standard x,y,z responsivities of the human eye
    scanner = colorSpec::product( illum, "soil", colorSpec::xyz1931.1nm, wavelength=wave )
    
    #   scale so the Perfect Reflecting Diffuser has Y=100; which is conventional for Munsell work
    scanner = colorSpec::calibrate( scanner, response=c(NA,100,NA), method='scaling' )
    
    #   compute white point of illum
    PRD     = colorSpec::neutralMaterial(1,wavelength=wave)    # the Perfect Reflecting Diffuser
    white   = colorSpec::product( PRD, scanner )
    
    #   compute XYZ for all soil samples; this is XYZ under illum
    XYZ     = colorSpec::product( soil_spec, scanner )
    
    #   convert XYZ to Munsell HVC
    #   XYZtoMunsell() internally does a chromatic adaptation to the white point of Illuminant C
    #   the default is to use the original NBS variant of the white point of Illuminant C
    HVC     = XYZtoMunsell( XYZ, white=white )   
    rownames(HVC)   = colorSpec::specnames(soil_spec)
    
    # create output data.frame and add 4 columns: HVC, Munsell notation, ISCC-NBS color name, and Lab
    out = data.frame( row.names=rownames(HVC) )
    out$HVC     = HVC
    out$Munsell = MunsellNameFromHVC( HVC )
    out[[ "ISCC-NBS Name" ]] = ColorBlockFromMunsell( HVC )$Name
    out$Lab     = spacesXYZ::LabfromXYZ( XYZ, white )
    
    return( out )
    }

## ----echo=TRUE, message=TRUE----------------------------------------------------------------------
dat_soil = soil_data( soil_spec, CCT=6500 )
dat_soil

## ----echo=TRUE, message=TRUE----------------------------------------------------------------------
dat_rnd     = roundHVC( dat_soil$HVC, books="soil" )
dat_rnd$HVC = NULL     # delete matrix HVC, because it is redundant in this context
dat_rnd

## ----echo=TRUE, message=TRUE, fig.width=7.3, fig.height=5.2,  fig.show='hold'---------------------
par( omi=c(0,0,0,0), mai=c(0.4,0.4,0.25,0) )
plotPatchesH( "2.5YR",   value=c(2.5,3:8), chroma=c(1,2,3,4,6,8) )
text( c(4,4,3), c(5,4,4), c("a6","a8","a15"), adj=c(0.5,0.5), col="white" )

## ----echo=TRUE, message=TRUE----------------------------------------------------------------------
tbl = data.frame( row.names=1:(2*nrow(dat_soil)) )
tbl$Sample  = as.character( rbind(rownames(dat_soil),'') )
tbl$Precise = as.character( rbind('',MunsellNameFromHVC( dat_soil$HVC, digits=3 )) )
tbl$Rounded = as.character( rbind('',dat_rnd$MunsellRounded ) )

library( flextable )
myrt <- regulartable( tbl )
myrt <- width( myrt, j=c(2,3), width=2 )
myrt <- align( myrt, j=1:3, align='center', part='all' )
myrt <- hline( myrt, i=seq( 2, nrow(tbl)-2, by=2 ), border=fp_border_default( color="black", width=2 ) )
myrt <- hrule( myrt, rule = "exact" )
idxcolor <- seq(1,nrow(tbl)-1,by=2)
myrt <- height( myrt, i=idxcolor, height=1 )
myrt <- bg( myrt, i=idxcolor, j=2, bg=rgb( round(MunsellTosRGB(dat_soil$HVC)$RGB), max=255 ) )
myrt <- bg( myrt, i=idxcolor, j=3, bg=rgb( round(MunsellTosRGB(dat_rnd$MunsellRounded)$RGB), max=255 ) )

## ----echo=FALSE, message=TRUE---------------------------------------------------------------------
myrt

## ----echo=FALSE, results='asis'-------------------------------------------------------------------
sessionInfo()

    
    
        

#   returns list with
#       omega.from.lambda()
#       lambda.from.omega()
#       responsivity.from.omega[[ ]]    a list with length = channels.  No longer needed, just take deriv=1 of the next one !
#       integral.from.omega[[ ]]        a list with length = channels
#   omega is in [0,1]

makeReparamFunctionList <- function( .lambda, .illum, .xyz, .mode='equalize', .pnorm=2, .Ymax=1 )
    {
    n   = length(.lambda)
    
    if( length(.illum) != n )
        {
        log_level( ERROR, "length(.illum) = %d is incorrect.", length(.illum) )
        return(NULL)
        }
        
    if( ! all( dim(.xyz) == c(n,3) ) )
        {
        log_level( ERROR, "dim(.xyz) = %d,%d is incorrect.", dim(.xyz)[1], dim(.xyz)[2] )
        return(NULL)
        }

    coredata    = matrix( .illum, n, 3 ) * .xyz
            
    #   remove rows that are all 0s
    mask    =   ( 0 < rowSums(abs(coredata)) )
    coredata    = coredata[ mask, ]
    .lambda     = .lambda[ mask ]
    
    step.wl = unique( diff(.lambda) )       
    if( length(step.wl) != 1 )
        {
        log_level( ERROR, "%d wavelengths are not regular.", length(.lambda) )        
        return(NULL)
        }      

    #   check that coredata is full-rank
    singular    = svd( coredata, nu=0, nv=0 )$d
    
    #   log_level( DEBUG, "SVD time = %g sec", as.double(Sys.time()) - time_svd )
    
    thresh  = max( dim(coredata) ) * singular[1] * 2^(-52)
    rank = sum( thresh < singular )
    if( rank <  min(dim(coredata)) )
        {
        log_level( ERROR, "The responsivity matrix   is rank-deficient (rank=%d < %d).", 
                                 rank, min(dim(coredata)) )        
        return(NULL)
        }    
        
    n           = nrow(coredata)
    channels    = ncol(coredata)
    
    lambda.center   = .lambda
    lambda.min      = lambda.center[1]  
    lambda.max      = lambda.center[n]  
        
    #   wavelengths defining the bins
    lambda.break    = seq( lambda.min - step.wl/2, lambda.max + step.wl/2, len=(n+1) )        
        
    out = list()
           
    if( .mode == 'equalize' )
        {
        if( .pnorm == 2 )
            omega   = sqrt( rowSums(coredata*coredata) )
        else if( .pnorm == 1 )
            omega   = rowSums( abs(coredata) )
        else
            omega   = rowSums( abs(coredata)^.pnorm )^(1/.pnorm)
        
        if( any(omega == 0) )
            {
            log_level( ERROR, "Responder has responsivity=0 at 1 or more wavelengths, which is invalid." )
            return(NULL)
            }
            
        omega   = c( 0, cumsum(omega) ) # n+1 of these.
        omega   = omega / omega[n+1]    # n+1 values from 0 to 1.  Not regular.  #;     print( omega )
        
        #   for v 2.6-* added ties=min to suppress warning from regularize.values()
        out$omega.from.lambda   = splinefun( lambda.break, omega, method='monoH.FC', ties=min )      # hyman  monoH.FC  natural
        out$lambda.from.omega   = splinefun( omega, lambda.break, method='monoH.FC', ties=min )      # hyman  monoH.FC  natural
        }
    else if( .mode == 'linear' )
        {
        omega   = (0:n) / n  # n+1 values in regular steps
        
        out$omega.from.lambda   = function( lambda )    { (lambda - lambda.break[1]) / (lambda.break[n+1] - lambda.break[1]) }
        
        out$lambda.from.omega   = function( omega )     {  (1-omega)*lambda.break[1]  +  omega*lambda.break[n+1] }
        }
    else
        {
        log_level( ERROR, ".mode='%s' is invalid.", .mode )
        return(NULL)
        }
        
    out$lambda.break    = lambda.break
    out$omega           = omega
             

    out$integral.from.omega = list()        
    out$omega.from.integral = list()     
    
    #   out$responsivity.from.omega = list()    
    
    s   = .Ymax / (sum( coredata[ , 2] ) )
    
    #   print( s * colSums(coredata) )
    
    for( j in 1:channels )
        {
        integral    = s * c( 0, cumsum( coredata[ , j ] ) )   #   ; print( range(integral) )
        
        #   for v 2.6-* added ties=min to suppress warning from regularize.values()        
        out$integral.from.omega[[j]] <- splinefun( omega, integral, method='monoH.FC', ties=min )   # hyman  monoH.FC  natural
        out$omega.from.integral[[j]] <- splinefun( integral, omega, method='monoH.FC', ties=min )   # hyman  monoH.FC  natural
        
        #   omega.center    = out$omega.from.lambda( lambda.center )
        #   out$responsivity.from.omega[[j]]    = splinefun( omega.center, coredata[ ,j] )  # a tiny bit of extrapolation
            
        #   the next line does not work -- only 1 new function is created  -- I think maybe it is a bug
        #   out$responsivity.from.omega[[j]]    <- function( om ) { return( out$integral.from.omega[[j]]( om, deriv=1 ) ) }
        #   print( str(out$responsivity.from.omega[[j]]) )
        }        
        
    return( out )
    }
    
    
    

MunsellSpecToColorlabFormat  <-  function( MunsellSpec )
    {
    if( is.character(MunsellSpec) )
        {
        rnames  = MunsellSpec
        MunsellSpec = HVCfromMunsellName( MunsellSpec )
        }
    else
        rnames = NULL
        
    HVC = prepareNx3(MunsellSpec)
    if( is.null(HVC) )  return(NULL)
    
    hue = HVC[ ,1] %% 100
    
    hue.rev = 100 - hue
    
    hue.idx     = as.integer( hue.rev / 10 )
    hue.frac    = hue.rev - 10 * hue.idx
    
    hue.idx     = ((hue.idx-3) %% 10) + 1
    hue.step    = 10 - hue.frac
    
    out = cbind( hue.step, HVC[ ,2:3,drop=FALSE], hue.idx )  #; print( str(out) )
    
    #   fix the neutrals, set both hue number and hue index to 0
    out[ HVC[ ,3]==0, c(1,4) ] = 0
    
    if( is.null(rnames) )   rnames = MunsellNameFromHVC(HVC)
    
    rownames(out)   = rnames
    colnames(out)   = c( 'HS', 'V', 'C', 'HI' )

    return(out)
    }
    
    
ColorlabFormatToMunsellSpec  <-  function( HVCH )    
    {
    HVCH    = prepareNx3( HVCH, M=4 )
    if( is.null(HVCH) ) return(NULL)
    
    hue.rev = 10 * HVCH[ ,4] - HVCH[ ,1]
    
    hue = 100 - ( (hue.rev+30) %% 100 )
    
    out = cbind( hue, HVCH[ ,2:3,drop=FALSE] )
    
    #   fix the neutrals, set hue to 0
    mask.neutral    = (HVCH[ ,3]==0)
    out[  mask.neutral, 1] = 0    
    
    #   invalidate any rows where the hue index is invalid
    mask.ok = (HVCH[ ,4]  %in%  1:10)  |  mask.neutral
    out[ ! mask.ok, ] = NA_real_
    
    rownames(out)   = rownames(HVCH)
    
    if( is.null( rownames(out) ) )
        rownames(out)   = MunsellNameFromHVC(out)
        
    colnames(out)   = c('H','V','C')
    
    return( out )
    }
    

#   'zonohedron' is an S3 class
#   It presents the zonohedron as an intersection of halfspaces.
#
#   It is stored as a list with
#       W           the original given nx3 matrix, defining n line segments in R^3  (with 1 endpoint at 0).  Must be rank 3.
#       center      middle gray = white/2
#       nonnegative logical.  All entries of W are non-negative
#       face        a data frame with n*(n-1)/2 rows.  Because of central symmetry, only 1/2 of the faces need to be stored.
#                   idx     2 columns with 2 indices to distinct rows of W
#                   normal  to a face of the zonohedron, the cross-product of the 2 rows, then unitized
#                   center  of the face, a parallelogram, *after* centering the zonohedron
#                   beta    the plane constant, *after* centering the zonohedron.  All are positive.

#   zonohedron() constructor for a zonohedron object
#       W       nx3 matrix with rank 3
#       tol     relative tolerance for degenerate faces
#
#   returns: a list as above
#
zonohedron  <- function( W, tol=1.e-6 )
    {
    ok  = is.numeric(W)  &&  is.matrix(W)  &&  3<=nrow(W)  &&  ncol(W)==3
    if( ! ok )
        {
        log_level( ERROR, "argument W is not an nx3 numeric matrix, with n>=3." )
        return(NULL)
        }
        
    if( any( is.na(W) ) )
        {
        log_level( ERROR, "matrix W is invalid because it has %d entries that are NA.", sum(is.na(W)) )
        return(NULL)
        }
    
    p12 = tcrossprod( W[ ,1,drop=F], W[ ,2,drop=F] )
    p13 = tcrossprod( W[ ,1,drop=F], W[ ,3,drop=F] )
    p23 = tcrossprod( W[ ,2,drop=F], W[ ,3,drop=F] )
    
    d12 = p12 - t(p12)
    d13 = t(p13) - p13
    d23 = p23 - t(p23)
    
    n   = nrow(W)
    
    idx = t( utils::combn(n,2) )   # matrix of pairs.  This is kind of slow, but OK for 81 wavelengths
    
    normal  = cbind( d23[idx], d13[idx], d12[idx] )
    
    #   set rows close to 0 to all NAs
    WW  = cbind( W[ idx[,1], ], W[ idx[,2], ] )
    WW2 = .rowSums( WW*WW, nrow(WW), ncol(WW) )
    
    normal2 = .rowSums( normal*normal, nrow(normal), ncol(normal) )
    bad     = normal2 <= tol*tol * WW2
    if( all(bad) )
        {
        log_level( ERROR, "argument W does not have rank 3, with relative tol=%g.", tol )
        return(NULL)
        }


    if( any(bad) )
        {    
        mess    = sprintf( "%d bad normals, out of %d.\n", sum(bad), length(bad) )
        cat(mess)
        #   log_level( INFO, "%d normals flagged as too small, out of %d.", sum(bad), length(bad) )
        normal[ bad, ]  = NA_real_        
        }
    
    #   now unitize
    normal  = normal / sqrt(normal2)    # sqrt(normal2) is replicated to all 3 columns

    out = list()
    
    out$W   = W
    
    out$center  = 0.5 * .colSums( W, nrow(W), ncol(W) )
    
    out$nonnegative = all( 0 <= W )
    
    out$face        = data.frame( row.names=1:nrow(idx) )
    #   out$face$idx    = idx       not really needed for inside()
    out$face$normal = normal
    
    
    if( FALSE )
        {
        #   cluster the unit normal vectors  [reminds me of something similar at Link]
        res = findRowClusters( normal )
        if( is.null(res) )  return(NULL)
        
        out$clusteridx  = res$clusteridx
        out$cluster     = res$cluster
        }    
    
    #   the next line is the biggest bottleneck
    #   the size of the output is on the order n^3
    functional  = tcrossprod( W, normal ) 
    #   pos = 0<S01 ;     S01[ pos ]  = 1 ;     S01[ !pos ] = 0
    
    #   functional is n x n(n-1)/2
    #   each column corresponds to an i<j pair, from array idx[]
    #   the corresponding entries in that column should be very close to 0
    if( TRUE )
        {
        #time_start = gettime()
        
        across  = 1:nrow(idx)
        idx2    = rbind( cbind(idx[,1],across),  cbind(idx[,2],across) )
        
        #   check   = functional[idx2]  ; print(check)      # all these are very small
        
        #   this next line has the effect of replacing an essentially random vertex of the parallelogram
        #   with the center of the parallelogram.
        #   but since all 4 vertices are in the same constant plane, the improvement in the plane constant is unnecessary
        functional[idx2] = 0
        
        #time_elapsed    = gettime() - time_start
        #cat( "time_elapsed =", time_elapsed, '\n' )    # takes  <1% to 5% of total time 
        }
    
    
    #   calculate the n(n-1)/2 face centers of the unit cube,
    #   but then all translated by -0.5 and therefore centered
    #   sign(0) = 0 exactly, so 2 (or more) of the 'spectrum' coords are 0
    face_center = 0.5 * sign(functional)      #;    print( str(S01) )  exact 0 maps to 0

    center = crossprod( face_center, W )        # of the face, a parallelogram, *after* centering the zonohedron
    
    #   out$face$center = center        not really needed for inside()
    
    #   calculate the n(n-1)/2 plane constants beta
    #   these plane constants are for the centered zonohedron
    out$face$beta   = .rowSums( normal * center, nrow(normal), ncol(normal) )
    
    #   since W has rank 3, 0 is in the interior of the _centered_ zonohedron,
    #   and this implies that all n(n-1)/2 beta's are positive, except the NAs.
    #   verify this
    betamin = min( out$face$beta, na.rm=TRUE )
    if( betamin <= 0 )
        {
        log_level( FATAL, "Internal Error.  min(beta)=%g <= 0.", tol, betamin )
        return(NULL)
        }
        
    class(out)  = c( 'zonohedron', class(out) )
        
    return(out)
    }
    
    
##----------        zonohedron methods    -------------##
    
#   x   a zonohedron object
#   g   an Nx3 matrix, etc.
#
#   value   a dataframe with columns
#           within  logical
#           delta   numeric, non-positive means within

inside.zonohedron <- function( x, g )
    {
    g   = prepareNx3( g, 3 )
    if( is.null(g) )    return(NULL)
    
    #   translate g to the centered zonohedron
    gcentered   = g - matrix( x$center, nrow(g), 3, byrow=TRUE ) #; print(gcentered)
    
    hg  = tcrossprod( x$face$normal, gcentered )    #; print( str(hg) )
    
    distance    = abs(hg) - matrix( x$face$beta, nrow(hg), ncol(hg) )
    
    distance    = base::apply( distance, 2, function(z) {suppressWarnings( max(z,na.rm=TRUE) ) } )  #;   print(distance)
    
    distance[ ! is.finite(distance) ]   = NA_real_
    
    if( x$nonnegative )
        {
        #   special override for black
        black   = apply( g, 1, function(v) { isTRUE(all(v==0)) } )  #; print(black)
        if( any(black) )
            distance[black] = 0
            
        #   special override for white.  Fortunately multiplication by 0.5 and 2 preserves all precision.
        white   = apply( g, 1, function(v) { isTRUE(all(v==2*x$center)) } )  #; print(white)
        if( any(white) )
            distance[white] = 0
        }
        
    out = data.frame( within=(distance<=0) )
    rownames(out)   = rownames(g)
    
    #   out$g           = g
    out$delta   = distance

    return(out)
    }
    
    
    
    
#--------       UseMethod() calls           --------------#    
    
inside <- function( x, g )
    {
    UseMethod("inside")
    }    

    
    
#gettime <- function()
#    {
#    return( microbenchmark::get_nanotime() * 1.e-9 )
#    }
    

#   returns time in seconds, from an arbitrary origin
gettime <- function()
    {
    if( p.microbenchmark )
        return( microbenchmark::get_nanotime() * 1.e-9 )
    else
        return( as.double( base::Sys.time() ) )
    }


#   projectiveMatrix()
#
#   .matrix     invertible matrix, for example a 3x3 matrix with columns the tristimulus coordinates of RGB primaries
#   .unit       non-zero vector.  For example the tristimulus coordinates of white.
#
#   return      square matrix  B, so that
#               B = matrix  %*%  diag(a)  <=>   each column of B is a multiple of the corresponding column in .matrix
#               B %*% 1  =  .unit.      (1 is the vector of all 1s)
#
#   so for colors, B maps RGB to XYZ
#
#    Another way to write these properties:
#        B %*% I = matrix     up to multiples of the columns
#        B %*% 1  =  .unit
#   So I and 1 are the *standard* projective basis,
#   and .matrix and .unit are a different one

projectiveMatrix  <-  function( .matrix, .unit )
    {
    a   = try( solve( .matrix, .unit ), silent=TRUE )

    if( ! is.numeric(a) ) return(NULL)

    ran = range( abs(a) )   #; print(ran)

    if( ran[1] <= 1.e-6 * ran[2] ) return(NULL)

    return( .matrix  %*%  diag(a) )
    }


#   return list with A and B
ABfromHC <- function( H, C )
    {
    theta   = H * pi/50
    list( A = C*cos(theta), B = C*sin(theta) )
    }

HCfromAB <- function( A, B )
    {
    theta   = atan2( B, A )
    list( H=(theta * 50/pi) %% 100, C = sqrt( A^2 + B^2 ) )
    }

hypot<-function(a, b){
# sqrt(a^2 + b^2) without under/overflow.  Author: Jose Gama **/
# http://www.java2s.com/Tutorial/Java/0120__Development/sqrta2b2withoutunderoverflow.htm
r<-0.0
if (abs(a) > abs(b)) {
         r <- b/a
         r <- abs(a)*sqrt(1+r^2)
      } else if (b != 0) {
         r <- a/b
         r <- abs(b)*sqrt(1+r^2)
      }
r
}

###########     argument processing     ##############
#
#   A   a non-empty numeric Nx3 matrix, or something that can be converted to be one
#
#   returns such a matrix, or NULL in case of error
#

#   A   a non-empty numeric NxM matrix, or something that can be converted to be one
#
#   returns such a matrix, or NULL in case of error
#
#   This is intended to check user-supplied A, so there is a lot of checking.
#
prepareNx3  <-  function( A, M=3 )
    {
    ok  = is.numeric(A) &&  0<length(A)  &&  (length(dim(A))<=2)

    ok  = ok  &&  ifelse( is.matrix(A), ncol(A)==M, ((length(A) %% M)==0)  )

    if( ! ok )
        {
        mess    = substr( as.character(A)[1], 1, 20 )

        Aname = deparse(substitute(A))

        #   notice .topcall assignment which makes log_level() print name of parent function, and *NOT* prepareNx3()
        log_level( ERROR, "Argument '%s' must be a non-empty numeric Nx%d matrix (with N>0). %s='%s...'",
                                    Aname, M, Aname, mess, .topcall=sys.call(-1L) )
        return(NULL)
        }

    if( ! is.matrix(A) )
        A = matrix( A, ncol=M, byrow=TRUE )

    return( A )
    }


#   varname name to search for, case-insensitive
#   ...     list of variables
#
#   if varname is found in ..., returns its value
#
#   if varname is not found in ..., returns NULL
#   matching is case-insensitive

intercept   <- function( varname, ... )
    {
    theList =  list( ... )  #; print( theList )

    n   = length(theList)
    if( n == 0 )
        {
        # log_level( ERROR, "No arguments." )
        return(NULL)
        }

    # logger::log_level( logger::DEBUG, "Found %d objects in '...'", n, namespace='munsellinterpol' )

    theNames    = names(theList)

    idx = match( tolower(varname), tolower(theNames) )

    if( is.na(idx) )    return(NULL)

    return( theList[[idx]] )
    }



#################       deadwood below      #########################

prepareNx3_old  <-  function( A, M=3 )
    {
    ok  = is.numeric(A)  &&  ((length(A) %% M)==0)  &&  0<length(A)
    if( ! ok )
        {
        #print( "prepareNx3" )
        #print( sys.frames() )
        mess    = substr( as.character(A)[1], 1, 10 )
        #arglist = list( ERROR, "A must be a non-empty numeric Nx3 matrix (with N>0). A='%s...'", mess )
        #do.call( log_level, arglist, envir=parent.frame(n=3) )
        #myfun   = log_level
        #environment(myfun) = parent.frame(3)
        log_level( ERROR, "Argument A must be a non-empty numeric Nx%d matrix (with N>0). A='%s...'", M, mess )
        return(NULL)
        }

    if( is.null(dim(A)) )
        A = matrix( A, ncol=M, byrow=TRUE )
    else if( ncol(A) != M )
        A = t(A)

    return( A )
    }

#   This is for all conversions EXCEPT   HVC  <--->  xyY   transforms

library( munsellinterpol )




options( width=144 )

printf <- function( msg, ... )
    {
    mess = sprintf( msg[1], ... )    # should this really be msg[1] ?
    cat( mess, '\n' )   #, file=stderr() )
    }

#   returns time in seconds, from an arbitrary origin
gettime <- function()
    {
    if( requireNamespace('microbenchmark') )
        return( microbenchmark::get_nanotime() * 1.e-9 )
    else
        return( as.double( base::Sys.time() ) )
    }




#   testing sRGB -> HVC -> sRGB
#   testing AdobeRGB -> HVC -> AdobeRGB
#   testing Lab -> HVC -> Lab
#   testing Luv -> HVC -> Luv


testMunsell <- function()
    {
    printf( "---------------------  testMunsell()  -----------------------" )

    set.seed(0)

    # make random RGBs in the cube
    count   = 1000
    RGB = matrix( runif(3*count,min=0,max=255), ncol=3 )
    rownames(RGB)   = sprintf( "%04d", 1:count )

    printf( "testing sRGB -> HVC -> sRGB with %d samples...", count )

    time_start  = gettime()
    HVC         = sRGBtoMunsell( RGB )          #; print( str(HVC) )
    RGB.back    = MunsellTosRGB( HVC )$RGB     #; print( str(RGB.back) )
    time_elapsed    = gettime() - time_start

    mask    = is.na( RGB.back[ ,1] )
    if( any( mask ) )
        {
        printf( "There were %d  sRGB -> HVC -> sRGB  failures (out of %d).", sum(mask), count )
        return(FALSE)
        }

    delta   = rowSums( abs(RGB - RGB.back) )  #; print( max(delta) )

    printf( "sRGB  max(delta)=%g   (out of %d)",  max(delta,na.rm=TRUE), sum(is.finite(delta)) )

    tol     = 3.e-2     # was 2.e-2 before v 2.3-1
    failures = sum( tol < delta ) #; print( failures )

    if( 0 < failures )
        {
        idx = which.max(delta)
        printf( "There were %d  sRGB -> HVC -> sRGB tolerance failures (out of %d).  Max error = %g > %g.",
                    failures, count, delta[idx], tol )

        df  = data.frame( row.names=1 )
        df$RGB      = RGB[idx, ,drop=FALSE]
        df$HVC      = HVC[idx, ,drop=FALSE]
        df$RGB.back = RGB.back[idx, ,drop=FALSE]
        print( df )

        return(FALSE)
        }

    printf( "sRGB passed.  %d round trips in %g sec,  %g/sample", count, time_elapsed, time_elapsed/count )


    #   test pure black
    black       = c(0,0,0)
    black.back  = as.numeric( MunsellTosRGB( sRGBtoMunsell(black) )$RGB )
    if( ! identical( black, black.back ) )
        {
        printf( "sRGB -> HVC -> sRGB .  pure black not preserved." )
        return(FALSE)
        }


    #---   AdobeRGB  -----#

    printf( "testing AdobeRGB -> HVC -> AdobeRGB with %d samples...", count )

    time_start      = gettime()
    HVC             = RGBtoMunsell( RGB, space='AdobeRGB' )         #; print( str(HVC) )
    RGB.back        = MunsellToRGB( HVC, space='AdobeRGB' )$RGB     #; print( str(RGB.back) )
    time_elapsed    = gettime() - time_start

    mask    = is.na( RGB.back[ ,1] )
    if( any( mask ) )
        {
        printf( "There were %d  AdobeRGB -> HVC -> AdobeRGB  failures (out of %d).", sum(mask), count )
        return(FALSE)
        }

    delta   = rowSums( abs(RGB - RGB.back) )  #; print( max(delta) )

    printf( "AdobeRGB  max(delta)=%g   (out of %d)",  max(delta,na.rm=TRUE), sum(is.finite(delta)) )

    tol     = 0.5
    failures = sum( tol < delta ) #; print( failures )

    if( 0 < failures )
        {
        idx = which.max(delta)
        printf( "There were %d  AdobeRGB -> HVC -> AdobeRGB tolerance failures (out of %d).  Max error = %g > %g.",
                    failures, count, delta[idx], tol )
        df  = data.frame( row.names=1 )
        df$RGB      = RGB[idx, ,drop=FALSE]
        df$HVC      = HVC[idx, ,drop=FALSE]
        df$RGB.back = RGB.back[idx, ,drop=FALSE]
        print( df )
        return(FALSE)
        }

    printf( "AdobeRGB passed.   %d round trips in %g sec,  %g/sample",
                count, time_elapsed, time_elapsed/count )




    #---   Lab   ----#
    printf( "testing Lab -> HVC -> Lab with %d samples...", count )
    Lab             = MunsellToLab( HVC )
    time_start      = gettime()
    Lab.back        = MunsellToLab( LabToMunsell( Lab ) )
    time_elapsed    = gettime() - time_start

    if( any( is.na(Lab.back) ) )
        {
        printf( "There were Lab -> HVC  failures (out of %d).", count )
        return(FALSE)
        }

    delta   = rowSums( abs(Lab - Lab.back) )

    printf( "Lab  max(delta)=%g   (out of %d)",  max(delta,na.rm=TRUE), sum(is.finite(delta)) )

    tol = 5.e-3
    failures = sum( tol < delta, na.rm=TRUE )
    if( 0 < failures )
        {
        idx = which.max(delta)
        printf(  "There were %d  Lab -> HVC -> Lab failures (out of %d).  Max error = %g",
                    failures, sum(is.finite(delta)), delta[idx] )

        df  = data.frame( row.names=1 )
        df$Lab      = Lab[idx, ,drop=FALSE]
        df$HVC      = HVC[idx, ,drop=FALSE]
        df$Lab.back = Lab.back[idx, ,drop=FALSE]
        print( df )

        return(FALSE)
        }

    printf( "Lab passed.  %d round trips in %g sec,  %g/sample", count, time_elapsed, time_elapsed/count )


    #---   Luv   ----#
    printf( "testing Luv -> HVC -> Luv with %d samples...", count )
    Luv             = MunsellToLuv( HVC )
    time_start      = gettime()
    Luv.back        = MunsellToLuv( LuvToMunsell( Luv ) )
    time_elapsed    = gettime() - time_start

    if( any( is.na(Luv.back) ) )
        {
        printf( "There were Luv -> HVC -> Luv failures (out of %d).", count )
        return(FALSE)
        }

    delta   = rowSums( abs(Luv - Luv.back) )

    printf( "Luv  max(delta)=%g   (out of %d)",  max(delta,na.rm=TRUE), sum(is.finite(delta)) )

    tol = 5.e-3
    failures = sum( tol < delta, na.rm=TRUE )
    if( 0 < failures )
        {
        idx = which.max(delta)
        printf(  "There were %d  Luv -> HVC -> Luv failures (out of %d).  Max error = %g",
                    failures, sum(is.finite(delta)), delta[idx] )

        df  = data.frame( row.names=1 )
        df$Luv      = Luv[idx, ,drop=FALSE]
        df$HVC      = HVC[idx, ,drop=FALSE]
        df$Luv.back = Luv.back[idx, ,drop=FALSE]
        print( df )

        return(FALSE)
        }

    printf( "Luv passed.  %d round trips in %g sec,  %g/sample", count, time_elapsed, time_elapsed/count )


    return( TRUE )
    }


testColorlab <- function()
    {
    printf( "---------------------  testColorlab()  -----------------------" )

    set.seed(0)
    count   = 1000

    # make random HVCs

    H   = runif( count, min=0, max=100 )
    V   = runif( count, min=0, max=10 )
    C   = runif( count, min=1, max=6 )

    HVC = cbind( H, V, C )

    # add a few some neutrals
    HVC.n   = cbind( 0, runif(20,min=0,max=10), 0 )
    HVC = rbind( HVC, HVC.n )
    HVC = rbind( HVC, c(0,0,0) )    # add pure black

    #   integers are the worst case, so make 1/2 the data integers
    idx = 1:(count/2)
    HVC[idx, ] = round( HVC[idx, ] )

    count   = nrow(HVC)
    printf( "testing HVC -> HVCH -> HVC with %d samples...", count )

    time_start    = gettime()

    HVCH        = MunsellSpecToColorlabFormat( HVC )
    HVC.back    = ColorlabFormatToMunsellSpec( HVCH )

    #   compare HVC and HVC.back
    HVC.delta   = abs(HVC - HVC.back)
    HVC.delta[ ,1]  = pmin( HVC.delta[ ,1], 100 - HVC.delta[ ,1] )
    delta = rowSums( HVC.delta )

    printf( "Colorlab max(delta) = %g.", max(delta) )

    tol = 1.e-16
    failures = sum( tol <= delta )
    if( 0 < failures )
        {
        idx = which.max( delta )
        df  = data.frame( row.names=1 )
        df$HVC      = HVC[ idx, , drop=F ]
        df$HVCH     = HVCH[ idx, , drop=F ]
        df$HVC.back = HVC.back[ idx, , drop=F ]
        printf(  "Colorlab maximum inversion error = %g.", delta[idx] )
        print( df )
        return(FALSE)
        }

    time_elapsed    = gettime() - time_start
    printf( "passed.  %d round trips in %g sec,  %g/sample", count, time_elapsed, time_elapsed/count )


    return( TRUE )
    }

x = gettime()   # load microbenchmark


if( ! testColorlab() )  stop( "testColorlab() failed !", call.=FALSE )

if( ! testMunsell() )   { stop( "testMunsell() failed !", call.=FALSE ) }



printf(  "\nPassed all Conversion tests !" )

    
#   put fn() between timestamp and the msg    
layout_mine <- structure(
    function(level, msg, namespace="munsellinterpol",
                                    .logcall = sys.call(), .topcall = sys.call(-1), .topenv = parent.frame())
        {
        fn  = deparse1( .topcall[[1L]] )
        
        paste0( attr(level, 'level'), ' [', format(Sys.time(), "%Y-%m-%d %H:%M:%S"), '] ', namespace, "::", fn, '(). ', msg )
        },
    generator = quote(layout_mine())
)



#   stop on FATAL
appender_mine <- structure(
    function(lines)
        {
        cat(lines, file = stderr(), sep = '\n' )
        
        #   test for STOP
        if( any( grepl("^(FATAL)",lines ) )  )
            {
            stop( "Stopping in package 'munsellinterpol', because level is FATAL.", call.=FALSE )
            }
        },
    generator = quote(appender_mine())
    )

#   HVC     Nx3 matrix with HVC in the rows
#   books   comma separated strings that designate searched books
#
#   the function depends on the data.frame  MunsellBooks   which is in munsellinterpol.rda

roundHVC    <- function( HVC, books )
    {
    if( is.character(HVC) )
        #   interpret as Munsell notation
        HVC = HVCfromMunsellName(HVC)
    else
        {
        HVC = prepareNx3(HVC)
        if( is.null(HVC) )  return(NULL)
        }

    #   check validity of books
    bookvec = strsplit( books, '[ ,]+' )[[1]]   #; cat( "bookvec =", bookvec, '\n' )
    if( length(bookvec) == 0 )
        {
        log_level( ERROR, "argument books is invalid." )
        return(NULL)
        }

    #   get the full and valid book names
    #   find the "FergusonName" column of MunsellBooks
    j  = match( "FergusonName", colnames(munsellinterpol::MunsellBooks) )
    if( is.na(j) )
        {
        log_level( FATAL, "Internal Error. column 'FergusonName' cannot be found." )
        return(NULL)
        }

    name_book   = colnames(munsellinterpol::MunsellBooks)[ (j+1):ncol(munsellinterpol::MunsellBooks) ]   #; cat( "name_book =", name_book, '\n' )

    idx = pmatch( tolower(bookvec), tolower(name_book), duplicates.ok=TRUE )    #; cat( "idx =", idx, '\n' )
    if( any( is.na(idx) ) )
        {
        i   = which( is.na(idx) )[1]
        log_level( ERROR, "string '%s' in argument 'books' is invalid; it does not match any book.", bookvec[i] )
        return(NULL)
        }

    #   make mask for the valid chips
    mask    = apply( munsellinterpol::MunsellBooks[ , name_book[idx], drop=FALSE ], 1, any )

    sample_search    = munsellinterpol::MunsellBooks[ mask, ]

    log_level( INFO, "searching %d books and %d samples...", length(idx), nrow(sample_search) )

    HVCrnd              = matrix( NA_real_, nrow(HVC), ncol(HVC) )
    colnames(HVCrnd)    = c('H','V','C')

    FergusonName    = rep( NA_character_, nrow(HVC) )

    #   search for closest sample one at a time
    for( k in 1:nrow(HVC) )
        {
        if( any( is.na(HVC[k, ]) ) )    next    # HVC[k, ] is NA
        
        HVCsample   = matrix( HVC[k, ], nrow=nrow(sample_search), ncol=ncol(HVC), byrow=TRUE )
        
        #   symmetric=FALSE results in fewer hue mismatches
        dist        = NickersonColorDifference(  HVCsample, sample_search$HVC, symmetric=FALSE )
        
        #   add very small tie-breaker,  Euclidean distance in V and C/2 multiplied by very small amount, Hue is ignored
        VCsample    = cbind( HVCsample[ ,2], 0.5*HVCsample[ ,3] )
        VCsearch    = cbind( sample_search$HVC[ ,2], 0.5*sample_search$HVC[ ,3] )
        
        dist        = dist +  1.e-6 * sqrt( rowSums( (VCsample - VCsearch)^2 ) )

        if( INFO <= log_threshold( namespace="munsellinterpol" ) )
            {
            #   log the closest 5
            log_level( INFO, "least 5 distances: %s.", paste( sort(dist)[1:5], collapse=", "  ) )
            }

        i   = which.min( dist )

        HVCrnd[k, ]     = sample_search$HVC[i, ]
        FergusonName[k] = sample_search$FergusonName[i]
        }


    rnames  = rownames(HVC)
    
    if( is.null(rnames) )   rnames =  MunsellNameFromHVC( HVC )

    if( any(is.na(rnames))  ||  anyDuplicated(rnames) )
        #   rnames is no good because some are NA or duplicated !  Use trivial names instead.
        #   this should be rare
        rnames = 1:nrow(HVC)   

    out = data.frame( row.names=rnames )

    out$HVC             = HVC
    colnames(out$HVC)   = c('H','V','C')

    out[[ "ISCC-NBS Name" ]]    = ColorBlockFromMunsell( HVC )$Name

    out$MunsellRounded  = MunsellNameFromHVC( HVCrnd, digits=3 )
    out$FergusonName    = FergusonName

    return(out)
    }

#   private global variables.  The initial 'p.' means private

#   This group is stored in sysdata.rda
#   p.LookupList        HVC -> xy lookup table(s).  It must be unlocked because the xy of the neutrals (xyC) may be changed during a session.
#   p.InversionCoeffs   3D table of coefficients for approximating A and B as polynomials of a,b.  It stays locked.
#   p.xyC               4x2 matrix of xy coords.  It stays locked.
#   p.sRGB2XYZ          3x3 matrix. It stays locked.
#   p.XYZ2sRGB          3x3 matrix. It stays locked.
#   p.OptimalHull       named list of convex hulls of optimal colors, indexed by 'C', 'D65', etc.  It stays locked.
#   p.xyz1931           the CIE 2-degree color matching functions, at 5nm step. It stays locked.
#   p.ACDs              illuminants A, C, D50, D55, D65, D75, at 5nm step. It stays locked.
#   p.System_ISCCNBS    the elementary blocks for the ISCC-NBS naming system. It stays locked.
#
#   This group is programmatically created during .onLoad()
#   p.VfromY            list of 3 splinefuns; not safe to store them in sysdata.rda.  It must be unlocked.
#   p.microbenchmark    logical value, whether the package microbenchmark is loaded.  It must be unlocked.
#   p.D65toC_CAT        CAT from Illuminant D65 to C.  Used in sRGB conversion.  It must be unlocked.
#   p.CtoD65_CAT        CAT from Illuminant C to D65.  Used in sRGB conversion.  It must be unlocked.

#   These are build options
#   p.vinterpOverride   if TRUE, change from 'cubic' to 'linear' when Value < 2


p.VfromY            = NULL
p.microbenchmark    = FALSE
p.D65toC_CAT        = NULL
p.CtoD65_CAT        = NULL

p.vinterpOverride   = FALSE


.onLoad <- function( libname, pkgname )
    {
    p.microbenchmark    <<- requireNamespace( 'microbenchmark', quietly=TRUE )  #;  cat( "p.microbenchmark=", p.microbenchmark, '\n' )
    
    if( requireNamespace( "logger", quietly=FALSE ) )
        {
        #   log_formatter( formatter_mine )
        #   layout_mine and appender_mine are defined in logger.R
        log_formatter( logger::formatter_sprintf, namespace=pkgname )   # force sprintf(), even if glue is installed
        log_layout( layout_mine, namespace=pkgname )                    # put fn() between timestamp and the msg
        log_appender( appender_mine, namespace=pkgname )                # maybe stop on ERROR or FATAL
        log_threshold( WARN, namespace=pkgname )                        # default is INFO
        }

    p.VfromY            <<- makeVfromYs()    #  this loads the list p.VfromY, and takes less than 0.25 seconds

    if( requireNamespace( "spacesXYZ", quietly=TRUE ) )
        {
        #   these 2 are not needed, the variables are unlocked during .onLoad()
        #unlockBinding( "p.D65toC_CAT", asNamespace('munsellinterpol') )
        #unlockBinding( "p.CtoD65_CAT", asNamespace('munsellinterpol') )

        white.D65   = c( 0.3127, 0.3290, 1 )    # xy are from the official sRGB standard
        white.C     = c( p.xyC['NBS',], 100 )

        white.D65   = spacesXYZ::XYZfromxyY( white.D65 )
        white.C     = spacesXYZ::XYZfromxyY( white.C )

        p.D65toC_CAT    <<- spacesXYZ::CAT( white.D65, white.C, method='Bradford' )
        p.CtoD65_CAT    <<- spacesXYZ::CAT( white.C, white.D65, method='Bradford' )
        }

    #assign( "LabToMunsell", function(...) { LabtoMunsell(...) }, pos='package:munsellinterpol' )   fails !
    #namespaceExport( 'munsellinterpol', "LabToMunsell" )
    }



.onAttach <- function( libname, pkgname )
    {
    #print( libname )
    #print( pkgname )

    if( FALSE )
        {
        info    = library( help='munsellinterpol' )        #eval(pkgname)
        info    = format( info )
        mask    = grepl( "^(Version|Author|Built)", info )     #Title
        info    = gsub( "[ ]+", ' ', info[mask] )
        mess    = sprintf( "This is %s", pkgname )
        mess    = paste( c( mess, info ), collapse='.  ' )   #; cat(mess)
        packageStartupMessage( mess )
        }

    for( p in c( "spacesXYZ", "spacesRGB" ) )
        {
        if( ! requireNamespace( p, quietly=TRUE ) )
            {
            mess    = sprintf( "ERROR.  Cannot load package %s.  Many functions will not work !", p )
            packageStartupMessage( mess )
            }
        }

    
    #   surprisingly, *exporting* these 2 functions is not necessary
    #namespaceExport( 'package:munsellinterpol', "LabToMunsell" )
    
    }

################      Munsell  <-->  RGB    #######################

RGBtoMunsell <- function( RGB, space='sRGB', which='scene', maxSignal=255, adapt='Bradford', ... )
    {
    for( p in c('spacesRGB','spacesXYZ') )
        {
        if( ! requireNamespace( p, quietly=TRUE ) )
            {
            log_level( ERROR, "required package '%s' could not be loaded.", p )
            return(NULL)
            }
        }
    
    #theSpace = spacesRGB::getRGB(space,full=FALSE)
    #if( is.null(theSpace) ) return(NULL)
    
    RGB = prepareNx3( RGB )
    if( is.null(RGB) )  return(NULL)    
    
    # Convert RGB input into CIE XYZ and xyY coordinates
    XYZ = spacesRGB::XYZfromRGB( RGB, space=space, which=which, maxSignal=maxSignal )$XYZ  # for white, Y=100
    if( is.null(XYZ) )  return(NULL)
    
    #   xyY = spacesXYZ::xyYfromXYZ(XYZ)

    # adapt xyY from the RGB space white to C
    white       = spacesRGB::getWhiteXYZ( space, which=which )  # theSpace$whiteXYZ
    
    white.C     = spacesXYZ::XYZfromxyY( c( p.xyC['NBS',], 100 ) )

    theCAT      = spacesXYZ::CAT( white, white.C, method=adapt )
    
    #   xyY.adapted = spacesXYZ::adaptxyY( theCAT, xyY )   #; print( xyY.adapted )

    XYZ.adapted = spacesXYZ::adaptXYZ( theCAT, XYZ ) #; print( xyY.adapted )
    
    # Convert adapted xyY coordinates to Munsell coordinates
    #tmp = xyYtoMunsell( xyY.adapted, ... )
    #HVC = tmp$HVC
    #rownames(HVC) = tmp$SAMPLE_NAME
    
    # Convert adapted XYZ coordinates to Munsell coordinates
    HVC = XYZtoMunsell( XYZ.adapted, ... )
    if( is.null(HVC) )  return(NULL)

    #   do an additional test for neutrals
    #   to correct possible roundoff error
    neutral = RGB[ ,1]==RGB[ ,2]    &    RGB[ ,2]==RGB[ ,3]

    neutral[ is.na(neutral) ]   = FALSE

    if( any(neutral) )
        {
        # when both R==G==B, then set H=0 and C=0
        HVC[neutral,c(1,3)]     = 0

        if( is.null(rownames(RGB)) )
            #   RGB has no rownames, which means that rownames(HVC) is MunsellName
            rownames(HVC)[neutral]  = MunsellNameFromHVC( HVC[neutral, ] )
        }
    
    return( HVC )
    }

MunsellToRGB <- function( MunsellSpec, space='sRGB', which='scene', maxSignal=255, adapt='Bradford', ... )
    {
    for( p in c('spacesRGB','spacesXYZ') )
        {
        if( ! requireNamespace( p, quietly=TRUE ) )
            {
            log_level( ERROR, "required package '%s' could not be loaded.", p )
            return(NULL)
            }
        }
    
    #theSpace = spacesRGB::getRGB(space,full=FALSE)
    #if( is.null(theSpace) ) return(NULL)
        
    # Convert from Munsell notation to CIE coordiantes
    tmp = MunsellToxyY( MunsellSpec, ... )
    if( is.null(tmp) )  return(NULL)

    out = tmp
    
    #   look for chroma==0 exactly
    neutral = out$HVC[ ,3] == 0
    neutral[ is.na(neutral) ]   = FALSE    
    
    out$HVC = NULL  # erase this column

    #xyY = tmp$xyY
    
    XYZ = spacesXYZ::XYZfromxyY( out$xyY )
    
    # adapt xyY from C to the RGB space white
    white.C =   spacesXYZ::XYZfromxyY( c( p.xyC['NBS',], 100 ) )    
    
    white   =   spacesRGB::getWhiteXYZ( space, which=which )     #theSpace$whiteXYZ
    if( is.null(white) )    return(NULL)
    
    theCAT      = spacesXYZ::CAT( white.C, white, method=adapt )
    
    XYZ.adapted = spacesXYZ::adaptXYZ( theCAT, XYZ )

    # XYZ.adapted = spacesXYZ::XYZfromxyY( xyY.adapted )

    # Convert CIE XYZ coordinates to RGB coordinates
    out = cbind( out, spacesRGB::RGBfromXYZ( XYZ.adapted, space=space, which=which, maxSignal=maxSignal ) )

    #rnames  = tmp$SAMPLE_NAME
    #if( anyDuplicated(rnames) ) rnames = 1:nrow(out)
    #rownames(out) = rnames
    
    if( any(neutral) )
        {
        RGB.mean = rowMeans( out$RGB[neutral, , drop=FALSE] )
        
        #   the vector on the right side is replicated over all columns on the left        
        out$RGB[neutral, ]  =  RGB.mean  # cbind( RGB.mean, RGB.mean, RGB.mean )
        }
    
    return( out )
    }

#   YfromV()
#
#   V       a numeric vector of Munsell Values
#           Each V should be in the interval [0,11]
#
#   return  vector of computed Ys, with NA wherever V is invalid
#           this is absolute reflectance Y, as a percentage
#           When V=10, Y=100 exactly.  This is the ASTM-D1535 convention.
#           When V=11, Y=128.7129 approximately
#
#   author: Glenn Davis

YfromV <- function( V, which='ASTM' )
    {
    if( ! missing(which) )
        {
        w  = pmatchYV( which )

        if( is.na(w) )
            {
            log_level( WARN, "which='%s' is invalid.", which )
            return( rep(NA_real_,length(V) ) )
            }
        which   = w
        }

    if( which == 'ASTM' )
        out = ( ((((81939*V - 2048400)*V  + 23352000)*V - 22533000)*V + 119140000)*V  ) / 1.e8
    else if( which=='OSA'  ||  which=='MGO' )
        {
        out = ((((8404*V - 210090)*V  + 2395100)*V  - 2311100)*V + 12219000)*V

        if( which == 'OSA' )
            out = out / 10256800
        else
            out = out / 1.e7
        }
    else if( which == "MUNSELL" )
        {
        b2  = 1474^2
        ac4 = pmin( 4*4740*V^2, b2 )
        out =  50 * ( (1474 - sqrt(b2 - ac4)) / 474 )
        }
    else if( which == "PRIEST" )
        out = V^2
    else
        {
        log_level( FATAL, "Internal Error. which='%s' is invalid.", as.character(which) )
        return(NULL)
        }

    return( out )
    }


#   VfromY()
#
#   Y   numeric vector of reflectances
#
#   make a spline function with a large number of 'knots'
#
VfromY  <- function( Y, which='ASTM' )
    {
    if( ! missing(which) )
        {
        w  = pmatchYV( which )
        if( is.na(w) )
            {
            log_level( WARN, "which='%s' is invalid.", which )
            return( rep(NA_real_,length(Y) ) )
            }
        which   = w
        }


    if( which %in% names(p.VfromY) )
        out = p.VfromY[[which]](Y)
    else if( which == "MUNSELL" )
        out = sqrt( 1.474*Y - 0.00474*Y^2 )
    else if( which == "PRIEST" )
        out = sqrt(Y)
    else
        {
        log_level( FATAL, "Internal Error. which='%s' is invalid.", as.character(which) )
        return(NULL)
        }

    return( out )
    }



makeVfromYs <- function()
    {
    whichvec = c( 'ASTM', 'OSA', 'MGO' )

    #log_level( INFO, "Making %d splinefuns...", length(whichvec) )
    #time_start  = gettime()

    out = list()

    #   these lookup Vs derived by some experimentation - see test-VandY.R for the number of digits of accuracy
    V1  = 0.025 * (-8:120)   # note that 0:3 are in V1 (this is important so that 0 -> 0).    Old sequence was seq(-0.2,2.98,len=121)
    V2  = seq( 3^(1/2), 10.5^(1/2), len=181 ) ^ (2)
    V2  = V2[-1]    # drop number approximately 3, since 3 is already in V1
    V   = sort( c( V1, V2, 4:10 ) )   # ensure that integers 1:10 are in V; these are the grid-point planes

    for( w in whichvec )
        {
        #mess    = sprintf("makeVfromYs().  DEBUG.  Making p.VfromY() for '%s'  (%d Values)...\n", w, length(V) )
        #cat( mess, file=stderr() )
        out[[w]]    = splinefun( YfromV(V,which=w), V, method='fmm' )
        }

    #log_level( INFO, "done.  [in %g sec]\n", gettime()-time_start )   # less than 0.25 seconds

    return( out )
    }



pmatchYV <- function( which )
    {
    full    = c( 'ASTM', 'OSA', 'MGO', 'MUNSELL', 'PRIEST' )

    idx = pmatch( toupper(which), full )

    if( is.na(idx) )    return( NA_character_ )

    return( full[idx] )
    }

hogs <- function( iPos=1 )
    {
    theNames    = ls(iPos)
    
    theList = sapply( theNames, function(x) get(x, pos=iPos )  )

    n = length(theList)    
    
    if( n == 0 ) { return(NULL) }    
    
    class1  <- function( x )    { return( class(x)[1] ) }
    
    out     = data.frame( name=theNames, size=sapply( theList, object.size ), 
                                mode=sapply(theList,mode),  class=sapply(theList,class1),  
                                stringsAsFactors=F )

    perm    = order( out$size )
    
    out     = out[ perm, ]
    
    out     = rbind( out, data.frame(name="Total:", size=sum(out$size), mode=NA, class=NA ) )
    
    #   row.names( out )    = theNames
    
    #z[ n+1 ] = sum(z)
    
    #names(z)[n+1] = "Total:"
    
    return( out )
    }
    
removeFunctions  <-  function( iPos=1, .exceptions=c("removeFunctions","hogs") )
    {
    theHogs = hogs()
    
    if( is.null(theHogs) )  return(FALSE)
    
    #   print( theHogs )
     
    df_sub  = subset( theHogs, mode=='function' )
    
    df_sub  = df_sub[ order(df_sub$name), ]
    
    #   print( df_sub )
    
    idx = match( .exceptions, df_sub$name )
    if( 0 < length(idx) )
        #   do not remove these
        df_sub = df_sub[ -idx, ]
        
    #	print( df_sub )

    n   = nrow( df_sub )
    
    if( n == 0 )
        {
        cat( "No functions to remove !\n" )
        return(FALSE)
        }
        
    mess    = sprintf( "About to remove %d functions...\n", n )
    cat( mess )
    
    print( df_sub$name )
    
    keydown <- function(key) { return( key )}

    
    key = readline( prompt="Proceed with removal ?  [Y for Yes]" )
    
    #   print(key)
    
    ok  = toupper(key) == "Y"
    
    if( ! ok )  return(FALSE)
    
    rm( list=df_sub$name, pos=iPos )
    
    return(TRUE)
    }
    

#   this file has functions related to the Munsell books,
#   and intended to be called from savePrivateDatasets()



#   this function reads 3 files and merges them into a single data.frame
#
#   returns a data.frame with 9 columns:
#       HVC             1753x3 matrix, with unique rows, sorted by H,V,C in that order. Neutrals get H=0 and come first.
#       FergusonName    the Ferguson color name; only defined for the first 5 books, and otherwise NA
#       bead            logical, present in the bead book
#       newstudent      <same>
#       plant           <same>
#       rock            <same>
#       soil            <same>
#       glossy          <same>
#       matte           <same>

readAllBooks <- function()
    {
    if( ! requireNamespace('zonohedra') )
        {
        cat( "Cannot load package zonohedra.\n" )
        return(NULL)
        }
    
    books5  = readBooks5()      #; print( str(books5) )

    glossyHVC   = readGlossyBook()

    matteHVC    = readMatteBook()

    ok  = ! is.null(books5)  &&  ! is.null(glossy)  &&  ! is.null(matte)
    if( ! ok )  return(NULL)


    #   add columns to books5 so all 3 can be rbound
    books5$glossy   = FALSE
    books5$matte    = FALSE

    #   add columns to glossy so all 3 can be rbound
    glossy  = data.frame( row.names=1:nrow(glossyHVC) )
    glossy$HVC          = glossyHVC
    glossy$FergusonName = NA_character_
    glossy$bead         = FALSE
    glossy$newstudent   = FALSE
    glossy$plant        = FALSE
    glossy$rock         = FALSE
    glossy$soil         = FALSE
    glossy$glossy       = TRUE      # the only one we truly know !
    glossy$matte        = FALSE


    #   add columns to matte so all 3 can be rbound
    matte  = data.frame( row.names=1:nrow(matteHVC) )
    matte$HVC           = matteHVC
    matte$FergusonName  = NA_character_
    matte$bead          = FALSE
    matte$newstudent    = FALSE
    matte$plant         = FALSE
    matte$rock          = FALSE
    matte$soil          = FALSE
    matte$glossy        = FALSE
    matte$matte         = TRUE      # the only one we truly know !

    df  = rbind( books5, glossy, matte )

    #   sort by H, V, C, in that order
    perm    = order( df$HVC[ ,1], df$HVC[ ,2], df$HVC[ ,3] )

    df     = df[ perm, ]

    #   now find duplicates and merge the duplicate rows. This is the hard part.

    dupvec  = zonohedra::grpDuplicated( df$HVC, MARGIN=1L )

    if( max(dupvec) == 0 )
        {
        #   there are no duplicates, all HVCs are singletons, so we are done
        return( df )
        }

    n   = length(dupvec)        # this is the number of rows, with duplicates

    #  Build output
    count   = attr( dupvec, "ngroups" )[1]       # total rows in output, with no duplicates

    #   idxlist maps from output index to list of input indexes
    idxlist = vector( count, mode='list' )

    #   i indexes the input data.frame df, and the increment is 1 or more
    #   k indexes the output data.frame out, and the increment is always 1
    i   = 1L
    for( k in 1:count )
        {
        if( dupvec[i] == 0 )
            #   just a singleton
            idxlist[[k]]    = i
        else
            idxlist[[k]]    = which( dupvec == dupvec[i] )

        i   = i + length( idxlist[[k]] )
        }

    #   sanity check, i should be n+1L
    if( i != n+1L )
        {
        mess    = sprintf( "ERR. index i=%d, but expected %d.\n", i, n+1L )
        cat( mess )
        return(NULL)
        }


    # print( dupvec[ 1:100 ] )

    #return( idxlist[ (count-10):count ] )

    #   allocate output variable, just to hold results, all will be overwritten !
    out = df[ 1:count, ]
    rownames(out)       = NULL

    out$HVC[ ]  = NA_real_  #  make sure all is overwritten
    rownames(out$HVC)   = NULL

    #   fill in rows of out, one at a time
    for( k in 1:count )
        {
        idx = idxlist[[k]]  # index into df

        if( length(idx) == 1 )
            #   a singleton, so just copy row from df
            out[k, ]    = df[idx, ]
        else
            {
            #   do real work
            dfrow   = data.frame( row.names=idx[1] )    #   make data.frame with 1 row, and the same columns as df

            dfrow$HVC           = df$HVC[ idx[1], , drop=FALSE]     # the HVC are all the same, so just use the first one, idx[1]

            #   find a FergusonName that is *not* NA
            namevec = df$FergusonName[idx]
            j       = which( ! is.na(namevec) )
            if( 0 < length(j) )
                dfrow$FergusonName  = namevec[ j[1] ]       #   actually length(j) should be either 0 or 1
            else
                dfrow$FergusonName  = NA_character_

            if( TRUE )
                {
                #   the last 7 columns are all logical and computed the same way, so use a loop
                for( j in (ncol(dfrow)+1L):ncol(df) )
                    {
                    thename = names(df)[j]
                    dfrow[[ thename ]] = any( df[[ thename ]][idx] )
                    }
                }
            else
                {
                dfrow$bead          = any( df$bead[idx] )
                dfrow$newstudent    = any( df$newstudent[idx] )
                dfrow$plant         = any( df$plant[idx] )
                dfrow$rock          = any( df$rock[idx] )
                dfrow$soil          = any( df$soil[idx] )
                dfrow$glossy        = any( df$glossy[idx] )
                dfrow$matte         = any( df$matte[idx] )
                }

            out[k, ]    = dfrow     # overwrite row k in output
            }
        }
        
    #   check that the rows of out$HVC are unique
    dupvec  = zonohedra::grpDuplicated( out$HVC, MARGIN=1L )

    if( 0 < max(dupvec) )
        {
        mess    = sprintf( "ERR. readAllBooks(). output HVC rows are not unique; there are %d non-singletons.\n",
                                max(dupvec) )
        cat( mess )
        return(NULL)
        }


    neutral = (out$HVC[ ,3] == 0)

    rownames( out )[ neutral ] = MunsellNameFromHVC( out$HVC[neutral, ], format='f', digits=2 )

    rownames( out )[ ! neutral ] = MunsellNameFromHVC( out$HVC[ !neutral, ] )

    #   check that all rownames are not NA
    if( any( is.na(rownames(out)) ) )
        {
        mess    = sprintf( "ERR. readAllBooks(). %d row names are NA.\n", sum( is.na(rownames(out)) ))
        cat( mess )
        return(NULL)
        }

    return( out )
    }


#   read data on the 5 Ferguson books

readBooks5   <- function( path="../inst/extdata/Supplement1_3.6.2024.csv" )
    {
    df  = read.table( path, sep='\t', row.names=NULL, col.names=c("Munsell","FergusonName","books"), fill=TRUE, na.strings='' )
    if( is.null(df) )
        {
        cat( "Cannot read", path, '\n', file=stderr() )
        return(NULL)
        }


    HVC = HVCfromMunsellName( df$Munsell )

    valid   = ! is.na( HVC[ ,2] )

    #   take out the non-valid rows
    row.names   = df$Munsell[valid]

    df  = df[ valid, ]
    rownames(df)    = row.names

    #   replace the first column with the matrix HVC
    df[[1]] = HVC[ valid, ]
    colnames(df)[1] = "HVC"

    #print( df[1:20, ] ) ; print( str(df) )

    #   replace the column 'books' with 5 distinct logical columns
    book    = c( B="bead", N="newstudent", P="plant", R="rock", S="soil" )

    for( nam in names(book) )
        {
        df[[ book[nam] ]] = grepl( nam, df$books )
        }

    #   remove the 'books' column, no longer needed
    df$books    = NULL

    #print( df[1:20, ] ) ; print( str(df) )

    return( invisible(df) )
    }



#   read 1269 chromatic colors
#   and then prepend 32 achromatic neutral colors

readMatteBook <- function( path="../inst/extdata/mattebook-Joensuu.txt" )
    {
    if( ! requireNamespace('munsellinterpol') )
        {
        cat( "Cannot load munsellinterpol.\n" )
        return(NULL)
        }

    linevec = readLines( path )

    #   get the number of colors from the last line
    colors  = sub( "^[ ]+([0-9]+).*", "\\1", linevec[length(linevec)] )  #;  print( colors )
    colors  = as.integer(colors)

    if( ! is.finite(colors)  ||  colors==0 )
        {
        cat( "Cannot determine chromatic color count, from last line.\n" )
        cat( linevec[length(linevec)], '\n' )
        return(NULL)
        }

    linevec = substring( linevec, 16 )  #   ignore everything after column 16

    #   find the line ending in "RED"
    mask    = grepl( "RED$", linevec )
    if( all( ! mask ) )
        {
        cat( "Cannot find start of data; i.e. the line ending in 'RED'.\n" )
        return(NULL)
        }

    linevec = linevec[ (which(mask)[1]+1):length(linevec) ]

    out = munsellinterpol::HVCfromMunsellName( linevec )

    #   remove any NAs
    mask    = is.finite( out[ ,1] )
    out     = out[mask, ]

    #   check length
    if( nrow(out) != colors )
        {
        mess    = sprintf( "ERR. Read %d chromatic colors, but expected %d.\n", nrow(out), colors )
        cat( mess )
        return(NULL)
        }

    #   cleanup the row names
    rownames(out)   = munsellinterpol::MunsellNameFromHVC( out )

    #   prepend 32 achromatic neutrals
    neutral             = cbind( 0, seq(1.75,9.5,by=0.25), 0 )
    rownames(neutral)   = munsellinterpol::MunsellNameFromHVC(neutral, format='f', digits=2 )

    out                 = rbind( neutral, out )

    return( out )
    }



#   read 1485 chromatic colors
#   and then prepend 37 achromatic neutral colors

readGlossyBook <- function( path="../inst/extdata/glossybook-Centore.txt" )
    {
    if( ! requireNamespace('munsellinterpol') )
        {
        cat( "Cannot load munsellinterpol.\n" )
        return(NULL)
        }

    df = read.csv( path, quote='', comment.char='#' )

    #   check length
    count_expected  = 1485L
    if( nrow(df) != count_expected )
        {
        mess    = sprintf( "ERR. Read %d chromatic colors, but expected %d.\n", nrow(df), count_expected )
        cat( mess )
        return(NULL)
        }

    out = munsellinterpol::HVCfromMunsellName( df$Name )

    #   check for NAs
    mask    = is.na( out[ ,2] )
    if( any(mask) )
        {
        mess    = sprintf( "ERR. Found %d NAs, out of %d names.\n", sum(mask), length(mask) )
        cat( mess )
        return(NULL)
        }

    #   prepend 37 achromatic neutrals
    neutral             = cbind( 0, seq(0.5,9.5,by=0.25), 0 )
    rownames(neutral)   = munsellinterpol::MunsellNameFromHVC( neutral, format='f', digits=2 )

    out                 = rbind( neutral, out )

    return( out )
    }


readGlossyNotations <- function( path="G:/color/historical/Munsell/SpectralReflectancesOf2007MunsellBookOfColorGlossy.txt" )
    {
    linevec = readLines( path )

    #   find the line starting with "Name,"
    mask    = grepl( "^Name,", linevec )
    if( all( ! mask ) )
        {
        cat( "Cannot find start of data; i.e. the line starting with 'Name,'.\n" )
        return(NULL)
        }

    linevec = linevec[ (which(mask)[1]+1):length(linevec) ]

    out = sub( "^([.0-9A-Z/]+).+", "\\1", linevec )

    return( out )
    }



#   HVC     Nx3 matrix with HVC in the rows. These should be unique and taken from a Munsell book.
#   check   check that there are no collisions
#
#   returns a numeric vector

hashHVC <- function( HVC, check=TRUE )
    {
    out = HVC %*% c( 8, 1, 810)

    dim(out) = NULL

    if( check  &&  length(unique(out)) != nrow(HVC) )
        {
        count   = length(unique(out))
        mess    = sprintf( "ERR. hashHVC().  Found %d collisions = %d - %d.\n",
                                nrow(HVC)-count, nrow(HVC), count )
        cat( mess )
        return(NULL)
        }

    names(out)  = rownames(HVC)

    return( out )
    }

nameddiff <- function( x, y )
    {
    mask = x %in% y

    return( x[ !mask ] )
    }

allcomparisons <- function( booklist )
    {
    n   = length(booklist)

    pair    = arrangements::combinations( n, 2 )

    for( k in 1:nrow(pair) )
        {
        j1  = pair[k,1]
        j2  = pair[k,2]

        book1   = booklist[[ j1 ]]
        book2   = booklist[[ j2 ]]

        if( nrow(book1) < nrow(book2) )
            {
            #   book1 is smaller
            cat( names(booklist)[j1], " - ", names(booklist)[j2],  ":  ", length( nameddiff(hashHVC(book1),hashHVC(book2)) ), '\n' )
            }
        else
            {
            #   book2 is smaller
            cat( names(booklist)[j2], " - ", names(booklist)[j1],  ":  ", length( nameddiff(hashHVC(book2),hashHVC(book1)) ), '\n' )
            }
        }

    return( invisible(TRUE) )
    }

#   tests xyYtoMunsell() and MunsellToxyY() over special colors

#   tests sRGBtoMunsell() and MunsellTosRGB() over special colors


library( munsellinterpol )

options( width=144 )

printf <- function( msg, ... )
    {
    mess = sprintf( msg[1], ... )    # should this really be msg[1] ?
    cat( mess, '\n' )   #, file=stderr() )
    }

#   returns time in seconds, from an arbitrary origin
gettime <- function()
    {
    if( requireNamespace('microbenchmark') )
        return( microbenchmark::get_nanotime() * 1.e-9 )
    else
        return( as.double( base::Sys.time() ) )
    }


testCentroids <- function()
    {
    #   path = "../inst/extdata/Centrals_ISCC_NBS.txt"
    path = system.file( "extdata/Centroids_ISCC-NBS.txt", package='munsellinterpol' )

    df = read.table( path, header=T,  sep='\t', stringsAsFactors=F )

    printf( "-----------  testCentroids()   %d samples -----------------------", nrow(df) )

    #   printf( "Testing centroids..." )

    # go forwards HVC -> xyY

    HVC = HVCfromMunsellName( df$MunsellSpec )

    time_start  = gettime()

    res = MunsellToxyY( HVC, warn=FALSE )

    time_elapsed = gettime() - time_start

    failures = sum( is.na(res$xyY[ ,1]) )

    printf(  "There were %d forward failures, out of %d samples.  time=%g sec.  %g/sample.",
                failures, nrow(df), time_elapsed, time_elapsed/nrow(HVC) )


    if( 0 < failures )  return(FALSE)

    #   now go backwards xyY -> HVC
    time_start  = gettime()

    res = xyYtoMunsell( res$xyY, perf=TRUE )

    time_elapsed = gettime() - time_start

    failures = sum( is.na(res$HVC[ ,1]) )

    printf(  "There were %d reverse failures, out of %d samples.  time=%g sec.  %g/sample.",
                            failures, nrow(HVC), time_elapsed, time_elapsed/nrow(HVC) )

    printf( "Performance:")
    cat( "Time Elapsed 5-number summary:  " )
    cat( fivenum(res$time.elapsed,na.rm=T), '\n' )
    cat( "Iterations 5-number summary:  " )
    cat( fivenum(res$iterations,na.rm=T), '\n' )
    cat( "Evaluations 5-number summary:  " )
    cat( fivenum(res$evaluations,na.rm=T), '\n' )

    #   compare HVC and res$HVC
    tol = 0.01
    HVC.delta   = abs(HVC - res$HVC)
    HVC.delta[ ,1]  = pmin( HVC.delta[ ,1], 100 - HVC.delta[ ,1] )
    err = rowSums( HVC.delta )
    idx = which.max( err )
    df  = data.frame( row.names=1 )
    df$HVC      = HVC[ idx, , drop=F ]
    df$xyY      = res$xyY[ idx, , drop=F ]
    df$HVC.back = res$HVC[ idx, , drop=F ]
    printf(  "Maximum round-trip error = %g  (tol=%g).", err[idx], tol )
    print( df )

    #   how many exceeded the tolerance
    count   = sum( tol < err )
    if( 0 < count )
        printf( "%d round-trip errors exceeded %g.", count, tol )

    if( 0 < failures  ||  0 < count )  return(FALSE)

    return( TRUE )
    }


testOptimals <- function()
    {
    path = system.file( "extdata/OptimalColorsForIlluminantC.txt", package='munsellinterpol' )

    df = read.table( path, header=T )

    printf( "---------  testOptimals()  %d samples -----------------------", nrow(df) )

    xyY = as.matrix( df )

    osname  = Sys.info()[ 'sysname' ]
    solaris = grepl( 'sun', osname, ignore.case=TRUE )

    out = TRUE

    for( hcinterp in c('bilin','bicub') )
        {
        for( vinterp in c('lin','cub') )
            {
            #   go backwards xyY -> HVC
            printf( "-----  hcinterp='%s',  vinterp='%s'  -----", hcinterp, vinterp )

            time_start  = gettime()

            res = xyYtoMunsell( xyY, warn=FALSE, perf=TRUE, hcinterp=hcinterp, vinterp=vinterp )

            time_elapsed = gettime() - time_start

            #   print( res )
            #   return(FALSE)

            HVC = res$HVC

            mask    = is.na( HVC[ ,1] )

            failures = sum( mask )

            printf( "testOptimals(). There were %d inversion failures, out of %d samples.  time=%g sec.  %g/sample.  OS=%s",
                                failures, nrow(xyY), time_elapsed, time_elapsed/nrow(xyY), osname )

            #   in 2018, solaris started to give 5 iteration failures in CRAN testing
            #   in 2019 (R v 3.6.1), the same thing happened with fedora (both clang and gcc).
            #   So for safety in the future, just set limit to 5 !
            limit   = 5     # ifelse( solaris, 5, 0 )   # with solaris, 5 errors were detected in CRAN testing

            if( limit < failures )
                {
                colnames(xyY)   = c('x','y','Y')
                colnames(HVC)   = c('H','V','C')
                df  = data.frame( row.names=1:failures )
                df$xyY  = xyY[mask, , drop=FALSE]
                df$HVC  = HVC[mask, , drop=FALSE]
                print( df )

                #print( res )
                out = FALSE
                }

            printf( "Performance:")
            cat( "Time Elapsed 5-number summary:  " )
            cat( fivenum(res$time.elapsed,na.rm=T), '\n' )
            cat( "Iterations 5-number summary:  " )
            cat( fivenum(res$iterations,na.rm=T), '\n' )
            cat( "Evaluations 5-number summary:  " )
            cat( fivenum(res$evaluations,na.rm=T), '\n' )

            if( max( res$iterations, na.rm=T ) == 100 )
                {
                idx =  which( res$iterations == 100 )   # ; print(idx)
                printf( "There were %d samples that failed to converge.", length(idx) )
                df  = res[ idx, , drop=FALSE ]
                print(df)
                }

            if( out )
                {
                #   do round-trip
                xyY.back    = MunsellToxyY( HVC, warn=FALSE,  hcinterp=hcinterp, vinterp=vinterp )$xyY

                delta   = rowSums( abs(xyY - xyY.back) )

                #   ignore pure black
                mask        = xyY[ ,3]==0  &  xyY.back[ ,3]==0
                delta[mask] = 0

                cat( "Round trip xyY -> HVC -> xyY  inversion error 5-number summary:  " )
                cat( fivenum(delta), '\n' )

                idx = which.max( delta )
                df  = data.frame( row.names=1 )
                df$xyY      = xyY[ idx, , drop=F ]
                df$HVC      = HVC[ idx, , drop=F ]
                df$xyY.back = xyY.back[ idx, , drop=F ]
                printf(  "Maximum inversion error occurs for this one:" )
                print( df )
                }
            }
        }

    return( out )
    }



testReals <- function()
    {
    df.real     = subset( Munsell2xy, real==TRUE )

    printf( "----------------  testReals()  %d samples -----------------------", nrow(df.real) )

    HVC         = cbind( df.real$H, df.real$V, df.real$C )
    xyY         = cbind( df.real$x, df.real$y, YfromV(df.real$V) )

    #   go backwards xyY -> HVC
    time_start  = gettime()

    res = xyYtoMunsell( xyY, warn=FALSE, perf=TRUE )

    time_elapsed = gettime() - time_start

    mask    = is.na( res$HVC[ ,1] )

    failures = sum( mask )

    printf( "testReals().  There were %d inversion failures, out of %d samples.  time=%g sec.  %g/sample.",
                        failures, nrow(xyY), time_elapsed, time_elapsed/nrow(xyY) )


    if( 0 < failures )
        {
        colnames(HVC)   = c('H','V','C')
        colnames(xyY)   = c('x','y','Y')
        df  = data.frame( row.names=MunsellNameFromHVC( HVC[mask, ] ) )
        df$HVC  = HVC[mask, ]
        df$xyY  = xyY[mask, ]
        print( df )
        }

    printf( "Performance:")
    cat( "Time Elapsed 5-number summary:  " )
    cat( fivenum(res$time.elapsed,na.rm=T), '\n' )
    cat( "Iterations 5-number summary:  " )
    cat( fivenum(res$iterations,na.rm=T), '\n' )
    cat( "Evaluations 5-number summary:  " )
    cat( fivenum(res$evaluations,na.rm=T), '\n' )



    #   compare HVC and res$HVC
    HVC.delta   = abs(HVC - res$HVC)
    HVC.delta[ ,1]  = pmin( HVC.delta[ ,1], 100 - HVC.delta[ ,1] )
    delta   = rowSums( HVC.delta )
    cat( "Round trip HVC -> xyY -> HVC  inversion error 5-number summary:  " )
    cat( fivenum(delta), '\n' )
    idx = which.max( delta )
    df  = data.frame( row.names=1 )
    df$HVC      = HVC[ idx, , drop=F ]
    df$xyY      = xyY[ idx, , drop=F ]
    df$HVC.back = res$HVC[ idx, , drop=F ]
    printf(  "Maximum inversion error occurs for this one:" )
    print( df )


    return( failures == 0 )
    }


testNeutrals <- function()
    {
    #   check that Chroma is small, and finite
    RGB = matrix( 0:255, 256, 3 )
    colnames(RGB)   = c('R','G','B')

    printf( "-----------  testNeutrals()   %d samples  -----------------------", nrow(RGB) )

    HVC = sRGBtoMunsell( RGB )

    chroma      = HVC[ ,3]
    mask        = is.na(chroma)
    failures    = sum(mask)
    if( 0 < failures )
        {
        printf( "testNeutrals().  There were %d failures to convert.", failures )
        df  = data.frame( row.names=which(mask) )
        df$RGB  = RGB[mask, ]
        print( df )
        }

    tol = 0     # 1.e-13    changed to 0 in v 3.0-0
    mask    = (is.finite(chroma) & tol < chroma)
    count   = sum( mask  )
    if( 0 < count )
        {
        printf( "testNeutrals().  There were %d chromas > %g.", count, tol )
        df  = data.frame( row.names=which(mask) )
        df$RGB  = RGB[mask, ]
        df$Chroma   = chroma[mask]
        print( df )
        }

    #   now go back
    res = MunsellTosRGB( HVC )

    RGB.delta   = abs(RGB - res$RGB)
    delta   = rowSums( RGB.delta )
    cat( "Round-trip error  (RGB -> HVC -> RGB)  5-number summary:  " )
    cat( fivenum(delta), '\n' )

    return( failures==0  &&  count==0 )
    }


testNearNeutrals <- function()
    {
    #   check that Chroma is small, and finite
    RGB = matrix( 0, 0, 3 )
    colnames(RGB)   = c('R','G','B')

    for( k in 10:255 )
        {
        RGB = rbind( RGB, matrix(k,3,3) - diag(3) )
        }

    printf( "-------  testNearNeutrals()  %d samples   -----------------------", nrow(RGB) )


    HVC = sRGBtoMunsell( RGB )

    chroma      = HVC[ ,3]
    mask        = is.na(chroma)
    failures    = sum(mask)
    if( 0 < failures )
        {
        printf( "testNearNeutrals().  There were %d failures to convert.", failures )
        df  = data.frame( row.names=which(mask) )
        df$RGB  = RGB[mask, ]
        print( df )
        return(FALSE)
        }

    tol = 0.30
    mask    = (is.finite(chroma) & tol < chroma)
    count   = sum( mask  )
    if( 0 < count )
        {
        printf( "testNearNeutrals().  There were %d chromas > %g.", count, tol )
        df  = data.frame( row.names=which(mask) )
        df$RGB      = RGB[mask, ]
        df$Chroma   = chroma[mask]
        print( df )
        return(FALSE)
        }


    #   now go back
    res         = MunsellTosRGB( HVC )
    RGB.delta   = abs(RGB - res$RGB)
    delta   = rowSums( RGB.delta )
    cat( "Round-trip error  (RGB->HVC->RGB)  5-number summary:  " )
    cat( fivenum(delta), '\n' )

    tol = 0.005
    if( tol < max(delta) )
        {
        printf( "Maximum round-trip error = %g > %g.", max(delta), tol )
        return(FALSE)
        }

    return( TRUE )
    }

testDarks <- function()
    {
    rgbmax  = 8  # 10    # 5

    RGB = as.matrix( expand.grid( R=0:rgbmax, G=0:rgbmax, B=0:rgbmax ) )

    printf( "---------------------  testDarks() %d samples   -----------------------", nrow(RGB) )

    HVC = sRGBtoMunsell( RGB )

    chroma      = HVC[ ,3]
    mask        = is.na(chroma)
    failures    = sum(mask)
    if( 0 < failures )
        {
        printf( "testDarks().  There were %d failures to convert.", failures )
        df  = data.frame( row.names=which(mask) )
        df$RGB  = RGB[mask, , drop=FALSE ]
        print( df )
        return(FALSE)
        }

    #   now go back
    res         = MunsellTosRGB( HVC )
    RGB.delta   = abs(RGB - res$RGB)
    delta   = rowSums( RGB.delta )
    cat( "Round-trip error  (RGB->HVC->RGB)  5-number summary:  " )
    cat( fivenum(delta), '\n' )

    return( failures==0 )
    }


{
x = gettime()   # load microbenchmark


if( testCentroids() )
    printf( "testCentroids() passed." )
else
    stop( "testCentroids() failed !", call.=FALSE )


if( testNeutrals() )
    printf( "testNeutrals() passed." )
else
    stop( "testNeutrals() failed !", call.=FALSE )

if( testNearNeutrals() )
    printf( "testNearNeutrals() passed." )
else
    stop( "testNearNeutrals() failed !", call.=FALSE )

if( testDarks() )
    printf( "testDarks() passed." )
else
    stop( "testDarks() failed !", call.=FALSE )

if( testReals() )
    printf( "testReals() passed." )
else
    stop( "testReals() failed !", call.=FALSE )

if( testOptimals() )
    printf( "testOptimals() passed." )
else
    stop( "testOptimals() failed !", call.=FALSE )



printf(  "Passed all Munsell Transforms tests !" )
}

################      Munsell  <-->  XYZ    #######################

#   Convert a Munsell specification into XYZ coordinates, by interpolating over the extrapolated Munsell renotation data.
#   MunsellSpec     an Nx3 matrix, or a character string
#   white           if NULL then XYZ must be already adapted to C, and with Y=100 for ref. white V=10
#   adapt           the chromatic adaptation method

MunsellToXYZ <- function( MunsellSpec, white=NULL, adapt="Bradford", ... )
    {
    p   = 'spacesXYZ'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }

    if( ! is.null(white) )
        {
        white   = process_white(white)
        if( any(is.na(white)) )
            {
            log_level( ERROR, "argument white is invalid" )
            return(NULL)
            }
        }

    tmp = MunsellToxyY( MunsellSpec, ... )
    if( is.null(tmp) )  return(NULL)

    XYZ = spacesXYZ::XYZfromxyY( tmp$xyY ) # copies rownames from xyY to XYZ

    if( ! is.null(white) )
        {
        # make CAT from illuminant C to given white

        # first check whether the user has selected a non-default xyC
        xyC  = intercept( "xyC", ... )

        if( is.null(xyC) )  xyC  = 'NBS'     # this is the default

        xy  = process_xyC( xyC )
        if( any(is.na(xy)) )
            {
            log_level( ERROR, "xyC=%s is invalid.", paste0(as.character(xyC),collapse=',') )
            return(NULL)
            }

        #   xy is now (x,y) chromaticity
        white.C = spacesXYZ::XYZfromxyY( c(xy,100) )

        theCAT  = spacesXYZ::CAT( white.C, white, method=adapt )
        if( is.null(theCAT) )   return(NULL)

        XYZ = spacesXYZ::adaptXYZ( theCAT, XYZ )

        if( is.null(XYZ) )  return(NULL)
        }

    return( XYZ )
    }



#   convert XYZ to an HVC matrix
#   XYZ     N x 3 matrix with XYZ vectors in the rows
#   white   if NULL then XYZ must be already adapted to C, and with Y=100 for ref. white V=10
#   adapt   the chromatic adaptation method

XYZtoMunsell <- function( XYZ, white=NULL, adapt='Bradford', ... )
    {
    p   = 'spacesXYZ'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }

    XYZ = prepareNx3( XYZ )
    if( is.null(XYZ) )  return(NULL)

    if( is.null(white) )
        {
        #   XYZ is already relative to Illuminant C, so nothing to do
        XYZ.adapted = XYZ
        }
    else
        {
        white   = process_white(white)
        if( any(is.na(white)) )
            {
            log_level( ERROR, "argument white is invalid"  )
            return(NULL)
            }

        # make CAT from given white to Illuminant C

        # first check whether the user has selected a non-default xyC
        xyC  = intercept( "xyC", ... )

        if( is.null(xyC) )  xyC  = 'NBS'     # this is the default

        xy  = process_xyC( xyC )
        if( any(is.na(xy)) )
            {
            log_level( ERROR, "xyC=%s is invalid.", paste0(as.character(xyC),collapse=',') )
            return(NULL)
            }

        xyC = xy

        #   xyC is now (x,y) chromaticity
        white.C = spacesXYZ::XYZfromxyY( c(xyC,100) )

        theCAT  = spacesXYZ::CAT( white, white.C, method=adapt )
        if( is.null(theCAT) )   return(NULL)

        XYZ.adapted = spacesXYZ::adaptXYZ( theCAT, XYZ )

        if( is.null(XYZ.adapted) )  return(NULL)
        }

    xyY = spacesXYZ::xyYfromXYZ( XYZ.adapted )
    if( is.null(xyY) )  return(NULL)

    tmp = xyYtoMunsell( xyY, ... )
    if( is.null(tmp) )  return(NULL)

    if( FALSE )
        {
        if( exists('xyC') )
            {
            cat( "xyC =", xyC, "   xyY.adapted =", xyY, '\n' )
            cat( "    delta =", xyC - xyY[1:2], '\n' )
            }

        print( tmp )
        }



    HVC = tmp$HVC

    rnames  = rownames(XYZ)     # if XYZ is not a matrix, then rnames is NULL

    if( is.null(rnames) )
        #   tmp$SAMPLE_NAME may have duplicates,
        #   but since HVC is a matrix that is OK
        rnames = tmp$SAMPLE_NAME

    rownames(HVC)   = rnames

    return( HVC )
    }




################      Munsell  <-->  Lab    #######################
#   Convert a Munsell specification into CIE Lab coordinates, by interpolating over the extrapolated Munsell renotation data.
MunsellToLab <- function( MunsellSpec, white='D65', adapt='Bradford', ... )
    {
    p   = 'spacesXYZ'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }

    if( is.character(MunsellSpec) )
        HVC = HVCfromMunsellName(MunsellSpec)
    else
        HVC = prepareNx3( MunsellSpec )

    if( is.null(HVC) )  return(NULL)

    white   = process_white(white)
    if( any(is.na(white)) )
        {
        log_level( ERROR, "argument white is invalid"  )
        return(NULL)
        }

    XYZ = MunsellToXYZ( HVC, white=white, adapt=adapt, ... )
    if( is.null(XYZ) )  return(NULL)

    Lab = spacesXYZ::LabfromXYZ( XYZ, white )   # copies rownames from XYZ to Lab

    #   do an additional test for neutrals
    #   to correct possible roundoff error
    neutral = HVC[ ,3]==0
    neutral[ is.na(neutral) ]   = FALSE

    if( any(neutral) )  Lab[neutral,2:3]  = 0     # when C==0, set both a=0 and b=0

    return( Lab )
    }


LabToMunsell <- function( Lab, white='D65', adapt='Bradford', ... )
    {
    p   = 'spacesXYZ'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }

    Lab = prepareNx3( Lab )
    if( is.null(Lab) )  return(NULL)

    white   = process_white(white)
    if( any(is.na(white)) )
        {
        log_level( ERROR, "argument white is invalid."  )
        return(NULL)
        }

    XYZ = spacesXYZ::XYZfromLab( Lab, white )

    HVC = XYZtoMunsell( XYZ, white=white, adapt=adapt, ... )


    #   do an additional test for neutrals
    #   to correct possible roundoff error
    neutral = Lab[ ,2]==0   &    Lab[ ,3]==0

    neutral[ is.na(neutral) ]   = FALSE

    if( any(neutral) )
        {
        # when both a==0 and b==0, then set H=0 and C=0
        HVC[neutral,c(1,3)]  = 0

        if( is.null(rownames(Lab)) )
            #   Lab has no rownames, which means that rownames(HVC) is MunsellName
            rownames(HVC)[neutral]  = MunsellNameFromHVC( HVC[neutral, ] )
        }

    return( HVC )
    }




################      Munsell  <-->  Luv    #######################
#Convert a Munsell specification into CIE Luv coordinates, by interpolating over the extrapolated Munsell renotation data.

MunsellToLuv <- function( MunsellSpec, white='D65', adapt='Bradford', ... )
    {
    p   = 'spacesXYZ'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }

    if( is.character(MunsellSpec) )
        HVC = HVCfromMunsellName(MunsellSpec)
    else
        HVC = prepareNx3( MunsellSpec )

    if( is.null(HVC) )  return(NULL)

    white   = process_white(white)
    if( any(is.na(white)) )
        {
        log_level( ERROR, "argument white is invalid"  )
        return(NULL)
        }

    XYZ = MunsellToXYZ( HVC, white=white, adapt=adapt, ... )
    if( is.null(XYZ) )  return(NULL)

    Luv = spacesXYZ::LuvfromXYZ( XYZ, white )   # copies rownames from XYZ to Luv

    #   do an additional test for neutrals
    #   to correct possible roundoff error
    neutral = HVC[ ,3]==0
    neutral[ is.na(neutral) ]   = FALSE

    if( any(neutral) )  Luv[neutral,2:3]  = 0     # when C==0, set both u=0 and v=0

    #   rownames(Luv)   = rownames(XYZ)  already done    #tmp$SAMPLE_NAME

    return( Luv )
    }


LuvToMunsell <- function( Luv, white='D65', adapt='Bradford', ... )
    {
    p   = 'spacesXYZ'
    if( ! requireNamespace( p, quietly=TRUE ) )
        {
        log_level( ERROR, "required package '%s' could not be loaded.", p )
        return(NULL)
        }

    Luv = prepareNx3( Luv )
    if( is.null(Luv) )  return(NULL)

    white   = process_white(white)
    if( any(is.na(white)) )
        {
        log_level( ERROR, "argument white is invalid"  )
        return(NULL)
        }

    XYZ = spacesXYZ::XYZfromLuv( Luv, white )

    HVC = XYZtoMunsell( XYZ, white=white, adapt=adapt, ... )

    #   do an additional test for neutrals
    #   to correct possible roundoff error
    neutral = Luv[ ,2]==0   &    Luv[ ,3]==0

    neutral[ is.na(neutral) ]   = FALSE

    if( any(neutral) )
        {
        # when both u==0 and v==0, then set H=0 and C=0
        HVC[neutral,c(1,3)]  = 0

        if( is.null(rownames(Luv)) )
            #   Luv has no rownames, which means that rownames(HVC) is MunsellName
            rownames(HVC)[neutral]  = MunsellNameFromHVC( HVC[neutral, ] )
        }

    return( HVC )
    }


#   this works, and avoids the call to assign() in .onAttach
#   however, this method requires an export entry in file NAMESPACE
LabtoMunsell    <- LabToMunsell
LuvtoMunsell    <- LuvToMunsell





################      Munsell  <-->  sRGB    #######################
# Convert sRGB input into CIE XYZ and xyY coordinates

sRGBtoMunsell <- function( sRGB, maxSignal=255, ... )
    {
    for( p in c('spacesRGB','spacesXYZ') )
        {
        if( ! requireNamespace( p, quietly=TRUE ) )
            {
            log_level( ERROR, "required package '%s' could not be loaded.", p )
            return(NULL)
            }
        }

    sRGB = prepareNx3( sRGB )
    if( is.null(sRGB) )  return(NULL)

    XYZ = spacesRGB::XYZfromRGB( sRGB, space='sRGB', which='scene', maxSignal=maxSignal )$XYZ
    if( is.null(XYZ) )  return(NULL)

    # adapt XYZ from D65 to C, using precomputed p.D65toC_CAT
    XYZ.adapted = spacesXYZ::adaptXYZ( p.D65toC_CAT, XYZ ) #; print( xyY.adapted )

    # Convert adapted XYZ coordinates to Munsell coordinates
    HVC = XYZtoMunsell( XYZ.adapted, ... )
    if( is.null(HVC) )  return(NULL)

    #   do an additional test for neutrals
    #   to correct possible roundoff error
    neutral = sRGB[ ,1]==sRGB[ ,2]    &    sRGB[ ,2]==sRGB[ ,3]

    neutral[ is.na(neutral) ]   = FALSE

    if( any(neutral) )
        {
        # when both R==G==B, then set H=0 and C=0
        HVC[neutral,c(1,3)]     = 0

        if( is.null(rownames(sRGB)) )
            #   sRGB has no rownames, which means that rownames(HVC) is MunsellName
            rownames(HVC)[neutral]  = MunsellNameFromHVC( HVC[neutral, ] )
        }

    return( HVC )
    }


MunsellTosRGB <- function( MunsellSpec, maxSignal=255, ... )
    {
    for( p in c('spacesRGB','spacesXYZ') )
        {
        if( ! requireNamespace( p, quietly=TRUE ) )
            {
            log_level( ERROR, "required package '%s' could not be loaded.", p )
            return(NULL)
            }
        }

    # Convert from Munsell notation to CIE coordinates
    tmp = MunsellToxyY( MunsellSpec, ... )
    if( is.null(tmp) )  return(NULL)

    out = tmp

    #   look for Chroma==0 exactly
    neutral = out$HVC[ ,3] == 0
    neutral[ is.na(neutral) ]   = FALSE

    out$HVC = NULL  # erase this column

    XYZ = spacesXYZ::XYZfromxyY( out$xyY )

    # adapt XYZ from C to D65, using precomputed p.CtoD65_CAT
    XYZ.adapted = spacesXYZ::adaptXYZ( p.CtoD65_CAT, XYZ )

    #   XYZ.adapted = spacesXYZ::XYZfromxyY( xyY.adapted )

    # Convert CIE XYZ coordinates to sRGB coordinates
    out = cbind( out, spacesRGB::RGBfromXYZ( XYZ.adapted, space='sRGB', maxSignal=maxSignal ) )

    #rnames  = tmp$SAMPLE_NAME
    #if( anyDuplicated(rnames) ) rnames = 1:nrow(out)
    #rownames(out) = rnames

    if( any(neutral) )
        {
        #   if Chroma=0 exactly, force RGBs to be exactly equal by computing their mean value
        RGB.mean = rowMeans( out$RGB[neutral, , drop=FALSE] )

        #   the vector on the right side is replicated over all columns on the left
        out$RGB[neutral, ]  =  RGB.mean  # cbind( RGB.mean, RGB.mean, RGB.mean )
        }

    return( out )
    }



#   return numeric XYZ with Y normalized to 100, and not 1
#
#   if white is invalid, returns NA_real_

process_white <- function( white )
    {
    if( is.character(white) )
        {
        if( length(white) != 1 )    return(NA_real_)

        #   first try lookup to XYZ
        out = spacesXYZ::standardXYZ(white)

        if( all( is.finite(out) ) ) return( 100*out )

        #   next try lookup to xy
        out = spacesXYZ::standardxy(white)

        if( all( is.finite(out) ) ) return(  spacesXYZ::XYZfromxyY( c(out,100) ) )

        return( NA_real_ )
        }

    if( is.numeric(white) )
        {
        if( length(white) == 3 )
            # interpret as XYZ
            return(white)

        if( length(white) == 2 )
            #   interpret as xy
            return( spacesXYZ::XYZfromxyY( c(white,100) ) )
        }

    return(NA_real_)
    }




#############              deadwood below                         ###############


#   xyz2srgb is here only because package 'colorscience' depends on it
xyz2srgb <- function(XYZ){
if ((length(XYZ) %% 3) != 0)  stop('XYZ matrix must be n x 3')
if (is.null(dim(XYZ))) if (length(XYZ)>2) XYZ<-matrix(XYZ, ncol=3,byrow=TRUE)
M <- matrix(c(3.2406, -1.5372, -0.4986, -0.9689, 1.8758, 0.0415, 0.0557, -0.2040, 1.0570),3,3,byrow=TRUE)
RGB <- t(M %*% t(XYZ))
# START: lines added March 2013 to set out-of-gamut flag.
# The out-of-gamut flag is a column vector of Boolean true/false values.  Each
# entry corresponds to one row of the input matrix XYZ.
NumberOfInputs <- dim(RGB)[1]
OutOfGamutFlag <- -99 * matrix(1,NumberOfInputs)
for (index in 1:NumberOfInputs){
if (RGB[index,1] < 0 || RGB[index,1] > 1 || RGB[index,2] < 0 || RGB[index,2] > 1 || RGB[index,3] < 0 || RGB[index,3] > 1) OutOfGamutFlag[index] = TRUE else OutOfGamutFlag[index] <- FALSE
}
# END: lines added March 2013 to set out-of-gamut flag
RGB[which(RGB<0)] <- 0
RGB[which(RGB>1)] <- 1
DACS = matrix(0,dim(XYZ)[1],3)
index <- RGB<=0.0031308
index[index==TRUE] <- 1
index[index==FALSE] <- 0
DACS <- DACS+index*(12.92*RGB)
DACS <- DACS+(1-index)*(1.055*RGB^(1/2.4)-0.055)
RGB <- ceiling(DACS*255)
RGB[which(RGB<0)] <- 0
RGB[which(RGB>255)] <- 255
return(list(Status.ind  = 1, sRGB=RGB, OutOfGamutFlag=OutOfGamutFlag))
}

srgb2xyz <- function(RGBmatrix){
if ((length(RGBmatrix) %% 3) != 0)  stop('RGBmatrix must be n x 3')
if (is.null(dim(RGBmatrix))) if (length(RGBmatrix)>2) RGBmatrix<-matrix(RGBmatrix, ncol=3,byrow=TRUE)
XYZmatrix <- matrix(0,dim(RGBmatrix)[1],3)
M <- matrix(c(0.4124, 0.3576, 0.1805, 0.2126, 0.7152, 0.0722, 0.0193, 0.1192, 0.9505),3,3,byrow=TRUE)
DACS <- RGBmatrix/255
RGBmatrix <- matrix(0,dim(DACS)[1],3)
index <- DACS <= 0.04045
index[index==TRUE] <- 1
index[index==FALSE] <- 0
RGBmatrix = RGBmatrix + (index)*(DACS/12.92)
RGBmatrix = RGBmatrix + (1-index)*((DACS+0.055)/1.055)^2.4
XYZmatrix <- t(M %*% t(RGBmatrix))
XYZmatrix
}


#############     Lightness in [0,100]  <-->  linear Y in [0,1]   ###############
#
#   these threshold=(24/116)^3 and coeff=(12/116)^3    are taken from CIE 15:2004 Section 8.2
#   see http://brucelindbloom.com/LContinuity.html.  Glenn Davis

#   linear Y    in the interval [0,1], nominally
#   returns L in [0,100]
#
#   Some of the x,y in all.dat are negative, and sometimes x+y > 1, which implies z<-.
#   So for Munsell purposes we must allow the tristimulus coordinates XYZ to be negative.
#   To accomodate this, Lightness_from_linear() is defined for negative numbers
#   so that it is an odd function.

Lightness_from_linear  <-  function( Y )
    {
    #   thresh = (24/116)^3

    s = sign(Y)
    Y = s * Y       # Y is now non-negative

    out = ifelse( Y < (24/116)^3, (116/12)^3 * Y, 116*Y^(1/3) - 16 )

    #   correct any negatives by multiplying by s
    return( s * out )
    }

#   nonlinear L         in the interval [0,100]
#   returns linear Y    in the interval [0,1]
linear_from_Lightness  <-  function( L )
    {
    s   = sign(L)
    L   = s * L     # L is now non-negative

    out = ifelse( L < 8, (12/116)^3 * L, ((L + 16)/116)^3 )

    #   correct any negatives by multiplying by s
    return( s * out )
    }

#   HVCfromMunsellName()
#
#   convert Munsell notation to numeric Hue,Value,Chroma
#
#   arguments:
#       MunsellName   a character vector of length N
#
#   return value: an Nx3 matrix with the computed HVCs placed in the rows
#       Hue     is ASTM Hue in the interval (0,100]  (except Hue=0 for neutrals)
#       Value   is in the interval [0,10]
#       Chroma  is positive  (except Chroma=0 for neutrals)
#
#   rownames(out) are set to MunsellName
#
#   author:  Glenn Davis

HVCfromMunsellName <- function( MunsellName )
    {
    n = length(MunsellName)

    if( n==0  ||  ! is.character(MunsellName) )
        {
        log_level( ERROR, "MunsellName is not a character vector with positive length." )
        return(NULL)
        }

    out = matrix( NA_real_, n, 3 )
    colnames(out)   = c( "H", "V", "C" )
    rownames(out)   = MunsellName

    # Remove whitespace from input Munsell string
    MunsellName <- gsub( ' |\t', '', MunsellName )

    # Make all letters in Munsell string upper case
    MunsellName <- toupper(MunsellName)

    #-----   chromatic colors   ----------#
    pattern = "^([0-9.]+)(R|YR|Y|GY|G|BG|B|PB|P|RP)([0-9.]+)/([0-9.]+)$"

    #   time_start  = as.double( Sys.time() )

    sub1234 = sub( pattern, "\\1 \\2 \\3 \\4", MunsellName )

    mask    = (nchar(MunsellName) < nchar(sub1234) )     # a matched string gets longer          #mask = grepl( pattern, MunsellName )   #   ; print( mask )

    #   change NAs to FALSE
    mask[ is.na(mask) ] = FALSE

    if( any(mask) )
        {
        #   make a little LUT
        hue = seq( 0, 90, by=10 )
        names(hue)  = c("R","YR","Y","GY","G","BG","B","PB","P","RP")

        sub1234    = sub1234[mask]

        dat     = unlist( strsplit( sub1234, ' ', fixed=T ) ) #;  print( dat )
        dat     = matrix( dat, length(sub1234), 4, byrow=T ) #;    print( dat )

        hue_minor   =  as.numeric(dat[ ,1])
        valid       =  (hue_minor <= 10)
        hue_minor[ ! valid ]    = NA

        out[ mask, 1]   = hue[ dat[ ,2] ]  +  hue_minor
        out[ mask, 2]   = as.numeric( dat[ ,3] )
        out[ mask, 3]   = as.numeric( dat[ ,4] )
        }

    #   cat( "Elapsed: ", as.double( Sys.time() ) - time_start, '\n' )

    #------   achromatic colors     ----#
    pattern = "^N([0-9.]+)(/|/0)?$"

    sub1    = sub( pattern, "\\1", MunsellName )

    mask    = (nchar(sub1) < nchar(MunsellName) ) # a matched string gets shorter.    #mask = grepl( pattern, MunsellName )   #   ; print( mask )

    #   change NAs to FALSE
    mask[ is.na(mask) ] = FALSE

    if( any(mask) )
        {
        out[ mask, 1]    = 0
        out[ mask, 2]    = as.numeric( sub1[mask] )
        out[ mask, 3]    = 0
        }

    #   now look for invalid Hue and Value
    #   we must check for Hue because hue_minor may have been set to NA above
    #   we do not really have to check for negative V and C, because the minus sign is not in the pattern
    valid   = is.finite(out[ ,1])  &  (out[ ,2] <= 10)

    out[ ! valid, ]  = NA

    return( out )
    }




#   for compatibility with package colorscience
#
#   MunsellName     character N-vector of Munsell notations
#
#   return value
#       Nx3 matrix of character strings
#
MunsellHVC  <- function( MunsellName )
    {
    out = strsplit( MunsellName, "[ ]+|/" )

    fixup <- function( vec )
        {
        if( vec[1]=='N'  &&  vec[length(vec)] != '0' )  vec = c( vec, '0' )

        if( length(vec) == 3 )
            return(vec)
        else
            return( rep(NA_character_,3) )
        }

    out = lapply( out, fixup )      #; print(out)

    n   = length(out)

    out = unlist( out )

    #   if( length(out) != 3*n )  out = rep( NA_character_, 3*n )

    out = matrix( out, nrow=n, 3, byrow=TRUE )

    colnames(out) = c('H','V','C')

    return(out)

    # return( HVCfromMunsellName( MunsellName )  )
    }




#   HueString   a character vector of length n, e.g.  c( '2.5RP', '3.2B' )
#
#   return value    numeric vector of length n
#                   numeric Hue is ASTM Hue in the interval (0,100]
#
HueNumberFromString  <-  function( HueString )
    {
    # Make all letters in Munsell string upper case
    HueString   = toupper(HueString)

    # Remove whitespace from input Munsell string
    HueString   = gsub( ' |\t', '', HueString )

    #-----   chromatic colors   ----------#
    pattern = "^([0-9.]+)(R|YR|Y|GY|G|BG|B|PB|P|RP)$"

    sub12 = sub( pattern, "\\1 \\2", HueString )

    mask    = (nchar(HueString) < nchar(sub12) )     # a matched string gets longer          #mask = grepl( pattern, MunsellSpecString )   #   ; print( mask )

    n = length(HueString)

    out = rep( NA_real_, n )

    if( any(mask) )
        {
        #   make a little LUT
        hue = seq( 0, 90, by=10 )
        names(hue)  = c("R","YR","Y","GY","G","BG","B","PB","P","RP")

        sub12    = sub12[mask]

        dat     = unlist( strsplit( sub12, ' ', fixed=T ) ) #;  print( dat )
        mat     = matrix( dat, length(sub12), 2, byrow=T ) #;    print( dat )

        hue_minor   =  as.numeric(mat[ ,1])
        valid       =  (hue_minor <= 10)
        hue_minor[ ! valid ]    = NA

        out[ mask ]   = hue[ mat[ ,2] ]  +  hue_minor
        }

    return( out )
    }

#   Hue       a numeric vector, with ASTM Hues, some may be NA
HueStringFromNumber <- function( Hue, format='g', digits=2 )
    {
    out = rep( NA_character_, length(Hue) )

    mask_good   = is.finite(Hue)

    if( ! any(mask_good) )  return(out)

    #   digits_given    = ( is.finite(digits)  &&  0<=digits )
    #   if( digits_given )  Hue = round( Hue, digits=digits )   not needed

    Hue   = Hue[mask_good] %% 100

    Hue[ Hue == 0 ] = 100

    idx     = as.integer( Hue / 10 )
    frac    = Hue - 10 * idx

    mask    = (frac == 0)
    if( any(mask) )
        {
        idx[mask]   = idx[mask]-1
        frac[mask]  = 10
        }

    namevec  = c("R","YR","Y","GY","G","BG","B","PB","P","RP")

    out[mask_good] = paste0( formatC(frac,digits=digits,format=format,width=1), namevec[idx+1] )

    #   out[mask_good] = sprintf( "%g%s", frac, namevec[idx+1] )

    return(out)
    }


#   HVC     a numeric vector of length 3
#           or a numeric matrix with 3 columns, with HVC in the rows
#
#   return value:  a character vector

MunsellNameFromHVC <- function( HVC, format='g', digits=2 )
    {
    HVC = prepareNx3(HVC)
    if( is.null(HVC) )  return(NULL)

    out = rep( NA_character_, nrow(HVC) )

    mask_finite = apply( HVC, 1, function(x) { all(is.finite(x)) } )

    if( ! any(mask_finite) )  return(out)

    #   digits_given    = ( is.finite(digits)  &&  0<=digits )

    if( format == 'f' )  HVC[ ,3] = round( HVC[ ,3], digits=digits )

    #   compute chromatic mask
    mask_chromatic  = (0 < HVC[ ,3] )   # negative chroma will convert to neutral
    mask_achromatic = ! mask_chromatic

    mask_chromatic[ ! mask_finite ]   = FALSE
    mask_achromatic[ ! mask_finite ]  = FALSE   #; print(mask_achromatic)

    #   use formatC()

    out[mask_chromatic]     = paste0( HueStringFromNumber(HVC[mask_chromatic,1], format=format, digits=digits),
                                        ' ', formatC( HVC[mask_chromatic,2], digits=digits, format=format, width=1),
                                        '/', formatC( HVC[mask_chromatic,3], digits=digits, format=format, width=1) )

    if( any(mask_achromatic) )
        out[ mask_achromatic ]  = paste0( "N ", formatC( HVC[mask_achromatic,2], digits=digits, format=format, width=1), '/' )

    names(out)  = rownames(HVC)

    return(out)
    }

#   HVC0        Nx3 or 1x3 matrix, or character vector of MunsellNotation
#   HVC1        Nx3 or 1x3 matrix, or character vector of MunsellNotation
#   symmetric   logical, whether to symmetrize the difference metric
#
#   returns     numeric N-vector of pairwise differences

NickersonColorDifference <- function( HVC0, HVC1, symmetric=TRUE, coeffs=c(0.4,6,3) )
    {
    if( is.character(HVC0) )
        HVC0 = HVCfromMunsellName(HVC0)
    else  
        {
        HVC0 = prepareNx3( HVC0 )
        if( is.null(HVC0) )  return(NULL)
        }
    
    if( is.character(HVC1) )
        HVC1 = HVCfromMunsellName(HVC1)
    else  
        {
        HVC1 = prepareNx3( HVC1 )
        if( is.null(HVC1) )  return(NULL)
        }
        
    if( nrow(HVC0)==1  &&  1<nrow(HVC1) )
        #   replicate HVC0
        HVC0    = matrix(HVC0,nrow(HVC1),3,byrow=TRUE)
        
    if( 1<nrow(HVC0)  &&  nrow(HVC1)==1 )
        #   replicate HVC1
        HVC1    = matrix(HVC1,nrow(HVC0),3,byrow=TRUE)
        
        
    if( nrow(HVC0) != nrow(HVC1) )
        {
        log_level( ERROR, "nrow(HVC0) = %d != %d = nrow(HVC1).",
                            nrow(HVC0), nrow(HVC1) )
        return(NULL)
        }
        
    ok  = is.numeric(coeffs)  &&  length(coeffs)==3  &&  all( is.finite(coeffs) )  &&  all( 0 < coeffs )
    if( ! ok )
        {
        log_level( ERROR, "argument coeffs is invalid." )
        return(NULL)
        }
        
        
    delta   = abs( HVC0 - HVC1 )
    
    #   scale the Delta Hue column by Chroma
    if( symmetric )
        s   = pmin( HVC0[ ,3], HVC1[ ,3] )
    else
        s   = HVC0[ ,3]
        
    # correct for Delta Hue wraparound
    dHue    = pmin( delta[ ,1], 100 - delta[ ,1] )
        
    delta[ ,1]  = s * dHue
    
    out = delta %*% coeffs
    
    dim(out)    = NULL
        
    return(out)
    }
    

#   the functions here are to be run pre-build, but with munsellinterpol already installed.
#   Note that the triple-colon trick ::: is used for
#       munsellinterpol:::makeLookupList()
#       munsellinterpol:::makeInversionCoeffs()
#       munsellinterpol:::gettime()
#
#   the file created are:
#       munsellinterpol.rda     for exported, and documented, data
#       sysdata.rda             for non-exported, and non-documented, data



#   These datasets are exported and can be inspected by the user.
#   Each one here must be documented.
#   They are all put in the same .RDA file
saveDatasets  <- function( pathout="../data/munsellinterpol.rda" )
    {
    savevec = character(0)

    Munsell2xy  = makeMunsell2xy()
    #   print( str(Munsell2xy) )

    attr( Munsell2xy, "comment" )   = readLines( "../inst/extdata/ReadMe.txt" )

    savevec = c( savevec, "Munsell2xy" )

    pathin    = "../inst/extdata/Centroids_ISCC-NBS.txt"
    CentroidsISCCNBS = read.table( pathin, sep='\t', header=T, stringsAsFactors=F )
    attr( CentroidsISCCNBS, "comment" )  = readLines( pathin, n=23 )

    savevec = c( savevec, "CentroidsISCCNBS" )

    #   the Munsell books: bead, soil, rock, etc.
    MunsellBooks    = readAllBooks()
    if( is.null(MunsellBooks) )   return(FALSE)
    savevec = c( savevec, "MunsellBooks" )


    ##  ready to save it
    save( list=savevec, file=pathout, compress='xz' )   #     'xz'  'gzip'  FALSE

    return(TRUE)
    }




#   an advantage of the private data in "sysdata.rda" is that these
#   do not have to be exposed, and therefore documented
savePrivateDatasets  <- function( .path="sysdata.rda" )
    {
    library( logger )

    log_threshold( INFO, namespace="munsellinterpol" )

    savevec = character(0)

    #   load( "../data/munsellinterpol.rda" )
    #   make little matrix
    p.xyC   = matrix( c( 0.31012,0.31631,  0.3101,0.3163, 0.310,0.316, 0.31006,0.31616 ),  4, 2,    byrow=T )
    # mat = cbind( mat, rep(100,4) )
    rownames(p.xyC) = c( 'JOSA', 'NBS', 'NTSC', 'CIE' )
    colnames(p.xyC) = c('x','y')
    savevec = c( savevec, "p.xyC" )

    if( FALSE )
    {
    #   these two 3x3 matrices are now built into package spacesXYZ
    if( ! requireNamespace('spacesXYZ') )   return(FALSE)

    primary     = matrix( c(0.64,0.33,  0.3,0.6, 0.15,0.06 ), 3, 2, byrow=T )   # from sRGB standard
    primary     = cbind( primary, 1 - rowSums(primary) )
    whiteXYZ    = as.numeric( spacesXYZ::XYZfromxyY( c(0.3127,0.3290,1) ) )       # from sRGB standard
    p.sRGB2XYZ  = projectiveMatrix( t(primary), whiteXYZ )        #; print( p.sRGB2XYZ )
    savevec = c( savevec, "p.sRGB2XYZ" )

    p.XYZ2sRGB  = solve(p.sRGB2XYZ)                             #;  print( p.XYZ2sRGB )
    savevec = c( savevec, "p.XYZ2sRGB" )
    }

    #   Illuminants A, C, D50 D55 D65 D75
    p.ACDs  = read.table( "../inst/extdata/ACDs.5nm.txt", header=TRUE  )
    savevec = c( savevec, "p.ACDs" )

    p.xyz1931   = read.table( "../inst/extdata/xyz1931.5nm.txt", header=TRUE  )
    savevec = c( savevec, "p.xyz1931" )

    p.OptimalHull       = list()

    for( iname in c("C","D65") )
        {
        ispec   = p.ACDs[[ iname ]]
        W   = ispec * as.matrix( p.xyz1931[ ,2:4] )
        W   = 100 * W / colSums(W)[2]
        p.OptimalHull[[ iname ]]  = zonohedron( W )
        }
    savevec = c( savevec, "p.OptimalHull" )



    pathin  = "../inst/extdata/System_ISCC-NBS.txt"
    p.System_ISCCNBS    = read.table( pathin, header=TRUE, sep='\t', stringsAsFactors=F )
    attr( p.System_ISCCNBS, "header" )  = readLines( pathin, n=50 )
    savevec = c( savevec, "p.System_ISCCNBS" )

    if( FALSE )
    {
    #   the Munsell books: bead, soil, rock, etc.
    p.Books = readAllBooks()
    if( is.null(p.Books) )   return(FALSE)
    savevec = c( savevec, "p.Books" )
    }
    
    
    
    
    if( TRUE )
    {
    #   these tables take the most time, by far
    Munsell2xy  = makeMunsell2xy()          # this one does not generate a warning

    p.LookupList = munsellinterpol:::makeLookupList( Munsell2xy, p.xyC['NBS', ], kfactor=c(0.70,0.50) )     # default is kfactor = c(0.9,0.5)
    savevec = c( savevec, "p.LookupList" )    # this compresses *very* well

    p.InversionCoeffs =  munsellinterpol:::makeInversionCoeffs( Munsell2xy, p.LookupList, warn=FALSE )
    savevec = c( savevec, "p.InversionCoeffs" )
    }

    ##  finally ready to save it
    save( list=savevec, file=.path, compress='xz' )   #     'xz'  'gzip'  FALSE

    mess    = sprintf( "Saved the following to '%s'.", .path )
    cat( mess, '\n', file=stderr() )
    cat( savevec, '\n', file=stderr() )

    return( invisible(TRUE) )
    }

pingDatasets  <- function( .path="sysdata.rda", .verbose=FALSE )
    {
    theName     = load(.path,verbose=.verbose)
    print( theName )

    if( 0 < length(theName)  &&  .verbose )
        {
        for( k in 1:length(theName ) )
            {
            obj = get( theName[k] )
            cat( '\n', theName[k], '\n' )
            print( str(obj) )
            }
        }

    return( invisible(T) )
    }


makeMunsell2xy <- function(  )
    {
    time_start  = munsellinterpol:::gettime()

    realMunsell = read.table( "../inst/extdata/real.dat", header=T, sep='', stringsAsFactors=F )    # ; print( str(realMunsell) )
    colnames(realMunsell)[1] = 'H'      # was 'h'
    realMunsell$H = HueNumberFromString( realMunsell$H )

    verydark = read.table( "../inst/extdata/verydark.dat", header=T, sep='', stringsAsFactors=F )   # ; print( str(verydark) )
    verydark$H = HueNumberFromString( verydark$H )

    allMunsell  = read.table( "../inst/extdata/all.dat", header=T, sep='', stringsAsFactors=F )     # ; print( str(allMunsell) )
    allMunsell$H = HueNumberFromString( allMunsell$H )

    #   diffMunsell( realMunsell, allMunsell )

    real    = contained( allMunsell, realMunsell )  |  contained( allMunsell, verydark )    # slow, not optimized

    allMunsell$C = as.numeric( allMunsell$C )   # do not want integer here

    out = allMunsell[ -6 ]  # drop Y

    out  = cbind( out, real=real  )

    #   make slight tweak(s)
    idx = which( out$H==72.5  &  out$V==10  &  out$C==2 )
    if( length(idx) == 1 )
        {
        out$x[idx]  = 0.2970
        out$y[idx]  = 0.3075
        }



    mess    = sprintf( "makeMunsell2xy(). INFO.  Made Munsell2xy in %g seconds.", munsellinterpol:::gettime() - time_start )
    cat( mess, '\n', file=stderr() )

    return( invisible(out) )
    }


contained <- function( .data1, .data2 )
    {
    #   brute force compare, not optimized
    n   = nrow(.data1)

    out     = logical(n)

    name1   = deparse( substitute( .data1 )  )
    name2   = deparse( substitute( .data2 )  )

    for( k in 1:n )
        {
        row1    = .data1[ k, ]

        mask    = .data2$H==row1$H  &  .data2$V==row1$V  &  .data2$C==row1$C
        row2    = .data2[ mask, ]

        if( nrow(row2) == 0 )   next

        if( 2 <= nrow(row2) )
            {
            mess    = sprintf( "'%s' is invalid. Duplicated data\n", name2 )
            cat( mess )
            print( row2 )
            return(NULL)
            }

        #   exactly 1 match
        out[k]  = TRUE
        }

    return(out)
    }






######################      deadwood below      #################################



diffMunsell <- function( .data1, .data2 )
    {
    #   brute force compare, not optimized
    out     = NULL
    count   = 0

    name1   = deparse( substitute( .data1 )  )
    name2   = deparse( substitute( .data2 )  )

    mat1    = matrix( 0, 1, 2 )
    colnames(mat1)  = c( 'x', 'y' )
    class(mat1) = "model.matrix"
    mat2    = mat1

    for( k in 1:nrow(.data1) )
        {
        row1    = .data1[ k, ]

        mask    = .data2$H==row1$H  &  .data2$V==row1$V  &  .data2$C==row1$C
        row2    = .data2[ mask, ]

        if( nrow(row2) == 0 )   next

        if( 2 <= nrow(row2) )
            {
            mess    = sprintf( "'%s' is invalid. Duplicated data\n", name2 )
            cat( mess )
            print( row2 )
            return(NULL)
            }

        #   exactly 1 match
        count   = count + 1

        mat1[ 1, 1 ]    = row1$x
        mat1[ 1, 2 ]    = row1$y
        mat2[ 1, 1 ]    = row2$x
        mat2[ 1, 2 ]    = row2$y

        ok  = all( mat1 == mat2 )

        if( ! ok )
            {
            delta   = mat1 - mat2
            #   colnames(delta) = c( "x", "y" )      redundant
            class(delta) = "model.matrix"
            out = rbind( out, data.frame( H=row1$H, V=row1$V, C=row1$C, xy1=mat1, xy2=mat2, delta=delta, stringsAsFactors=F ) )
            }
        }


    colnames( out )[4] = name1
    colnames( out )[5] = name2


    mess    = sprintf( "Chips that %s and %s have in common = %d\n", name1, name2, count )
    cat(mess)
    mess    = sprintf( "Chips that are in %s, but not in %s = %d\n", name1, name2, nrow(.data1) - count )
    cat(mess)
    mess    = sprintf( "Chips that are in %s, but not in %s = %d\n", name2, name1, nrow(.data2) - count )
    cat(mess)

    return( out )
    }

% lightness.rnw
% Time-stamp: "lightness.rnw"

\documentclass[11pt]{article}

% Set margins to be narrow
\RequirePackage[left=1in,top=0.75in,right=1in,bottom=0.75in]{geometry}

%\VignetteIndexEntry{Uniform Lightness Scales}
%\VignetteEngine{knitr::knitr}

\RequirePackage{color}
\definecolor{darkblue}{rgb}{0,0,0.5}
\definecolor{blue}{rgb}{0,0,0.8}
\definecolor{lightblue}{rgb}{0.2,0.2,0.9}
\definecolor{darkred}{rgb}{0.6,0.0,0.0}
\definecolor{red}{rgb}{0.7,0,0}
\definecolor{darkgreen}{rgb}{0.0,0.4,0.0}
\definecolor{lightgray}{rgb}{0.7,0.7,0.7}
\definecolor{darkorange}{rgb}{0.75, 0.45, 0}
\definecolor{purple}{rgb}{0.65, 0, 0.75}
\definecolor{goldenrod}{rgb}{0.80, 0.61, 0.11}
\definecolor{lightyellow}{rgb}{0.98,0.94,0.83}


\RequirePackage{fancyvrb}
\RequirePackage[T1]{fontenc}
\RequirePackage{ae}       % ComputerModern Fonts
\RequirePackage{fancyhdr}
\RequirePackage{float}
\RequirePackage{hyperref}
\usepackage{lastpage}

\pagestyle{fancy}
\cfoot{page \thepage\ of \pageref{LastPage}}
\renewcommand{\headrulewidth}{0pt}

% \code mini environment ttfamily->texttt
\newcommand\code{\bgroup\@codex}
\def\@codex#1{{\color{darkred} \normalfont\ttfamily\hyphenchar\font=-1 #1}\egroup}

% This environment defines the look of R ouput
\DefineVerbatimEnvironment{Soutput}{Verbatim}{
  fontsize=\small,
  formatcom=\color{darkblue}
}

\begin{document}
% \SweaveOpts{concordance=TRUE}

\title{ {\Huge Uniform Lightness Scales} }
\author{Glenn Davis}
\maketitle
% \thispagestyle{fancy}

% Setup stuff.
<<setup, echo=FALSE, results="hide">>=
require("knitr",quietly=TRUE)
opts_chunk$set(fig.path="figs/ag2-", fig.align="center",
  fig.width=7, fig.height=7, comment="")
knit_hooks$set(output = function(x, options) {
  paste('\\begin{Soutput}\n', x, '\\end{Soutput}\n', sep = '')
})
options(width=90)
par( omi=c(0,0,0,0), mai=c(0.2,0.2,0.2,0.2) )
if(!file.exists("figs")) dir.create("figs")
@

% ----------------------------------------------------------------------------
\section*{Introduction}

A \textit{Uniform Lightness Scale} is a reparameterization of luminance factor,
so that equal numerical steps in lightness produce equal perceptual steps.
The goal of this \textbf{munsellinterpol} vignette is to 
treat Munsell Value (V) as a Uniform Lightness Scale,
compare the different definitions of Value,
and compare Munsell Value with CIE Lightness.
Featured functions in this vignette are: 
\code{VfromY()} and
\code{YfromV()}.


<<packs, echo=TRUE, message=FALSE>>=
library( munsellinterpol )
@

\setcounter{figure}{0}  

% ----------------------------------------------------------------------------

\section*{Munsell Value and Luminance Factor}

In the search for a perceptually uniform Munsell Value scale,
there have been many refinements:
from \cite{Priest}, to \cite{Munsell1933}, to \cite{Newhall1943}
to \cite{ASTM-D1535-08}.
For definitions of these, see the man page for \code{VfromY()}.


<<lee11, echo=TRUE, fig.pos="H", fig.height=5, out.width='1.0\\linewidth', fig.cap='Value V  vs.  Luminance Factor Y' >>=
par( omi=c(0,0,0,0), mai=c(0.5,0.5,0.1,0.1) )
plot( c(0,100), c(0,10), type='n', xlab='', ylab='', las=1, tcl=0,
        lab=c(10,8,7), mgp=c(3,0.25,0) )
title( xlab='Y', line=1.5 ) ;  title( ylab='Value', line=1.5 )
grid( lty=1 ) ;  abline( h=0, v=0 )
V = seq( 0, 10, by=0.125 )
color = unlist( list(ASTM='black',OSA='black',MgO='black',Munsell='red',Priest='blue') )
for( w in names(color) )
    lines( YfromV(V,w), V, col=color[w], lty=ifelse(w=='MgO',2,1), lwd=0.75 )
lty = ifelse( names(color)=='MgO', 2, 1 )
legend( "bottomright", names(color), bty='n', lty=lty, lwd=1.5, col=color, inset=0.1 )
@

Note that for MgO, Y=102.568 when V=10. 
This luminance factor
for this curve is relative to MgO instead of the perfect reflecting diffuser.

The curves for ASTM and OSA quintics are indistinguishable at this scale.
Plot the difference between them.

<<lee12, echo=TRUE, fig.pos="H", fig.height=3.5, out.width='1.0\\linewidth', fig.cap='The Difference between ASTM and OSA Definitions of Value' >>=
par( omi=c(0,0,0,0), mai=c(0.5,1,0.1,0.1) )
Y = seq( 0, 100, by=0.5 )
delta = VfromY(Y,'OSA') - VfromY(Y,'ASTM')
plot( range(Y), range(delta), type='n', xlab='', ylab='', las=1, tcl=0,
        lab=c(10,8,7), mgp=c(3,0.25,0) )
title( xlab='Y', line=1.5 ) ;  title( ylab='{OSA Value} - {ASTM Value}', line=3 )
grid( lty=1 ) ;  abline( h=0, v=0 )
lines( Y, delta, lwd=0.75 )
@
This difference is negligible.



% ----------------------------------------------------------------------------

\section*{CIE Lightness and Munsell Value}

Most modern work uses the CIE Lightness function  ,
which is actually a simplified version of Munsell Value,
but on different domains.
The linear domain for Y is [0,1] and the perceptually uniform domain
for Lightness is [0,100].
Here is the definition:

<<Lightness, echo=TRUE, message=FALSE>>=

Lightness_from_linear  <-  function( Y )  {
ifelse( Y < (24/116)^3, (116/12)^3 * Y, 116*Y^(1/3) - 16 )
}
@

And here is the comparison plot, properly scaled for the difference in domains.

<<lee13, echo=TRUE, fig.pos="H", fig.height=5, out.width='1.0\\linewidth', fig.cap='Comparing Lightness and Munsell Value' >>=
par( omi=c(0,0,0,0), mai=c(0.5,0.75,0.1,0.1) )
Y = (0:100)/100
L = Lightness_from_linear( Y )
plot( range(Y), range(L), type='n', xlab='', ylab='', las=1, tcl=0,
        lab=c(10,8,7), mgp=c(3,0.25,0) )
title( xlab='Y   (luminance factor)', line=1.5 ); title( ylab='Lightness', line=2 )
grid( lty=1 ) ;  abline( h=0, v=0 )
lines( Y, L, lwd=0.75 )
V = VfromY( 100 * Y, 'ASTM' )
lines( Y, 10*V, lty=2 )
legend(  "bottomright", c("Lightness (CIE)","10*Value (ASTM)"), lty=c(1,2),
            bty='n', inset=0.1 )
@
The agreement is very good.
Let's examine the ratio.

<<lee14, echo=TRUE, fig.pos="H", fig.height=3, out.width='1.0\\linewidth', fig.cap='Comparing Lightness and Munsell Value' >>=
par( omi=c(0,0,0,0), mai=c(0.5,0.75,0.1,0.1) )
quotient = L / V
plot( range(Y), range(quotient,na.rm=T), type='n', xlab='', ylab='', las=1, tcl=0,
        lab=c(10,8,7), mgp=c(3,0.25,0) )
title( xlab='Y   (luminance factor)', line=1.5 )
title( ylab='Lightness / Value', line=3 )
grid( lty=1 ) ;  abline( h=0, v=0 )
lines( Y, quotient )
@
So Value $\simeq$ Lightness/10  is a good approximation.


% \pagebreak

\bibliographystyle{apalike}
\bibliography{bibliography}


% ----------------------------------------------------------------------------


\section*{Appendix}
This document was prepared \today \quad with the following configuration:
<<finish, echo=FALSE, results="asis">>=
knit_hooks$set(output = function(x, options) { x })
toLatex(sessionInfo(), locale=FALSE)
@


\end{document}

% lightness.rnw
% Time-stamp: "lightness.rnw"

\documentclass[11pt]{article}

% Set margins to be narrow
\RequirePackage[left=1in,top=0.75in,right=1in,bottom=0.75in]{geometry}

%\VignetteIndexEntry{Loci of Constant Hue and Chroma}
%\VignetteEngine{knitr::knitr}

\RequirePackage{color}
\definecolor{darkblue}{rgb}{0,0,0.5}
\definecolor{blue}{rgb}{0,0,0.8}
\definecolor{lightblue}{rgb}{0.2,0.2,0.9}
\definecolor{darkred}{rgb}{0.6,0.0,0.0}
\definecolor{red}{rgb}{0.7,0,0}
\definecolor{darkgreen}{rgb}{0.0,0.4,0.0}
\definecolor{lightgray}{rgb}{0.7,0.7,0.7}
\definecolor{darkorange}{rgb}{0.75, 0.45, 0}
\definecolor{purple}{rgb}{0.65, 0, 0.75}
\definecolor{goldenrod}{rgb}{0.80, 0.61, 0.11}
\definecolor{lightyellow}{rgb}{0.98,0.94,0.83}


\RequirePackage{fancyvrb}
\RequirePackage[T1]{fontenc}
\RequirePackage{ae}       % ComputerModern Fonts
\RequirePackage{fancyhdr}
\RequirePackage{float}
\RequirePackage{hyperref}
\usepackage{lastpage}

\pagestyle{fancy}
\cfoot{page \thepage\ of \pageref{LastPage}}
\renewcommand{\headrulewidth}{0pt}

% \code mini environment ttfamily->texttt
\newcommand\code{\bgroup\@codex}
\def\@codex#1{{\color{darkred} \normalfont\ttfamily\hyphenchar\font=-1 #1}\egroup}

% This environment defines the look of R ouput
\DefineVerbatimEnvironment{Soutput}{Verbatim}{
  fontsize=\small,
  formatcom=\color{darkblue}
}

\begin{document}
% \SweaveOpts{concordance=TRUE}

\title{ {\Huge Loci of Constant Hue and Chroma} }
\author{Glenn Davis}
\maketitle
% \thispagestyle{fancy}

% Setup stuff.
<<setup, echo=FALSE, results="hide">>=
require("knitr",quietly=TRUE)
opts_chunk$set(fig.path="figs/ag2-", fig.align="center",
  fig.width=7, fig.height=7, comment="")
knit_hooks$set(output = function(x, options) {
  paste('\\begin{Soutput}\n', x, '\\end{Soutput}\n', sep = '')
})
options(width=90)
par( omi=c(0,0,0,0), mai=c(0.2,0.2,0.2,0.2) )
if(!file.exists("figs")) dir.create("figs")
@

% ----------------------------------------------------------------------------
\section*{Introduction}

The plots that Dorothy Nickerson made in \cite{Newhall1943}
and later in \cite{Wyszecki1982}, are beautiful and a labor of love.
But they often do not cover regions of high Chroma.
And obviously one cannot zoom in to inspect problem areas in detail,
which is often useful in development.
The function \code{plotLociHC()} is designed to improve this situation.


<<packs, echo=TRUE, message=FALSE>>=
library( munsellinterpol )
@

\setcounter{figure}{0}  

% ----------------------------------------------------------------------------

\section*{Zoom Out}

Pick value 8/ for demonstration

<<lee11, echo=TRUE, fig.pos="H", fig.height=5, out.width='1.0\\linewidth', fig.cap='Loci of Constant Hue and Chroma for Value 8/' >>=
par( omi=c(0,0,0,0), mai=c(0.41,0.7,0.6,0.2) )
plotLociHC( value=8 )
@
The gridpoints marked with a black disk are published in \cite{Newhall1943}.
The gridpoints marked with a open circle are from \code{all.dat},
and almost certainly described in \cite{Schleter1958}, but not published.
And the other gridpoints are pre-extrapolated by this package;
see the \textbf{munsellinterpol} User Guide for details.
The blue curve is the CIE inverted-U,
and the red curve is the MacAdam limit for the given Value=8.

In \code{plotLociHC()}
by default the largest "real" chroma is determined ($C_{max}$=24 for 8/),
the complete oval at the chroma is computed,
and the plot limits are set for that oval.

Plotting higher chroma in the 10GY area is easy:
<<lee12, echo=TRUE, fig.pos="H", fig.height=5, out.width='1.0\\linewidth', fig.cap='Loci of Constant Hue and Chroma for Value 8/' >>=
par( omi=c(0,0,0,0), mai=c(0.41,0.7,0.6,0.2) )
plotLociHC( value=8, hue=seq(5,55,by=2.5), chroma=seq(2,34,by=2) )
@
And so we see that \code{all.dat} goes up to C=32 for 10GY, 2.5G, and 5G.


% ----------------------------------------------------------------------------

\section*{Zoom In}

For inversion, the region that gave me the most grief was V < 1 and H near 5P (85).
Here is a sample
<<lee13, echo=TRUE, fig.pos="H", fig.height=5, out.width='1.0\\linewidth', fig.cap='Loci of Constant Hue and Chroma for Value 0.4/' >>=
par( omi=c(0,0,0,0), mai=c(0.41,0.7,0.6,0.2) )
plotLociHC( value=0.4, hue=seq(80,90,by=2.5), chroma=seq(2,12,by=2) )
@
It is clear that the curves are not passing through the points from \code{all.dat}.
This is by design, because these points on the opposite side of the hue circle
produced badly injective forward maps using cubic interpolation in V.
The gridlines for non-real points with V<1 were extrapolated using a thin-plate spline
with the function \code{mgcv::gam()}.
It is also clear that this spline is not extraplating lines of constant hue
into straight lines - they bend.
Finding a better way to handle extrapolation with V<1 is an area for a future
version of the package.


% ----------------------------------------------------------------------------

\section*{Bilinear vs Bicubic Interpolation}

In the function \code{MunsellToxyY()} 
the default interpolation method for Hue and Chroma is bicubic.
This interpolation method computes a weighted average of points on a 4x4 grid.
The alternative is bilinear, which only requires a 2x2 grid.
The difference in "flops" is roughly a factor of 4
(the difference in time only about 2 or 3) - but is that really worth it ?
Here is the original plot, except using bilinear interpolation
<<lee14, echo=TRUE, fig.pos="H", fig.height=5, out.width='1.0\\linewidth', fig.cap='Loci of Constant Hue and Chroma for Value 8/, with bilinear interpolation' >>=
par( omi=c(0,0,0,0), mai=c(0.41,0.7,0.6,0.2) )
plotLociHC( value=8, hcinterp='bilinear' )
@
Most of this plot is similar to the first one,
but in the high curvature region near 10GY the lines are perfectly straight
so the plot looks like a spider web.
It is certainly not $C^1$
and does not match Dorothy Nickerson's plots well,
so 'yes' I think that bicubic is worth it.

\bibliographystyle{apalike}
\bibliography{bibliography}


% ----------------------------------------------------------------------------


\section*{Appendix}
This document was prepared \today \quad with the following configuration:
<<finish, echo=FALSE, results="asis">>=
knit_hooks$set(output = function(x, options) { x })
toLatex(sessionInfo(), locale=FALSE)
@


\end{document}

---
title: "Soil Colors"
author: "Glenn Davis"
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 2
    number_sections: false
bibliography: bibliography.bib

# csl: iso690-numeric-brackets-cs.csl
csl: personal.csl
# csl: institute-of-mathematical-statistics.csl
# csl: transactions-on-mathematical-software.csl
vignette: >
  %\VignetteIndexEntry{Soil Colors}
  %\VignetteEngine{knitr::rmarkdown}
---

```{css, echo=FALSE}
body {
  max-width: 725px;     /* make wider, default is 700px */
}
h1{
  font-size: 18pt;    /* make the level 1 headers smaller */
}
```

<script>
function ToggleDiv(ID) {
  var D = document.getElementById("D-" + ID);
  var B = document.getElementById("B-" + ID);
  if (D.style.display === "none") {
    // open the div and change button text
    D.style.display = "block";
    B.innerText = "Hide " + ID;
  } else {
    // close the div and change button text
    D.style.display = "none";
    B.innerText = "Show " + ID;
  }
}
</script>


```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
options( width=100 )
```

<br><br>

#  Introduction



```{r echo=TRUE, message=TRUE}
library(munsellinterpol)
```

The goal of this vignette is to give examples of rounding to the closest official color chip in
one or more Munsell chart sets.
The five chart sets are 
**Soil** @MunsellSoil, **Rock** @MunsellRock, **Bead** @MunsellBead, **Plant** @MunsellPlant, 
and **The New Student Color Set** @Long2011.
For the contents of these color charts we depend completely on Ferguson @Ferguson2014.
The supplementary material from this paper was put in computer readable form by Willie Ondricek,
who greatly improved the wording of this vignette.
One of the fields in this supplementary material is a color name for each chip;
we call this name the `FergusonName`.
In this vignette, we only use examples from the Munsell Soil-Color Charts.

The key function in this vignette is `roundHVC()`.
Other featured functions from **munsellinterpol** are
`XYZtoMunsell()`,
`MunsellNameFromHVC()`,
`ColorBlockFromMunsell()`, and
`MunsellTosRGB()`.
Other packages that play a very important part are
**OxSR** @OxSR, **colorSpec** @colorSpec, and **spacesXYZ** @spacesXYZ.
But we will not attach these and prefix their functions with the namespace
to make it easier to locate them.
The package **flextable** @flextable is used in one section to display some RGB color patches.


<br><br>

# Load and Plot Selected Soil Spectra

The package **OxSR** has some high resolution (0.5 nm) reflectance spectra for 23 soil samples;
these spectra are stored in the object `OxSR::soil_refle`.
Of these 23 we have pre-selected a subset of 6 samples with high variability.
The following function also adds an ideal neutral sample with constant reflectance of 18% -
an 18% gray card @graycard.

```{r echo=TRUE, warning=TRUE }
load_soil <- function( soil_raw=OxSR::soil_refle, subset=c( "a4", "a6", "a8", "a11", "a14", "a15" ),
                                         wave=380:780 )
    {
    #   locate only the wavelengths in OxSR::soil_refle that are relevant for color
    #   the given data.frame OxSR::soil_refle goes way into the infra-red and these are not needed
    #   We already know that all relevant multiples of 1nm are in soil_raw
    idx     = match( wave, soil_raw$wavelength_nm )

    #   extract just a subset of samples with high variability
    #   the reflectances in OxSR::soil_refle are percentages, so divide by 100    
    mat     = as.matrix(soil_raw)[ idx, subset ] / 100
    
    #   convert the matrix mat to a colorSpec object
    soil_spec = colorSpec::colorSpec( mat, wavelength=wave, quantity='reflectance' )
    
    #   add neutral gray as a test sample
    gray      = colorSpec::neutralMaterial( 0.18, wavelength=wave )
    soil_spec = colorSpec::bind( soil_spec, gray )
   
    return( soil_spec ) 
    }
```    

Load the spectra and plot them.
```{r echo=TRUE, warning=TRUE,  fig.width=7.3, fig.height=5.2,  fig.show='hold' }
soil_spec = load_soil()
par( omi=c(0,0,0,0), mai=c(0.6,0.6,0.3,0.3) )
plot( soil_spec, legend="topleft" )
```
The graphed colors here are computed for Illuminant D65.
This may not always be appropriate, so here is another function that
can create and use illuminants of any Correlated Color Temperature (CCT).


<br><br>

# A Function to Compute XYZ and HVC from Reflectance Spectra

In the following function `soil_data()`, argument `soil_spec` is a `colorSpec` object with
reflectance spectra of the soil, and `CCT` is the desired Correlated Color Temperature of the illuminant.

<button id="B-soil_data()" onclick="ToggleDiv('soil_data()')">Show soil_data()</button>
<div id="D-soil_data()" style="display: none">
```{r  echo=TRUE, warning=TRUE, results="hide" }
soil_data <- function( soil_spec, CCT )
    {
    wave    = colorSpec::wavelength(soil_spec) 
    
    #   make illuminant from the CCT. This is the reference illuminant in IES Standard TM-30.
    illum   = colorSpec::referenceSpectraTM30( CCT, wavelength=wave )
    
    #   make scanner from this illuminant and the standard x,y,z responsivities of the human eye
    scanner = colorSpec::product( illum, "soil", colorSpec::xyz1931.1nm, wavelength=wave )
    
    #   scale so the Perfect Reflecting Diffuser has Y=100; which is conventional for Munsell work
    scanner = colorSpec::calibrate( scanner, response=c(NA,100,NA), method='scaling' )
    
    #   compute white point of illum
    PRD     = colorSpec::neutralMaterial(1,wavelength=wave)    # the Perfect Reflecting Diffuser
    white   = colorSpec::product( PRD, scanner )
    
    #   compute XYZ for all soil samples; this is XYZ under illum
    XYZ     = colorSpec::product( soil_spec, scanner )
    
    #   convert XYZ to Munsell HVC
    #   XYZtoMunsell() internally does a chromatic adaptation to the white point of Illuminant C
    #   the default is to use the original NBS variant of the white point of Illuminant C
    HVC     = XYZtoMunsell( XYZ, white=white )   
    rownames(HVC)   = colorSpec::specnames(soil_spec)
    
    # create output data.frame and add 4 columns: HVC, Munsell notation, ISCC-NBS color name, and Lab
    out = data.frame( row.names=rownames(HVC) )
    out$HVC     = HVC
    out$Munsell = MunsellNameFromHVC( HVC )
    out[[ "ISCC-NBS Name" ]] = ColorBlockFromMunsell( HVC )$Name
    out$Lab     = spacesXYZ::LabfromXYZ( XYZ, white )
    
    return( out )
    }
```
</div>
This one is fairly long, but it does a lot of work.


<br><br>

#  Pick an Illuminant and Compute the Munsell HVC for It

When viewing soil samples, these are the official recommendations:

> The visual impression of color from the standard color chips is accurate only
> under standard conditions of light intensity and quality.
> Comparison of soil color to the standard Munsell chips should be done
> without sunglasses and in normal sunlight. ...
> If artificial light is used ... the light source used must be as near to
> the white light of midday as possible.
> --- from _Conditions for measuring color_ in @MunsellSoil

Instead of "normal sunlight" I believe the authors actually meant "normal daylight",
as in the next sentence "light of midday".
The American Cinematographer Manual, p 199 of @Detmers1986,
says that a reasonable Correlated Color Temperature (CCT)
for "Average Summer Daylight" is 6500 K.
So we'll choose the CIE-standardized daylight illuminant with CCT=6500 for this calculation.
Note that the white point of this illuminant is also the white point for the color space sRGB.

Some teams of soil scientists may use viewing conditions that are different,
and also want agreement between human and instrumental Munsell colors.
Here are some possible alternatives to the daylight illuminant with CCT=6500 (D65).
An unofficial page @Seelinger, recommends that viewing is done under
"Natural light conditions (preferably on a cloudy day or in shaded area)".
Reference @Detmers1986 states that a reasonable CCT for "Average Summer Shade" is 8000K.
So if viewing is done in a shaded area, CCT=8000 might be more appropriate.
For viewing indoors under incandescent lamps, warm light with CCT=3000 K might be more appropriate.
For viewing indoors under LEDs, an actual LED spectrum might be more appropriate.
There are many such LED spectra in packages
**photobiologyLEDs**  @photobiologyLEDs 
and 
**photobiologyLamps** @photobiologyLamps.
These can be converted to `colorSpec` objects using 
`photobiologyInOut::as.colorSpec()` @photobiologyInOut.


```{r echo=TRUE, message=TRUE}
dat_soil = soil_data( soil_spec, CCT=6500 )
dat_soil
```
Note that all 6 pre-selected soil samples are within the **YR** (Yellow-Red) Hues.
Actually all 23 soil samples are **YR**, and not just these selected 6 !
Note that for the `Neutral0.18` test sample (not soil), `C=a=b=0` as they should be.



<br><br>

#  Round to the Nearest Chip in the Munsell Soil Charts

The above samples have high precision, but unfortunately that precision makes it
inconvenient to report in a way that researchers familiar with the Munsell Soil Color Charts can relate to.
For the sake of consistency with the traditional way of communicating colors, it is sometimes better
to provide the Munsell notation for the closest color chip in the official Soil Charts.

```{r echo=TRUE, message=TRUE}
dat_rnd     = roundHVC( dat_soil$HVC, books="soil" )
dat_rnd$HVC = NULL     # delete matrix HVC, because it is redundant in this context
dat_rnd
```
The Munsell notations in the `MunsellRounded` column are guaranteed to be in the Soil Color Charts;
note that all 6 are distinct.
The `ISCC-NBS Name`s are computed from the original and precise HVC,
and the `FergusonName`s are computed from `MunsellRounded`.
Note that these names are all different.

Note that 3 of the samples are on the same Hue chart **2.5YR**.
Plot that Hue chart and label those 3 samples:

```{r echo=TRUE, message=TRUE, fig.width=7.3, fig.height=5.2,  fig.show='hold'}
par( omi=c(0,0,0,0), mai=c(0.4,0.4,0.25,0) )
plotPatchesH( "2.5YR",   value=c(2.5,3:8), chroma=c(1,2,3,4,6,8) )
text( c(4,4,3), c(5,4,4), c("a6","a8","a15"), adj=c(0.5,0.5), col="white" )
```
The Value and Chroma vectors match the actual soil color chart, but the chips
here are a superset of those in the chart.
The top-right chip, **2.5YR 8/8**, is inside the object color gamut, but outside the sRGB gamut.
These colors are best viewed on a display that is calibrated for sRGB.


<br><br>

#  Comparisons

Finally, we make a table with side-by-side color comparisons of
the precisely computed color in Munsell notation,
and the Munsell notation after rounding to the closest color chip in the Munsell Soil Color Charts.

<button id="B-table code" onclick="ToggleDiv('table code')">Show table code</button>
<div id="D-table code" style="display: none">
```{r echo=TRUE, message=TRUE}
tbl = data.frame( row.names=1:(2*nrow(dat_soil)) )
tbl$Sample  = as.character( rbind(rownames(dat_soil),'') )
tbl$Precise = as.character( rbind('',MunsellNameFromHVC( dat_soil$HVC, digits=3 )) )
tbl$Rounded = as.character( rbind('',dat_rnd$MunsellRounded ) )

library( flextable )
myrt <- regulartable( tbl )
myrt <- width( myrt, j=c(2,3), width=2 )
myrt <- align( myrt, j=1:3, align='center', part='all' )
myrt <- hline( myrt, i=seq( 2, nrow(tbl)-2, by=2 ), border=fp_border_default( color="black", width=2 ) )
myrt <- hrule( myrt, rule = "exact" )
idxcolor <- seq(1,nrow(tbl)-1,by=2)
myrt <- height( myrt, i=idxcolor, height=1 )
myrt <- bg( myrt, i=idxcolor, j=2, bg=rgb( round(MunsellTosRGB(dat_soil$HVC)$RGB), max=255 ) )
myrt <- bg( myrt, i=idxcolor, j=3, bg=rgb( round(MunsellTosRGB(dat_rnd$MunsellRounded)$RGB), max=255 ) )
```
</div>
<br>
```{r echo=FALSE, message=TRUE}
myrt
```
These colors are best viewed on a display that is calibrated for sRGB.


<br><br>

# References

<div id="refs"></div>



<br><br>

# Session Information

<pre>
```{r, echo=FALSE, results='asis'}
sessionInfo()
```
</pre>

---
title: "munsellinterpol User Guide"
author: "Glenn Davis and Jose Gama"
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 2
    number_sections: false
bibliography: bibliography.bib

# csl: iso690-numeric-brackets-cs.csl
csl: personal.csl
# csl: institute-of-mathematical-statistics.csl
# csl: transactions-on-mathematical-software.csl
vignette: >
  %\VignetteIndexEntry{munsellinterpol User Guide}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
options( width=100 )
library(munsellinterpol)
```


**munsellinterpol** is an **R** package that transforms
from Munsell HVC to CIE xyY, and back again.
Our package was inspired by Paul Centore's
**An Open-Source Inversion Algorithm for the Munsell Renotation** @Centore.

The goals of this package are:

<ol>
<li>in the forward conversion HVC $\to$ xyY, if HVC is in the published tables
(@Newhall1943 and @Judd1956) then the output xyY must match the tables</li>
<li>accurate round-trips HVC $\to$ xyY $\to$ HVC,
and xyY $\to$ HVC$\to$ xyY</li>
<li>conversion for all colors within the MacAdam limits,
including Munsell Values < 1 </li>
<li>both conversions HVC $\leftrightarrow$ xyY are class $C^1$ (continuously differentiable), except at the neutrals where $C^0$ (continuous) is acceptable
</ol>

Goals 1, 2, and 3 have been met.
This means that extensive tests were passed, with many color samples.
There were initial problems with very dark samples where V $<$ 1,
so it is possible there _could_ be some other dark samples
that were missed, and where goal 3 fails.

In attempting to meet goal 4, the forward map HVC $\to$ xyY uses
Catmull-Rom spline interpolation in V (by default).
But because of a change in Value spacing,
the function HVC $\to$ xyY is *not* $C^1$ on the plane V=1.
When V $\ge$ 1 the Value spacing is 1, but when V $\le$ 1 the Value spacing is 0.2.
So goal 4 has *not* been met.
For H and C bicubic interpolation is used, again with Catmull-Rom splines;
see @wiki-spline and @wiki-bicubic.
There are options to use simpler linear interpolation for V,
and bilinear interpolation for H and C.
These are useful for comparison with other algorithms, such as @Rheinboldt1958.
Of course, to meet goal 2 the options for forward and inverse maps must be the same.

Note that goal 1 does **not** include points from `all.dat`.
This goal is actually met for V $\ge$ 1, but for V $<$ 1 we had to ignore the points
from `all.dat` and re-extrapolate to make inversion work reliably.
In general, in this package the xy for these non-real points is further from white
than the corresponding points in `all.dat`.
For the meaning of `all.dat` see **Appendix A**.

The package's strong dependencies are:

<ul>
<li>**rootSolve** @rootSolve - for inverting the forward conversion HVC $\to$ xyY</li>
<li>**spacesXYZ** @spacesXYZ - for conversion between CIE xyY and other CIE spaces</li>
<li>**spacesRGB** @spacesRGB - for conversion between RGB and CIE XYZ</li>
<li>**logger** @logger - for all logging to the console</li>
</ul>
The package **microbenchmark**  @microbenchmark  is suggested,
for its high-precision timer.

Similar packages:
**munsell** @munsell also does HVC $\leftrightarrow$ sRGB conversion,
but only for discrete HVCs from the Munsell Book;
there is no interpolation.


We assume that the reader is familiar with Munsell Hue, Value, and Chroma,
which we abbreviate by HVC.
A good introduction is @Munsell_color_system.
All 3 are stored as floating-point numbers:
H is in the interval (0,100],
V is in [0,10], and
C is non-negative with upper limit a complex function of H and V.
These are cylindrical coordinates on the irregular Munsell object color solid.
We assume that the reader is familiar with the Munsell
character string notation `H V/C` for chromatic colors,
and `N V/` for achromatics (neutrals).

We assume that the reader is familiar with sRGB; a good reference is @sRGB.

We assume that the reader is familiar with the CIE spaces
xyY, XYZ, and Lab.
Y is the _luminance factor_ and in this domain
is considered to be a percentage in the interval [0,100].
A perfectly black surface has Y=0,
and a perfectly reflecting diffuser has Y=100.
An excellent reference is @lindbloom.

There are, and have been, many other software programs that do these conversions.
The earliest one we know of is @Rheinboldt1960 (1960),
which ran on an IBM 704.
See @Centore for a discussion of many other software programs and algorithms.



<br><br>

# Basics
The forward conversion HVC $\to$ xyY is as simple as:

```{r echo=TRUE}
MunsellToxyY( '4.2RP 5.5/8' )
```
The return value is a `data.frame`.
The first column `SAMPLE_NAME` is the input Munsell notation.
The second column `HVC` is its numerical version.
And the 3rd column is the output xyY.
If the input is a character vector of length N, then the `data.frame` has N rows.
The input can also be a numeric matrix HVC with N rows.
There are many options for advanced users, see the man page for details.

And here is an example of the reverse conversion xyY  $\to$  HVC:
```{r echo=TRUE}
xyY = MunsellToxyY( '4.2RP 5.5/8' )$xyY
xyYtoMunsell( xyY )
```
And so the round-trip has returned us to `4.2RP 5.5/8`.
In general, the round-trip is accurate to about 2 decimal places;
the worst case is the Hue at near neutrals.
Note that the return value is also a `data.frame`,
but now `SAMPLE_NAME` comes at the end.

Other color spaces available are: sRGB, XYZ, Lab, Luv, and Colorlab.
All the conversion functions accept multiple "samples".
Some of the functions return a `data.frame` and some return a plain matrix.
See the **Reference Guide** for details.


<br><br>

# Plotting

The most obvious thing to plot is a simulation of a page from the Munsell Book of Color:
```{r, echo=TRUE,  message=TRUE, results='hold', fig.width=7, fig.height=6,  fig.show='hold'}
par( omi=c(0,0,0,0), mai=c(0.6,0.7,0.4,0.2) )
plotPatchesH( "10GY", back='#f7f7f7' )
```

This is the forward conversion 
$\text{HVC} \to \text{xyY}_C \to \text{XYZ}_C \to \text{XYZ}_{D65} \to \text{sRGB}$.
There is no interpolation here, unless the Hue is not a multiple of 2.5, which is allowed.
For each Value, the Chroma is extended up to the MacAdam limit,
and this determines the limit of the Chroma axis.
If the sRGB for a patch is outside the cube (before clamping),
it is not drawn and the sRGB values are printed instead.
This makes it easy to see *why* the patch is outside the sRGB gamut.
The chromatic adaption method, the background color, and the main title can be controlled.

Note that the plot is titled with sRGB, and this plot is best viewed on a display
calibrated for sRGB.
Adobe RGB space is also available.

```{r, echo=TRUE,  message=TRUE, results='hold', fig.width=7, fig.height=6,  fig.show='hold'}
par( omi=c(0,0,0,0), mai=c(0.6,0.7,0.4,0.2) )
plotPatchesH( "10GY", space='AdobeRGB', back='#f7f7f7' )
```
Of course, this one is best viewed on a display calibrated for Adobe RGB.
Note that the gamut is quite a bit larger in this figure.
Users can add custom RGB spaces too.

<br><br>
<p>
In the Munsell arena,
another standard plot is the curves of constant Hue and Chroma.

```{r, echo=TRUE,  message=TRUE, results='hold', fig.width=7, fig.height=6,  fig.show='hold'}
par( omi=c(0,0,0,0), mai=c(0.6,0.7,0.6,0.2) )
plotLociHC( value=8 )
```

The black-filled circles are from `real.dat` (see **Appendix A**),
and the open circles are extrapolated from the "real" ones.
The blue curve is the CIE inverted-U,
and the red curve is the MacAdam limit for the given Value=8.
<p>
One can also plot in the a,b plane.
```{r, echo=TRUE,  message=TRUE, results='hold', fig.width=7, fig.height=6,  fig.show='hold'}
par( omi=c(0,0,0,0), mai=c(0.6,0.7,0.6,0.2) )
plotLociHC( value=8, coords='ab' )
```
The interpolation takes place in x,y and is _then_ mapped to a,b.
For similar plots, see the
<a href="lociHC.pdf">Loci of Contant Hue and Chroma</a> vignette.




<br><br>

# ISCC-NBS Color Names

The ISCC-NBS System is a partition of Munsell Color Solid into 267 color blocks,
see @kelly1976.
Each block has an _ISCC-NBS Name_;
one of the goals of these names is that they should be
"simple enough to be understood by the average person on the street".
Each block is a disjoint union of _elementary blocks_ (of which there are 932)
where an elementary block is defined
by its minimum and maximum limits in Hue, Value, and Chroma.
The _peripheral blocks_ (of which there are 120)
have arbitrary large chroma (maximum Chroma is $\infty$).
Each block has an associated _centroid color_, see @kelly1958,
which is in the interior of the block (even though some blocks are non-convex).
Given a query point HVC, the function `ColorBlockFromMunsell()` searches 
for the one elementary block that contains that point.

Since 2000, the Pantone Color Institute has declared a "Color of the Year",
see @pantone.
And they publish the sRGB coordinates of the that color.
We thought it would be interesting to compare the Pantone
colors with the corresponding ISCC-NBS Names and centroids.
Read the data for all colors since 2010.

```{r echo=TRUE}
path = system.file( 'extdata/PantoneCoY.txt', package='munsellinterpol' )
pantone = read.table( path, header=TRUE, sep='\t', strings=FALSE )
pantone
```
Add the block names and centroids.
```{r echo=TRUE}
pantone$Year = NULL ;   pantone$Code = NULL
RGB = as.matrix( pantone[ , c('R','G','B') ] )
HVC = RGBtoMunsell( RGB, space='sRGB' )
pantone$Munsell = MunsellNameFromHVC( HVC )
block = ColorBlockFromMunsell( HVC )
pantone[[ "ISCC-NBS Name" ]] = block$Name
pantone[[ "ISCC-NBS Centroid" ]] = block$Centroid
pantone
```

Note that `Emerald` is unfortunately outside the sRGB cube.
Now compare the original color, and the centroid approximation to it.
The color match is poor.
```{r echo=TRUE}
color.pant = rgb( RGB, max=255 )
color.cent = rgb( MunsellToRGB( block$Centroid, space='sRGB' )$RGB, max=255 )
tbl = data.frame( row.names=1:nrow(pantone) )
tbl[[ "Pantone Name" ]]  = pantone$Name
tbl[[ "Pantone Color" ]] = '' ;    tbl[[ "Centroid Color" ]] = ''
tbl[[ "ISCC-NBS Name" ]] = block$Name
library( flextable )
myrt <- regulartable( tbl )
myrt <- align( myrt, j=4, align='left', part='all' )
myrt <- align( myrt, j=2:3, align='center', part='all' )
myrt <- hrule( myrt, rule='exact' )
myrt <- height( myrt, height=0.75 )
myrt <- width( myrt, j=c(1,4), width=2 ) ; myrt <- width( myrt, j=2:3, width=2.5 )
myrt <- fontsize( myrt, size=14, part='all' ) ; myrt <- fontsize( myrt, size=16, part='header' )
for( i in 1:nrow(tbl) )
{ myrt <- bg(myrt, i=i, j=2, bg=color.pant[i]) ; myrt <- bg(myrt, i=i, j=3, bg=color.cent[i]) }
myrt
```
<br>
This table is best viewed on a display calibrated for sRGB.
In the 6 levels of color designation
(from coarsest to finest) in  @kelly1976,
ISCC-NBS is Level 3 and Pantone would most likely be Level 4.



<br><br>

# Possible Future Work

Here are a few possible improvements and additions.

<p>
The conversions are already pretty fast.
On my PC with 3GHz Intel Core Duo,
the forward conversion HVC $\to$ xyY takes about 1 msec/sample,
and the reverse xyY $\to$ HVC takes about 8 msec/sample.
This assumes a large number of samples per function call.
The next obvious speed-up is to move the interpolation code
for the forward conversion from **R**  to compiled C.
It would have the added benefit that the 2x2 Jacobian could also be computed
during the inversion xyY $\to$ HVC, 
which would reduce the number of function evaluations 
per Newton-Raphson iteration from 3 to 1.

<p>
For V $<$ 1, it might be possible to restore the extrapolated points from `all.dat`.
But it will probably take better 3D visualization to assist
in viewing the xy irregularities in these very dark colors.



<br><br>

# References

<div id="refs"></div>




<br><br>

\Appendix


<br><br>

## Appendix A - Munsell Renotation Details

For history prior to 1943, please consult @Nickerson1976-I,
@Nickerson1976-II, and  @Nickerson1976-III.
 
The OSA Subcommitte on the Spacing of the Munsell Colors
issued their Final Report in 1943 @Newhall1943.
There were 2734 published coordinates (what might be called _aim points_),
for Munsell Values from 1 to 9.
The work was based on visual interpolation and smoothing on
xy graphs of spectrophotometrically measured physical samples
from the _Munsell Book of Color (1929)_. 
There were 421 samples from @Kelly1943 and 561 samples from @Granville1943.
There were a few obvious transcription errors in this data, which will be discussed below.
After 1943, the Munsell Book of Color used these aim points,
which are called the _renotations_.

In 1956 Judd and Wyszecki extended the renotation
to Values 0.2/, 0.4/, 0.6/, and 0.8/, in  @Judd1956.
The Chromas are /1, /2, /3, /4, /6, and /8.
However, Chroma /1 was obtained by averaging /2 with white,
and /3 was obtained by averaging /2 with /4.
If Chromas /1 and /3 are dropped there are 355 very dark samples that remain.
There were now 2734 + 355 = 3089 renotations, with even Chroma, in 1956.

In 1958, at the Forty-Third Annual Meeting of the Optical Society of America,
a 15-minute talk  @Schleter1958 was given on extrapolation of these renotations
for a computer program in development.
Since the abstract of @Schleter1958 is so short, 
we think it is appropriate to quote it in full:

<blockquote>
<p align="justify">
<font face='serif'>
<b>
TA15.
Extension of the Munsell Renotation System. 
</b>
<font style="font-feature-settings: 'smcp'">
J . C. SCHLETER, DEANE B. JUDD, AND H. J. KEEGAN, 
</font>
<i>National Bureau of Standards, Washington, D. C.</i> -
In order to convert tristimulus data to Munsell renotations by
using the tri-dimensional, linear, 
interpolation program developed at the NBS for the high-speed, digital IBM 704 computer, 
it was necessary to extend the published data of the 
Munsell re-notation system<sup>1,2</sup>
by extrapolation so that, for each point in Munsell color space, 
the enclosing six-sided volume would be defined by its eight corners. 
The approximately 2000 corners to be defined by CIE chromaticity coordinates,
<i>x</i>, <i>y</i>, 
are at a given value level the intersections of the hue and chroma loci that cross the MacAdam limit<sup>3</sup> at that value level and at the next lower level. 
Tentative definitions were found by graphical extrapolation on plots at each of the
14 value levels similar to those published. 
A second set of tentative definitions were found by plotting 
<i>x</i> and <i>y</i> against Munsell value 
for each of the 40 Munsell hues and extrapolating to value 10/. 
The final definitions were found by adjustment of the two tentative sets to produce 
smooth loci in both types of plots. 
Plots at selected value levels will be shown illustrating the extended data. 
(15 min.) <br>
<sup>1</sup> Newhall, Nickerson, and Judd, J. Opt. Soc. Am. <bold>33</bold>, 385 (1943). <br>
<sup>2</sup> D. B. Judd and G. Wyszecki, J. Opt. Soc. Am. <bold>46</bold>, 281 (1956). <br>
<sup>3</sup> D. MacAdam, J. Opt. Soc. Am. <bold>25</bold>, 361 (1935). <br>
</font>
</p>
</blockquote>
The "2000 corners" is the number of points in the extension;
we will call these the _Schleter points_.

The next talk @Rheinboldt1958 was a preliminary report on the above-mentioned software program.
The authors note that
1) the extended grid from @Schleter1958  "consists of approximately 5000 points",
2) the program stores these points on magnetic tape, and
3) an xyY $\to$ HVC conversion takes "approximately 20 seconds per sample".

In 1960 there was an updated report @Rheinboldt1960 on the program.
Rheinboldt and Menard report that the 1956 grid of 3089 points
was extended by 1907 points to yield 4996 points.
So the exact number of Schleter points is 1907.
We also learn that the memory of the NBS IBM 704
"had been increased to 8000 words" for "the second version of the code".
This made it possible to store the entire grid in memory.
They do not report the improved conversion time per sample.

Skip now to contemporary times.
The  _Program of Color Science_  at the Rochester Institute Technology
has posted 2 files: `real.dat` and `all.dat`, see @renotationRIT.
The first file `real.dat` is described as
"those colors listed the original 1943 renotation article" @Newhall1943.
There are 2734 samples in the file, so this is an exact match !
Paul Centore found that before June 2010 there were 5 samples missing,
but this has now been corrected (in 2018).
Note that `real.dat` does *not* include the very dark colors,
even though they are inside the MacAdam limits.

The second file `all.dat` is described this way:
<blockquote>
These are all the Munsell data, including the extrapolated colors. Note that extrapolated colors are in some cases unreal. That is, some lie outsize the Macadam [sic] limits.

This file should be used for those performing multidimensional interpolation to/from Munsell data. You will need the unreal colors in order to completely encompass the real colors, which is required to do the interpolation when near the Macadam limits.
</blockquote>

There are 4995 samples in `all.dat`,
which differs from the count in @Rheinboldt1960 by only 1.
Perhaps the latter count includes Illuminant C ?
These two counts are so close that we are confident
that the extended grid in the NBS software from 1958 survived,
and made its way to the _Program of Color Science_ at RIT,
and then into this package.
The above-quoted abstract is likely all we will ever know about the methods used
to extrapolate the values in `all.dat`.
Note that `all.dat` *does* include the very dark colors.

Now return to the NBS computer program in @Rheinboldt1960.
For the forward conversion HVC $\to$ xy they first determine the two
adjacent planes of constant V that the given point is between.
In both of these two HC planes they use bilinear interpolation to determine xy.
They then linearly interpolate these 2 xy vectors. 
So a single conversion requires a table lookup of 8 points in general.
With this method, the plotted ovoids of constant chroma are 40-sided polygons,
and not the smooth ovoids in the published plots by Dorothy Nickerson.
For an example, which looks more like a spiderweb, see the
<a href="lociHC.pdf">Loci of Contant Hue and Chroma</a> vignette.

To get smooth ovoids,
the default method in **munsellinterpol** uses bicubic interpolation in the HC plane.
The forward conversion is then class $C^1$ in H and C (except at C=0).
To get $C^1$ continuity in V as well, the default conversion uses cubic interpolation
for 4 planes of constant value -  2 above the input value and 2 below.
This requires a table lookup of 64 points in general,
and it requires extending the renotations even more !
In the terminology of _mathematical morphology_ the required binary image
is the _dilation_ of the 4995 points with a 3x3x3 structuring element.
The grid in **munsellinterpol** currently has 7606 points.

The extrapolation was done, again from the grid of 4995 points,
using the function `gam()` in the package **mgcv** 
which is preinstalled with **R** itself.
It is a thin-plate spline model.
A separate model was computed for each of the Munsell values: 0.2,0.4,0.6,0,8,1,...10.
Unfortunately, for V=1.5 (on almost any V near 1.5) the forward conversion HVC $\to$ xy
was badly non-injective for small Chroma in the area of the hue `GY`.
It is because the extrapolated xy values at V=0.8 seem to diverge a lot from V=1 and V=2
(they went too far into halfplane x+y>1).
Recall that a conversion at V=1.5 requires lookup at 4 planes V=0.8,1,2, and 3.
Also note that the weight at V=0.8 is negative.
Cubic interpolation has overshoot and undershoot - 
and linear interpolation (where weights are non-negative) does not.

So for dark colors V<1 we decided to ignore the extrapolations in @Schleter1958,
and re-extrapolate, using a `mgcv::gam()` model, from all _real_ points with V $\le$ 1.
This extrapolation was satisfactory.

<br><br>

## Appendix B - Munsell Renotation Discrepancies

There have been at least 4 versions of the renotation tables.
In chronological order they are:

<ul>
<li>original tables in @Newhall1943</li>
<li>tables in the "Color Bible" @Wyszecki1982</li>
<li>tables in ASTM standard @ASTM-D1535-97</li>
<li>digital file `all.dat` at @renotationRIT mentioned in the previous Appendix
</ul>
I have noticed 4 discrepancies, and there might be more.
They appear to be either simple transcription errors,
or adjustments to make the data smoother.
This package uses `all.dat`, except for **2.5PB 10/2** noted below.

For samples **7.5BG 4/16** and **5BG 1/4** it looks like a simple
transcription error in @Wyszecki1982,
which was corrected in @ASTM-D1535-97 and @renotationRIT.
For **10PB 8/4**  it looks like @ASTM-D1535-97 made a smoothing adjustment,
which was carried over to @renotationRIT.
For **2.5R 9/2** it looks like @renotationRIT made a smoothing adjustment.

Finally, in the extrapolated data in @renotationRIT, we have made a small
smoothing adjustment at **2.5PB 10/2** for this package.
Before the adjustment, the forward mapping HVC $\to$ xyY was not injective
near **N 10/**, which caused the root-finding iteration to cycle.
This was discovered during testing phase;
more ambitious "vetting" software,
which specifically searches for non-injectivity, might uncover more of these.


<br><br>

## Appendix C - Inversion Details

Given an xy target point,
the method in @Rheinboldt1960 searches for the exact quadrilateral that contains xy.
It then solves a pair of quadratic equations to determine H and C.
For bicubic interpolation, a similar explicit solution is not really feasible.
The method in @Centore starts with an initial approximation HC$_0$,
and modifies H and C alternately to converge to an HC that maps to xy,
within a certain tolerance.

We decided to use a general root-solver from the package **rootSolve**
that uses Newton-Raphson iteration.
In our case, this iteration really does double-duty:
the first iteration or two locate the quadrilateral that contains xy,
and the remaining iterations polish the root.
It typically takes about 4 or 5 total iterations.
The theory of Newton's Method only covers functions of class $C^2$ (see @Newton)
and our function is at best $C^1$.
However, it is $C^\infty$ on the quadrilateral interiors,
so if the root is in the interior of a quadrilateral,
and the iteration gets close enought to the root,
then it typically stays inside that quadrilateral and converges.
Even when a root is on the boundary of quadrilateral,
we have found that the _basin of attraction_ is still large enough to find it.
The usual trouble is when the intial estimate is too far from the root,
and then the iterations wanders outside the lookup table itself.
Cycling is another possible problem - the iteration could jump around
from quadrilateral to quadrilateral.
But we have only seen this happen once, when the function was non-injective
near white (mentioned at the end of the previous section).
The quadrilaterals become small triangular slivers there,
which causes more trouble near *all* neutrals.

Instead of HC itself, we decided to use a rectangular system:
$$A = C  \cos \left( \frac{\pi}{50} H \right)   ~~~~~~~~   B = C  \sin \left(  \frac{\pi}{50} H  \right)$$
with the inverse:
$$H=\left( \frac{50}{\pi} \right) \text{atan2}(B,A)  ~~~~~~~~  C = \text{sqrt}( A^2 + B^2 )$$


$A$ and $B$ are analogous to $a$ and $b$ in $Lab$ space.
The symbols $a''$ and $b''$ are used for the same coordinate system
in @Simon1980 page 207.
So now we are trying to invert $AB \to HC \to xy$.
Each iteration requires an extra conversion $AB \to HC$,
but this is very fast.

For the initial estimate, the algorithm in @Centore recommends:
$$A_0 = a/5.5   ~~~~~~~~~  B_0 = b/5.5$$
We use a cubic polynomial in $a$ and $b$ for better accuracy.
A separate polynomial is computed for each V-plane, using **R**'s built-in
function for linear models with no intercept:
```
stats::lm( A ~ polym(a,b,degree=3,raw=TRUE) + 0, ... )
```
and similarly for $B$.
We precompute the model, but only save the coefficients.
For V < 1, a cubic polynomial in $a$ and $b$ was still not good enough,
and it was found that a cubic in $\Delta x$ and $\Delta y$ worked better;
where $\Delta x = x - x_C$ and $\Delta y = y - y_C$,
and $x_C,y_C$ are the chromaticity of Illuminant C.
Evaluation of the polynomial takes much longer than division by 5.5,
but it is still negligible compared to the iterations to come.

The root-solver is called like this:
```
rootSolve::multiroot( forwardfun, AB0, rtol=1.e-8, atol=1.e-6, ctol=0, xy_target=xy_target, ...)
```
The function `forwardfun()` performs the equivalent of `MunsellToxyY()`
but does not call it explicitly.
The argument `AB0` is the initial estimate.
It usually takes 4 or 5 iterations to converge with the indicated tolerances,
which are in the xy plane.







<br><br>

## Appendix D - Glossary

<dl>
<dt>luminance factor</dt>
<dd>
The _luminance factor_ of an object is the quotient of the luminance of
the object (for a specific illuminant and in a specific viewing direction)
divided by that of the perfect reflecting diffuser.
In Munsell work this quotient is denoted by Y and is
a percentage in the interval [0,100].
For Lambertian surfaces, Y does not depend on the viewing direction.
</dd>

<dt>uniform lightness scale</dt>
<dd>
A _uniform lightness scale_ is a reparameterization of luminance factor,
so that equal numerical steps in lightness produce equal perceptual steps.
The Munsell Value is an example of such a scale;
in early investigations it was found
that the luminance factor of the background affects the perception.
Another popular scale is CIE Lightness (1976).
For more on these, see the
<a href="lightness.pdf">Uniform Lightness Scales</a> vignette.
</dd>
</dl>




<br><br>

## Appendix E - Logging

Logging is performed using the package **logger**  @logger.
This is a powerful package that allows a separate configuration
for logging from within **munsellinterpol**, and that is what I have done.
During package loading, the logging threshold is changed from `INFO` to `WARN`.
To change it back again, one can execute:  
`logger::log_threshold( logger::INFO, namespace="munsellinterpol" )     # preferred`  
or  
`library( logger )                                                # not preferred`  
`log_threshold( INFO, namespace="munsellinterpol" )`

The layout callback function is customized;
it adds the name of the calling function to the message.
To change it back again, one can execute:  
`log_layout( layout_simple, namespace="munsellinterpol" )`  
or to install ones own layout function, one can execute:  
`log_layout( <your function>, namespace="munsellinterpol" )`.

The appender callback functions is also customized;
it comes to an immediate stop if the log event level is `FATAL`.
To continue execution after such a log event, one can execute:  
`log_appender( appender_console, namespace="munsellinterpol" )`

The formatter callback function is initialized to be `formatter_sprintf()`;
this should not be changed.





<br><br>

## Session Information

<pre>
```{r, echo=FALSE, results='asis'}
sessionInfo()
```
</pre>

---
title: "Dental Shade Guides"
author: "Glenn Davis"
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 2
    number_sections: false
bibliography: bibliography.bib

# csl: iso690-numeric-brackets-cs.csl
csl: personal.csl
# csl: institute-of-mathematical-statistics.csl
# csl: transactions-on-mathematical-software.csl
vignette: >
  %\VignetteIndexEntry{Dental Shade Guides}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
options( width=100 )
```

A *Dental Shade Guide* is a set of simulated teeth used to select prosthetic teeth by color.
The simulated teeth are made of plastic or porcelain.
Commercial shade guides have existed for almost a century.
In 1933, Clark @CLARK1933 discussed the manufacture and use of porcelain shade guides based on the
cylindrical color dimensions of _hue_, _brilliance_, and _saturation_,
which correspond to Munsell's _Hue_, _Value_, and _Chroma_.

In @OBRIEN1989, the 24 teeth in a master Bioform shade guide were measured with a spectrophotometer.
The spectral measurements were converted to xyY assuming Illuminant C,
and then to both CIE Lab and Munsell HVC.
The goal of this vignette is to plot the 24 colors as square patches,
and then check the calculations from the article.

The spectrophotometer was an ACTA CIII from Beckman Instruments.
Relative reflectance data were recorded from 410 to 700nm at 10-nm intervals.
The conversion from such data to xyY and Lab is standard,
but conversion to Munsell HVC is not explained in the paper.
However, in a similar article @OBRIEN1990 by the same authors they state that:

> The chromaticity coordinates were converted to Munsell notation by means of graphs ...
> and the method described by ASTM standard D 1535-80.

so we assume that the same method was used here.
Note that the graphical method only applies to Munsell Hue and Chroma;
for Munsell Value lookup tables are available in @ASTM-D1535-97.
Below we check the published conversions against numerical conversions using **munsellinterpol**.

Load the required R packages.

```{r echo=TRUE, message=FALSE}
library(munsellinterpol)
library(spacesRGB)  # for converting to RGB and plotting the patches
library(spacesXYZ)  # for xyY<->XYZ and Chromatic Adaptation Transform
```
Featured functions from **munsellinterpol** are
`XYZtoMunsell()`,
`MunsellNameFromHVC()`,
`NickersonColorDifference()`, and
`ColorBlockFromMunsell()`.



## Reading and Plotting the Data

Read the published data table.

```{r echo=TRUE}
path = system.file( 'extdata/dental.txt', package='munsellinterpol' )
dental = read.table( path, header=TRUE, sep='\t', stringsAsFactors=FALSE )
dental
```
Extract xyY, adapt from Illuminant C to D65, convert XYZ to sRGB, and display as a 6x4 grid of patches.

```{r echo=TRUE,  fig.width=7.2, fig.height=5.2,  fig.show='hold'}
xyY = as.matrix( dental[ c('x','y','Y') ] )
XYZ = XYZfromxyY( xyY ) / 100
# adapt from Illuminant C to the whitepoint of sRGB, which is D65
# make the Chromatic Adaptation Transform
theCAT = spacesXYZ::CAT( 'C', getWhiteXYZ('sRGB',which='display') )
XYZ = adaptXYZ( theCAT, XYZ )
# create data.frame obj for plotting
obj = expand.grid( LEFT=1:6, TOP=1:4 )
obj$WIDTH = 0.9
obj$HEIGHT = 0.9
obj$RGB = RGBfromXYZ( XYZ, space='sRGB' )$RGB     # convert to sRGB
rownames(obj) = rownames(dental)
# plot as square patches
par( omi=c(0,0,0,0), mai=c(0.1,0.1,0.1,0.1) )
plotPatchesRGB( obj, which='signal', labels="bottomleft", adj=c(-0.2,-0.5), cex=0.7 )
```
This figure is best viewed on a display calibrated for sRGB.



## Checking the Conversions

We now recompute Lab and Munsell values, and check against the published values.

```{r echo=TRUE}
Lab  = as.matrix( dental[ c('L','a','b') ] )
XYZ = XYZfromxyY( xyY )
Lab2 = LabfromXYZ( XYZ/100, 'C' )     # recompute Lab
HVC  = HVCfromMunsellName( dental$Munsell )
HVC2 = XYZtoMunsell( XYZ )            # recompute Munsell HVC
comp = data.frame( row.names=rownames(dental) )
comp$Y  = dental$Y
comp$L  = Lab[ ,1]
comp$L2 = round(Lab2[ ,1],4)
comp$Ldiff = round( comp$L - comp$L2, 4 )
comp$DeltaE = round( DeltaE( Lab, Lab2 ), 4 )   # DeltaE is the pairwise color difference
comp$Munsell = dental$Munsell
comp$Munsell2 = MunsellNameFromHVC( HVC2, format='f', digits=2 )
comp$NickersonCD = round( NickersonColorDifference( HVC, HVC2 ), 4 )
comp
```

The Lab agreement is good, but the published Lightness values are consistently too large
and the reason for this is unknown.
The exception is B-83 whose published Lightness=71.52 is too small and with the largest DeltaE by far;
it appears to be a transcription error.
<br>
The Munsell agreement is not bad, but the pubished Munsell Value is too small.
This could be due to using magnesium oxide instead of the perfect reflecting diffuser.
We can test this by recomputing the value component of `HVC2`.
The newly recomputed Munsell notation is denoted `Munsell3`.

```{r echo=TRUE}
HVC3 = HVC2
HVC3[ ,2] = VfromY( dental$Y, which='MgO' )
comp$Munsell2 = NULL
comp$NickersonCD = NULL
comp$Munsell3 = MunsellNameFromHVC( HVC3, format='f', digits=2 )
comp$NickersonCD = round( NickersonColorDifference( HVC, HVC3 ), 4 )
comp
```
The Munsell Value agreement is now much better.
The worst Nickerson difference is for B-65 and this is largely because
the published Munsell Value is 7.30 instead of 7.40.
Note that the Y values for B-65 and B-54 are almost identical
and that the Munsell Value for B-54 is correct.
So again I think that the problem with B-65 is transcription error.
For Munsell Hue and Chroma the agreement is good.
It must have been tedious to use the graphical method for all 24 samples.
<br>


## ISCC-NBS Color Names

I thought it would be interesting to display the ISCC-NBS names
for each of the 24 dental shades.
```{r echo=TRUE}
obj = data.frame( row.names=rownames(dental) )
obj$Munsell2 = MunsellNameFromHVC( HVC2, format='f', digits=2 )
block = ColorBlockFromMunsell( HVC2 )
obj[[ "ISCC-NBS Name" ]] = block$Name
obj
```
All the dental shades are in the same block, except for 2.
It would be interesting to turn this into a 3D scatterplot,
with the color block boundaries displayed with transparency.


## References

<div id="refs"></div>



## Session Information

<pre>
```{r, echo=FALSE, results='asis'}
sessionInfo()
```
</pre>

---
title: "Soil Colors"
author: "Glenn Davis"
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 2
    number_sections: false
bibliography: bibliography.bib

# csl: iso690-numeric-brackets-cs.csl
csl: personal.csl
# csl: institute-of-mathematical-statistics.csl
# csl: transactions-on-mathematical-software.csl
vignette: >
  %\VignetteIndexEntry{Soil Colors}
  %\VignetteEngine{knitr::rmarkdown}
---

```{css, echo=FALSE}
body {
  max-width: 725px;     /* make wider, default is 700px */
}
h1{
  font-size: 18pt;    /* make the level 1 headers smaller */
}
```

<script>
function ToggleDiv(ID) {
  var D = document.getElementById("D-" + ID);
  var B = document.getElementById("B-" + ID);
  if (D.style.display === "none") {
    // open the div and change button text
    D.style.display = "block";
    B.innerText = "Hide " + ID;
  } else {
    // close the div and change button text
    D.style.display = "none";
    B.innerText = "Show " + ID;
  }
}
</script>


```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
options( width=100 )
```

<br><br>

#  Introduction



```{r echo=TRUE, message=TRUE}
library(munsellinterpol)
```

The goal of this vignette is to give examples of rounding to the closest official color chip in
one or more Munsell chart sets.
The five chart sets are 
**Soil** @MunsellSoil, **Rock** @MunsellRock, **Bead** @MunsellBead, **Plant** @MunsellPlant, 
and **The New Student Color Set** @Long2011.
For the color chip data we depend completely on Ferguson @Ferguson2014.
The supplementary material from this paper was put in computer readable form by Willie Ondricek,
who greatly improved the wording of this vignette.
One of the fields in this supplementary material is a color name for each chip;
we call this name the `FergusonName`.
In this vignette, we only use examples from the Munsell Soil-Color Charts.

The key function in this vignette is `roundHVC()`.
Other featured functions from **munsellinterpol** are
`XYZtoMunsell()`,
`MunsellNameFromHVC()`, and
`ColorBlockFromMunsell()`.
Other packages that play a very important part are
**OxSR** @OxSR, **colorSpec** @colorSpec, and **spacesXYZ** @spacesXYZ.
But we will not attach these and prefix their functions with the namespace
to make it easier to locate them.
The package **flextable** @flextable is used in one section to display some RGB color patches.


<br><br>

# Load and Plot Selected Soil Spectra

The package **OxSR** has some high resolution (0.5 nm) reflectance spectra for 23 soil samples;
these spectra are stored in the object `OxSR::soil_refle`.
Of these 23 we have pre-selected a subset of 6 samples with high variability.
The following function also adds an ideal neutral sample with constant reflectance of 18% -
an 18% gray card @graycard.

```{r echo=TRUE, warning=TRUE }
load_soil <- function( soil_raw=OxSR::soil_refle, subset=c( "a4", "a6", "a8", "a11", "a14", "a15" ),
                                         wave=380:780 )
    {
    #   locate only the wavelengths in OxSR::soil_refle that are relevant for color
    #   the given data.frame OxSR::soil_refle goes way into the infra-red and these are not needed
    #   We already know that all relevant multiples of 1nm are in soil_raw
    idx     = match( wave, soil_raw$wavelength_nm )

    #   extract just a subset of samples with high variability
    #   the reflectances in OxSR::soil_refle are percentages, so divide by 100    
    mat     = as.matrix(soil_raw)[ idx, subset ] / 100
    
    #   convert the matrix mat to a colorSpec object
    soil_spec = colorSpec::colorSpec( mat, wavelength=wave, quantity='reflectance' )
    
    #   add neutral gray as a test sample
    gray      = colorSpec::neutralMaterial( 0.18, wavelength=wave )
    soil_spec = colorSpec::bind( soil_spec, gray )
   
    return( soil_spec ) 
    }
```    

Load the spectra and plot them.
```{r echo=TRUE, warning=TRUE,  fig.width=7.3, fig.height=5.2,  fig.show='hold' }
soil_spec = load_soil()
par( omi=c(0,0,0,0), mai=c(0.6,0.6,0.3,0.3) )
plot( soil_spec, legend="topleft" )
```
The graphed colors here are computed for Illuminant D65.
This may not always be appropriate, so here is another function that
can create and use illuminants of any Correlated Color Temperature (CCT).


<br><br>

# A Function to Compute HVC and Lab from Reflectance Spectra

In the following function `soil_data()`, argument `soil_spec` is a `colorSpec` object with
reflectance spectra of the soil, and `CCT` is the desired Correlated Color Temperature of the illuminant.

<button id="B-soil_data()" onclick="ToggleDiv('soil_data()')">Show soil_data()</button>
<div id="D-soil_data()" style="display: none">
```{r  echo=TRUE, warning=TRUE, results="hide" }
soil_data <- function( soil_spec, CCT )
    {
    wave    = colorSpec::wavelength(soil_spec) 
    
    #   make illuminant from the CCT. This is the reference illuminant in IES Standard TM-30.
    illum   = colorSpec::referenceSpectraTM30( CCT, wavelength=wave )
    
    #   make scanner from this illuminant and the standard x,y,z responsivities of the human eye
    scanner = colorSpec::product( illum, "soil", colorSpec::xyz1931.1nm, wavelength=wave )
    
    #   scale so the Perfect Reflecting Diffuser has Y=100; which is conventional for Munsell work
    scanner = colorSpec::calibrate( scanner, response=c(NA,100,NA), method='scaling' )
    
    #   compute white point of illum
    white   = colorSpec::product( colorSpec::neutralMaterial(1,wavelength=wave), scanner )
    
    #   compute XYZ for all soil samples; this is XYZ under illum
    XYZ     = colorSpec::product( soil_spec, scanner )
    
    #   compute Lab for all soil samples, under illum
    Lab     = spacesXYZ::LabfromXYZ( XYZ, white )

    #   convert Lab to Munsell HVC, which automatically does a chromatic adaptation to Illuminant C
    HVC     = LabtoMunsell( Lab, white=white )   #  the default is to use the original NBS variant of C
    rownames(HVC)   = colorSpec::specnames(soil_spec)
    
    # create output data.frame and add 4 columns: HVC, Munsell notation, ISCC-NBS color name, and Lab
    out = data.frame( row.names=rownames(HVC) )
    out$HVC     = HVC
    out$Munsell = MunsellNameFromHVC( HVC )
    out[[ "ISCC-NBS Name" ]] = ColorBlockFromMunsell( HVC )$Name
    out$Lab     = Lab
    
    return( out )
    }
```
</div>
This one is fairly long, but it does a lot of work.


<br><br>

#  Pick an Illuminant and Compute the Munsell HVC for It

When viewing soil samples, these are the official recommendations:

> The visual impression of color from the standard color chips is accurate only
> under standard conditions of light intensity and quality.
> Comparison of soil color to the standard Munsell chips should be done
> without sunglasses and in normal sunlight. ...
> If artificial light is used ... the light source used must be as near to
> the white light of midday as possible.
> --- from _Conditions for measuring color_ in @MunsellSoil

Instead of "normal sunlight" I believe the authors actually meant "normal daylight",
as in the next sentence "light of midday".
The American Cinematographer Manual, p 199 of @Detmers1986,
says that a reasonable Correlated Color Temperature (CCT)
for "Average Summer Daylight" is 6500 K.
So we'll choose the CIE-standardized daylight illuminant with CCT=6500 for this calculation.
Note that the white point of this illuminant is also the white point for the color space sRGB.

Some teams of soil scientists may use viewing conditions that are different,
and also want agreement between human and instrumental Munsell colors.
Here are some possible alternatives to the daylight illuminant with CCT=6500 (D65).
An unofficial page @Seelinger, recommends that viewing is done under
"Natural light conditions (preferably on a cloudy day or in shaded area)".
Reference @Detmers1986 states that a reasonable CCT for "Average Summer Shade" is 8000K.
So if viewing is done in a shaded area, CCT=8000 might be more appropriate.
For viewing indoors under incandescent lamps, warm light with CCT=3000 K might be more appropriate.
For viewing indoors under LEDs, an actual LED spectrum might be more appropriate.
There are many such LED spectra in packages
**photobiologyLEDs**  @photobiologyLEDs 
and 
**photobiologyLamps** @photobiologyLamps.
These can be converted to `colorSpec` objects using 
`photobiologyInOut::as.colorSpec()` @photobiologyInOut.


```{r echo=TRUE, message=TRUE}
dat_soil = soil_data( soil_spec, CCT=6500 )
dat_soil
```
Note that all 6 pre-selected soil samples are within the **YR** (Yellow-Red) Hues.
Actually all 23 soil samples are **YR**, and not just these selected 6 !
Note that for the `Neutral0.18` test sample (not soil), `C=a=b=0` as they should be.



<br><br>

#  Round to the Nearest Chip in the Munsell Soil Charts

The above samples have high precision, but unfortunately that precision makes it
inconvenient to report in a way that researchers familiar with the Munsell Soil Color Charts can relate to.
For the sake of consistency with the traditional way of communicating colors, it is sometimes better
to provide the Munsell notation for the closest color chip in the official Soil Charts.

```{r echo=TRUE, message=TRUE}
dat_rnd     = roundHVC( dat_soil$HVC, books="soil" )
dat_rnd$HVC = NULL     # delete matrix HVC, because it is redundant in this context
dat_rnd
```
The Munsell notations in the `MunsellRounded` column are guaranteed to be in the Soil Color Charts;
note that all 6 are distinct.
The `ISCC-NBS Name`s are computed from the original and precise HVC,
and the `FergusonName`s are computed from `MunsellRounded`.
Note that these names are all different.

Note that 3 of the samples are on the same Hue chart **2.5YR**.
Plot that Hue chart and label those 3 samples:

```{r echo=TRUE, message=TRUE, fig.width=7.3, fig.height=5.2,  fig.show='hold'}
par( omi=c(0,0,0,0), mai=c(0.4,0.4,0.25,0) )
plotPatchesH( "2.5YR",   value=c(2.5,3:8), chroma=c(1,2,3,4,6,8) )
text( c(4,4,3), c(5,4,4), c("a6","a8","a15"), adj=c(0.5,0.5), col="white" )
```
The Value and Chroma vectors match the actual soil color chart, but the chips
here are a superset of those in the chart.
The top-right chip, **2.5YR 8/8**, is inside the object color gamut, but outside the sRGB gamut.
These colors are best viewed on a display that is calibrated for sRGB.


<br><br>

#  Comparisons

Finally, we make a table with side-by-side color comparisons of
the precisely computed color in Munsell notation,
and the Munsell notation after rounding to the closest color chip in the Munsell Soil Color Charts.

<button id="B-table code" onclick="ToggleDiv('table code')">Show table code</button>
<div id="D-table code" style="display: none">
```{r echo=TRUE, message=TRUE}
tbl = data.frame( row.names=1:(2*nrow(dat_soil)) )
tbl$Sample  = as.character( rbind(rownames(dat_soil),'') )
tbl$Precise = as.character( rbind('',MunsellNameFromHVC( dat_soil$HVC, digits=3 )) )
tbl$Rounded = as.character( rbind('',dat_rnd$MunsellRounded ) )

library( flextable )
myrt <- regulartable( tbl )
myrt <- width( myrt, j=c(2,3), width=2 )
myrt <- align( myrt, j=1:3, align='center', part='all' )
myrt <- hline( myrt, i=seq( 2, nrow(tbl)-2, by=2 ), border=fp_border_default( color="black", width=2 ) )
myrt <- hrule( myrt, rule = "exact" )
idxcolor <- seq(1,nrow(tbl)-1,by=2)
myrt <- height( myrt, i=idxcolor, height=1 )
myrt <- bg( myrt, i=idxcolor, j=2, bg=rgb( round(MunsellTosRGB(dat_soil$HVC)$RGB), max=255 ) )
myrt <- bg( myrt, i=idxcolor, j=3, bg=rgb( round(MunsellTosRGB(dat_rnd$MunsellRounded)$RGB), max=255 ) )
```
</div>
<br>
```{r echo=FALSE, message=TRUE}
myrt
```
These colors are best viewed on a display that is calibrated for sRGB.


<br><br>

# References

<div id="refs"></div>



<br><br>

# Session Information

<pre>
```{r, echo=FALSE, results='asis'}
sessionInfo()
```
</pre>

\name{roundHVC}
\alias{roundHVC}

\title{Round a Given Munsell HVC to the Closest Chip in One or More Books of Chips}

\description{
The functions in this package compute Munsell HVC coordinates with high precision.
But when reporting the Munsell notation it is often desirable,
for historical consistency or other reasons,
to report a notation that is actually appears in one or more published books of
Munsell chips.
This rounding operation requires exact knowledge of such books,
which is taken from the data frame \code{\link{MunsellBooks}}.
}

\usage{roundHVC( HVC, books )}

\arguments{
\item{HVC}{a numeric Nx3 matrix with Munsell HVC values in the rows, or a vector that can be converted to such a matrix, by row.
\cr
\code{HVC} can also be a character N-vector with Munsell Notations,
which is converted to an Nx3 matrix using \code{\link{HVCfromMunsellName}()}. 
}

\item{books}{a comma-delimited string that specifies one or more books of chips.
The words in the string can be 
\code{'soil'}, \code{'rock'}, \code{'bead'}, \code{'plant'}, \code{'newstudent'},
\code{'glossy'}, and \code{'matte'}.
Matching is partial (initial substring only) and case-insensitive.
See \bold{References} and \bold{Examples}.
}
}

\value{
A \code{data.frame} with a N rows and these columns:

\item{HVC}{the given Nx3 input matrix HVC.}

\item{ISCC-NBS Name}{the ISCC-NBS Name (the color name) of the given input chip, see \code{\link{ColorBlockFromMunsell}()}. }

\item{MunsellRounded}{the Munsell notation that is rounded from HVC.
It is guaranteed to be in one of the books given by the argument \code{books}.}

\item{FergusonName}{the color name of \code{MunsellRounded}, as assigned by \bold{Ferguson}.
If the chip does not appear in one of the 5 specialized books in \code{\link{MunsellBooks}},
it is set to \code{NA}.}

The row names are set to the row names of HVC.
But if these are \code{NULL} they are set to the Munsell notation of the input HVC.
Finally, if these row names have duplicates, they are set to \code{1:N}.
}


\details{
First, the set of chips over all selected books is formed - a sub-frame of \code{\link{MunsellBooks}}.
Then the closest chip in this set to the given HVC is found.
For the color distance, the function
\code{\link{NickersonColorDifference}()} is used.
Unfortunately, the particular coefficients in this difference equation can lead to large regions
of Munsell color space with exact ties for the two closest chips.
These exact ties make the final result unpredictable and sometime unintuitive.
As a tie-breaker, a very small multiple (\code{1.e-6}) of plain
Eucliden distance is added to the Nickerson difference, which has proven effective.
}


\references{
\bold{Ferguson, Jonathan}.
\emph{Munsell notations and color names: Recommendations for archaeological practice}.
\bold{Journal of Field Archaeology}. 39:4. 327-335. (2014).
Online Supplement #1.
}

\author{Glenn Davis}


\seealso{
\code{\link{MunsellBooks}},
\code{\link{ColorBlockFromMunsell}()}
}

\examples{
## search the soil book for the best match for 3 chips
roundHVC( c("7.7YR 3.4/6.1", "2.6PB 6.1/4.5", "N 6.6/" ), books='soil' )

##               HVC.H HVC.V HVC.C ISCC-NBS Name MunsellRounded Ferguson Name
## 7.7YR 3.4/6.1  17.7   3.4   6.1  strong brown      7.5YR 4/6  Strong brown
## 2.6PB 6.1/4.5  72.6   6.1   4.5     pale blue        5PB 6/1   Bluish gray
## N 6.6/          0.0   6.6   0.0    light gray           N 7/    Light gray

## The middle chip has a poor match, since "purplish-blue" with Chroma=4.5
## is not well-matched by soil color chips.
## Search again, but this time in 5 books.

roundHVC( c("7.7YR 3.4/6.1", "2.6PB 6.1/4.5", "N 6.6/" ), books='s,b,r,p,n' )

##               HVC.H HVC.V HVC.C ISCC-NBS Name MunsellRounded Ferguson Name
## 7.7YR 3.4/6.1  17.7   3.4   6.1  strong brown      7.5YR 4/6  Strong brown
## 2.6PB 6.1/4.5  72.6   6.1   4.5     pale blue        5PB 6/4     Pale blue
## N 6.6/          0.0   6.6   0.0    light gray           N 7/    Light gray

## This time, for the middle chip, a better match was found.
## The other 2 chip are unchanged.
}

\name{MunsellToxyY}
\alias{MunsellToxyY}

\title{Convert Munsell HVC to xyY coordinates}

\description{\code{MunsellToxyY} Converts Munsell HVC to xyY coordinates, by interpolating over the
extrapolated Munsell renotation data.

Note that the returned \code{xyY} is relative to the white point of Illuminant C.
Since Illuminant C is obsolete, it is almost certain that if \code{xyY} is used in subsequent
calculations, it must be chromatically adapted to a different white point.
This chromatic adaptation can be done using
\code{\link[spacesXYZ:adaptxyY]{spacesXYZ::adaptxyY()}}.
Alternatively, one can use \code{\link{MunsellToXYZ}()},
which does the chromatic adaptation internally.
}

\usage{MunsellToxyY( MunsellSpec, xyC='NBS', hcinterp='bicubic', vinterp='cubic',
                     YfromV='ASTM', warn=TRUE ) }

\arguments{
\item{MunsellSpec}{ a numeric Nx3 matrix or a vector that can be converted to
such a matrix.
Each row has Munsell HVC, where H is Hue Number,
and V and C are the standard Munsell Value and Chroma.
The Hue is automatically wrapped to the interval (0,100].
\code{MunsellSpec} can also be a character N-vector with standard Munsell notation; it is converted to an Nx3 matrix.}

\item{xyC}{ a numeric 2-vector with xy chromaticity of Illuminant C.
It can also be one of the strings in the first column of this table;
it is then replaced by the corresponding xy in the second column.
\tabular{lll}{
             \tab   xy white point \tab   reference \cr
\code{'NBS'} \tab  \code{c(0.3101,0.3163)} \tab Kelly, et. al. [RP1549] (1943). Rheinboldt et al. (1960) \cr
\code{'JOSA'} \tab  \code{c(0.31012,0.31631)} \tab Judd, Deane B. (1933) \cr
\code{'NTSC'} \tab  \code{c(0.310,0.316)}  \tab  NTSC (1953) \cr
\code{'CIE'}  \tab  \code{c(0.31006,0.31616)}  \tab  CIE:15 2004\cr
}
The default \code{'NBS'} is probably what is intended by \cite{Newhall et. al.}
although no xy for C appears in that paper.
This is the C used in the first computer program for conversion:
\emph{Rheinboldt et al. (1960)}.
The other options are provided so that a neutral Munsell chip has the xy that
the user expects.  Alternative values of \code{xyC} should not be \emph{too far}
from the above.
If \code{hcinterp} is \code{'bicubic'},
this parameter only affects chips with Chroma \eqn{<} 4 (except Chroma=2).
If \code{hcinterp} is \code{'bilinear'},
this parameter only affects chips with Chroma \eqn{<} 2.
}

\item{hcinterp}{the type of 2D interpolation for H and C; either \code{'bicubic'} or \code{'bilinear'} (partial matching enabled).
In the bicubic case, for a general input point, the output value is interpolated
using a 4x4 subgrid of the lookup table, and the interpolation function is class \eqn{C^1}
(except at the neutrals).
In the bilinear case, the interpolation uses a 2x2 subgrid,
and the function is class \eqn{C^0}.
}

\item{vinterp}{the type of interpolation for V; either \code{'cubic'} or \code{'linear'} (partial matching enabled).
In the cubic case, for a general input point, the output value is interpolated
using 4 planes of constant Value, and the interpolation function is class \eqn{C^1}.
In the linear case, the interpolation uses 2 planes and the function is class \eqn{C^0}.
}

\item{YfromV}{passed as the parameter \code{which} to the function \code{YfromV()}.
See \code{\link{YfromV}()} for details.
Option \code{'MGO'} is not allowed because then Y>100 when V=10.
}

\item{warn}{if a chip cannot be mapped (usually because the Chroma is too large),
its x and y are set to \code{NA} in the returned
\code{data.frame}.
Just before returning, if any rows have \code{NA}, and this argument is \code{TRUE},
then a warning is logged.
}
}

\note{
When making plots in planes of constant Value,
option \code{hcinterp='bicubic'} makes fairly smooth ovals,
and \code{hcinterp='bilinear'} makes polygons.
The ovals are smooth even when \code{vinterp='linear'},
but the function is not class \eqn{C^1} at the planes of integer Value.
To get a fully \eqn{C^1} function (except at the neutrals and on the plane V=1),
\code{hcinterp} and \code{vinterp} must be set to the defaults.
}

\section{Warning}{
Even when \code{vinterp='cubic'} the function HVC \rarrow xyY is not \eqn{C^1}
on the plane V=1.
This is because of a change in Value spacing:
when V\eqn{\ge}1 the Value spacing is 1, but when V\eqn{\le}1 the Value spacing is 0.2.
}

\value{
a \code{data.frame} with these columns:
\item{SAMPLE_NAME}{the original \code{MunsellSpec} if that was a character vector.
Or the Munsell notation string converted from the input matrix HVC.}
\item{HVC}{the input Nx3 matrix, or the HVC matrix converted from the input Munsell notation}
\item{xyY}{the computed output matrix, with CIE xyY of MunsellSpec,
relative to the white point of Illuminant C.
In case of error, x and y are set to \code{NA}.
The rownames of xyY are set to those of HVC, unless they are NULL when they are set to \code{SAMPLE_NAME}.}
}

\details{
In case \code{hcinterp='bicubic'} or \code{vinterp='cubic'} a Catmull-Rom spline is used;
see the article \emph{Cubic Hermite spline}.
This spline has the nice property that it is \emph{local} and requires at most 4 points.
And if the knot spacing is uniform:
1) the resulting spline is \eqn{C^1},
2) if the knots are on a line, the interpolated points are on the line too.
}

\source{
Paul Centore 2014
The Munsell and Kubelka-Munk Toolbox
\MKMTB\cr
\url{https://www.rit.edu/science/munsell-color-lab}\cr
\url{https://www.rit-mcsl.org/MunsellRenotation/all.dat}\cr
\url{https://www.rit-mcsl.org/MunsellRenotation/real.dat}
}

\references{
Judd, Deane B.
\bold{The 1931 I.C.I. Standard Observer and Coordinate System for Colorimetry}.
Journal of the Optical Society of America.
Vol. 23. pp. 359-374.
October 1933.

Newhall, Sidney M.,
Dorothy Nickerson,
Deane B. Judd.
\bold{Final Report of the O.S.A. Subcommitte on the Spacing of the Munsell Colors}.
Journal of the Optical Society of America.
Vol. 33.  No. 7.  pp. 385-418.
July 1943.

Kelly, Kenneth L.
Kasson S. Gibson.
Dorothy Nickerson.
\bold{Tristimulus Specification of the Munsell Book of Color from Spectrophometric
Measurements}
National Bureau of Standards RP1549
Volume 31. August 1943.

Judd, Deane B. and Günther Wyszecki.
\bold{Extension of the Munsell Renotation System to Very Dark Colors}.
Journal of the Optical Society of America.
Vol. 46.  No. 4.  pp. 281-284.
April 1956.

National Television System Committee.
[Report and Reports of Panel No. 11, 11-A, 12-19,
with Some supplementary references cited in the Reports,
and the Petition for adoption of transmission standards
for color television before the Federal Communications Commission]
(1953)

Rheinboldt, Werner C. and
John P. Menard.
\bold{Mechanized Conversion of Colorimetric Data to Munsell Renotations}.
Journal of the Optical Society of America.
Vol. 50, Issue 8, pp. 802-807.
August 1960.

Wikipedia.
Cubic Hermite spline.
\url{https://en.wikipedia.org/wiki/Cubic_Hermite_spline}

Paul Centore 2014
The Munsell and Kubelka-Munk Toolbox
\MKMTB
}

\author{Jose Gama and Glenn Davis}

\seealso{
\code{\link{xyYtoMunsell}()},
\code{\link{MunsellToXYZ}()},
\code{\link[spacesXYZ:adaptxyY]{spacesXYZ::adaptxyY()}}
}


\examples{
MunsellToxyY( '7.6P 8.9/2.2' )
##    SAMPLE_NAME  HVC.H HVC.V HVC.C      xyY.x      xyY.y      xyY.Y
##  1 7.6P 8.9/2.2  87.6   8.9   2.2  0.3109520  0.3068719 74.6134498
}

\keyword{Munsell}

\name{MunsellToRGB}
\alias{MunsellToRGB}

\title{Convert a Munsell specification to RGB coordinates}
\description{\code{MunsellToRGB} Converts a 
Munsell specification to RGB coordinates, by interpolating over the 
extrapolated Munsell renotation data
}

\usage{MunsellToRGB( MunsellSpec, space='sRGB', which='scene', maxSignal=255, 
    adapt='Bradford', ... ) }
\arguments{
\item{MunsellSpec}{ a numeric Nx3 matrix with HVC values in the rows, or a vector that can be converted to such a matrix, by row.\cr
\code{MunsellSpec} can also be a character vector with Munsell Notations,
which is converted to an Nx3 matrix using \code{\link{HVCfromMunsellName}()}. }

\item{space}{the name of an installed RGB space.
Spaces \code{'sRGB'} and \code{'AdobeRGB'} are pre-installed,
and others can be installed with \code{\link[spacesRGB:installRGB]{spacesRGB::installRGB}()}.
}

\item{which}{either \code{'scene'} or \code{'display'}; passed to \code{\link[spacesRGB:RGBfromXYZ]{spacesRGB::RGBfromXYZ}()} }

\item{maxSignal}{maximum of the non-linear signal RGB; Other popular values are 1, 100, 1023, and 65535}

\item{adapt}{method for chromatic adaptation, see \code{\link[spacesXYZ:CAT]{spacesXYZ::CAT}()} for valid values.
Also see \bold{Details}. }

\item{...}{other parameters passed to \code{\link{MunsellToxyY}()} }       
}

\value{
a \code{data.frame} with these columns:

\item{SAMPLE_NAME}{a character N-vector equal to the original \code{MunsellSpec} if that was a character vector.
Or the Munsell notation string converted from the input matrix HVC.}

\item{xyY}{ an Nx3 matrix with xyY values in the rows, that are adapted to Illuminant C.
This is an intermediate result that is sometimes useful, e.g. it can be passed
to \code{\link{IsWithinMacAdamLimits}()}.
The rownames of xyY are copied from those of HVC, unless the rownames of HVC are \code{NULL}
when they are set to \code{SAMPLE_NAME}. }

\item{RGB}{an Nx3 matrix with non-linear RGB signal values in the rows. 
All values are clamped to the appropriate cube, e.g. \eqn{[0,255]^3}.
The rownames of RGB are copied from those of xyY. }

\item{OutOfGamut}{logical vector, TRUE means the result was out of gamut (the cube) before clamping it}

In case of error, it returns \code{NULL}.
}


\details{
The conversion is done with these steps:
\itemize{
\item HVC  \rarrow  xyY using \code{\link{MunsellToxyY}()} with \code{...}. This xyY is for Illuminant C.

\item xyY  \rarrow  XYZ using \code{\link[spacesXYZ:toXYZ]{spacesXYZ::XYZfromxyY}()} 

\item XYZ is adapted from Illuminant C to the white-point (with \code{which}) of the RGB \code{space},
using \code{\link[spacesXYZ:adaptXYZ]{spacesXYZ::adaptXYZ}()}, with the given chromatic adaptation method \code{adapt}

\item XYZ  \rarrow  RGB using \code{\link[spacesRGB:RGBfromXYZ]{spacesRGB::RGBfromXYZ}()} with the given \code{space},
\code{which}, and \code{maxSignal}
}
}

\references{
Paul Centore 2014
The Munsell and Kubelka-Munk Toolbox
\MKMTB
}

\author{Jose Gama and Glenn Davis}

\seealso{
\code{\link{MunsellToXYZ}()},
\code{\link{IsWithinMacAdamLimits}()},
\code{\link[spacesRGB:installRGB]{spacesRGB::installRGB}()},
\code{\link[spacesRGB:RGBfromXYZ]{spacesRGB::RGBfromXYZ}()},
\code{\link[spacesXYZ:toXYZ]{spacesXYZ::XYZfromxyY}()},
\code{\link[spacesXYZ:CAT]{spacesXYZ::CAT}()}
}

\examples{
MunsellToRGB( c('7.6P 8.9/2.2', 'N 3/'), space='AdobeRGB' )
##     SAMPLE_NAME      xyY.x      xyY.y      xyY.Y     RGB.R     RGB.G     RGB.B OutOfGamut
##  1 7.6P 8.9/2.2  0.3109520  0.3068719 74.6134498 227.72419 220.18659 229.23297      FALSE
##  2         N 3/  0.3101000  0.3163000  6.3911778  73.01793  73.01793  73.01793      FALSE

}

\keyword{RGB}

\name{Munsell2xy}
\alias{Munsell2xy}
\docType{data}
\title{The Munsell HVC to xy 3D Lookup Table}

\description{
This is the discrete data for the Munsell Renotation System,
which is often considered to be the most perceptually uniform
color atlas.
It was created by the NBS and OSA from "3,000,000 color judgments"
by 40 observers.
}

% \usage{<none>}

\format{
\code{Munsell2xy} is a data.frame with 4995 rows and these 6 columns:
\itemize{
\item{\code{H}} the Munsell Hue. Each \code{H} is a multiple of 2.5 and in the interval (0,100].
\item{\code{V}} the Munsell Value.  Each \code{V} is an integer from 1 to 10, or one of 0.2, 0.4, 0.6, 0.8
\item{\code{C}} the Munsell Chroma.  Each \code{C} is a positive even integer.
\item{\code{x}} the x chromaticity coordinate, for Illuminant C.
\item{\code{y}} the y chromaticity coordinate, for Illuminant C.
\item{\code{real}} a logical value. If \code{TRUE} then x,y were published;
            otherwise they have been extrapolated from the published values.
}
Note that the luminance factor Y is *not* here,
since Y is a simple function of V,
see \code{\link{YfromV}()}.
}

\details{
All the (x,y) data here comes from the file \code{all.dat} downloaded from
Rochester Institute of Technology, see \bold{Source}.
The file \code{real.dat} is a subset, and contains the (x,y) published
in \cite{Newhall, et. al. (1943)}.
These rows have \code{real=TRUE} and are only for Value \eqn{\ge} 1.
There are 2734 of these.
\cr
Similarly, for Value<1 (very dark colors), (x,y) data from
the paper \cite{Judd et. al. (1956)} also have \code{real=TRUE}.
There are 355 of these.
\cr
So \code{all.dat} has 4995 colors,
of which 2734+355=3089 are "real" colors,
and the remaining 1906 are extrapolated.
I am confident that the extrapolation was done by
\emph{Schleter et. al. (1958)} at the NBS,
and put online by the Rochester Institute of Technology.
For more details, and the abstract of the 1958 article,
see the \bold{munsellinterpol User Guide}.
}

\note{
For the purpose of this package, I have found that the extrapolated (x,y)
for V\eqn{\ge}1 work well.
But for V<1 they did not work so well, and I was able to get better results with
my own extrapolation.
Moreover, to get reliable results in this package for high Chroma,
it was necessary to extrapolate past the data in \code{all.dat}.
}

\author{
Glenn Davis
}

\source{
Rochester Institute of Technology.
Program of Color Science.
Munsell Renotation Data.
\url{https://www.rit.edu/science/munsell-color-lab}
}

\references{
Newhall, Sidney M.,
Dorothy Nickerson,
Deane B. Judd.
\bold{Final Report of the O.S.A. Subcommitte on the Spacing of the Munsell Colors}.
Journal of the Optical Society of America.
Vol. 33.  No. 7.  pp. 385-418.
July 1943.

Judd, Deane B. and Gunter Wyszecki.
\bold{Extension of the Munsell Renotation System to Very Dark Colors}.
Journal of the Optical Society of America.
Vol. 46.  No. 4.  pp. 281-284.
April 1956.


Schleter, J. C, D. B. Judd, D. B., H. J. Keegan.
\bold{Extension of the Munsell Renotation System (Abstract)}.
J. Opt. Soc. Am.
Vol 48. Num. 11.
pp. 863-864.
presented at the Forty-Third Annual Meeting of the Optical Society of America.
Statler Hilton Hotel, Detroit, Michigan.  October 9, 10, and 11, 1958.
}

\seealso{
\code{\link{YfromV}()}
}

\examples{
str(Munsell2xy)

##  'data.frame':	4995 obs. of  6 variables:
##   $ H   : num  32.5 35 37.5 37.5 40 40 42.5 42.5 45 45 ...
##   $ V   : num  0.2 0.2 0.2 0.2 0.2 0.2 0.2 0.2 0.2 0.2 ...
##   $ C   : num  2 2 2 4 2 4 2 4 2 4 ...
##   $ x   : num  0.713 0.449 0.262 -0.078 0.185 -0.257 0.144 -0.235 0.117 -0.209 ...
##   $ y   : num  1.414 1.145 0.837 2.16 0.676 ...
##   $ real: logi  FALSE FALSE FALSE FALSE TRUE FALSE ...
}
\keyword{datasets}

\name{CentroidsISCCNBS}
\alias{CentroidsISCCNBS}
\docType{data}
\title{Centroid Notations for the Revised ISCC-NBS Color-Name Blocks}

% \usage{CentroidsISCCNBS}

\description{\code{CentroidsISCCNBS} is a table with the centroids
of the revised ISCC-NBS Color-Name Blocks, taken from \emph{Kelly (1958)}}

\format{
This \code{data.frame} has 267 rows and these columns:
\describe{
\item{Number}{ISCC-NBS number (an integer from 1 to 267)}
\item{Name}{ISCC-NBS name}
\item{MunsellSpec}{Munsell specification of the centroid of the block a (character string)}
}
}

\details{
The earliest paper I am aware of is by \emph{Nickerson, et. al.} in 1941.
After the big Munsell renotation in 1943, the name blocks were revised in 1955.
\cr
When the central colors were recomputed in
\emph{Kelly (1958)}, they were called the "Central Colors",
though the text makes it clear that most are truly centroids,
which were computed from the centroid of an "elementary shape",
which is a "sector of a right cylindrical annulus".
For the "peripheral blocks" of high chroma, the centroids were
"estimated graphically by plotting the MacAdam limits".
\cr
In \emph{Kelly (1965)} these were called "centroid colors",
and that is the name we will use here.
}



\references{
Nickerson, Dorothy and Sidney M. Newhall.
\bold{Central Notations for ISCC-NBS Color names}.
J Opt. Soc. Am.
Vol 31  Iss. 9.
pp. 597-591.
1941.

Newhall, Sidney M.,
Dorothy Nickerson,
Deane B. Judd.
\bold{Final Report of the O.S.A. Subcommitte on the Spacing of the Munsell Colors}.
Journal of the Optical Society of America.
Vol. 33.  No. 7.  pp. 385-418.
July 1943.

Kelly, Kenneth L. and 
Deane B. Judd
\bold{The ISCC-NBS Method of Designating Colors and a Dictionary of Color Names}.
National Bureau of Standards Circular 553.
Washington DC: US Government Printing Office.
November 1, 1955.

Kelly, Kenneth Low.
\bold{Central Notations for the Revised ISCC-NBS Color-Name Blocks}.
Journal of Research of the National Bureau of Standards.
Research Paper 2911.
Vol. 61  No. 5.
pp. 427-431.
November 1958.

Kelly, Kenneth Low.
\bold{A Universal Color Language}.
Color Engineering.
Vol. 3  No. 2.
pp. 2-7.
March-April, 1965.
}

\section{Contributor}{Glenn Davis}

\examples{
print( CentroidsISCCNBS[ 1:5, ] )

##    Number          Name   MunsellSpec
##  1      1    vivid pink     1.5R 7/13
##  2      2   strong pink  1.5R 7.5/9.1
##  3      3     deep pink 1.9R 6.0/11.1
##  4      4    light pink  2.5R 8.6/5.2
##  5      5 moderate pink  2.5R 7.2/5.2
}

\keyword{datasets}

\name{sRGBtoMunsell}
\alias{sRGBtoMunsell}
\title{Convert sRGB coordinates to Munsell HVC}

\description{Converts non-linear sRGB coordinates to Munsell HVC, 
by interpolating over the extrapolated Munsell renotation data
}

\usage{sRGBtoMunsell( sRGB, maxSignal=255, ... ) }

\arguments{
\item{sRGB}{ a numeric Nx3 matrix with signal sRGB coordinates in the rows,
or a vector that can be converted to such a matrix, by row.
These are non-linear signal values, but they are not required to be integers.}

\item{maxSignal}{maximum value of signal sRGB.
Other popular values are 1, 100, 1023, and 65535.
Even when 1, they are still taken to be non-linear signal values.}

\item{...}{other parameters passed to \code{\link{XYZtoMunsell}()}    }
}

\value{
a numeric Nx3 matrix with HVC coordinates in the rows.
The rownames are copied from input to output, unless the rownames are \code{NULL}
when they are set to the Munsell notations for HVC.\cr
In case of error, it returns \code{NULL}.
}


\details{
The conversion is done in these steps:
\itemize{
\item sRGB  \rarrow  XYZ using \code{\link[spacesRGB:XYZfromRGB]{spacesXYZ::XYZfromRGB}()} with the given \code{maxSignal}

\item XYZ is adapted from Illuminant D65 (from the sRGB standard) to Illuminant C,
using \code{\link[spacesXYZ:adaptXYZ]{spacesXYZ::adaptXYZ}()} with the \emph{Bradford} chromatic adaptation method

\item XYZ  \rarrow  HVC using \code{\link{XYZtoMunsell}()} and the given \code{...}
}
}

\note{
The more general function \code{\link{RGBtoMunsell}()} also performs this conversion.
This function has the advantage that it takes a little less time,
since the CAT (using the \emph{Bradford} method) is precomputed during
\code{\link[base:.onLoad]{base::.onLoad}()}.
But it has the disadvantage that the chromatic adaptation method cannot be changed.
}


\references{
Wikipedia. \bold{sRGB}.
\url{https://en.wikipedia.org/wiki/SRGB}.

Paul Centore 2014
The Munsell and Kubelka-Munk Toolbox
\MKMTB
}

\author{Jose Gama and Glenn Davis}

\seealso{
\code{\link{RGBtoMunsell}()},
\code{\link{XYZtoMunsell}()},
\code{\link[spacesXYZ:CAT]{spacesXYZ::CAT}()},
\code{\link[spacesXYZ:adaptXYZ]{spacesXYZ::adaptXYZ}()},
\code{\link[spacesRGB:XYZfromRGB]{spacesXYZ::XYZfromRGB}()}
}

\examples{
sRGBtoMunsell( c(255,45,67) )
##                     H        V        C
##  5.4R 5.5/18 5.401135 5.477315 18.01984

sRGBtoMunsell( c(1,0,1), maxSignal=1 )
##                 H        V        C
##  8P 6/26 87.98251 5.981297 25.64534
}

\keyword{sRGB}

\name{LabToMunsell}
\alias{LabToMunsell}
\alias{LabtoMunsell}

\title{Convert CIE Lab coordinates to Munsell HVC}

\description{\code{LabToMunsell} Converts CIE Lab coordinates to Munsell HVC,
by interpolating over the extrapolated Munsell renotation data
}

\usage{LabToMunsell( Lab, white='D65', adapt='Bradford', ... )}
\arguments{
  \item{Lab}{CIE Lab coordinates.  An Nx3 matrix, or a vector that can be converted to such a matrix. Each row of the matrix has Lab.}
  
  \item{white}{XYZ for the source white - a numeric 3-vector with scaling irrelevant.
    \code{white} can also be a numeric 2-vector; which is interpreted as xy chromaticity and converted to XYZ with Y=100.
    \code{white} can also be the name of any standard illuminant recognized by
    \code{\link[spacesXYZ:standardXYZ]{spacesXYZ::standardXYZ}()} which is scaled so that Y=100.
    The default \code{'D65'} is converted to \code{c(95.047,100,108.883)}. }
  
  \item{adapt}{method for chromatic adaptation, see \code{\link[spacesXYZ:adaptation]{spacesXYZ::CAT}()} for valid values.
    Also see \bold{Details}.}
  
  \item{...}{other parameters passed to \code{\link{XYZtoMunsell}()} }
}

\value{
An Nx3 matrix with the Munsell HVC coordinates in each row.
The rownames are set to those of Lab.
}

\details{
The conversion is done in these steps.
\itemize{
\item Lab  \rarrow  XYZ using \code{\link[spacesXYZ:toXYZ]{spacesXYZ::XYZfromLab}()} with the given \code{white}.
\item XYZ \rarrow  HVC using \code{\link{XYZtoMunsell}()}. Before the Munsell conversion, XYZ is internally adapted from the given \code{white} to Illuminant C using the given \code{adapt} method.
}
}

\note{
The case of the letter 't' in the function name was recently changed from lower to upper.
The function \code{LabtoMunsell()} is equivalent but deprecated and provided for a limited time.
}

\references{
Paul Centore 2014
The Munsell and Kubelka-Munk Toolbox
\MKMTB
}

\seealso{
\code{\link{MunsellToLab}()},
\code{\link{XYZtoMunsell}()},
\code{\link[spacesXYZ:standardXYZ]{spacesXYZ::standardXYZ}()},
\code{\link[spacesXYZ:adaptation]{spacesXYZ::CAT}()},
\code{\link[spacesXYZ:toXYZ]{spacesXYZ::XYZfromLab}()}
}

\author{Jose Gama and Glenn Davis}
\examples{
LabToMunsell( c(74.613450, -20.4, 10.1,    80, 0, 0) )
##                      H        V       C
##  3.1G 7.4/3.6 43.13641 7.379685 3.62976
##  N 7.9/        0.00000 7.945314 0.00000
}

\keyword{datasets}

\name{ColorBlockFromMunsell}
\alias{ColorBlockFromMunsell}

\title{Get ISCC-NBS Number and ISCC-NBS Name from Munsell HVC or Munsell Notation}

\description{Get ISCC-NBS Number and ISCC-NBS Name from Munsell HVC or Munsell Notation.
}

\usage{ColorBlockFromMunsell( MunsellSpec ) }

\arguments{
\item{MunsellSpec}{ a numeric Nx3 matrix or a vector that can be converted to
such a matrix.
Each row has Munsell HVC, where H is Hue Number,
and V and C are the standard Munsell Value and Chroma.
The Hue is automatically wrapped to the interval (0,100].
\code{MunsellSpec} can also be a character N-vector with standard Munsell notation; it is converted to an Nx3 matrix.}
}

\details{
The ISCC-NBS System is a partition of Munsell Color Solid into 267 color blocks.
Each block is a disjoint union of \emph{elementary blocks},
where an elementary block is defined
by its minimum and maximum limits in Hue, Value, and Chroma.
Some blocks are non-convex.
The \emph{peripheral blocks}, of which there are 120,
have arbitrary large chroma and are considered semi-infinite for this function;
there is no consideration of the MacAdam limits.
For each query vector HVC,
the function searches a private \code{data.frame} with 932 elementary blocks,
for the one elementary block that contains it.
}

\section{History}{
The \bold{Munsell Book of Color} was published in 1929.
The first ISCC-NBS partition, in 1939, had 319 blocks and names (including 5 neutrals).
There were no block numbers.
The aimpoints of the Munsell samples were thoroughly revised in 1943.
The ISCC-NBS partition was revised in 1955,
and this is the version used here.
}

\section{Future Work}{
It might be useful to compute the distance from the query point
to the boundary of the containing color block.
}

\value{
a \code{data.frame} with N rows and these columns:
\item{HVC}{the input Nx3 matrix, or such a matrix converted from Munsell notation}
\item{Number}{the corresponding ISCC-NBS color number - an integer from 1 to 267}
\item{Name}{the corresponding ISCC-NBS color name - a character string}
\item{Centroid}{the centroid of the block in Munsell Notation - a character string; see \code{\link{CentroidsISCCNBS}}}
The \code{rownames} are set to the input \code{MunsellSpec}.
}

\references{
Munsell Color Company, A.H. Munsell, and A.E.O. Munsell.
\bold{Munsell book of color: defining, explaining, and illustrating the fundamental characteristics of color}.
1929.

Judd, Deane B. and 
Kenneth L. Kelly.
\bold{Method of Designating Colors}.
Journal of Research of the National Bureau of Standards.
Research Paper 1239.
Volume 23  Issue 3.
pp. 355-385.
September 1939.

Newhall, Sidney M.,
Dorothy Nickerson,
Deane B. Judd.
\bold{Final Report of the O.S.A. Subcommitte on the Spacing of the Munsell Colors}.
Journal of the Optical Society of America.
Vol. 33.  No. 7.  pp. 385-418.
July 1943.

Kelly, Kenneth L. and 
Deane B. Judd
\bold{The ISCC-NBS Method of Designating Colors and a Dictionary of Color Names}.
National Bureau of Standards Circular 553.
Washington DC: US Government Printing Office.
November 1, 1955.
}

\examples{
ColorBlockFromMunsell( c( "3R 8/3", "7.4YR 3/4" ) )

##            HVC.H HVC.V HVC.C Number           Name      Centroid
##  3R 8/3      3.0   8.0   3.0      4     light pink  2.5R 8.6/5.2
##  7.4YR 3/4  17.4   3.0   4.0     58 moderate brown 5.5YR 3.5/3.9
}


\seealso{
\code{\link{CentroidsISCCNBS}}
}

\author{Glenn Davis}
\keyword{ISCC-NBS}

% right arrow, in either format
\newcommand{\rarrow}{ \if{latex}{\out{$\rightarrow$}} \if{html}{\out{&rarr;}} }

% inner-product angular brackets, in either format
\newcommand{\langle}{ \if{latex}{\out{$\langle$}} \if{html}{\out{&langle;}} }
\newcommand{\rangle}{ \if{latex}{\out{$\rangle$}} \if{html}{\out{&rangle;}} }

\newcommand{\MKMTB}{https://www.munsellcolorscienceforpainters.com/MunsellAndKubelkaMunkToolbox/MunsellAndKubelkaMunkToolbox.html}

\name{MunsellBooks}

\alias{MunsellBooks}
\docType{data}
\title{The Munsell Chips present in Specialized and Full Munsell Books}

\description{
In addition to the standard glossy and matte books,
Munsell publishes books for specialized fields.
For 5 of these specialized books, we use a paper by \bold{Ferguson}, which gives the
chip contents of books for these fields:
\bold{soil}, \bold{rocks}, \bold{beads}, and \bold{plants}.
It also includes \bold{newstudent} \emph{Color Set}, see \bold{References}.
\cr
For the full \bold{glossy} and \bold{matte} books, see \bold{Source}.
}


\section{Acknowledgements}{
The conversion of the 5 specialized books data to 
\code{./inst/extdata/Supplement1_3.6.2024.csv} was done by Willie Ondricek,
who provided all the motivation for this function.
He also kindly provided photographs of the soil book, and much other material.
}


\format{
\code{MunsellBooks} is a data.frame with 1753 rows and these columns:

\tabular{ll}{
{\code{HVC}} \tab 
a matrix with 3 columns: Hue, Value, and Chroma.
For the chromatic chips, Hue is a positive multiple of 2.5,
except for 3 chips in the bead book which are odd multiples of 1.25.
For the achromatic neutral chips, Hue is really undefined, but here it is set to 0.
Value is always a positive multiple of 0.25, and between 0.5 and 9.5.
Chroma is a non-negative integer between 0 and 16, except 13 and 15 are omitted.
\cr
\cr
{\code{FergusonName}} \tab
the color name of the chip, as assigned by \bold{Ferguson}.
If a chip does not appear in one of the 5 specialized books, it is set to \code{NA}.
\cr
\cr
{\code{bead}}         \tab
a logical value. If \code{TRUE} then the chip appears in the \bold{bead} book
\cr
\cr
{\code{newstudent}}         \tab
a logical value. If \code{TRUE} then the chip appears in the \bold{newstudent} book
\cr
\cr
{\code{plant}}         \tab
a logical value. If \code{TRUE} then the chip appears in the \bold{plant} book
\cr
\cr
{\code{rock}}         \tab
a logical value. If \code{TRUE} then the chip appears in the \bold{rock} book
\cr
\cr
{\code{soil}}         \tab
a logical value. If \code{TRUE} then the chip appears in the \bold{soil} book
\cr
\cr
{\code{glossy}}         \tab
a logical value. If \code{TRUE} then the chip appears in the \bold{glossy} book
\cr
\cr
{\code{matte}}         \tab
a logical value. If \code{TRUE} then the chip appears in the \bold{matte} book
}
The rows are sorted first by Hue, then Value, and then Chroma, all in increasing order.
Thus the achromatic neutral chips come first, followed by the chromatics.
}


\details{
The chip counts in the 7 books are:

\tabular{lccc}{
\bold{Book}         \tab  \bold{| Neutral Chips |} \tab  \bold{Chromatic Chips |} \tab \bold{Total Chips |} \cr
Soil                \tab           10          \tab          427            \tab         437  \cr
Rock                \tab            9          \tab          106            \tab         115  \cr
Bead                \tab            5          \tab          171            \tab         176  \cr
Plant               \tab            0          \tab          332            \tab         332  \cr
New Munsell Student \tab            8          \tab          238            \tab         246  \cr
Glossy              \tab           37          \tab         1485            \tab        1522  \cr
Matte               \tab           32          \tab         1269            \tab        1301  \cr
}
The union of all 7 books has 1753 chips.
There are only 3 inclusions among the books:
Rock \eqn{\subset} Glossy,  Plant \eqn{\subset} Glossy,  and New_Munsell_Student \eqn{\subset} Matte.
}


\source{
The data on the first 5 books of chips in the following table was converted from the
file
\cr
\code{yjfa_a_11710256_sm0001.docx},
which is \bold{Online Supplement #1} from \bold{Ferguson}.

For the Matte book, the 1269 chromatic chips
are taken from:
\cr
\code{https://sites.uef.fi/spectral/databases-software/munsell-colors-matt-spectrofotometer-measured/}.
\cr
For the Glossy book, the 1485 chromatic chips
are taken from:
\cr
\code{https://www.munsellcolorscienceforpainters.com/MunsellResources/MunsellResources.html}
\cr
\code{/SpectralReflectancesOf2007MunsellBookOfColorGlossy.txt}.

}


\references{
\bold{Ferguson, Jonathan}.
\emph{Munsell notations and color names: Recommendations for archaeological practice}.
\bold{Journal of Field Archaeology}. 39:4. 327-335. (2014).
Online Supplement #1.

\bold{Long, J. T.}
\emph{The New Munsell Student Color Set}.
New York: Fairchild Books. 2011.

\bold{Munsell Color}. 
\emph{Munsell Soil-Color Charts}.
Grand Rapids, MI:  2009.

\bold{Munsell Color}. 
\emph{Geological Rock-Color Chart}.
GrandRapids, MI: 2009.

\bold{Munsell Color}.
\emph{Munsell Bead Color Book}.
Grand Rapids, MI: 2012.

\bold{Munsell Color}.
\emph{Munsell Color Charts for Plant Tissues}.
New Windsor, NY:  1977.
}


\seealso{
\code{\link{roundHVC}()}
}


\examples{
str(MunsellBooks)

## 'data.frame':   1753 obs. of  9 variables:
## $ HVC         : num [1:1753, 1:3] 0 0 0 0 0 0 0 0 0 0 ...
##  ..- attr(*, "dimnames")=List of 2
##  .. ..$ : NULL
##  .. ..$ : chr [1:3] "H" "V" "C"
## $ FergusonName: chr  NA NA "Black" NA ...
## $ bead        : logi  FALSE FALSE TRUE FALSE FALSE FALSE ...
## $ newstudent  : logi  FALSE FALSE FALSE FALSE FALSE FALSE ...
## $ plant       : logi  FALSE FALSE FALSE FALSE FALSE FALSE ...
## $ rock        : logi  FALSE FALSE TRUE FALSE FALSE FALSE ...
## $ soil        : logi  FALSE FALSE FALSE FALSE FALSE FALSE ...
## $ glossy      : logi  TRUE TRUE TRUE TRUE TRUE TRUE ...
## $ matte       : logi  FALSE FALSE FALSE FALSE FALSE TRUE ...
}

\keyword{datasets}

\name{plotLociHC}
\alias{plotLociHC}
\title{Plot Curves of Constant Munsell Hue and Chroma}

\description{Plot Curves of Constant Munsell Hue and Chroma}

\usage{plotLociHC( value=5, hue=seq(2.5,100,by=2.5), chroma='auto', coords='xy',
                        main="Value \%g/", est=FALSE, ... ) }

\arguments{
\item{value}{a Munsell value for which the plot is created.
It must be in the interval (0,10].
\code{value} can also be a numeric vector of such numbers,
and then a separate plot is made for each element of the vector.}

\item{hue}{a numeric vector for which curves of contant Hue are plotted.
Each of these radial curves starts at Munsell Chroma = \code{min(chroma,1)}
and extends to \code{max(chroma)}.
\code{hue} can also be a character vector,
which is then converted to a numeric vector using
\code{\link{HueNumberFromString}()}.
}

\item{chroma}{a numeric vector for which ovoids of constant Chroma are plotted.
Each of these ovoids is closed; i.e. goes full circle from Hue=0 to Hue=100.
If \code{chroma='auto'} then a vector is chosen appropriate for the current \code{value}.
}

\item{coords}{either \code{'xy'} or \code{'ab'}.
If  \code{coords='xy'} then the plots are in the standard xy chromaticity plane.
If  \code{coords='ab'} then the plots are in the ab chrominance plane, from Lab.
Even when it is \code{'ab'}, the interpolation of loci takes place in \code{'xy'}
\emph{before} transformation to Lab.
}

\item{main}{a string used to set the main title of the plot.
The optional placeholder \code{'\%g'} is replaced by the current \code{value}.
}

\item{est}{if \code{TRUE}, initial estimates for the iteration used in
\code{\link{xyYtoMunsell}()} are plotted
}

\item{...}{other arguments passed to the function \code{MunsellToxyY()}.
This includes \code{hcinterp}, \code{vinterp}, and \code{xyC}.
However \code{warn=FALSE} is forced.
}
}

\details{
The plot limits (\code{xlim} and \code{ylim}) are set to include
all points where the Hue radials intersect the Chroma ovoids,
plus the white point.
\cr
If \code{value} is one of \{0.2,0.4,0.6,0.8,1,2,3,4,5,6,7,8,9,10\}
then published points from \code{real.dat}
are plotted with filled black points (real points),
and extrapolated points from \code{all.dat} are drawn with open circles (unreal points).
}

\note{
The option \code{hcinterp='bicubic'} makes fairly smooth ovoids,
and \code{hcinterp='bilinear'} makes 40-sided polygons (when \code{coords='xy'}).
Compare with the plots in  
\emph{Newhall et. al. (1943)},
\emph{Judd, et. al. (1956)},
and \emph{Judd, et. al. (1975)} p. 263.
}

\value{
\code{TRUE} for success and \code{FALSE} for failure.
}


\references{
Newhall, Sidney M.,
Dorothy Nickerson,
Deane B. Judd.
\bold{Final Report of the O.S.A. Subcommitte on the Spacing of the Munsell Colors}.
Journal of the Optical Society of America.
Vol. 33.  No. 7.  pp. 385-418.
July 1943.

Judd, Deane B. and Günther Wyszecki.
\bold{Extension of the Munsell Renotation System to Very Dark Colors}.
Journal of the Optical Society of America.
Vol. 46.  No. 4.  pp. 281-284.
April 1956.

Judd, Deane B. and Günther Wyszecki.
\bold{Color in Business, Science, and Industry}.
3rd edition.
John Wiley & Sons.
1975.
}

\author{Glenn Davis}

\seealso{
\code{\link{MunsellToxyY}()},
\code{\link{HueNumberFromString}()}
}

\keyword{plot}

\name{HVCfromMunsellName}
\alias{HVCfromMunsellName}
\alias{MunsellHVC}
\alias{HueNumberFromString}

%- Also NEED an '\alias' for EACH other topic documented here.
\title{Convert Munsell Notation to numerical Munsell HVC}

\description{Convert Munsell Notation to numerical Munsell HVC}

\usage{
HVCfromMunsellName( MunsellName )
MunsellHVC( MunsellName )
HueNumberFromString( HueString )
}

\arguments{
\item{MunsellName}{
a character vector of length N > 0,
where each string should be a valid Munsell notation, e.g.
\code{'2.3P 5/2.3',  '9.2YR 3/6',  'N 2.3/'}, and \code{'N 4/0'}. 
Whitespace is optional and ignored.
It is OK for a neutral to end in either \code{'/'} or \code{'/0'}.}

\item{HueString}{
a character vector of length N > 0,
where each string should be the initial hue part of a Munsell notation,
e.g. '4.5GY', '2.5R', '10.3B', etc.
Whitespace is optional and ignored.
Neutrals, denoted by \code{'N'}, are invalid because the hue is undefined.}
}


\value{
\code{HVCfromMunsellName()}
returns a numeric Nx3 matrix with HVC in the rows.
For neutral colors, both H and C are set to 0.
If a string cannot be parsed, the entire row is set to \code{NA}s.
The rownames are set to \code{MunsellName}.
\cr
\code{MunsellHVC()}
returns a character Nx3 matrix with HVC in the rows,
and is there for backward compatibility with older versions of the package.
For neutral colors, H is set to \code{'N'} and C is set to \code{'0'}.
\cr
\code{HueNumberFromString()} returns the hue number H (in (0,100]).
If the string cannot be parsed, or the color is neutral, the output is set to \code{NA}.
\cr
For all functions the Hue Number is wrapped to (0,100].
}

\note{
Ever since the \emph{Munsell Book of Color (1929)},
the Munsell hue circle has been divided into 5 \emph{principal hues} and
5 \emph{intermediate hues}, to define a total of 10 arcs.
Each of these hues has been assigned a 10-point scale, with 5 at the midpoint of the arc.
Moreover, the hue "origin" has been at '10RP'.
So a 100-point scale (with no letters) for the entire hue circle is obvious and trivial
to construct,
but I have been unable to determine the first explicit mention of such a scale.
The earliest I have have found is from \emph{Nimeroff (1968)},
Figures 20 and 21 on page 27.

\if{html}{\out{<div style="text-align: center">}}

\if{html}{\figure{hue-circle.jpg}{options: alt="Hue Circle"}}

\if{html}{\out{</div>}}

\if{latex}{\figure{hue-circle.jpg}}

There is a reference to \emph{ASTM D 1535} in the \bold{References} of \emph{Nimeroff},
but it is not dated, and the 2 figures are not attributed to it.
There was an \emph{ASTM D 1535} in 1968 but I have not been able to locate it;
it is possible that the 100-point scale first appeared in \emph{ASTM D-1535 (1968)},
or even earlier in \emph{ASTM D 1525-58T (1958)}.
\cr
Interestingly, in the \emph{Atlas of the Munsell Color System (1915)} there were only
5 principal hues (with no intermediate hues) and each of the 5 arcs was assigned a 10-point scale.
If the entire hue circle of 1915 were assigned a scale, it would have been a 50-point scale.

}



\references{
Nimeroff, I.
\bold{Colorimetry}.
National Bureau of Standards Monograph 104.
January 1968. 35 cents.

ASTM D 1535-80.
Standard Practice for Specifying Color by the Munsell System.
1980.

Munsell Book of Color: defining, explaining, and illustrating the fundamental characteristics of color.
Munsell Color Co.
1929.

Atlas of the Munsell Color System.
Malden, Mass., Wadsworth, Howland & Co., inc., Printers.
1915.
}

\author{
Glenn Davis
}


%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{MunsellNameFromHVC}()},
\code{\link{HueStringFromNumber}()}
}

\examples{
HVCfromMunsellName( c( "4.2P 2.9/3.8", "N 2.3/", "N 8.9/0" ) )
##                  H   V   C
##  4.2P 2.9/3.8 84.2 2.9 3.8
##  N 2.3/        0.0 2.3 0.0
##  N 8.9/0       0.0 8.9 0.0

HueNumberFromString( c('4B','4.6GY','10RP','N') )
##  [1]  64.0  34.6 100.0    NA
}

% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{MunsellName}% __ONLY ONE__ keyword per line

\name{MunsellToLuv}
\alias{MunsellToLuv}

\title{Convert a Munsell specification to CIE Luv coordinates}

\description{\code{MunsellToLuv} Converts a 
Munsell specification to CIE Luv coordinates, by interpolating over the 
extrapolated Munsell renotation data
}
\usage{MunsellToLuv( MunsellSpec, white='D65', adapt='Bradford', ... ) }

\arguments{
  \item{MunsellSpec}{ a numeric Nx3 matrix with HVC values in the rows, or a vector that can be converted to such a matrix, by row.\cr
\code{MunsellSpec} can also be a character vector with Munsell Notations,
which is converted to an Nx3 matrix using \code{\link{HVCfromMunsellName}()}}

  \item{white}{XYZ for the source white - a numeric 3-vector with scaling irrelevant.
    \code{white} can also be a numeric 2-vector; which is interpreted as xy chromaticity and converted to XYZ with Y=100.
    \code{white} can also be the name of any standard illuminant recognized by
    \code{\link[spacesXYZ:standardXYZ]{spacesXYZ::standardXYZ}()} which is scaled so that Y=100.
    The default \code{'D65'} is converted to \code{c(95.047,100,108.883)}. }
    
  \item{adapt}{ method for chromatic adaptation, see \code{\link[spacesXYZ:adaptation]{spacesXYZ::CAT}()} for valid values.
  Also see \bold{Details}. }
  
  \item{...}{other parameters passed to \code{\link{MunsellToXYZ}()}  }
}

\value{
An Nx3 matrix with the Luv coordinates in each row.
The rownames of Luv are copied from the input HVC matrix,
unless the rownames are \code{NULL} and
then the output rownames are the Munsell notations for HVC.
}

\details{
The conversion is done in these steps:
\itemize{
\item HVC  \rarrow  XYZ using \code{\link{MunsellToXYZ}()}.
In this function, XYZ is internally adapted from Illuminant C to the given \code{white} 
using \code{\link[spacesXYZ:adapt]{spacesXYZ::adaptXYZ}()} with the given chromatic adaptation method.
\item XYZ  \rarrow  Luv using \code{\link[spacesXYZ:fromXYZ]{spacesXYZ::LuvfromXYZ}()} with the given \code{white}
}
}


\references{
Paul Centore 2014
The Munsell and Kubelka-Munk Toolbox
\MKMTB
}

\author{Jose Gama and Glenn Davis}
\examples{
MunsellToLuv( c('7.6P 8.9/2.2', 'N 5/' ) )
##                      L        u         v
##  7.6P 8.9/2.2 89.19097 5.247155 -5.903808
##  N 5/         51.00375 0.000000  0.000000
}

\seealso{
\code{\link{LuvToMunsell}()},
\code{\link{MunsellToXYZ}()},
\code{\link[spacesXYZ:standardXYZ]{spacesXYZ::standardXYZ}()},
\code{\link[spacesXYZ:adaptation]{spacesXYZ::CAT}()},
\code{\link[spacesXYZ:adapt]{spacesXYZ::adaptXYZ}()},
\code{\link[spacesXYZ:fromXYZ]{spacesXYZ::LuvfromXYZ}()}
}

\name{IsWithinMacAdamLimits}
\alias{IsWithinMacAdamLimits}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Test xyY Coordinates for being Inside the MacAdam Limits
}
\description{
Test xyY Coordinates for being Inside the MacAdam Limits for Illuminants C and D65
}
\usage{
IsWithinMacAdamLimits( xyY, Illuminant='C' )
}
%- maybe also 'usage' for other objects documented here.
\arguments{
\item{xyY}{a numeric Nx3 matrix with CIE xyY coordinates in the rows,
or a vector that can be converted to such a matrix, by row. 
The reference white is asssumed to have Y=100.}

\item{Illuminant}{either \code{'C'} or \code{'D65'}. Partial matching is enabled, and is case-insensitive.}
}

\details{
The \emph{MacAdam Limit} is the boundary of the
\emph{optimal color solid} (also called the \emph{Rösch Farbkörper}),
in XYZ coordinates.
The optimal color solid is convex and depends on the illuminant.
Points on the boundary of the solid are called \emph{optimal colors}.
% This color solid has a smooth boundary (in the physical sense), except at black and white.
It is symmetric about the midpoint of the segment joining black and white
(the 50\% gray point).
It can be expressed as a \emph{zonohedron} - a convex polyhedron with a special form;
for details on zonohedra, see \cite{Centore}.

For each of the 2 illuminants, a zonohedron \eqn{Z} is pre-computed
(and stored in \code{sysdata.rda}).
The wavelengths used are 380 to 780 nm with 5nm step (81 wavelengths).
Each zonohedron has 81*80=6480 parallelogram faces, though some of them are coplanar.
\eqn{Z} is expressed as the intersection of 6480 halfspaces.
The plane equation of each parallelogram is pre-computed,
but redundant ones are not removed (in this version).

For testing a query point xyY, a pseudo-distance metric \eqn{\delta} is used.
Let the zonohedron \eqn{Z} be the intersection of the halfspaces
\langle \eqn{h_i,x} \rangle \eqn{\le b_i ~~ i=1,...,n}, 
where each \eqn{h_i} is a unit vector.
The point xyY is converted to XYZ, and \eqn{\delta}(XYZ) is computed as:
\eqn{\delta}(XYZ) := max( \langle \eqn{h_i},XYZ \rangle - \eqn{b_i} )
where the maximum is taken over all \eqn{i=1,...,n}.
This calculation can be optimized;
because the zonohedron is centrally symmetric, only half of the planes actually have to be stored,
and this cuts the memory and processing time in half.
It is clear that XYZ is within the zonohedron iff \eqn{\delta}(XYZ) \eqn{\le} 0,
and that XYZ is on the boundary iff \eqn{\delta}(XYZ)=0.
This pseudo-distance is part of the returned \code{data.frame}.

An interesting fact is that if \eqn{\delta}(XYZ)>0,
then \eqn{\delta}(XYZ) \eqn{\le} dist(XYZ,\eqn{Z}),
with equality iff the segment from \eqn{XYZ} to the
point \eqn{z} on the boundary of \eqn{Z} closest to XYZ is
normal to one of the faces of \eqn{Z} that contains \eqn{z}.
This is why we call \eqn{\delta} a \emph{pseudo-distance}.
Another interesting fact is that if \eqn{\delta}(XYZ) \eqn{\le} 0,
then \eqn{\delta}(XYZ) = -min( \eqn{\Psi_Z(u)} - \langle \eqn{u},XYZ \rangle ),
where the minimum is taken over all unit vectors \eqn{u} and
where \eqn{\Psi_Z} is the \emph{support function} of \eqn{Z}.
}


\value{
A \code{data.frame} with N rows and these columns:
\item{within}{a logical which is \code{TRUE} iff the corresponding row in \code{xyY}
is inside the optimal color solid for the illuminant.
If a point is exactly on the boundary (unlikely), \code{within=TRUE}.
Explicitly, \code{within = (delta<=0)}.}
\item{delta}{the pseudo-distance \eqn{\delta} discussed in \bold{Details}}

The row names of the output value are set equal to the row names of \code{xyY}.
}



\references{
Centore, Paul.
\emph{A zonohedral approach to optimal colours}.
\bold{Color Research & Application}.
Vol. 38.
No. 2.
pp. 110-119.
April 2013.

Rösch, S.
\bold{Darstellung der Farbenlehre für die Zwecke des Mineralogen}.
Fortschr. Mineral. Krist. Petrogr.
Vol. 13 No. 143.
1929.

MacAdam, David L.
\bold{Maximum Visual Efficiency of Colored Materials}.
Journal of the Optical Society of America.
Vol 25, No. 11.
pp. 361-367.
November 1935.

Wikipedia.
\bold{Support Function}.
\url{https://en.wikipedia.org/wiki/Support_function}
}


\author{
Glenn Davis and Jose Gama
}



\examples{
IsWithinMacAdamLimits( c(0.6,0.3,10, 0.6,0.3,20, 0.6,0.3,30, 0.6,0.3,40  ), 'C' )

##    within  delta
##  1   TRUE -1.941841
##  2   TRUE -1.332442
##  3  FALSE  3.513491
##  4  FALSE 12.826172
}


% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ MacAdam }

\name{plotPatchesH}
\alias{plotPatchesH}
\title{Plot Colored Patches for a fixed Munsell Hue}

\description{This plot simulates a page from the Munsell Book of Color.
The colors are best viewed on a display calibrated for
the RGB space given as the second argument.}

\usage{plotPatchesH( hue, space='sRGB', adapt='Bradford', background='gray50',
                        main="Hue \%s  (H=\%g)      [\%s   adapt=\%s]",
                        value=NULL, chroma=NULL, ... ) }

\arguments{

\item{hue}{a Munsell hue for which the plot is created.
It is automatically wrapped to the interval (0,100].
It does not have to be a multiple of 2.5.
\code{hue} can also be a numeric vector of such numbers,
and then a separate plot is made for each element of the vector.
\code{hue} can also be a character vector of Hue Names,
which is then converted to a numeric vector using
\code{\link{HueNumberFromString}()}.
}

\item{space}{the name of an installed RGB space.
Spaces \code{'sRGB'} and \code{'AdobeRGB'} are pre-installed,
and others can be installed using \code{\link[spacesRGB:installRGB]{spacesRGB::installRGB}()} 
}

\item{adapt}{method used to adapt xyY for Illuminant C to xyY for Illuminant D65.
It is passed to \code{\link{MunsellToRGB}()}.
}

\item{background}{background color for the plot.
It is passed to \code{par()} as argument \code{bg}.
}

\item{main}{a string used to set the main title of the plot.
The optional placeholder \code{'\%s'} is replaced by the Hue Name,
\code{'\%g'} is replaced by the Hue Number,
the next \code{'\%s'} is replaced by \code{space}, and
the last \code{'\%s'} is replaced by \code{adapt}.
}

\item{value}{a vector of Munsell Values to use for the plot.
Values must be in the interval [0,10], and in strictly increasing order.
Non-integers, such as 2.5 and 8.5, are valid.
If \code{value=NULL}, then it is set to \code{0:10}.
}

\item{chroma}{a vector of Munsell Chromas to use for the plot.
Chromas must be non-negative and in strictly increasing order.
Odd integers and non-integers are valid.
If \code{chroma=NULL}, then a sequence of even numbers is chosen,
starting at 0 and ending at the Chroma limit for the given Hue page, see \bold{Details}.
}

\item{...}{other arguments passed to the function \code{MunsellToRGB()}.
This includes \code{hcinterp}, \code{vinterp}, and \code{xyC}.
}
}

\details{
If \code{chroma=NULL},
for the closest discrete Hue in \code{real.dat},
the patches in \code{real.dat} are transformed to xyY using simple lookup with no interpolation.
These are then tested against the MacAdam Limits for Illuminant C
using \code{\link{IsWithinMacAdamLimits}()}.
The patches outside the limits are discarded,
and the maximum Chroma of the remaining patches,
which is always an even integer, determines \code{chroma}.
\cr
Patches inside the MacAdam Limits can still be outside the RGB cube;
those patches are drawn in outline only, and with the clamped RGB coordinates printed inside.
}

\value{
\code{TRUE} for success and \code{FALSE} for failure.
}


\author{Glenn Davis}

\seealso{
\code{\link{MunsellToRGB}()},
\code{\link{HueNumberFromString}()},
\code{\link{IsWithinMacAdamLimits}()},
\code{\link[spacesRGB:installRGB]{spacesRGB::installRGB}()}
}

\keyword{plot}

\name{MunsellNameFromHVC}
\alias{MunsellNameFromHVC}
\alias{HueStringFromNumber}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Convert Munsell Numeric to Munsell String Notation
}
\description{
Convert Munsell Numeric to Munsell String Notation
}

\usage{
MunsellNameFromHVC( HVC, format='g', digits=2 )

HueStringFromNumber( Hue, format='g', digits=2 )
}

%- maybe also 'usage' for other objects documented here.
\arguments{
\item{HVC}{a numeric Nx3 matrix or a vector that can be converted to
such a matrix.
Each row has an HVC vector, where H is Hue Number,
and V and C are the standard Munsell Value and Chroma.
The Hue is automatically wrapped to the interval (0,100].}

\item{Hue}{a numeric vector of Hue Numbers,
which are automatically wrapped to the interval (0,100].}

\item{format}{determines the meaning of the argument \code{digits},
and usually equal to \code{'g'} or \code{'f'} though other options are available,
see \code{\link[base]{formatC}()} for details.
When \code{format='f'} trailing 0s might be displayed.}

\item{digits}{
when \code{format='g'} the number of significant digits,
and when \code{format='f'} the number of digits displayed after the decimal point.
Both \code{format} and \code{digits} are passed as arguments to
\code{\link[base]{formatC}()}.}
}


\value{
Both functions return  a character vector of length N.
\code{MunsellNameFromHVC()} returns the full notation.
\code{HueStringFromNumber()} returns just initial the hue part;
which is useful for labeling plots.
}

\note{
If \code{format='f'}, then Chroma is first \emph{rounded} to to the given \code{digits}.
Chromas close to 0 may then become 0 and be displayed as a neutral, see \bold{Examples}.
\cr
The \code{width} argument of \code{\link[base]{formatC}()} is always set to 1,
to suppress leading spaces.
}

\references{
ASTM D 1535-97.
Standard Practice for Specifying Color by the Munsell System.
1997
}

\author{Glenn Davis}


%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link[base]{formatC}()},
\code{\link{HVCfromMunsellName}()},
\code{\link{HueNumberFromString}()}
}

\examples{
MunsellNameFromHVC( c(39,5.1,7.3,  0,5.1234,0.003 )  )
##  [1] "9GY 5.1/7.3"    "10RP 5.1/0.003"

MunsellNameFromHVC( c(39,5.1,7.34,  0,5.1234,0.003 ), format='f' )
##  [1] "9.00GY 5.10/7.34" "N 5.10/"

HueStringFromNumber( seq( 2.5, 100, by=2.5 ) )   # make nice labels for a plot
##   [1] "2.5R"  "5R"    "7.5R"  "10R"   "2.5YR" "5YR"   "7.5YR" "10YR"  "2.5Y" 
##  [10] "5Y"    "7.5Y"  "10Y"   "2.5GY" "5GY"   "7.5GY" "10GY"  "2.5G"  "5G"   
##  [19] "7.5G"  "10G"   "2.5BG" "5BG"   "7.5BG" "10BG"  "2.5B"  "5B"    "7.5B" 
##  [28] "10B"   "2.5PB" "5PB"   "7.5PB" "10PB"  "2.5P"  "5P"    "7.5P"  "10P"  
##  [37] "2.5RP" "5RP"   "7.5RP" "10RP" 
}


% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ MunsellName }% use one of  RShowDoc("KEYWORDS")

\name{LuvToMunsell}
\alias{LuvToMunsell}
\alias{LuvtoMunsell}
\title{Convert CIE Luv coordinates to Munsell HVC}

\description{\code{LuvToMunsell} Converts CIE Luv coordinates to Munsell HVC,
by interpolating over the extrapolated Munsell renotation data
}

\usage{LuvToMunsell( Luv, white='D65', adapt='Bradford', ... ) }
\arguments{
  \item{Luv}{CIE Luv coordinates An Nx3 matrix, or a vector that can be converted to such a matrix. Each row of the matrix has Luv.}
  
  \item{white}{XYZ for the source white - a numeric 3-vector with scaling irrelevant.
    \code{white} can also be a numeric 2-vector; which is interpreted as xy chromaticity and converted to XYZ with Y=100.
    \code{white} can also be the name of any standard illuminant recognized by
    \code{\link[spacesXYZ:standardXYZ]{spacesXYZ::standardXYZ}()} which is scaled so that Y=100.
    The default \code{'D65'} is converted to \code{c(95.047,100,108.883)}. }
  
  \item{adapt}{method for chromatic adaptation, see \bold{Details}}
  
  \item{...}{other parameters passed to \code{\link{XYZtoMunsell}()} }  
}

\value{
An Nx3 matrix with the Munsell HVC coordinates in each row.
The rownames are set to those of Luv.
}

\details{
The conversion is done in these steps:
\itemize{
\item Luv \rarrow XYZ using \code{\link[spacesXYZ:toXYZ]{spacesXYZ::XYZfromLuv}()} with the given \code{white}.
\item XYZ \rarrow HVC using \code{\link{XYZtoMunsell}()}. 
Before the conversion to HVC, XYZ is internally adapted from the given \code{white} to Illuminant C using
the given chromatic adaptation method, see \code{\link[spacesXYZ:adaptation]{spacesXYZ::CAT}()}.
}
}

\note{
The case of the letter 't' in the function name was recently changed from lower to upper.
The function \code{LuvtoMunsell()} is equivalent but deprecated and provided for a limited time.
}


\references{
Paul Centore 2014
The Munsell and Kubelka-Munk Toolbox
\MKMTB
}

\seealso{
\code{\link{XYZtoMunsell}()},
\code{\link[spacesXYZ:toXYZ]{spacesXYZ::XYZfromLuv}()},
\code{\link[spacesXYZ:standardXYZ]{spacesXYZ::standardXYZ}()},
\code{\link[spacesXYZ:adaptation]{spacesXYZ::CAT}()}
}


\author{Jose Gama and Glenn Davis}
\examples{
LuvToMunsell( c( 74.613450, -5.3108, 10.6,  55, 0, 0 ) )
##                      H        V        C
##  8.7GY 7.4/1.2 38.6599 7.383948 1.229376
##  N 5.4/         0.0000 5.395003 0.000000
}

\name{MunsellTosRGB}
\alias{MunsellTosRGB}

\title{Convert a Munsell specification to sRGB coordinates}

\description{\code{MunsellTosRGB} Converts a
Munsell specification to non-linear sRGB coordinates, by interpolating over the
extrapolated Munsell renotation data
}

\usage{MunsellTosRGB( MunsellSpec, maxSignal=255, ... ) }
\arguments{
  \item{MunsellSpec}{ a numeric Nx3 matrix with HVC values in the rows, or a vector that can be converted to such a matrix, by row.\cr
\code{MunsellSpec} can also be a character N-vector with Munsell notations,
which is converted to an Nx3 matrix using \code{\link{HVCfromMunsellName}()}. }
  \item{maxSignal}{maximum of signal sRGB; Other popular values are 1, 100, 1023, and 65535}

  \item{...}{other parameters passed to \code{\link{MunsellToxyY}()} }
}

\value{
a \code{data.frame} with these columns:

\item{SAMPLE_NAME}{the original \code{MunsellSpec} if that was a character vector.
Or the Munsell notation string converted from the input matrix HVC.}

\item{xyY}{ an Nx3 matrix with xyY values in the rows, that are adapted to Illuminant C.
This is an intermediate result that is sometimes useful, e.g. it can be passed
to \code{\link{IsWithinMacAdamLimits}()}.
The rownames of xyY are copied from those of HVC, unless the rownames of HVC are \code{NULL}
when they are set to \code{SAMPLE_NAME}. }

\item{RGB}{an Nx3 matrix with non-linear signal sRGB values in the rows.
All values are clamped to the appropriate cube, e.g. \eqn{[0,255]^3}.
The rownames of RGB are copied from those of xyY.}

\item{OutOfGamut}{logical vector, TRUE means the RGB result was out of gamut (the cube) before clamping it}
}


\details{
The conversion is done in these steps.
\itemize{
\item HVC  \rarrow  xyY using \code{\link{MunsellToxyY}()} and the given \code{...}. This xyY is for Illuminant C.

\item xyY  \rarrow  XYZ  using \code{\link[spacesXYZ:XYZfromxyY]{spacesXYZ::XYZfromxyY}()}

\item XYZ is adapted from Illuminant C to Illuminant D65 (from the sRGB standard) using
\code{\link[spacesXYZ:adaptXYZ]{spacesXYZ::adaptXYZ}()} and the \emph{Bradford} chromatic adaptation method

\item XYZ  \rarrow  sRGB using \code{\link[spacesRGB:RGBfromXYZ]{spacesRGB::RGBfromXYZ}()} with the given \code{maxSignal}
}
}

\note{
The more general function \code{\link{MunsellToRGB}()} also performs this conversion.
The main reason to use \code{MunsellTosRGB()} is that it takes a little less time,
since the CAT (using the \emph{Bradford} method) is precomputed during
\code{\link[base:.onLoad]{base::.onLoad}()}.
}

\references{
Paul Centore 2014
The Munsell and Kubelka-Munk Toolbox
\MKMTB
}

\author{Jose Gama and Glenn Davis}

\seealso{
\code{\link{MunsellToXYZ}()},
\code{\link{MunsellToRGB}()},
\code{\link[spacesXYZ:XYZfromxyY]{spacesXYZ::XYZfromxyY}()},
\code{\link[spacesXYZ:adaptation]{spacesXYZ::CAT}()},
\code{\link[spacesXYZ:adaptXYZ]{spacesXYZ::adaptXYZ}()},
\code{\link[spacesRGB:RGBfromXYZ]{spacesRGB::RGBfromXYZ}()},
\code{\link{IsWithinMacAdamLimits}()}
}


\examples{
MunsellTosRGB( c('7.6P 8.9/2.2', 'N 3/') )
##     SAMPLE_NAME      xyY.x      xyY.y      xyY.Y     RGB.R     RGB.G     RGB.B OutOfGamut
##  1 7.6P 8.9/2.2  0.3109520  0.3068719 74.6134498 231.35746 221.14207 230.35011      FALSE
##  2         N 3/  0.3101000  0.3163000  6.3911778  71.50491  71.50491  71.50491      FALSE
}

\name{RGBtoMunsell}
\alias{RGBtoMunsell}
\title{Convert RGB coordinates to Munsell HVC}

\description{\code{RGBtoMunsell} Converts RGB coordinates to Munsell HVC,
by interpolating over the extrapolated Munsell renotation data
}

\usage{RGBtoMunsell( RGB, space='sRGB', which='scene', maxSignal=255, adapt='Bradford', ... ) }

\arguments{
\item{RGB}{ a numeric Nx3 matrix with RGB coordinates in the rows,
or a vector that can be converted to such a matrix, by row.
These are non-linear display values, but they are not required to be integers.}

\item{space}{the name of an installed RGB space.
Spaces \code{'sRGB'} and \code{'AdobeRGB'} are pre-installed,
and others can be installed using
\code{\link[spacesRGB:installRGB]{spacesRGB::installRGB}()}.}

\item{which}{either \code{'scene'} or \code{'display'}; passed to \code{\link[spacesRGB:XYZfromRGB]{spacesRGB::XYZfromRGB}()} }

\item{maxSignal}{maximum value of signal RGB.
Other popular values are 1, 100, 1023, and 65535.
Even when 1, they are still taken to be non-linear signal values.}

\item{adapt}{method for chromatic adaptation,
see \code{\link[spacesXYZ:CAT]{spacesXYZ::CAT}()} for valid values.
Also see \bold{Details}.}
  
\item{...}{other parameters passed to \code{\link{XYZtoMunsell}()} }
}

\value{
a numeric Nx3 matrix with HVC coordinates in the rows.
The rownames are copied from input RGB to output HVC, unless the rownames are \code{NULL}
when they are set to the Munsell notations for HVC.\cr
In case of error, it returns \code{NULL}.
}


\details{
The conversion is done in these steps:
\itemize{
\item RGB  \rarrow  XYZ using \code{\link[spacesRGB:XYZfromRGB]{spacesRGB::XYZfromRGB}()} with the given
\code{space}, \code{which}, and \code{maxSignal}

\item XYZ is adapted from the white-point (with \code{which}) of \code{space} to Illuminant C,
using \code{\link[spacesXYZ:adaptXYZ]{spacesXYZ::adaptXYZ}()},
with the given chromatic adaptation method \code{adapt}

\item XYZ  \rarrow  HVC using \code{\link{XYZtoMunsell}()} with \code{...}
}
}


\references{
Wikipedia. \bold{sRGB}.
\url{https://en.wikipedia.org/wiki/SRGB}.

Paul Centore 2014
The Munsell and Kubelka-Munk Toolbox
\MKMTB
}

\author{Jose Gama and Glenn Davis}

\seealso{
\code{\link{XYZtoMunsell}()},
\code{\link[spacesXYZ:CAT]{spacesXYZ::CAT}()},
\code{\link[spacesXYZ:adaptXYZ]{spacesXYZ::adaptXYZ}()},
\code{\link[spacesRGB:XYZfromRGB]{spacesRGB::XYZfromRGB}()}
}

\examples{
RGBtoMunsell( c(255,45,67) )
##                     H        V        C
##  5.4R 5.5/18 5.401135 5.477315 18.01984

RGBtoMunsell( c(255,45,67), space='Adobe' )
##                     H        V        C
##  5.9R 6.2/22 5.924749 6.214155 21.83907
}

\keyword{RGB}

\name{MunsellToXYZ}
\alias{MunsellToXYZ}

\title{Convert a Munsell specification to CIE XYZ coordinates}

\description{\code{MunsellToXYZ} Converts a 
Munsell specification to XYZ coordinates, by interpolating over the 
extrapolated Munsell renotation data.

Note that \emph{by default} the returned \code{XYZ} is relative to the white point of Illuminant C.
Since Illuminant C is obsolete, it is almost certain that if this \code{XYZ} is used in subsequent
calculations, it must be chromatically adapted to a different white point.

But if the white point for subsequent calculations is supplied,
then the chromatic adaptation is done internally by
\code{\link[spacesXYZ:adaptXYZ]{spacesXYZ::adaptXYZ()}}.
}

\usage{MunsellToXYZ( MunsellSpec, white=NULL, adapt='Bradford', ... ) }
\arguments{
\item{MunsellSpec}{ a numeric Nx3 matrix with HVC values in the rows, or a vector that can be converted to such a matrix, by row.\cr
\code{MunsellSpec} can also be a character vector with Munsell Notations,
which is converted to an Nx3 matrix using \code{\link{HVCfromMunsellName}()}. }


\item{white}{XYZ for the output white point.  
  If \code{NULL} then the output XYZ are for viewing in an environment with Illuminant C,
  and the output \code{XYZ} is not chromatically adapted.
  If \code{white} is not \code{NULL}, then the interpolated \code{XYZ} is chromatically adapted from
  the white point of Illuminant C to the given white point.  
  \code{white} can be a numeric 3-vector; by convention \code{Y=100} for Munsell work.
  \code{white} can also be a numeric 2-vector; which is interpreted as xy chromaticity and converted to XYZ with Y=100.
  \code{white} can also be the name of any standard illuminant recognized by
  \code{\link[spacesXYZ:standardXYZ]{spacesXYZ::standardXYZ}()} which is then scaled so that Y=100.
  }

\item{adapt}{method for chromatic adaptation; for valid values see \code{\link[spacesXYZ:adaptation]{spacesXYZ::CAT}()}.
  Also see \bold{Details}.
  If \code{white=NULL} then \code{adapt} is ignored.
  }

\item{...}{other parameters passed to \code{\link{MunsellToxyY}()}        
}
}

\value{
an Nx3 matrix with XYZ values in the rows.
The rownames of XYZ are copied from the input HVC matrix,
unless the rownames are \code{NULL} and
then the output rownames are the Munsell notations for HVC.
}

\details{
The conversion is done in these steps:
\itemize{
\item HVC  \rarrow  xyY using \code{\link{MunsellToxyY}()} and \code{...}. This xyY is for Illuminant C.
\item xyY  \rarrow  XYZ using \code{\link[spacesXYZ:toXYZ]{spacesXYZ::XYZfromxyY}()} 
\item If \code{white} is not \code{NULL}, XYZ is chromatically adapted from the white point of Illuminant C to the given \code{white} using the given \code{adapt} method.
}
}

\references{
Paul Centore 2014
The Munsell and Kubelka-Munk Toolbox
\MKMTB
}

\author{Jose Gama and Glenn Davis}
\examples{
## make expected XYZ for a Perfect Reflecting Diffuser, under Illuminant D65
white.D65 = 100 * as.numeric( spacesXYZ::standardXYZ( 'D65' ) )
white.D65
##  [1]  95.047 100.000 108.883

## convert Munsell 'N 10/' - the Perfect Reflecting Difuser - to XYZ
MunsellToXYZ( "N 10/" )
##               X   Y        Z
##  N 10/ 98.03984 100 118.1157
##  this is the white point of Illuminant C, and not equal to white.D65

## Redo and supply the true white point.
MunsellToXYZ( "N 10/", white=white.D65 )
##             X   Y       Z
##  N 10/ 95.047 100 108.883
##
##  now we have a match


MunsellToXYZ('7.6P 8.9/2.2')
##                     X        Y        Z
##  7.6P 8.9/2.2 75.6055 74.61345 92.92308
}

\seealso{
\code{\link{MunsellToxyY}()},
\code{\link[spacesXYZ:toXYZ]{spacesXYZ::XYZfromxyY}()},
\code{\link[spacesXYZ:adaptXYZ]{spacesXYZ::adaptXYZ()}}
}


\keyword{XYZ}

\name{NickersonColorDifference}
\alias{NickersonColorDifference}

\title{Calculate the Nickerson Color Difference between two Colors}

\description{Calculate the Nickerson Color Difference between two colors, given in Munsell HVC;
see \cite{Nickerson}.
}

\usage{NickersonColorDifference( HVC0, HVC1, symmetric=TRUE, coeffs=c(0.4,6,3) ) }

\arguments{
\item{HVC0}{a numeric Nx3 matrix with HVC values in the rows, or a vector that can be converted to such a matrix, by row.
\code{HVC0} can also be a numeric 3-vector with a single HVC, and it is then replicated
to match the size of \code{HVC1}.
\cr
\code{HVC0} can also be a character N-vector with Munsell Notations,
which is converted to an Nx3 matrix using \code{\link{HVCfromMunsellName}()}. 
}

\item{HVC1}{a numeric Nx3 matrix with HVC values in the rows, or a vector that can be converted to such a matrix, by row.
\code{HVC1} can also be a numeric 3-vector with a single HVC, and it is then replicated
to match the size of \code{HVC0}.
\cr
\code{HVC1} can also be a character N-vector with Munsell Notations,
which is converted to an Nx3 matrix using \code{\link{HVCfromMunsellName}()}.
}

\item{symmetric}{if \code{FALSE} then use the original Nickerson difference formula,
and if \code{TRUE} then use a symmetrized version;
see \bold{Details}.
}

\item{coeffs}{the coefficients for Hue, Value, and Chroma in the difference equation,
see \bold{Details}.  The defaults are the ones used by Nickerson, but tweaking them is possible
in special circumstances.  The argument must have length 3 and all coefficents positive.
}
}

\value{
A numeric N-vector with the pairwise differences, 
i.e. between row i of \code{HVC0} and  row i of \code{HVC1}.
}

\details{
If \code{HVC0}=\eqn{H_0,V_0,C_0} and If \code{HVC1}=\eqn{H_1,V_1,C_1} then the original Nickerson formula is:
\deqn{NCD(HVC0,HVC1) = 0.4 C_0 \Delta H ~+~ 6 \Delta V  ~+~  3 \Delta C}
where \eqn{ \Delta H = |H_0 - H_1| } (on the circle),
\eqn{ \Delta V = |V_0 - V_1| } and
\eqn{ \Delta C = |C_0 - C_1| }.
Unfortunately, if \code{HVC0} and \code{HVC1} are swapped, the color difference is different.
The first color is considered to be the \emph{reference color} and the second one is the \emph{test color}.
The difference is not symmetric.
\cr
Another problem is that the difference is not continuous when the second color is a neutral gray,
for rectangular coordinates on a plane of constant V.

Both of these problems are fixed with a slightly modified formula:
\deqn{NCD(HVC0,HVC1) = 0.4 \min(C_0,C_1) \Delta H ~+~ 6 \Delta V  ~+~  3 \Delta C}

For the first formula set \code{symmetric=FALSE} and for the second formula set \code{symmetric=TRUE}.
}

\references{
Nickerson, Dorothy.
The Specification of Color Tolerances.
\bold{Textile Research}.
Vol 6. pp. 505-514.
1936.
}

\author{Jose Gama and Glenn Davis}



\examples{
NickersonColorDifference( '7.6P 8.9/2.2', '8P 8.2/3'  )
##  [1] 6.952
}
\keyword{datasets}

\name{MunsellSpecToColorlabFormat}
\alias{MunsellSpecToColorlabFormat}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Convert Munsell Specification to Colorlab Format
}
\description{
Convert Munsell Specification to Colorlab Format
}
\usage{
MunsellSpecToColorlabFormat( MunsellSpec )
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{MunsellSpec}{
a numeric Nx3 matrix, or a vector that can be converted to such a matrix, by row.
Each row of the matrix contains an HVC vector.
H is automatically wrapped to the interval (0,100].\cr
\code{MunsellSpec} can also be a character vector with Munsell Notation;
which is converted to an Nx3 matrix using \code{\link{HVCfromMunsellName}()}.
}
}

\details{
Colorlab Munsell format uses 4 numbers.
\enumerate{
\item Hue Step, in the interval (0,10], or 0 for neutrals.
In Colorlab documentation it is called the \emph{hue shade}.
It is also the Hue Number H mod 10 (unless H is an exact multiple of 10).
\item Munsell Value, in the interval [0,10]
\item Munsell Chroma, non-negative
\item Hue Index, an integer from 1 to 10, or 0 for neutrals.
This index defines the \emph{principal hue}, see \bold{Details}.
}

}

\value{
an Nx4 matrix, with rows as described in \bold{Details}.
Value and Chroma are simply copied unchanged.
The complex part is conversion of Hue Number to Colorlab Hue Step and Hue Index.
If Chroma is 0, both the Hue Step and Hue Index are set to 0.
Invalid input values are converted to \code{NA}s.\cr
If the input is a character vector,
the rownames of the returned matrix are set to that vector.
}

\references{
Color Processing Toolbox.      
Colorlab 1.0.
\url{https://www.uv.es/vista/vistavalencia/software/colorlab.html}

Cleland, T. M.
\bold{A Practical description of the Munsell Color System with Suggestions for its Use}.
(1921)
}

\author{
Jose Gama and Glenn Davis
}

\note{
The Colorlab format is closer to the
\emph{Munsell Book of Color (1929)} than HVC.
In the book the hue circle is divided into 10
\emph{principal hues} - 5 simple and 5 compound.
The 10 hue labels are: \bold{R, YR, Y, GY G, BG, B, PB, P}
(simple are 1 letter and compound are 2 letters).
In Colorlab these labels are replaced by the Hue Index.
WARNING: In the Munsell System, see \emph{Cleland}, there is a different Hue Index -
\bold{R} is 1, \bold{YR} is 2, ..., \bold{P} is 10.
The Colorlab index has a different origin, and goes around the circle in a different
direction !
\cr
Each one of these major hues corresponds to an arc on the circle,
with a 10-point hue scale.
The midpoint of the arc has hue step = 5.
Fortunately this 10-point hue scale is exactly the same as the Colorlab Hue Step.
}

%% ~Make other sections like Warning with \section{Warning }{....} ~




\seealso{
\code{\link{HVCfromMunsellName}()},
\code{\link{ColorlabFormatToMunsellSpec}()}
}

\examples{
MunsellSpecToColorlabFormat( c(100,5,5, 10,3,4, 90,4,3, 77,1,2, 66,2,0, 0,1,2 ) )
##                    HN V C HI
##  10.00RP 5.00/5.00 10 5 5  8
##  10.00R 3.00/4.00  10 3 4  7
##  10.00P 4.00/3.00  10 4 3  9
##  7.00PB 1.00/2.00   7 1 2 10
##  N 2.00/            0 2 0  0
##  10.00RP 1.00/2.00 10 1 2  8
}
 

% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ Colorlab }% use one of  RShowDoc("KEYWORDS")

\name{xyYtoMunsell}
\alias{xyYtoMunsell}

\title{Convert xyY coordinates to Munsell HVC}

\description{\code{xyYtoMunsell} Convert xyY coordinates to Munsell HVC,
by interpolating over the tabulated Munsell renotation data.\cr

Note that the input \code{xyY} must be relative to the white point of Illuminant C.
Since Illuminant C is obsolete, it is almost certain that \code{xyY}
in modern calculations \emph{must} be chromatically adapted
to the white point of Illuminant C \emph{before} passing to this function.
This chromatic adaptation can be done using
\code{\link[spacesXYZ:adaptxyY]{spacesXYZ::adaptxyY()}}.
Alternatively, one can use \code{\link{XYZtoMunsell}()},
which does the chromatic adaptation internally.

HVC \eqn{\to} xyY is the forward direction, and xyY \eqn{\to} HVC is the inverse direction.
This inversion requires requires root-finding,
which is done using \code{\link[rootSolve:multiroot]{rootSolve::multiroot()}}.
}

\usage{xyYtoMunsell( xyY, xyC='NBS', hcinterp='bicubic', vinterp='cubic',
                     VfromY='ASTM', rtol=1.e-8, atol=1.e-6, warn=TRUE, perf=FALSE ) }

\arguments{
\item{xyY}{a numeric Nx3 matrix with CIE xyY coordinates in the rows,
or a vector that can be converted to such a matrix, by row.  
These are for viewing in an environment with the white point of Illuminant C.
Y must be scaled so that Y=100 for the perfect reflecting diffuser.}

\item{xyC}{ a numeric 2-vector with xy chromaticity of Illuminant C.
It can also be one of the strings given in \code{\link{MunsellToxyY}()}. }

\item{hcinterp}{the type of 2D interpolation for H and C; either \code{'bicubic'} or \code{'bilinear'} (partial matching enabled).
See \code{\link{MunsellToxyY}()} for details. }

\item{vinterp}{the type of interpolation for V; either \code{'cubic'} or \code{'linear'} (partial matching enabled).
See \code{\link{MunsellToxyY}()} for details. }

\item{VfromY}{passed as the parameter \code{which} to the function \code{VfromY()}.
See \code{\link{VfromY}()} for details.
Option \code{'MGO'} is not allowed because then Y>100 when V=10.
}

\item{rtol}{relative tolerance passed to \code{\link[rootSolve:multiroot]{rootSolve::multiroot()}} }

\item{atol}{absolute tolerance passed to \code{\link[rootSolve:multiroot]{rootSolve::multiroot()}} }

\item{warn}{if an xyY cannot be mapped
(usually because the root finder has wandered afar),
its H and V are set to \code{NA} in the returned \code{data.frame}.
Just before returning, if any rows in HVC have an \code{NA}, and \code{warn == TRUE},
then a warning is logged.
}

\item{perf}{if \code{perf} is \code{TRUE}, then extra performance related metrics
are appended to the returned \code{data.frame}, see \bold{Value}.
}
}

\value{
a \code{data.frame} with N rows and these columns:
\item{xyY}{The input xyY}
\item{HVC}{the computed HVC. H is automatically wrapped to (0,100].  In case of failure, H and C are set to \code{NA}.
The rownames of HVC are set to those of xyY, unless they are \code{NULL} when they are set to \code{SAMPLE_NAME}.}
\item{SAMPLE_NAME}{the Munsell notation for HVC, a character vector}
If \code{perf} is \code{TRUE} then there are these additional columns:
\item{time.elapsed}{elapsed time in seconds. If available, the function \code{microbenchmark::get_nanotime()} is used.}
\item{iterations}{the number of iterations of \code{rootSolve::multiroot()}}
\item{evaluations}{the number of forward (HVC  \rarrow  xyY) function evaluations}
\item{estim.precis}{the estimated precision from \code{rootSolve::multiroot()}.
This is in the HC plane for the Munsell Value computed from Y.}

%If the rownames of xyY are not \code{NULL} and have no duplicates,
%they are copied to the returned data frame.
}

\details{
See \code{\link{MunsellToxyY}()} and the \bold{User Guide - Appendix C}.
}

\source{
Paul Centore 2014
The Munsell and Kubelka-Munk Toolbox
\MKMTB\cr
\url{https://www.rit.edu/science/munsell-color-lab}\cr
\url{https://www.rit-mcsl.org/MunsellRenotation/all.dat}\cr
\url{https://www.rit-mcsl.org/MunsellRenotation/real.dat}
}

\section{Warning}{
Even when \code{vinterp='cubic'} the function xyY  \rarrow HVC is not \eqn{C^1}
on the plane V=1.
This is because of a change in Value spacing:
when V\eqn{\ge}1 the Value spacing is 1, but when V\eqn{\le}1 the Value spacing is 0.2.
}

\references{
Judd, Deane B.
\bold{The 1931 I.C.I. Standard Observer and Coordinate System for Colorimetry}.
Journal of the Optical Society of America.
Vol. 23. pp. 359-374.
October 1933.

Newhall, Sidney M.,
Dorothy Nickerson,
Deane B. Judd.
\bold{Final Report of the O.S.A. Subcommitte on the Spacing of the Munsell Colors}.
Journal of the Optical Society of America.
Vol. 33.  No. 7.  pp. 385-418.
July 1943.

Kelly, Kenneth L.
Kasson S. Gibson.
Dorothy Nickerson.
\bold{Tristimulus Specification of the Munsell Book of Color from Spectrophometric
Measurements}
National Bureau of Standards RP1549
Volume 31. August 1943.

Judd, Deane B. and Günther Wyszecki.
\bold{Extension of the Munsell Renotation System to Very Dark Colors}.
Journal of the Optical Society of America.
Vol. 46.  No. 4.  pp. 281-284.
April 1956.

Paul Centore 2014
The Munsell and Kubelka-Munk Toolbox
\MKMTB
}


\author{Jose Gama and Glenn Davis}

\seealso{
\code{\link{MunsellToxyY}()},
\code{\link[rootSolve:multiroot]{rootSolve::multiroot()}},
\code{\link[spacesXYZ:adaptxyY]{spacesXYZ::adaptxyY()}},
\code{\link[microbenchmark:microbenchmark]{microbenchmark::get_nanotime()}}
}


\examples{
xyYtoMunsell(c(0.310897, 0.306510, 74.613450))
##       xyY.1     xyY.2     xyY.3     HVC.H     HVC.V     HVC.C    SAMPLE_NAME
## 1  0.310897  0.306510 74.613450 87.541720  8.900000  2.247428   7.5P 8.9/2.2


##  The chromaticity of D65 is xy = (0.3127,0.3290) to 4 digits.
##  If this is passed without chromatic adaptation, the result is not neutral.
munsellinterpol::xyYtoMunsell( c(0.3127,0.3290,18) )

##      xyY.x   xyY.y   xyY.Y      HVC.H      HVC.V      HVC.C    SAMPLE_NAME
##  1  0.3127  0.3290 18.0000 37.6679246  4.8519274  0.6461961 7.7GY 4.9/0.65

##  The Chroma is 0.65 instead of 0.
}


\keyword{datasets}

\name{VandY}
\alias{VfromY}
\alias{YfromV}
\title{Convert Munsell Value V to Luminance Factor Y, and back again}

\description{Convert non-linear Munsell Value V to linear Luminance Factor Y, and back again
}
\usage{
YfromV( V, which='ASTM' )
VfromY( Y, which='ASTM' )}

\arguments{
\item{V}{a numeric vector with elements in the interval [0,10]}

\item{Y}{a numeric vector with elements in the interval [0,100]}

\item{which}{ one of the strings in the first column of this table
\tabular{ll}{
              \tab    reference \cr
\code{'ASTM'} \tab    ASTM D-1535 (2008) \cr
\code{'OSA'} \tab     Newhall, et. al. (1943) \cr
\code{'MgO'} \tab     Newhall, et. al. (1943) \cr
\code{'Munsell'} \tab Munsell, et. al. (1933)\cr
\code{'Priest'} \tab Priest, et. al. (1920)\cr
}
Partial matching is enabled, and it is case insensitive.
}
}



\details{
\code{'Priest'} is the earliest (1920) transfer function in this package.
It is implemented as:
\deqn{V = sqrt(Y)  ~~~~~~and~~~~~~   Y = V^2 }
One readily checks that when V=10, Y=100, and vice-versa.
This transfer function has been implemented in colorimeters,
using analog electric circuits.
It is used in Hunter Lab - the precursor of CIE Lab.

\code{'Munsell'} is the next (1933) transfer function,
and was proposed by Munsell's son (Alexander Ector Orr Munsell) and co-workers.   
It is implemented as:
\deqn{V = sqrt( 1.474*Y - 0.00474*Y^2 )}
\deqn{Y = 50 * ( (1474 - sqrt(1474^2 - 4*4740*V^2)) / 474 )}
One readily checks that when V=10, Y=100, and vice-versa.
The luminance factor Y is \emph{absolute},
AKA \emph{relative to the perfect reflecting diffuser}.

\code{'Priest'} and \code{'Munsell'} are included in this package
for historical interest only.

The remaining three define Y as a quintic polynomial in V.

The next one historically - \code{'MgO'} -  is implemented as:
\deqn{ Y = (((((8404*V - 210090)*V  + 2395100)*V  - 2311100)*V + 10000000)*V ) / 10000000}
One readily checks that when V=10, Y=102.568.
This Y is larger than 100, because the authors decided to make Y relative
to a clean surface of MgO, instead of the perfect reflecting diffuser.
In their words:

\itemize{
\item It should be noted that the
reflectances indicated are not absolute but relative to magnesium oxide;
whereas the maximum at value 10/ was formerly 100 percent, it is now 102.57.
Use of this relation facilitates results and
also avoids the somewhat dubious conversion to
absolute scale, by permitting Y determinations
with a MgO standard to be converted directly to Munsell value.
}

Nowadays, the perfect reflecting diffuser is preferred over MgO.
For users who would like to modify this quintic as little as possible,
with the perfect reflecting diffuser in mind
(going back to \code{'MUNSELL'}),
I offer \code{'OSA'}, which is given by this quintic of my own design
\deqn{Y = (((((8404*V - 210090)*V  + 2395100)*V  - 2311100)*V + 10000000)*V ) / 10256800}

ASTM had a similar modification in mind,
but did it a little differently by scaling each coefficient.
\code{'ASTM'} is given by this quintic:
\deqn{Y = ( ((((81939*V - 2048400)*V  + 23352000)*V - 22533000)*V + 119140000)*V  ) / 1.e8 }
One readily checks that when V=10, Y=100 exactly (for both \code{'OSA'} and \code{'ASTM'}).

The inverses - from Y to V - of all 3 quintics are implemented as 
3 \code{splinefun()}s at a large number (about 300) of points.
These inverses are both fast and accurate.
The round-trip Y  \rarrow  V  \rarrow  Y is accurate to 7 digits after the decimal.
The round-trip V  \rarrow  Y  \rarrow  V is accurate to 8 digits after the decimal.
}

\value{
a numeric vector the same length as the input
}

\note{
The quintic functions \code{'ASTM'} and \code{'OSA'} are very close.
They agree at the endpoints 0 and 10 exactly, and the largest difference is near
V=6.767 where they differ by about 0.0007.
}


\references{
Priest, I. G. Gibson, K. S. and McNicholas, H. J.
\bold{An Examination of the Munsell Color System.
I. Spectral and and Total Reflection and the Munsell Scale of Value}.
Technologic Papers of the Bureau of Standards, No. 167.
pp. 1-33.
Washington D.C.
1920.

Munsell, A. E. O.,
L. L. Sloan,
and I. H. Godlove.
\bold{Neutral Value Scales. I. Munsell Neutral Value Scale}.
Journal of the Optical Society of America.
Vol. 23. Issue 11. pp. 394-411.
November 1933.

Newhall, Sidney M.,
Dorothy Nickerson,
Deane B. Judd.
\bold{Final Report of the O.S.A. Subcommitte on the Spacing of the Munsell Colors}.
Journal of the Optical Society of America.
Vol. 33.  No. 7.  pp. 385-418.
July 1943.

ASTM D 1535-08.
Standard Practice for Specifying Color by the Munsell System.
2008
}

\author{Glenn Davis}

\seealso{
\code{\link{MunsellToxyY}()},
\code{\link{xyYtoMunsell}()}
}

\examples{
VfromY( c(0,50,100) )
##  [1]  0.00000  7.53772 10.00000
}

\keyword{datasets}

\name{ColorlabFormatToMunsellSpec}
\alias{ColorlabFormatToMunsellSpec}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Convert Colorlab Munsell Format to Munsell HVC}
\description{Convert Colorlab Munsell Format to Munsell HVC}

\usage{
ColorlabFormatToMunsellSpec( HVCH )
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{HVCH}{
a numeric Nx4 matrix, or a vector that can be converted to such a matrix, by row.
Each row of the matrix contains an HVCH vector.\cr
}
}
 
\value{
an Nx3 matrix, with each row an HVC vector.
Value and Chroma are simply copied unchanged.
The complex part is conversion of Colorlab Hue Step and Hue Index
to Hue Number.
For neutrals, both Hue Step and Hue Index are ignored.
Invalid input values, such as a Hue Index that is not
an integer from 0 to 10 (except for neutrals), are converted to \code{NA}s.
The rownames of the input are copied to the output,
but if these are \code{NULL}, the rownames are set to the Munsell notations.
}

\details{
Colorlab Munsell format uses 4 numbers.
\enumerate{
\item Hue Step, in the interval (0,10], or 0 for neutrals
\item Munsell Value, in the interval (0,10]
\item Munsell Chroma, non-negative
\item Hue Index, an integer from 1 to 10, or 0 for neutrals
}
}


\references{
Color Processing Toolbox.      
Colorlab 1.0.
\url{https://www.uv.es/vista/vistavalencia/software/colorlab.html}
}
\author{
Jose Gama and Glenn Davis
}


%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{MunsellSpecToColorlabFormat}()}
}

\examples{
ColorlabFormatToMunsellSpec( c( 3.2,3,2,1,  2,5.1,0,0, 2,5.1,0.1,0 ) )
##                     H   V  C
##  3.20B 3.00/2.00 63.2 3.0  2
##  N 5.10/          0.0 5.1  0
##  <NA>              NA  NA NA
}

% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ Colorlab }% use one of  RShowDoc("KEYWORDS")

\name{XYZtoMunsell}
\alias{XYZtoMunsell}

\title{Convert XYZ coordinates to Munsell HVC}

\description{\code{XYZtoMunsell()} converts XYZ coordinates to Munsell HVC,
by interpolating over the  Munsell renotation data.

Note that \emph{by default} the input \code{XYZ} is relative to the white point of Illuminant C.
Since Illuminant C is obsolete, it is almost certain that an \code{XYZ} appearing
in modern calculations \emph{must} be chromatically adapted
to the white point of Illuminant C before interpolating.
This chromatic adaptation is done internally if the white point of XYZ is supplied as argument \code{white}.
The adaptation is performed by \code{\link[spacesXYZ:adaptXYZ]{spacesXYZ::adaptXYZ()}}.

If the input XYZ is computed from a reflectance spectrum,
then ideally the white point should be set to the XYZ computed when that reflectance spectrum
has constant value 1 - the so-called \emph{Perfect Reflecting Diffuser}.
The white points in published standards, e.g. D65 white = (95.047, 100, 108.883),
can differ from spectrally-derived white points
because of different wavelength step sizes and ranges.
For an example, see the vignette \href{../doc/soil.html}{\bold{Soil Colors}}.
}

\usage{XYZtoMunsell( XYZ, white=NULL, adapt='Bradford', ... ) }
\arguments{
\item{XYZ}{a numeric Nx3 matrix with CIE XYZ coordinates in the rows,
or a vector that can be converted to such a matrix, by row.
Ideally, XYZ should be scaled so that Y=100 for the perfect reflecting diffuser.
}

\item{white}{XYZ for the source white point.  
  If \code{NULL} then the input XYZ are for viewing in an environment with Illuminant C, and the input \code{XYZ} is not chromatically adapted.
  If \code{white} is not \code{NULL}, then the input \code{XYZ} is chromatically adapted from
  the given \code{white} to the white point of Illuminant C before conversion to HVC.  
  \code{white} can be a numeric 3-vector; by convention \code{Y=100} for Munsell work.
  \code{white} can also be a numeric 2-vector; which is interpreted as xy chromaticity and converted to XYZ with Y=100.
  \code{white} can also be the name of any standard illuminant recognized by
  \code{\link[spacesXYZ:standardXYZ]{spacesXYZ::standardXYZ}()} which is then scaled so that Y=100.
  
  If the input XYZ is computed from a reflectance spectrum,
  then ideally \code{white} should be set to the XYZ computed
  when that reflectance spectrum has constant value 1 (see \bold{Description}).
  }

\item{adapt}{method for chromatic adaptation, see \code{\link[spacesXYZ:adaptation]{spacesXYZ::CAT}()} for valid values.
  Also see \bold{Details}.
  If \code{white=NULL} then \code{adapt} is ignored.
  }

\item{...}{other parameters passed to \code{\link{xyYtoMunsell}()} }
}

\value{
an Nx3 matrix with Munsell HVC in the rows.
The rownames are copied from input to output, unless the input rownames are \code{NULL}
when the ouput rownames are set to the Munsell notations for HVC.\cr
In case of error, it returns \code{NULL}.
}

\details{
The conversion is done in these steps:
\itemize{
\item If \code{white} is not \code{NULL}, XYZ is chromatically adapted from the given \code{white} to the white point of Illuminant C using the given \code{adapt} method.
\item XYZ  \rarrow  xyY using \code{\link[spacesXYZ:xyYfromXYZ]{spacesXYZ::xyYfromXYZ}()} 
\item xyY  \rarrow  HVC using \code{\link{xyYtoMunsell}()} and \code{...}.  This function does the interpolation over the  Munsell renotation data.
}
}

\references{
Paul Centore 2014.
The Munsell and Kubelka-Munk Toolbox.
\MKMTB
}

\author{Jose Gama and Glenn Davis}

\seealso{
\code{\link{MunsellToXYZ}()},
\code{\link[spacesXYZ:xyYfromXYZ]{spacesXYZ::xyYfromXYZ}()},
\code{\link[spacesXYZ:adaptXYZ]{spacesXYZ::adaptXYZ()}}
}

\examples{
## make XYZ for an 18% gray card, under Illuminant D65
white.D65 = 100 * as.numeric( spacesXYZ::standardXYZ( 'D65' ) )
gray18 = 0.18 * white.D65
XYZtoMunsell( gray18 )
##                       H        V         C
##  7.6GY 4.9/0.65 37.6456 4.851927 0.6466169

## This is not neutral gray, as expected.  The hue is green-yellow.
## Chroma = 0.6466 is too large; it should be 0, up to numerical truncation.
## The problem is that the white points are not equal.
## Redo and supply the true white point.

XYZtoMunsell( gray18, white=white.D65 )
##         H        V C
##  N 4.9/ 0 4.851927 0        #  much better; an exact neutral

##  Although the hue H=0 here, it is actually undefined because C=0
##  (as in polar coordinates, theta is undefined when r=0).


## When XYZ=(0,0,0) then Value should be exactly 0.
## And when XYZ is equal to the white point, then Value should be exactly 10.
XYZtoMunsell( c( 0,0,0, white.D65 ), white=white.D65 )
##        H  V C
##  N 0/  0  0 0
##  N 10/ 0 10 0


##  By convention, Y=100 for the reference white; but this is not strictly required.
gray18
##              X  Y        Z
##  [1,] 17.10846 18 19.59894

XYZtoMunsell( gray18, white=gray18 )
##        H  V C
##  N 10/ 0 10 0
##  The Value is still exactly 10.
}

\keyword{XYZ}
